import os, sys, getopt, subprocess, time

def usage():
	print """
Usage: getSignedServerCert.py

    -d <CERTDIR>    Directory where rootCa and other certs are stored (required)

    -n <CERTNAME>   The name of the cert (required)

    -c <COMMONNAME> The CommonName for the cert. This should match the DNS
                    name, if DNS is not available then the IP will suffice.
                    It must be one of these two. (optional)

    -l <KEYLEN>     Specifiy the length of RSA key to generate. Default 1024

    -p              Prompt for other arguments in the cert. without prompt it
                    fills in CN and org name (optional)

    SPLUNK_HOME environment variable must be set before running.
"""
	sys.exit(1)

DEBUG = False

def execCmdDebug(arglist):
	print int(time.time()),
	for arg in arglist:
		print " %s" % arg,
	print

def execCmd(arglist):
	if DEBUG:
		execCmdDebug(arglist)
	subprocess.call(arglist)

SPLUNK_HOME = os.environ.get("SPLUNK_HOME")
if not SPLUNK_HOME:
	usage()

CERTDIR = ""
CERTNAME = ""
COMMONNAME = ""
PROMPT = True
KEYLEN = "1024"

try:

	opts, args = getopt.getopt(sys.argv[1:], "d:n:c:l:p", [])

	for o, a in opts:
		if o == '-d':
			CERTDIR = a
		elif o == '-n':
			CERTNAME = a
		elif o == '-c':
			COMMONNAME = a
		elif o == '-l':
			KEYLEN = a
		elif o == '-p':
			PROMPT = False
except getopt.GetoptError, exc:
	usage()

cert_pem__file = "%scert.pem" % CERTNAME
key_pem_file  = "%skey.pem" % CERTNAME
req_pem_file  = "%sreq.pem" % CERTNAME
output_file = "%s.pem" % CERTNAME
ca_pem_file = "ca.pem"
cacert_pem_file = "cacert.pem"

openssl_cnf = None
if sys.platform == "win32":
	# on windows the config file is stored in different location from UNIX
	openssl_cnf = os.path.join(SPLUNK_HOME, "openssl.cnf")
	os.environ["OPENSSL_CONF"] = openssl_cnf
else:
	openssl_cnf = os.path.join(SPLUNK_HOME, "openssl", "openssl.cnf")

if CERTDIR == "" or CERTNAME == "":
	# required arguments
	usage()

if COMMONNAME == "":
	print "No commonName specified, if client side checks are enabled connection negotiation may fail."
	COMMONNAME = CERTNAME

os.chdir(CERTDIR)

if not (os.path.isfile(ca_pem_file) and os.access(ca_pem_file, os.R_OK)):
	print "Unable to find %s. You must supply the location of your CA in order " \
		"to sign this cert. Aborting generation." % ca_pem_file
	sys.exit(1)

print "* Create certificate %s.pem signed by the root CA" % CERTNAME
print "* Store the %s.pem key file locally with your client/server application" % CERTNAME
print "* Enter a secret pass phrase when requested"
print "* The pass phrase is used to access %s.pem in your application" % CERTNAME
print "* Enter the application's host name as the Common Name when requested"
print "* Enter the root CA pass phrase (Getting CA Private Key) to sign the key file"
print "* The key file will expire after one year or sooner when the root CA expires"

if PROMPT == False:
	# Create a certificate and signing request
	# Runs: openssl req -newkey rsa:KEYLEN -sha1 -keyout "${CERTNAME}key.pem" -out "${CERTNAME}req.pem" -config "${SPLUNK_HOME}/openssl/openssl.cnf"
	args = [
		"openssl",
		"req",
		"-newkey",
		"rsa:" + KEYLEN,
		"-sha1",
		"-keyout",
		key_pem_file,
		"-out",
		req_pem_file,
		"-config",
		openssl_cnf
	]
	execCmd(args)
else:
	# Runs: openssl req -newkey rsa:KEYLEN -passout pass:password -sha1 -batch  -subj /commonName="${CN}"/organizationName=SplunkUser/  -keyout "${CERTNAME}key.pem" -out "${CERTNAME}req.pem" -config "${SPLUNK_HOME}/openssl/openssl.cnf"
	args = [
		"openssl",
		"req",
		"-newkey",
		"rsa:" + KEYLEN,
		"-passout",
		"pass:password",
		"-sha1",
		"-batch",
		"-subj",
		"/commonName=%s/organizationName=SplunkUser/" % COMMONNAME,
		"-keyout",
		key_pem_file,
		"-out",
		req_pem_file,
		"-config",
		openssl_cnf
	]
	execCmd(args)

# Sign the certificate with the root CA
# Runs: openssl x509 -req -in "${CERTNAME}req.pem" -sha1 -passin pass:password -extensions usr_cert -CA ca.pem -CAkey ca.pem -CAcreateserial -out "${CERTNAME}cert.pem" -days 1095
args = [
   	"openssl",
	"x509",
	"-req",
	"-in",
	"%sreq.pem" % CERTNAME,
	"-sha1",
	"-passin",
	"pass:password",
	"-extensions",
	"usr_cert",
	"-CA",
	ca_pem_file,
	"-CAkey",
	ca_pem_file,
	"-CAcreateserial",
	"-out",
	cert_pem__file,
	"-days",
	"1095",
]
execCmd(args)

# Runs: cat "${CERTNAME}cert.pem"  "${CERTNAME}key.pem"  cacert.pem > "${CERTNAME}.pem"
outfile = open(output_file, "w")
for file in [cert_pem__file, key_pem_file, cacert_pem_file]:
	infile = open(file, "r")
	outfile.writelines(infile.readlines())
	infile.close()
outfile.close()

# Runs: openssl x509 -subject -issuer -dates -passin pass:password -noout -in "${CERTNAME}.pem"
args = [
   	"openssl",
	"x509",
	"-subject",
	"-issuer",
	"-dates",
	"-passin",
	"pass:password",
	"-noout",
	"-in",
	output_file
]
execCmd(args)

# Runs: rm "${CERTNAME}cert.pem" "${CERTNAME}key.pem" "${CERTNAME}req.pem"
for file in [cert_pem__file, key_pem_file, req_pem_file]:
	os.remove(file)

