################################################################################
#                            Application Management                            #
################################################################################

from __future__ import with_statement
from contextlib import closing, nested

import splunk
import splunk.bundle as bundle
import splunk.clilib.bundle_paths as bundle_paths
import splunk.rest.format as format
import splunk.appbuilder as appbuilder

import logging as logger
import lxml.etree as etree
import os
import platform
import re
import time
import urllib
import urllib2

import splunk.clilib.cli_common as cli

# Overriden by settings in server.conf -> [applicationsManagement]
DEFAULT_URL = "http://www.splunkbase.com/api/apps"
LOGIN_URL   = "http://www.splunkbase.com/api/account:login/"

APPS_LOCAL  = "local"
APPS_REMOTE = "remote"

HTTP_METHOD_GET     = "GET"
HTTP_METHOD_POST    = "POST"
HTTP_METHOD_DELETE  = "DELETE"

HTTP_ACTION         = "action"
HTTP_ACTION_INSTALL = "install"


HTTP_GET_COUNT      = "count"
HTTP_GET_OFFSET     = "offset"
HTTP_GET_QUERY      = "q"
HTTP_GET_SORTBY     = "sort_by"
HTTP_GET_SORTDIR    = "sort_dir"

HTTP_AUTH_TOKEN     = "auth"
HTTP_AUTH_HEADER    = "X-Auth-Token"

ATOM_NS = "{%s}" % format.ATOM_NS
NSMAP   = {'a' : format.ATOM_NS}

HTTP_POST_LOCATION  = "location"
HTTP_POST_NAME      = "name"
HTTP_POST_TEMPLATE = "template"

'''
/services/apps/install
'''
class AppsHandlerInstall(splunk.rest.BaseRestHandler):

	"""
	Installs an application from a local filepath or URI
	"""

	def handle_POST(self):
		if not self.args.has_key(HTTP_POST_LOCATION):
			logger.error('Filepath or URI not found')
			raise splunk.BadRequest('Filepath or URI not found')

		location = self.args[HTTP_POST_LOCATION]

		appbuilder.installApp(location)
		

class AppsHandlerList(splunk.rest.BaseRestHandler):

	"""
	Generate links to local and remote application management.
	"""

	def handle_GET(self):
		links = { }
		links[APPS_LOCAL] = "Local Applications"
		links[APPS_REMOTE] = "Remote Applications"
		return links

		

'''
A simple xml response message generator
'''	
def _buildResponseMessage(type, text):
	msg = "<msg type='%s'>%s</msg>" % (type.upper(), text)
	xml = ['<?xml version="1.0" encoding="UTF-8"?>']
	xml.append('<response>')
	xml.append('\t<messages>')
	xml.append('\t\t'+ msg)
	xml.append('\t</messages>')
	xml.append('</response>')
	return "\n".join(xml)

################################################################################
#                         Local Application Management                         #
################################################################################


class LocalAppsManager(splunk.rest.BaseRestHandler):

	"""
	Manage locally-installed applications.
	"""

	def handle_GET(self):
		"""
		Respond to an HTTP GET with information about locally-installed
		applications.
		"""
		parts = len(self.pathParts)
		feed = AtomFeed(self.request['path'])

		if parts == 3:
			'''
			Show all apps  that match the given filter.
			'''
			filter = Filter(self.args)
			results = filter.filter(appbuilder.getAppsIterator())
			for b in results:
				if b.name() == 'appbuilder': continue
				feed.add_bundle(b, False)
			self.response.write(feed.toXml())
			
		elif parts == 4:
			'''
			Show one bundle.
			'''
			b = self.self_to_bundle()
			if b is None:
				raise splunk.ResourceNotFound
			entry = feed.bundle_to_entry(b, True)
			self.response.write(entry.toXml())

		elif parts == 5:
			if self.pathParts[-1] == 'package':
				'''
				Package app
				'''
				appName = urllib.unquote(self.pathParts[-2])
				if not appbuilder.appExists(appName):
					raise splunk.ResourceNotFound("Application doesn't exist")

				pkgPath = appbuilder.packageApp(appName)
				
				# Serve the file from /static
				entry = format.AtomEntry(id = 'download',
								title = pkgPath,
								updated = format.strftime(time.localtime())
								)
				entry.links = [('alternate', pkgPath)]
				
				feed.entries.append(entry)
				self.response.write(feed.toXml())
				
			elif self.pathParts[-1] == 'components':
				'''
				Return app components
				'''
				appName = urllib.unquote(self.pathParts[-2])
				if not appbuilder.appExists(appName):
					raise splunk.ResourceNotFound("Application doesn't exist")

				components = appbuilder.getAppComponents(appName)
				
				for comp in components:
					feed.add_component(comp)
				
				feed.title = '%s Components' % appName.capitalize()
				self.response.write(feed.toXml())
			
			else:
				raise splunk.BadRequest
				
		elif parts == 6:
			'''
			Returns a list all different components of a given type
			'''
			componentType = self.pathParts[-1]
			appName = urllib.unquote(self.pathParts[-3])
			
			if not appbuilder.appExists(appName):
				raise splunk.ResourceNotFound("Application doesn't exist")
	
			components = appbuilder.getComponentsOfType(appName, componentType)
			
			if len(components) > 0:
				for comp in components:
					feed.add_component(comp)
				feed.title = '%s Endpoints' % appName.capitalize()
				self.response.write(feed.toXml())
			else:
				raise splunk.BadRequest('This component type cannot be managed yet')
			
		elif parts == 7:
			'''
			Returns a dictionary of parameters of the specified component
			'''
			appName = self.pathParts[-4]
			if not appbuilder.appExists(appName):
				raise splunk.ResourceNotFound("Application doesn't exist")

			componentType = self.pathParts[-2]
			componentName = self.pathParts[-1]
			
			compInfo = appbuilder.getComponentProps(appName, componentType, componentName)
			
			if len(compInfo.keys()) > 0 :
				return compInfo
			else:
				raise splunk.BadRequest('This component type cannot be managed yet')
		else:
			raise splunk.BadRequest		

		self.response.setHeader('content-type', 'text/xml')		

	def handle_POST(self):
		"""
		Enables / disables an application in response to an HTTP POST.
		"""

		parts = len(self.pathParts)
		
		if parts == 3:
			'''
			Create a new application
			Note: requires 'template' directory under /etc/apps/appbuilder/
			'''
			if not self.args.has_key(HTTP_POST_NAME):
				raise splunk.BadRequest("Application name is missing")		
			
			template = 'default'
			if HTTP_POST_TEMPLATE in self.args:
				template = self.args[HTTP_POST_TEMPLATE].lower().strip()
			
			appName = self.args[HTTP_POST_NAME]
			if appbuilder.appExists(appName):
				raise splunk.RESTException(409, "Application name already exists")

			appbuilder.createApp(appName, template)
			
			msg = _buildResponseMessage('info', "Application '%s' was created" % appName)
		
			self.response.setHeader('Location', '/services/apps/local/%s' % appName)
			self.response.setStatus(201)
			self.response.write(msg)					
			
		elif parts == 4:
			'''
			Enables/disables application
			'''
			if not self.args.has_key(HTTP_ACTION):
				raise splunk.BadRequest
			action = self.args[HTTP_ACTION]			
			
			appName = urllib.unquote(self.pathParts[-1])
			if not appbuilder.appExists(appName):
				raise splunk.ResourceNotFound("Application doesn't exist")
				
			appbuilder.controlApp(appName, action)
			
			self.addMessage("INFO", "%sd application: %s" %
								(action.capitalize(), appName))

		elif parts == 6:
			'''
			Adds a new component of type component_type to application app_name
			'''	
			componentType = self.pathParts[-1]
			
			appName = urllib.unquote(self.pathParts[-3])
			if not appbuilder.appExists(appName):
				logger.error("Application doesn't exist")
				raise splunk.ResourceNotFound("Application doesn't exist")

			if not HTTP_POST_NAME in self.args.keys():
				componentName = 'new_component'
			else:
				componentName = self.args[HTTP_POST_NAME]
			
			try:				
				if not appbuilder.addAppComponent(appName, componentType, componentName):
					raise splunk.BadRequest
					
			except splunk.SplunkdException, e:
				raise splunk.RESTException(409, e)

			msg = _buildResponseMessage('info', "Component '%s' was added to application '%s'" % (componentType, appName))
		
			self.response.setHeader('Location', '/services/apps/local/%s/components/%s/%s' % (appName, componentType, componentName))
			self.response.setStatus(201)
			self.response.write(msg)					
			
		elif parts == 7:
			'''
			Edit a component
			'''		
			pass 
			
		else:
			raise splunk.BadRequest			
			

	def handle_DELETE(self):
		
		parts = len(self.pathParts) 

		if parts == 4:
			'''
			Delete an application in response to an HTTP DELETE.
			'''
			appName = urllib.unquote(self.pathParts[-1])
			
			try:
				appbuilder.deleteApp(appName)
				
				self.addMessage("INFO", "Deleted application: %s" % appName)
			except Exception, e:
				raise splunk.InternalServerError(e)
				
				
		elif parts == 7:
			'''
			Deletes a component
			'''
			pass
			
		else:
			raise splunk.BadRequest

	def self_to_bundle(self):
		name = urllib.unquote(self.pathParts[-1])
		return bundle_paths.get_bundle_by_prettyname(name)



class Filter:

	"""
	Filter an <interable> of bundles based on a dict of HTTP GET arguments.
	"""

	def __init__(self, args):
		self.q = None
		if args.has_key(HTTP_GET_QUERY):
			self.q = args[HTTP_GET_QUERY].lower()

		self.offset = 0
		if args.has_key(HTTP_GET_OFFSET):
			self.offset = int(args[HTTP_GET_OFFSET])

		self.count = 0
		if args.has_key(HTTP_GET_COUNT):
			self.count = int(args[HTTP_GET_COUNT])

		self.sortby = "name"
		if args.has_key(HTTP_GET_SORTBY):
			sortby = args[HTTP_GET_SORTBY]
			if (sortby == "name") or (sortby == "updatedTime"):
				self.sortby = sortby
			else:
				raise splunk.BadRequest("Invalid value for %s: %s" %
										(HTTP_GET_SORTBY, sortby))

		self.reverse = False
		if args.has_key(HTTP_GET_SORTDIR):
			asc = args[HTTP_GET_SORTDIR]
			if asc == "asc":
				self.reverse = False
			elif asc == "desc":
				self.reverse = True
			else:
				raise splunk.BadRequest("Invalid value for %s: %s" %
										(HTTP_GET_SORTDIR, asc))

	def filter(self, iterable):
		"""
		Apply the entire filter.
		"""
		results = [ ]
		for b in self.match(iterable):
			results.append(b)
		results.sort(self.cmp)
		return self.chop(results)

	def match(self, iterable):
		"""
		Match against query.
		"""
		for b in iterable:
			if self._match_q(b.name()) or self._match_q(b.description()):
				yield b

	def _match_q(self, txt):
		if (self.q is None):
			return True
		return (txt.lower().find(self.q) != -1)

	def cmp(self, x, y):
		"""
		Apply sort_by and sort_dir.
		"""
		val = 0
		if self.sortby == "name":
			val = cmp(x.name().lower(), y.name().lower())
		elif self.sortby == "updatedTime":
			val = cmp(x.ctime(), y.ctime())
		if self.reverse:
			val *= -1
		return val

	def chop(self, list):
		"""
		Apply offset and count.
		"""
		end = self.count + self.offset
		if (self.count > 0) and (end < len(list)):
			return list[self.offset:end]
		else:
			return list[self.offset:]



class AtomFeed(format.AtomFeed):

	"""
	Generate an AtomFeed of information about locally-installed applications.
	"""

	def __init__(self, relpath):
		format.AtomFeed.__init__(self)
		self._hostpath = splunk.mergeHostPath()
		self._relpath = relpath
		self.id = self._hostpath + self._relpath
		self.title = self._relpath.split('/')[-1]

	def bundle_to_entry(self, b, inherit_feed_id):
		if inherit_feed_id:
			id = self.id
		else:
			id = "%s/%s" % (self.id, urllib.quote(b.prettyname()))
		return b.to_atom(id)

	def add_bundle(self, b, inherit_feed_id):
		self.entries.append(self.bundle_to_entry(b, inherit_feed_id))
		
	def add_component(self, component):
		id = "%s/%s" % (self.id, component)
		entry = format.AtomEntry(id = id,
								title = component,
								updated = format.strftime(time.localtime())
								)
		self.entries.append(entry)



################################################################################
#                        Remote Application Management                         #
################################################################################



class RemoteAppsHandlerList(splunk.rest.BaseRestHandler):

	"""
	Generate links to remote applications and categories.
	"""

	def handle_GET(self):
		links = { }
		links["entries"] = "Remote Applications"
		links["categories"] = "Remote Application Categories"
		return links



class RemoteAppsSetup(splunk.rest.BaseRestHandler):

	"""
	Prepare remote applications management based on configuration settings.
	"""

	def __init__(self, method, requestInfo, responseInfo, sessionKey):
		splunk.rest.BaseRestHandler.__init__(self,
											method,
											requestInfo,
											responseInfo,
											sessionKey)
		# Default values
		self._login = LOGIN_URL
		self._base = DEFAULT_URL
		self._agent = None
		try:
			platform_info = platform.platform()
			with open(os.path.join(bundle_paths.etc(), "splunk.version")) as f:
				for i in f:
					if i.startswith("VERSION"):
						version = i.split("=")[1].strip().strip('"')
					elif i.startswith("BUILD"):
						build = i.split("=")[1].strip()
			self._agent = "%s-%s-%s" % (version, build, platform_info)
		except Exception, e:
			logger.exception(e)
		# Manual overrides in server.conf
		try:
			conf = bundle.getConf("server", self.sessionKey)
			s = conf["applicationsManagement"]
			if not s.isDisabled():
				if s.has_key("loginUrl"):
					self._login = s["loginUrl"]
				if s.has_key("url"):
					self._base = s["url"]
				if s.has_key("useragent"):
					self._agent = s["useragent"]
		except:
			pass
		logger.debug("applicationsManagement.loginUrl = %s" % self._login)
		logger.debug("applicationsManagement.url = %s" % self._base)
		logger.debug("applicationsManagement.useragent = %s" % self._agent)



class RemoteAppsLogin(RemoteAppsSetup):

	"""
	Handle login to remote applications provider.
	"""

	def handle_POST(self):
		try:
			post_args = urllib.urlencode(self.request["form"])
			logger.debug("Logging into %s" % self._login)
			# Forward post arguments, including username and password.
			with closing(urllib2.urlopen(self._login, post_args)) as f:
				root = etree.parse(f).getroot()
				token = root.xpath("a:id", namespaces=NSMAP)[0].text
				# Generate response.
				response = etree.Element("response")
				sessionKey = etree.SubElement(response, "sessionKey")
				sessionKey.text = token
				self.response.setHeader('content-type', 'text/xml')
				self.response.write(etree.tostring(response, pretty_print=True))
				logger.debug("Login successful")
		except urllib2.HTTPError, e:
			logger.exception(e)
			raise splunk.RESTException(e.code, e.msg)
		except Exception, e:
			logger.exception(e)
			raise splunk.AuthenticationFailed



class RemoteAppsManager(RemoteAppsSetup):

	"""
	Interact with remote applications provider.
	"""

	# /services/apps/remote/ -> "base depth" of 3 URL parts
	BASE_DEPTH = 3

	def handle_GET(self):
		"""
		Respond to an HTTP GET with information about remote applications.
		"""
		try:
			url = self._native_to_foreign_url()
			# Create a new ElementTree based on the root element of the remote
			# feed. This strips things like processing instructions and the XML
			# manifest.
			tree = etree.ElementTree(self._get_feed_root(url))
			root = tree.getroot()
			self._transform_feed(root)
			for h in format.getAtomStyleNodes():
				root.addprevious(h)
			str = etree.tostring(tree,
								xml_declaration=True,
								encoding='UTF-8',
								pretty_print=True)
			self.response.setHeader('content-type', 'text/xml')
			self.response.write(str.decode('UTF-8'))
		except Exception, e:
			if len(self.pathParts) == self.BASE_DEPTH:
				# If we're handling the base endpoint, suppress exceptions.
				# They probably aren't caused by user error here.
				logger.exception(e)
				# We can't get to the remote source, or it isn't providing
				# parseable XML back to us. Display no remote apps.
				return { }
			else:
				# Otherwise, allow exceptions to reach user, as they are likely
				# caused by a bad URL or application name.
				raise

	def handle_POST(self):
		"""
		Install a remote application in response to an HTTP POST.
		"""
		parts = len(self.pathParts)
		if parts == self.BASE_DEPTH + 2:
			default_version = True
		elif parts == self.BASE_DEPTH + 3:
			default_version = False
		else:
			raise splunk.BadRequest
		if not self.args.has_key(HTTP_AUTH_TOKEN):
			raise splunk.BadRequest("Missing argument: %s" % HTTP_AUTH_TOKEN)
		if not self.args.has_key(HTTP_ACTION):
			raise splunk.BadRequest("Missing argument: %s" % HTTP_ACTION)
		if self.args[HTTP_ACTION] != HTTP_ACTION_INSTALL:
			raise splunk.BadRequest("Invalid value '%s' for argument '%s'" %
									(self.args[HTTP_ACTION], HTTP_ACTION))
		url = self._native_to_foreign_url()
		root = self._get_feed_root(url)
		if default_version:
			root = self._get_latest_version(root)
		href = self._parse_link(root)
		try:
			# Package up a Request with auth information.
			req = urllib2.Request(href)
			# XXX: Converting the auth token from a POST arg to a header
			# requires us to unquote() it. If the client did not correctly
			# quote() the token, login will fail.
			req.add_header(HTTP_AUTH_HEADER,
						urllib.unquote(self.args[HTTP_AUTH_TOKEN]))
			# Install using this Request object.
			installer = bundle_paths.BundleInstaller()
			b, status = installer.install_from_url(req)
			self.response.setStatus(status)
			if ((status == bundle_paths.BundleInstaller.STATUS_INSTALLED) or
				(status == bundle_paths.BundleInstaller.STATUS_UPGRADED)):
				# Migrate old-style bundles.
				logger.debug("Configuring application contents")
				try:
					b.migrate()
				except Exception, e:
					logger.exception(e)
					self.addMessage("WARN",
									"Error during configuration: %s" % e)
				# Save application metadata.
				logger.debug("Saving application metadata")
				try:
					if default_version:
						metadata_url = url
					else:
						# Strip version number from URL.
						metadata_url = url.strip("/")
						index = metadata_url.rfind("/")
						metadata_url = metadata_url[:index]
					feed = self._get_feed_root(metadata_url)
					entry = self._get_latest_version_entry(feed)
					dict = self._convert_content(entry)
					self._get_from_xml(feed, "a:title", dict,
									bundle_paths.BundleMetadata.PRETTYNAME)
					self._get_from_xml(entry, "a:title", dict, "version")
					self._get_from_xml(feed, "a:author/a:name", dict, "author")
					try:
						del dict["islatest"]
					except:
						pass
					b.set_postinstall_metadata(dict)
				except Exception, e:
					logger.exception(e)
					self.addMessage("WARN",
						"Error writing extra application information: %s" % e)
				# Redirect to local application.
				self.response.setHeader("Location", self._redirect_to_local(b))
			if status == bundle_paths.BundleInstaller.STATUS_INSTALLED:
				self.addMessage("INFO", "Installed application: %s" % b.name())
			elif status == bundle_paths.BundleInstaller.STATUS_UPGRADED:
				self.addMessage("INFO", "Upgraded application: %s" % b.name())
			else:
				self.addMessage("WARN",
								"Could not install application: %s" % b.name())
		except splunk.ResourceNotFound:
			raise
		except splunk.InternalServerError:
			raise
		except Exception, e:
			logger.exception(e)
			raise splunk.InternalServerError(e)

	def _native_to_foreign_url(self):
		"""
		Convert this endpoint's URL into a remote-provider URL.
		"""
		url = self._base
		for part in self.pathParts[self.BASE_DEPTH:]:
			url += "/" + part
		return url

	def _foreign_to_native_url(self, url):
		"""
		Convert a remote-provider URL into a URL pointing to this endpoint.
		"""
		if not url.startswith(self._base):
			return url
		converted_base = splunk.mergeHostPath()
		for part in self.pathParts[:self.BASE_DEPTH]:
			converted_base += '/' + part
		return converted_base + url[len(self._base):]

	def _get_feed_root(self, url):
		"""
		Get an Atom feed of application information from the remote provider.
		"""
		try:
			target_url = url
			# Forward GET arguments, and add user-agent.
			args_dict = { }
			args_dict.update(self.request["query"])
			if self._agent:
				args_dict["agent"] = self._agent
			args = urllib.urlencode(args_dict)
			if args != "":
				target_url += ("?" + args)
			logger.debug("Getting feed from: %s" % target_url)
			f = urllib2.urlopen(target_url)
		except urllib2.HTTPError, e:
			raise splunk.RESTException(e.code, e.msg)
		except urllib2.URLError, e:
			raise splunk.RESTException(400, "Timed out connecting to SplunkBase")   
		except Exception, e:
			raise splunk.RESTException(404, "Resource not found")
		try:
			root = etree.parse(f).getroot()
			f.close()
			return root
		except Exception, e:
			logger.exception(e)
			raise splunk.InternalServerError(e)

	# XXX: This is very brittle. Assumptions:
	# - The local apps endpoint is named 'local'.
	# - If the remote apps endpoint is at: /services/apps/remote,
	#   then the local apps endpoint is at: /services/apps/local
	# - Local apps are located at /services/apps/local/<app_name>
	def _redirect_to_local(self, b):
		url = splunk.mergeHostPath()
		for part in self.pathParts[:(self.BASE_DEPTH - 1)]:
			url += '/' + part
		url += '/' + 'local'
		url += '/' + urllib.quote(b.prettyname())
		return url

	def _transform_feed(self, xml):
		"""
		Make an Atom feed from the remote provider look as though we generated
		it by rewriting certain URLs in the feed.
		"""
		try:
			self._convert_remote_elements(xml)
			for entry in xml.xpath("//a:entry", namespaces=NSMAP):
				self._convert_remote_elements(entry)
		except Exception, e:
			logger.exception(e)
			raise splunk.InternalServerError(e)

	def _convert_remote_elements(self, root):
		"""
		Make URLs that point to the remote provider point to us instead.
		"""
		# Rewrite URL in ID.
		for id in root.xpath("a:id", namespaces=NSMAP):
			id.text = self._foreign_to_native_url(id.text)
		# Rewrite hrefs in links.
		for link in root.xpath("a:link", namespaces=NSMAP):
			href = link.get("href")
			rel = link.get("rel")
			# Leave download links alone.
			if href and (rel != "download"):
				link.set("href", self._foreign_to_native_url(href))

	def _get_latest_version(self, xml):
		"""
		Given a feed of version entries for an application, get the feed for
		the latest version.
		"""
		try:
			entry = self._get_latest_version_entry(xml)
			href = entry.xpath("a:link/@href", namespaces=NSMAP)[0]
			return self._get_feed_root(href)
		except Exception, e:
			logger.exception(e)
			msg = "Could not find latest version of application"
			raise splunk.ResourceNotFound(msg)

	def _get_latest_version_entry(self, xml):
		"""
		Given a feed of version entries for an application, get the entry for
		the latest version.
		"""
		for entry in xml.xpath("//a:entry", namespaces=NSMAP):
			try:
				contents = self._convert_content(entry)
				if contents["islatest"] == "True":
					return entry
			except:
				pass
		return None

	def _parse_link(self, xml):
		"""
		Given a feed of application files, get the URL of the installer.
		"""
		msg = "Could not find application download location"
		try:
			for entry in xml.xpath("//a:entry", namespaces=NSMAP):
				try:
					contents = self._convert_content(entry)
					if contents["fileclass"] in ["bundle", "other"]:
						return entry.xpath("a:link/@href", namespaces=NSMAP)[0]
				except:
					pass
			raise splunk.ResourceNotFound(msg)
		except Exception, e:
			logger.exception(e)
			raise splunk.ResourceNotFound(msg)

	def _convert_content(self, xml):
		"""
		Convert an Atom entry's <content> node into a Python datastructure.
		"""
		try:
			return format.nodeToPrimitive(xml.xpath("a:content", namespaces=NSMAP)[0][0])
		except:
			return None

	def _get_from_xml(self, xml, location_path, contents, key):
		"""
		Set the value for contents[key] to the text at location_path in xml.
		"""
		try:
			contents[key] = xml.xpath(location_path, namespaces=NSMAP)[0].text
		except Exception, e:
			logger.exception(e)
			pass
