import splunk
import splunk.clilib.bundle_paths as bundle_paths
import splunk.clilib.cli_common as comm
import splunk.util as util, traceback

import os
import time
import urllib2
import string
import tarfile
import sys


APPS_PATH = os.path.join(comm.splunk_home, 'etc', 'apps')
PACKAGE_PATH = os.path.join(bundle_paths.get_system_bundle_path(), 'static', 'app-packages')
TEMPLATES_PATH = os.path.join(comm.splunk_home, 'share', 'splunk', 'appbuilder', 'templates')

TEXT_EXTENSIONS = ['txt', 'html', 'htm', 'xhtml', 'css', 'py', 'pl', 'ps1', 'bat', 'sh', 'conf', 'js', 'xml', 'xsl', 'conf', 'meta']
TXT_PREFIX = '__'

DEFAULT_APP_INFO = {}

'''
Needed to prevent collisions between mako and appbuilder templates
'''
class SafeTemplate(string.Template):
    delimiter = '$$'

'''
Returns Splunkd uri in form of 'hostname:port'
'''
def _getSplunkdUri():
    return comm.getMgmtUri()
    
'''
Returns Splunkweb uri in form of 'proto://hostname:port'
'''
def _getSplunkWebUri():
    return comm.getWebUri()

''' 
Checks if the dir exists and creates it if doesn't
Returns dir path or None
'''
def _getPackageDir():
    if bundle_paths.maybe_makedirs(PACKAGE_PATH):
        return PACKAGE_PATH
    return None
    
'''
Returns local path of an application 
'''
def _getAppPath(appName):
    return os.path.join(APPS_PATH, appName)

'''
Returns a string 'version-build-product-platform' of the current Splunk instance
'''
def _getSplunkVersion():
    try:
        f = open(os.path.join(bundle_paths.etc(), "splunk.version"))
        for i in f:
            if i.startswith("VERSION"):
                version = i.split("=")[1].strip().strip()
            elif i.startswith("BUILD"):
                build = i.split("=")[1].strip()
            elif i.startswith("PRODUCT"):
                product = i.split("=")[1].strip()
            elif i.startswith("PLATFORM"):
                platform = i.split("=")[1].strip()
                
        return "%s %s %s %s" % (version, build, product, platform)
    except Exception, e:
        raise e
    

def _getDefaultAppInfo():
    DEFAULT_APP_INFO = {
        'version': '1.0', 
        'creator' : 'Splunk', 
        'description' : 'Splunk apps FTW',
        'license' : 'New BSD License',
        'copyright' : '2009'
        }	

    DEFAULT_APP_INFO['compatible'] = _getSplunkVersion()
    DEFAULT_APP_INFO['created'] = util.getISOTime(time.localtime())
    return DEFAULT_APP_INFO

'''
We assume files that have one of TEXT_EXTENSIONS or whose name starts with TXT_PREFIX to be templates
Returns a tuple: (filename, isItText)
'''
def _isTextFile(fn):
    if fn.startswith(TXT_PREFIX):
        fn = fn[len(TXT_PREFIX):]
        return (fn, True)
    ext = os.path.splitext(fn)[1][1:]
    return (fn, True) if ext in TEXT_EXTENSIONS else (fn, False)

'''
Insert parameter values into each template file and save them under the selected directory
chosenFiles - list of template files to be used
template_params - dictionary of parameters
'''	
def _populate(dstPath, template, templateParams):

    bundle_paths.maybe_makedirs(dstPath)
    os.chdir(dstPath)	
    
    templatePath = os.path.join(TEMPLATES_PATH, template)
    
    if not os.path.exists(templatePath):
        raise splunk.InternalServerError("Template not found")	
    
    for root, dirs, files in os.walk(templatePath):
        # determine relative path  
        relPath = root[len(templatePath)+1:]

        # create subdirs
        for dir in dirs:
            bundle_paths.maybe_makedirs(os.path.join(relPath,dir))

        # Read template files and apply param values then save
        for fn in files:

            try:
                # use params to create custom file names
                inFilePath = os.path.join(root, fn)
                # filter by file type
                fn, isText = _isTextFile(fn)
                outFilePath = os.path.join(dstPath, relPath, fn)
                if not isText:
                    comm.copyItem(inFilePath, outFilePath)
                    continue

                f = open(inFilePath, 'r')
                content = f.read()
                f.close()

                content = SafeTemplate(content).substitute(templateParams)

                f = open(outFilePath, 'w')
                f.write(content)
                f.close()
            except:
                print traceback.print_exc(file=sys.stderr)
                pass

    return True


'''
This needs to handle this a bit better but to support adding static assets by upload this 
method will copy in file that have been donwloaded: todo: find a way to call this method 
with the cgi fs instaed of saveing to a temp dir since its not thread save
'''
def addUploadAssets(appName):
    appPath = _getAppPath(_toCamelCase(appName))

    if not os.path.exists(appPath):
        raise splunk.InternalServerError("app does not exist")

    tempPath = os.path.join(os.environ['SPLUNK_HOME'], 'var', 'run', 'splunk', 'apptemp')
    # if does not exist then it means no assets exist for moving
    if not os.path.exists(tempPath):
        return

    dstPath = os.path.join(appPath, 'appserver', 'static')
    bundle_paths.maybe_makedirs(dstPath)
    comm.mergeDirs(tempPath, dstPath)

    # clean up
    bundle_paths.safe_remove(tempPath)

    
''' 
Creates a tar archive of the app
Returns tar file path
'''
def _package( appName, merge=True ):
    packageDir = _getPackageDir()
    tarFile = "%s.tar.gz" % appName
    tarPath = os.path.join(packageDir, tarFile)
    z = tarfile.open(tarPath, 'w')
    
    if merge:
        appPath = mergeApp(appName)
    else:
        appPath = _getAppPath(appName)
    
    # walk through files in directory and package them up
    def _tarWalker( tar, appPath, files, root=appPath ):
        for file in files:
            file = os.path.join( appPath, file )
            archiveName = os.path.join(appName, file[len(os.path.commonprefix( (root, file) ))+1:]) 
            tar.add( file, archiveName, False )
                
    os.path.walk( appPath , _tarWalker, z  )

    z.close()
    
    # cleanup tmp dir
    if merge:
        bundle_paths.safe_remove(appPath)
    
    return _getSplunkdUri() + "/static/app-packages/" + tarFile

    
'''
Checks if the manifest file exists in the app 
If not, it creates a new one and populates with default values
Returns path to the manifest file
'''
def _setManifest( appName ):
    manifest = {}

    # read manifest file if exists or create a new one
    appconfPath = os.path.join(_getAppPath(appName), 'default', 'app.conf')
    defAppInfo = _getDefaultAppInfo()
    
    info = comm.readConfFile(appconfPath)

    if not info:
        manifest[appName] = defAppInfo
    else:
        for key in info:
            manifest[key] = info[key]    

        newinfo = info[appName] if appName in info else {}

        for key in defAppInfo:
            if key not in newinfo:
                newinfo[key] = defAppInfo[key]

        ver = newinfo['version'].split('.')
        newinfo['version'] = "%s.%s" % (int(ver[0])+1,ver[1])
        newinfo['last-update'] = util.getISOTime(time.localtime())

        manifest[appName] = newinfo

    filesPath = os.path.join(_getAppPath(appName), 'FILES-%s' % manifest[appName]['version'])            
    f = file(filesPath, 'w')
    f.writelines( ['%s, %s\n' % (entry[0], entry[1]) for entry in _getAppContents(appName)] )
    f.close()
    
    comm.writeConfFile(appconfPath , manifest)


'''
Returns a list of tuples of filenames and their last modification time 
'''	
def _getAppContents( appName ):
    fileNames = []

    appPath = _getAppPath(appName)
    for root, dirs, files in os.walk(appPath):
        relPath = root[len(appPath)+1:]
        for file in files:	
            fileNames.append( (os.path.join(relPath,file), util.getISOTime(time.localtime(os.path.getmtime(os.path.join(root,file))))) )

    return fileNames

'''
Return the members of a package as a list of their names.
'''
def _getPackageNames( pkgPath ):
    if not os.path.exists(pkgPath):
        return []
    t = tarfile.TarFile(pkgPath)
    return t.getnames()
    
  
'''
Converts a space delimited string to a CamelCase string
'''
def _toCamelCase(str):
    strings = str.split()
    if len(strings) == 1:
        return str
        
    result = ''
    for string in strings:
        result += string[0].upper() + string[1:]
    return result
    
'''
--------------------------------------------------------------
Public methods
--------------------------------------------------------------
'''

'''
Returns true if app with such name exists, false otherwise
'''
def appExists(appName):
    return appName in [name for name in os.listdir(APPS_PATH)]

'''
Returns an iterator of all locally installed apps. Filters results if corresponding arguments provided
'''
def appInfoIterator():
    for appName in [f for f in os.listdir(APPS_PATH) if os.path.isdir(os.path.join(APPS_PATH, f))]:
        label = ''
        enabled = False
        visible = False
        
        try:
            appconfPath  = os.path.join( _getAppPath(appName), 'default', 'app.conf' )
            conf = comm.readConfFile(appconfPath)
            label = conf['ui']['label']
            visible = conf['ui']['is_visible']
        except:
            pass
        
        try:
            enabled = isAppEnabled(appName)
        except:
            pass
       
        yield { 
                'name' : appName, 
                'label' : label, 
                'enabled' : str(enabled),
                'visible' : str(visible).lower().capitalize()
              }

'''
Returns a list of available app templates
'''
def getTemplates():
    return [f.lower() for f in os.listdir(TEMPLATES_PATH) if os.path.isdir(os.path.join(TEMPLATES_PATH, f))]
        
        
'''
Creates skeleton application from a template
Returns url to the new app
'''
def createApp(appName, appLabel, visible, manageable, template):
    # Generate app directory
    appPath = _getAppPath(_toCamelCase(appName))
    
    if os.path.exists(appPath):
        raise splunk.RESTException(409, "Application %s already exists" % appName)
        
    if appLabel in ['', None]:
        appLabel = appName
    
    templateParams = {
            'appLabel' : appLabel, 
            'appName' : appName,
            'visible' : visible,
            'manageable' : manageable
            }

    if template not in getTemplates():
        raise splunk.BadRequest("Template %s doesn't exist" % template)

    # Generate files for the app 			
    _populate(appPath, template, templateParams)

    return _getSplunkWebUri() + '/app/' + _toCamelCase(appName)
        
'''
Installs an app from the location which could be a url or local path string
'''
def installApp(location):
    installer = bundle_paths.BundleInstaller()
    
    location = location.strip().lower()
    
    if location.startswith('http'):
        req = urllib2.Request(url=location)
        installer.install_from_url(req)
    else:
        if not os.path.exists(location):
            raise splunk.ResourceNotFound('The package wasn\'t found')
        if os.path.isdir(location):
            raise splunk.BadRequest('Path is a directory')
        if appExists(_getPackageNames(location)[0].split('/')[0]):
            raise splunk.RESTException(409, "Application name already exists")

        installer.install_from_tar(location)


'''
Packages the appName application, providing manifest and returning the url of the package
By default merges contents of local and default directories with higher precedence of local.
'''
def packageApp(appName, needsMerging=True):
    _setManifest(appName)
    pkgPath = _package(appName, needsMerging)

    return pkgPath


'''
Merges local and default parts of an app into default 
Returns the path to the merged app.
'''
def mergeApp(appName):
    appPath = _getAppPath(appName)
    tmpPath = os.path.join(PACKAGE_PATH, 'DELETEME_' + appName)
    
    # this should copy app dir to tmp dir
    bundle_paths.maybe_makedirs(tmpPath)
    comm.mergeDirs(appPath, tmpPath)

    localPath = os.path.join(tmpPath, 'local')
    defaultPath = os.path.join(tmpPath, 'default')
    
    # check if the app is allowed to be merged

    if os.path.exists(localPath) and os.path.exists(defaultPath):
        # merge local and default dirs in tmp, result in local
        b = bundle_paths.Bundle('dummy', 'bundle')
        comm.mergeDirs(defaultPath, localPath, False, b._merger)

        # remove default
        bundle_paths.safe_remove(defaultPath)

        # move local to default
        comm.moveItem(localPath, defaultPath)
    
    return tmpPath
    
