import cherrypy
from splunk.appserver.mrsparkle import *
import os, os.path, copy, cgi
import splunk
import copy, pprint
import splunk.entity as en
import splunk.rest as rest
import logging
import urllib
import formencode
import json
import splunk.appserver.mrsparkle.lib.module as module
import lxml.etree as et
from splunk.appserver.mrsparkle.lib.eai import DynamicUIHelper, cpUnquoteEntity, cpQuoteEntity
from splunk.appserver.mrsparkle.lib import util, i18n, message, cached
from splunk.util import parseISO
from splunk.util import normalizeBoolean
from splunk.util import safeURLQuote
import splunk.input
from splunk import auth
from formencode import validators
from splunk.appserver.mrsparkle.lib.msg_pool import MsgPoolMgr, UI_MSG_POOL
import re
import splunk.clilib.bundle_paths as bundle_paths
import urlparse

logger = logging.getLogger('splunk.appserver.controllers.admin')

# permissions constants
NOACCESS = 0
READONLY = 1
READWRITE = 2

DEFAULT_CANCEL_URL = '/manager/index'

# View that holds the navigation modules
ADMIN_VIEW_NAME = '_admin'

CONTROL_ACTIONS = {'enable': 'Enabled',
                   'disable': 'Disabled',
                   'remove': 'Deleted'}

class SplunkbaseException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)                      
                   
def _unpackOnChange(onChange):
    onChange_dict = {}
    for k in onChange:
        if k.tag == 'group_set':
            onChange_dict['_groupset'] = [group.attrib['name'] for group in k]
        elif k.tag == 'value_map':
            onChange_dict['value_map'] = dict([(key.attrib['name'], key.text if key.text else '') for key in k])
        elif k.tag == 'chained_action':
            onChange_dict['_chained_action'] = _unpackOnChange(k)
        else:
            onChange_dict[k.attrib['name']] = k.text

    return onChange_dict

    
def _unpackOpt(opt):
    type = opt.attrib.get('type')
    if type:
        if type == 'dict':
            return_dict = {}
            for o in opt:
                return_dict[o.attrib['name']] = _unpackOpt(o)
            return return_dict
        elif type  == 'list':
            return_list = []
            for item in opt:
                return_list.append(_unpackOpt(item))
            return return_list

    return opt.text

def _unpackElement(element):
    element_dict = {}
    if 'name' in element.attrib:
        element_dict['elementName'] = element.attrib['name']

    for attr in ('type', 'label'):
        if not attr in element.attrib:
            continue

        element_dict[attr] = element.attrib[attr]

    onChanges = element.findall('onChange')
    if len(onChanges) == 1:
        element_dict['onChange'] = _unpackOnChange(onChanges[0])
    elif len(onChanges) > 1:
        element_dict['onChange'] = [_unpackOnChange(onChange) for onChange in onChanges]
        

    for o_type in ('options', 'extraOptions'):
        o_node = element.find(o_type)

        if not o_node:
            continue

        element_dict[o_type] = [{'value': opt.attrib['value'], 'label': _(opt.attrib['label'])} for opt in o_node]

    views = element.findall('view')
    if views:
        element_dict['showInView'] = [view.attrib['name'] for view in views]

    opts = element.findall('key')
    for opt in opts:
        element_dict[opt.attrib['name']] = _unpackOpt(opt)
            
    elements = element.find('elements')
    if elements:
        element_dict['elements'] = [_unpackElement(e) for e in elements]

    # type=repeatable uses this
    subelement = element.find('element')
    if subelement:
        element_dict['element'] = _unpackElement(subelement)

        
    return element_dict

def _unpack(root_str):
    parser = et.XMLParser(remove_blank_text=True, remove_comments=True)
    root = et.XML(root_str, parser)

    if root.tag != 'endpoint':
        return None

    contents = {}

    contents['redirect'] = normalizeBoolean(root.attrib['redirect']) if 'redirect' in root.attrib else True
    if 'hideEnabledColumn' in root.attrib:
        logger.debug("should hide enabled column.")
        contents['hideEnabledColumn'] = normalizeBoolean(root.attrib['hideEnabledColumn'])
    if 'hidePermissionsColumn' in root.attrib:
        logger.debug("should hide permissions column.")
        contents['hidePermissionsColumn'] = normalizeBoolean(root.attrib['hidePermissionsColumn'])
    if 'hideActionsColumn' in root.attrib:
        logger.debug("should hide actions column.")
        contents['hideActionsColumn'] = normalizeBoolean(root.attrib['hideActionsColumn'])

    contents['header'] = root.findtext('header')

    breadcrumb = root.find('breadcrumb')
    if breadcrumb:
        breadcrumb_parent_node = breadcrumb.find('parent')
        if breadcrumb_parent_node is not None:
            contents['breadcrumb_parent'] = breadcrumb_parent_node.text
            contents['breadcrumb_hide_current'] = normalizeBoolean(breadcrumb_parent_node.attrib['hidecurrent']) if 'hidecurrent' in breadcrumb_parent_node.attrib else False
        breadcrumb_noentity = breadcrumb.findtext('noentity')
        contents['breadcrumb_noentity'] = normalizeBoolean(breadcrumb_noentity) if breadcrumb_noentity else False
        contents['breadcrumb_name'] = breadcrumb.findtext('name')
        contents['breadcrumb_entityname'] = breadcrumb.findtext('entityname')

    if 'showAppContext' in root.attrib:
        contents['showAppContext'] = root.attrib['showAppContext']

    menu = root.find('menu')
    if menu:
        menu_dict = {}
        if 'name' in menu.attrib:
            menu_dict['name'] = menu.attrib['name']
        for menu_opt in ('label', 'url', 'description'):
            menu_opt_node = menu.find(menu_opt)
            if menu_opt_node != None:
                menu_dict[menu_opt] = menu_opt_node.text

        order_node = menu.find('order')
        menu_dict['order'] = 0
        if order_node != None:
            try:
                menu_dict['order'] = int(order_node.text)
            except:
                pass

        link_nodes = menu.findall('link')
        for link in link_nodes:
            link_dict = {}
            for opt in link.findall('key'):
                link_dict[opt.attrib['name']] = _unpackOpt(opt)
            menu_dict.setdefault('links', []).append(link_dict)

        contents['menu'] = menu_dict


    elements = root.find('elements')

    if elements:
        contents['elements'] = [_unpackElement(element) for element in elements]
    
    return (root.attrib['name'], contents)


def _extractPathFromURI(uri):
    '''Takes a string that should look like a uri, parses out any scheme and returns just the path and arguments.'''
    if uri == None:
        return uri
    elif isinstance(uri, basestring):
        parsed_uri = urlparse.urlsplit(uri)
        if parsed_uri:
            parsed_uri = list(parsed_uri)
            parsed_uri[0] = ''
            parsed_uri[1] = ''
            return urlparse.urlunsplit(parsed_uri)
    return None


class AdminController(BaseController):
    """
    /admin
    Handle entities stuff... TK mgn 06/19/09
    """

    def __init__(self):
        super(AdminController, self).__init__()
        logger.debug("INIT.")

    def generateBreadcrumbs(self, namespace, endpoint_base=None, entity_name=None, final=None):
        """
        Generate the list of breadcrumb urls used by the layout template

        This is generally [root] -> [endpoint1] -> [endpoint2] -> [entity]

        Endpoints represent the url for the entity list/summary view.

        There's only two endpoints if endpoint2 supplies a parent endpoint in it's breadcrumb
        definition in uiHelper.

        The uiHelper config can optionally specify a few parameters to affect the breadcrumb definition:
        <breadcrumb>

            <!-- specify that a parent breadcrumb should be inserted and optionally don't display
                 a breadcrumb for the current endpoint -->
            <parent hidecurrent="True">parent_endpoint</parent>

            <!-- override the name used for the endpoint - By default it will use the menu label
                 or the value specified in the header node -->
            <name>List Entities</name>

            <!-- override the name used for entities contained by this endpoint.
                 By default the entity's own name is used, setting this will make them all the same -->
            <entityname>System Setings</entityname>

            <!-- don't display a breadcrumb at all for the entity view -->
            <noentity>True</noentity>

        </breadcrumb>
        """

        chain = []
        # the final url will be generated via make_url() by the template
        base_url = '/manager/%s/' % namespace

        if endpoint_base:
            # not the manager index page
            if final:
                # permissions editor is one level deeper than the entity itself
                chain.append( (_(final), None) )

            if entity_name:
                # viewing an individual entity
                if entity_name == en.NEW_EAI_ENTITY_NAME:
                    entity_name = _('Add New')
                chain.append( (entity_name, '%s%s/%s?action=edit' % (base_url, endpoint_base, cpQuoteEntity(entity_name))) )

            uiHelpers = self.fetchUIHelpers(namespace=namespace)
            while endpoint_base:
                entry = uiHelpers.get(endpoint_base)
                if not entry:
                    break
                if entry.get('breadcrumb_noentity') and chain:
                    # Remove the entity name from the breadcrumb
                    chain.pop()

                elif entry.get('breadcrumb_entityname') and entity_name:
                    # Override the entity name supplied by the endpoint
                    name = _(entry['breadcrumb_entityname']) # this tends to be a name that needs translating
                    if name[0] == '$': # variable entity name; typicaly $namespace
                        name = locals().get(name[1:], entity_name)
                    chain[0] = (name, None)

                if not entry.get('breadcrumb_hide_current'):
                    if entry.has_key('menu'):
                        # use the menu label by default
                        title = entry['menu']['label']
                        url = entry['menu']['url'] % {'namespace':namespace}

                    elif entry.get('breadcrumb_name'):
                        # else see if there's a breadcrumb override
                        title = entry['breadcrumb_name']
                        url = base_url+endpoint_base

                    elif entry.has_key('header'):
                        # else use the name defined by the header tag
                        title = entry['header']
                        url = base_url+endpoint_base

                    chain.append( (_(title), url) )
                # see if there's a link to a parent endpoint
                endpoint_base = entry.get('breadcrumb_parent')

        chain.append( (_('Manager'), base_url) )

        if chain[0][1]:
            # Last entry should never be linked.
            chain[0] = (chain[0][0], None) 

        # reverse into the correct order
        chain.reverse()
        return chain


    def generateUIHelperFromEntity(self, entity, namespace='search'):
        uiHelper = {}
        uiHelper["header"] = "%s setup" % namespace
        uiHelper["elements"] = []
        setupXml = entity.get("eai:setup", None)
        # logger.debug(setupXml)
        import xml.etree.cElementTree as et
        root = et.fromstring(setupXml)
        block_nodes = root.findall('block')
        
        i = 0
        for block in block_nodes:
            
            fieldsetElement = {
                "elementName"   : "blockFieldset-%s" % i,
                "showInView"    : [ "edit" ],
                "type"          : "fieldset",
                "legend"        : block.get("title", "I am Legend"),
                "elements"      : []
            }
            
            j = 0
            for node in block:
                
                # go through everything in the block and add to this fieldset
                logger.debug(node.tag)
                if node.tag == "text":
                    # for <text> nodes, put them into a type: helpstring
                    element = {
                        "elementName"   : "textNode-%s-%s" % (i, j),
                        "type"          : "helpstring",
                        "helpText"      : node.text,
                        "showInView"    : [ "edit" ]
                    }
                    fieldsetElement["elements"].append(element)
                    
                elif node.tag == "input":
                    # for <input> nodes, put them into the appropriate type per the input
                    type = node.find("type").text
                    if type == "bool":
                        elementType = "checkbox"
                    else:
                        elementType = "textfield"
                    elementLabel = node.find("label").text
                    element = {
                        "elementName"   : node.get("id"),    
                        "type"          : elementType,
                        "label"         : elementLabel,
                        "showInView"    : [ "edit" ]
                    }
                    fieldsetElement["elements"].append(element)
                    
                j = j + 1
            
            uiHelper["elements"].append(fieldsetElement)
            
            i = i + 1
        
        return uiHelper
        
            
    def fetchUIHelpers(self, namespace='search', search=None, refresh=0):
        """
        Helper routine to get the ui helpers from EAI.
        """
        
        try:
            if refresh:
                helper_en = en.getEntities('data/ui/manager', count=-1, namespace=namespace, refresh=1)
            else:
                helper_en = cached.getEntities('data/ui/manager', count=-1, namespace=namespace)
                
        except splunk.ResourceNotFound:
            msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', 'Unable to locate the configuration for this URL.')
            return self.redirect_to_url("/manager/%s" % splunk.getDefault('namespace'), _qs={
                "msgid" : msgid
            } )

        uiHelper = {}

        for e in helper_en:
            (key, val) = _unpack(helper_en[e].get('eai:data'))
            if search:
                if key == search:
                    return val
            else:
                uiHelper[key] = val

        return uiHelper
    
    def getElementPermissions(self, endpoint_path, entity_name, element_name):
        """
        Return the current user's permissions for accessing a given form element
        XXX This obviously needs some work ;-)
        """
        return READWRITE


    def getCancelURL(self):
        """
        Get the cancel URI for a given admin section page.
        Attempts to elegantly retrieve a best fit redirect url for the cancel button.
        """
        # redirect url
        url = None
        # the referer
        referer = cherrypy.request.headers.get('Referer', None)
        # the session stored manager_cancel_redirect or default if None
        session_url = cherrypy.session.get('manager_cancel_url', DEFAULT_CANCEL_URL)
        # no referer use session or default
        if referer is None:
            url = session_url
        # has a referer
        else:
            # this a referal from itself use session or default
            if referer.find(safeURLQuote(cherrypy.request.path_info)) > -1:
                url = session_url
            # this is a referal from the same machine update session cancel url with referer
            elif referer.find(cherrypy.request.base) > -1:
                trim_referer = referer.replace(cherrypy.request.base, '')
                # strip /prefix/en-US from the url so it's not added back by redirect later
                url = cherrypy.session['manager_cancel_url'] = self.strip_url(trim_referer)
            # not a referal from itself or the same machine use default    
            else:
                url = DEFAULT_CANCEL_URL
        return url
        

    def saveUploadFileEntity(self, args, namespace):
        """
        Stream a file upload to Splunk for the data/inputs/monitor endpoint
        """
        fs = args.get('spl-ctrl_remotefile')
        if not (isinstance(fs, cgi.FieldStorage) and fs.file):
            return False, _('No file was uploaded')

        host = args.get('host')
        host_regex = args.get('host_regex')
        host_segment = args.get('host_segment')
        sourcetype = args.get('sourcetype')
        index = args.get('index', 'main')

        try:
            line = fs.file.readline()
            if not line:
                return False, _("File is empty or doesn't exist")
        except Exception, e:
            return False, str(e)


        try:
            try:
                fh = None
                fh = splunk.input.open(hostname=host, host_regex=host_regex, host_segment=host_segment, sourcetype=sourcetype, index=index, source=os.path.basename(fs.filename) if fs.filename else 'webform')
                fh.write(line)
                for line in fs.file:
                    fh.write(line)
            except Exception, e: 
                raise
            finally:
                if fh:
                    fh.close() #Bad things can happen here
        except Exception, e: # Would be good to be smarter here
            return False, str(e)
        else:
            return True, ''



    def saveUploadAppAsset(self, args, namespace):
        """
        Stream a file upload to Splunk for the data/inputs/monitor endpoint
        """
        fs = args.get('spl-ctrl_bg_app')
        if isinstance(fs, cgi.FieldStorage) and fs.file and fs.filename != '':
            self.writeAsset(args, fs, 'bg_app.png', namespace)

        fs = args.get('spl-ctrl_image')
        if isinstance(fs, cgi.FieldStorage) and fs.file and fs.filename != '':
            self.writeAsset(args, fs, '',  namespace)

        fs = args.get('spl-ctrl_application_css')
        if isinstance(fs, cgi.FieldStorage) and fs.file and fs.filename != '':
            self.writeAsset(args, fs, 'application.css', namespace)

        fs = args.get('spl-ctrl_upload1')
        if isinstance(fs, cgi.FieldStorage) and fs.file and fs.filename != '':
            self.writeAsset(args, fs, '', namespace)

        fs = args.get('spl-ctrl_upload2')
        if isinstance(fs, cgi.FieldStorage) and fs.file and fs.filename != '':
            self.writeAsset(args, fs, '', namespace)


    def writeAsset(self, args, fs, filename, namespace):
        try:
            logger.debug('about to copy to file %s ' % fs.filename)
            tempPath = os.path.join(os.environ['SPLUNK_HOME'],'var', 'run','splunk','apptemp')
            if not os.path.exists(tempPath):
                os.makedirs(tempPath)

            if not filename or filename =='':
                filename = fs.filename

            # make sure we have no funny chars
            # in practice protecting for windows filenames is byeond my interest at this time. 
            # mitch showed me this and i got scared http://msdn.microsoft.com/en-us/library/aa365247.aspx
            # some python code to validate window filenames would be nice.
            if filename.find('/')>-1 or filename.find('\\')>-1 or filename.startswith('.'):
                logger.warn('Not going to upload files with funny chars %s' % filename)
                # should probably throw something other than return
                return
            newPath = os.path.join(tempPath, filename)
            newFile = open ( newPath,'wb')
            logger.debug('about to copy across %s to %s ' % (filename, newPath ))
            while 1:
                buf = fs.file.read(1024)
                if buf:
                    newFile.write(buf)
                else:
                    break
            #for line in fs.file:
            #    newFile.write(line)
            newFile.close()

        except Exception, e: # Would be good to be smarter here
            logger.warn('Could not get uploaded file %s' % str(e))
            return False, str(e)

        return True, ''

    def saveUploadLookupTableFile(self, args, namespace):
        """
        Write a lookup file into $SPLUNK_HOME/etc/apps/*/lookups.
        """
        filename = args.get('name')
        if not filename or (len(filename) == 0):
            return False, _('No filename was given')

        fs = args.get('spl-ctrl_lookupfile')
        if not (isinstance(fs, cgi.FieldStorage) and fs.file):
            return False, _('No file was uploaded')

        try:
            line = fs.file.readline()
            if not line:
                return False, _("File is empty or doesn't exist")
            lookupDir = os.path.join(bundle_paths.get_base_path(), namespace, "lookups")
            if not os.path.exists(lookupDir):
                os.makedirs(lookupDir, 0755)
            dst = os.path.join(lookupDir, filename)
            if os.path.exists(dst):
                return False, _('Lookup table file already exists: %s' % dst)
            with open(dst, "w") as out:
                for line in fs.file:
                    out.write(line)
        except Exception, e:
            return False, str(e)

        return True, ''

    def saveEntity(self, endpoint_path, entity_name, args, namespace, entity_owner, entity_uri):
        logger.debug('Saving Entity with %s to namespace=%s, owner=%s endpoint_path=%s entity_name=%s' % (args, namespace, entity_owner, endpoint_path, entity_name) )
        err = ""
        bolSaved = False

        if endpoint_path=='data/inputs/monitor' and entity_name==en.NEW_EAI_ENTITY_NAME and args.get('spl-ctrl_switcher')=='uploadfile':
            return self.saveUploadFileEntity(args, namespace)

        if endpoint_path=='apps/local':
            self.saveUploadAppAsset(args, namespace)
            
        """
        if endpoint_path=='data/lookup-table-files':
            bolSaved, err = self.saveUploadLookupTableFile(args, namespace)
            if not bolSaved:
                return bolSaved, err
        """
        
        #construct the entity to update
        updatedEnt = en.getEntity(endpoint_path,entity_name, namespace=namespace, uri=_extractPathFromURI(entity_uri), sessionKey=cherrypy.session['sessionKey'])
        
        """
        for item in updatedEnt:
            logger.debug("%s : %s" % (item, updatedEnt[item]) )
        """
        updatedEnt.properties = {}
        # set the entity owner to the current user so it will save down to their namespace.
        # currently the editor saves into their namespace, they now own it.  this may need
        # to be expanded into a function when an admin will have the ability to save into
        # someone else's namespace.
        if entity_owner:
            updatedEnt.owner = entity_owner
        else:
            pwnr = auth.getCurrentUser()['name']
            updatedEnt.owner = pwnr
        
        for arg in args:
            # logger.debug("Value from args is %s=%s" % (arg, args[arg]) )
            
            # if ( (not arg.startswith('spl-ctrl_')) and (arg!='splunk_session_id') and (arg!=None) and (arg!='confirm_password') and (args[arg] != "") ):
            if ( (not arg.startswith('spl-ctrl_')) and (arg!='splunk_session_id') and (arg!=None) and (arg!='confirm_password') ):
                # don't send empty password string
                if not (arg=='password' and args[arg]==''):
                    updatedEnt[arg] = args[arg]
                # logger.debug("updatedEnt[%s]=%s" % (arg, updatedEnt[arg]) )
        
        try:
            # bolSaved returns True if successful, False if not.
            logger.debug("updatedEnt %s." % updatedEnt)
            msgObj = {'messages':None}
            bolSaved = en.setEntity(updatedEnt, sessionKey=cherrypy.session['sessionKey'], uri=_extractPathFromURI(entity_uri), msgObj=msgObj)
            logger.debug("msgObj was updated in setEntity: %s" % msgObj)
            if (msgObj['messages'] is not None):
                err = msgObj
                
        except splunk.RESTException, e:
            
            if e.extendedMessages:
                if e.extendedMessages[0].has_key("text"):
                    err = e.extendedMessages[0].get("text")
                    logger.debug("Fail in saveEntity. e.extendedMessages[0].get('text'): %s." % err)
            elif "text" in e.msg[0]:
                err = e.msg[0].get("text")
                logger.debug("Fail in saveEntity. e.msg[0].get('text'): %s." % err)
            else:
                err = e
                logger.debug("Fail in saveEntity. no text in e.msg[0]: %s." % e)
                
        except Exception, e:
            err = e
            logger.exception(e)
            
        return bolSaved, err
    
    def flattenElements(self, elements, result):
        """Build a elementname=>element mapping"""
        for element in elements:
            if element.has_key('elements'):
                if element.has_key('elementName'):
                    result[element['elementName']] = { 'type': 'fieldset' }
                self.flattenElements(element['elements'], result)
            else:
                result[element['elementName']] = element
        return result

    def loadUniqueWidgets(self):
        """
        Find an example of each kind of widget used in uiHelpers, along 
        with the configuration for each
        Returns a dictionary of widget configs keyed on widget type name
        """
        uiHelpers = self.fetchUIHelpers()
        widgets = {}
        for elements in [endpoint['elements'] for endpoint in uiHelpers.values() if endpoint.has_key('elements')]:
            for element in elements:
                widgets.setdefault(element['type'], element)
        return widgets
    
    def mergeUiHelperWithData(self, uiHelper, entity_name):
        pass
    
    def getSingleEntity(self, endpoint_path, uri=None, namespace=splunk.getDefault('namespace'), viewFilter="edit"):
        # return values
        form_defaults = None
        uiHelper = None
        entity_name = None
        fetchStaticHelpers = True
        statusCode = 0
        msgid = None
        
        # note this does not send an owner of None, it just uses the default system user
        # this gets overridden for app setup to use the owner "nobody"
        pwnr = None
        logger.debug("endpoint_path:: %s." % endpoint_path)
        segments = endpoint_path.split("/")
        # chop off the last one to use for the entity name, use the rest for the path.
        entity_name = cpUnquoteEntity(segments.pop())
        endpoint_base = "/".join(segments)
        # logger.debug("endpoint_base: %s." % endpoint_base)

        if entity_name == en.NEW_EAI_ENTITY_NAME:
            if namespace == '-':
                namespace = 'search'
            if pwnr == '-':
                pwnr = None
                
        try:
            entity = en.getEntity(endpoint_base, entity_name, uri=_extractPathFromURI(uri), namespace=namespace, sessionKey=cherrypy.session['sessionKey'])

            # now that we are using the cpUnquoteEntity method to get the entity name from the url we don't need to do this.
            # entity_name = entity.name
        except (splunk.ResourceNotFound, splunk.AuthorizationFailed), e:
            entity = {}
            error = "Fail: %s." % e.msg
            msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', error)
            logger.exception(e)
            return self.redirect_to_url("/manager/%s" % namespace, _qs={
                "msgid" : msgid
            } )
        except (splunk.RESTException), e:
            logger.debug("RESTException.")
            logger.debug(e.statusCode)
            statusCode = e.statusCode
            if (statusCode == 402):
                msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', '%s' % e.msg[0].get('text') )
            form_defaults = {}
            entity = {}
            uiHelper = self.fetchUIHelpers(search=endpoint_base, namespace=namespace)
            
            return uiHelper, form_defaults, entity, statusCode, msgid
        
        form_defaults = entity
        
        if endpoint_base.startswith("apps/local/") and entity_name == "setup":
            fetchStaticHelpers = False
        
        if fetchStaticHelpers:
            uiHelper = self.fetchUIHelpers(search=endpoint_base, namespace=namespace)
        else:
            uiHelper = self.generateUIHelperFromEntity(entity, namespace=namespace)    
    
        if len(uiHelper) == 0:
            logger.error('listProperties - unable to load the section form definition for endpoint=%s' % endpoint_base)
    
        # Merge uihelper and entity data to create dictionary for template use.
        if "elements" in uiHelper:
            # XXX We'll probably want to change this once the roles stuff is figured out
            def filterElements(elements):
                elementsToFilter = []
                elnum = 0
                
                for element in elements:
                    # logger.debug("elementName:: %s." % str(element.get("elementName", None)) )
                    # logger.debug("showInView:: %s." % str(element.get("showInView", [])) )
                    
                    if element.has_key('elementName') and (viewFilter in element.get("showInView", [])):
                        
                        if self.getElementPermissions(endpoint_path, entity_name, element['elementName']) == NOACCESS:
                            elementsToFilter.append(element["elementName"])
                        else:
                            
                            eaiAttributes = entity.get("eai:attributes")
                            if 'disableIfNotInEAIAttributes' in element:
                                disableElement = False
                                if ((element.get('type') != 'fieldset') and
                                    ( not element.get('elementName').startswith("__") ) and
                                    (element.get('elementName') not in eaiAttributes['optionalFields'])  and
                                    (element.get('elementName') not in eaiAttributes['requiredFields']) ):
                
                                    disableElement = True
        
                                    if eaiAttributes.has_key('wildcardFields'):
                
                                        matches = 0
                                        for regex in eaiAttributes['wildcardFields']:
                                            result = re.search(regex, element.get('elementName'))
                                            if result:
                                                matches = matches + 1
                                        ## if there were any regex matches don't disable the field, it is a wildcard field.
                                        if matches > 0:
                                            disableElement = False
        
                                ## if the disableElement bool was toggled, disable the element
                                if disableElement:
                                    element["disabled"] = 1
                            
                            value = form_defaults.get(element['elementName'])
                            # logger.debug("value from form_defaults: %s." % value)
                            processValue = 'processValueAdd' if (entity_name == en.NEW_EAI_ENTITY_NAME and not uri) else 'processValueEdit'
                            if element.has_key(processValue):
                                try:
                                    value = eval(element[processValue].strip())
                                    logger.debug("processed value: %s" % value)
                                except Exception, e:
                                    logger.error('uiHelper %s operator failed for endpoint_path=%s elementName=%s: %s' % (processValue, endpoint_path, element['elementName'], str(e)))
                                form_defaults[element['elementName']] = value
                            element["value"] = value
                                
                        if element.has_key("disabledInView"):
                            if viewFilter in element["disabledInView"]:
                                element["disabled"] = "1"
                        if element.has_key('elements'):
                            # recurse through elements that contain other elements (eg. fieldset)
                            filterElements(element['elements'])
                    else:
                        elementsToFilter.append(elements[elnum]["elementName"])
                    elnum += 1
                
                def removeFromHelper(elements,elementsToFilter):
                    for i in range(len(elements)-1,-1,-1):
                        if elements[i]["elementName"] in elementsToFilter:
                            # fieldsets must contain the view for ALL subfields.
                            del elements[i]
                        elif "elements" in elements[i]:
                            # rerun for grouped elements
                            removeFromHelper(elements[i]["elements"],elementsToFilter)
                    
                removeFromHelper(uiHelper["elements"], elementsToFilter)
                
            # filter out elements that shouldn't be displayed
            # logger.debug("entity_name: %s" % entity_name)
            # logger.debug("viewFilter: %s" % viewFilter)
            
            filterElements(uiHelper['elements'])
        
        else:
            uiHelper["elements"] = []
            
        return uiHelper, form_defaults, entity, statusCode, msgid
    

    def render_admin_template(self, template_name, template_args=None):
        """
        Render a template including arguments required to display module based navigation
        """
        root = cherrypy.request.app.root
        namespace = template_args.get('namespace', splunk.getDefault('namespace'))
        # set build_nav=False below as manager doesn't display any modules requiring nav/saved searches
        args =  root.app.buildViewTemplate(namespace, ADMIN_VIEW_NAME, render_invisible=True, build_nav=False)
        if template_args:
            args.update(template_args)
        return self.render_template(template_name, args)
        
    
    @expose_page(must_login=True, methods=['GET'])
    def index(self, **kwargs):
        return self.redirect_to_url('/manager/%s/' % splunk.getDefault('namespace'))

    @route('/:namespace')
    @expose_page(must_login=True, methods=['GET'])
    @set_cache_level('never')
    def nsindex(self, namespace, **kwargs):
        """
        displays list of available entities which can be viewed / managed through this inerface
        or some other type of static / dynamic admin home page
        """

        
        uiHelpers = self.fetchUIHelpers(namespace=namespace, refresh=1)
        blueLinks = {'system_configurations': {'description': _("Manage configuration of this Splunk server."),
                                               'menuItems': []},
                     'app_configurations': {'description': _("View or edit your app configs."),
                                            'menuItems': []}}

        for endpoint_name in uiHelpers:
            endpoint = uiHelpers[endpoint_name]
            if 'menu' not in endpoint:
                continue

            menu_cfg = endpoint['menu']
            menu_name = menu_cfg.pop('name')
            menu = blueLinks.setdefault(menu_name, {'description': '', 'menuItems': []})
            
            menu['menuItems'].append(menu_cfg)

        for menu in blueLinks:
            blueLinks[menu]['menuItems'].sort((lambda x, y: x['order'] - y['order']))
        
        return self.render_admin_template('admin/index.html', {
            'namespace'             : namespace,
            'breadcrumbs'           : self.generateBreadcrumbs(namespace),
            'blueLinks'             : blueLinks,
            'msgid'                 : kwargs.get('msgid', None)
        })

    
    @route('/:namespace/*endpoint_path/', methods=['GET'])
    @expose_page(must_login=True)
    @set_cache_level('never')
    def listEntities(self, namespace, endpoint_path, **kwargs):
        """
        displays a list of entities that exist at a given endpoint,
        provides management links to edit / delete / create entities.
        """
        logger.debug('listEntities endpoint_path == %s' % endpoint_path)
        error = None
        entities = None
        entity = None
        form_defaults = None
        statusCode = 0
        nagwareMessaging = {}
        entity_name = None
        uiHelper = None
        uiHelper_elements ={}
        
        noContainer = kwargs.get('noContainer', None)
        eleOnly = kwargs.get('eleOnly', None)
        action = kwargs.get('action', False)
        
        search = kwargs.get('search', None)
        count = kwargs.get('count', 25)
        offset = kwargs.get('offset', 0)
        sort_key = kwargs.get('sort_key', None)
        sort_dir = kwargs.get('sort_dir', None)
        msgid = kwargs.get('msgid', None)
        pwnr = kwargs.get('pwnr', None)
        app_only = normalizeBoolean(kwargs.get('app_only', '0'))
        
        ns = kwargs.get('ns', namespace)
        if ns == "" or ns == None:
            # use the url part for the namespace... default app
            ns = namespace

        endpoint_base = endpoint_path
        appOptionList = []
        pwnrOptionList = []
        showNewButton = True
        isCloneAction = False
        
        if action == "edit":
            # logger.debug("edit action of listEntities")
            segments = endpoint_path.split("/")
            entity_name = segments.pop()
            endpoint_base = "/".join(segments)

            defaultView = "create" if entity_name == en.NEW_EAI_ENTITY_NAME else "edit"
            viewFilterMode = kwargs.get('viewFilter', defaultView)
            # switch for uri or template for endpoint inference
            uri = kwargs.get('uri', None)
            template = kwargs.get('template', None)
            
            if (entity_name == en.NEW_EAI_ENTITY_NAME and uri is not None):
                logger.debug("this is a clone. do some stuff differently.")
                isCloneAction = True
            

            if not uri and template:
                uri = en.buildEndpoint(endpoint_base, entityName=template, namespace=ns, owner=auth.getCurrentUser()['name'])
                
            # logger.debug("view filter mode:%s." % viewFilterMode)
            try:
                uiHelper, form_defaults, entity, statusCode, msgid = self.getSingleEntity(endpoint_path, uri=uri, namespace=ns, viewFilter=viewFilterMode)
                if endpoint_path.startswith('admin/win-wmi-collections') and not '_new' in endpoint_path:
                    if entity['classes'] == [] and entity['fields'] == []:
                        msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('warn', 'This query has additional parameters set in its configuration file. If you save changes from the UI, these will be lost.')                
            except (splunk.AuthenticationFailed), e:
                logger.debug("authentication on splunkd side failed, logout of browser session.")
                return self.redirect_to_url("/account/logout/")
            
            self.flattenElements(uiHelper['elements'], uiHelper_elements)
            
            try:
                entity_name = entity.name
            except (AttributeError):
                entity_name = None
            if isCloneAction:
                entity_name = "Clone from %s" % entity_name
                # reach into the entity and add name to the eai:attributes[requiredFields] list
                # logger.debug(entity.get('eai:attributes').get('requiredFields'))
                if type(entity.get('eai:attributes')) == type({}):
                    if type(entity.get('eai:attributes').get('requiredFields')) == type([]):
                        entity.get('eai:attributes').get('requiredFields').append("name")
                
                # logger.debug(entity.get('eai:attributes').get('requiredFields'))
                # logger.debug("*" * 50)
            
            breadcrumbs = self.generateBreadcrumbs(namespace, endpoint_base, entity_name)
        else:
            
            try:
                logger.debug("fetching ui helper for %s." % endpoint_path)
                uiHelper = self.fetchUIHelpers(search=endpoint_path, namespace=namespace)

                # determine if this list show show a 'New' button or not
                elems = uiHelper.get('elements',[])
                if len(elems) > 0:
                    views = elems[0].get('showInView', [])
                    showNewButton = 'create' in views
                
            except (splunk.AuthenticationFailed), e:
                logger.debug("authentication on splunkd side failed, logout of browser session.")
                return self.redirect_to_url("/account/logout/")
                
            except (splunk.ResourceNotFound, splunk.RESTException) , e:
                raise cherrypy.NotFound()
            
            # pick up the application and owner lists
            try:
                appList = cached.getEntities("apps/local", search=['disabled=false','visible=true'], count=-1)
                appOptionList = [  {'label': '%s (%s)' % ( appList[x].name, appList[x].get('label') ), 'value': appList[x].name} for x in appList ]
            except Exception, e:
                logger.error("exception accessing apps local endpoint.")
                logger.exception(e)
            
            try:
                # per SPL-24818 ... cap large user environments to 250... should we try to cache this or do something smarter here?
                pwnrList = en.getEntities("authentication/users", count=250, search="roles=*")
                pwnrOptionList = [  {'label': "%s (%s)" % (pwnrList[x].name, pwnrList[x]['realname']), 'value': pwnrList[x].name} for x in pwnrList ]
                pwnrOptionList.append({'label': _("No owner"), 'value': 'nobody'})
            except Exception, e:
                logger.error("exception accessing users endpoint:")
                logger.exception(e)

            if pwnr and pwnr != "-":
                owner_search = "eai:acl.owner=%s" % pwnr
                if not search:
                    search = owner_search
                else:
                    search = [search, owner_search]

            if app_only and ns != '-':
                app_search = "eai:acl.app=%s" % ns
                if not search:
                    search = app_search
                else:
                    search = [search, app_search]
                    
            try:
                entities = en.getEntities(endpoint_path, namespace=ns, owner="-", search=search, count=count, offset=offset, sort_key=sort_key, sort_dir=sort_dir, unique_key='id', sessionKey=cherrypy.session['sessionKey'])

                # in this case we need to augment the entities with their endpoint_bases
                if endpoint_path == 'admin/directory':
                    BASE_MAP = {'eventtypes': 'saved/eventtypes', 'savedsearch': 'saved/searches', 'views': 'data/ui/views', 'nav': 'data/ui/nav', 'commands': 'admin/commands'}
                    for e in entities.values():
                        e_type = e['eai:type']
                        e['endpoint_base'] = BASE_MAP.get(e_type, None)
                    
            except (splunk.RESTException), e:
                logger.exception(e)
                statusCode = e.statusCode
                # logger.debug("*"*20)
                # logger.debug( "type of e.msg: %s" % type(e.msg) )
                # logger.debug("*"*20)
                errMsgStr = ""
                if ( type(e.msg) == type([]) ):
                    if ( type(e.msg[0]) == type({}) ):
                        errMsgStr = e.msg[0].get("text")
                elif ( type(e.extendedMessages) == type([]) ):
                    if ( type(e.extendedMessages[0]) == type({}) ):
                        errMsgStr = e.extendedMessages[0].get("text")
                else:
                    errMsgStr = _("An exception occurred, but no message string was extracted.")
                        
                msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', '%s' % errMsgStr )
                
            except (splunk.InternalServerError), e:
                # when the params are stripped away (action=edit) this occurs and resulted in an ugly stack trace
                # this only happens if they manually type in an incorrect URL, or for some reason the params get removed
                # which was occuring from the logout.
                # catching and making them start over.
                logger.debug("500 internal server error.")
                logger.exception(e)
                msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', 'An error occured completing this request: %s.' % e.extendedMessages[0].get('text') )
                # provide an empty entities so we can get to teh list view page.
                entities = {}
                
                # return self.redirect_to_url("/manager/%s" % (namespace), _qs={
                #     "msgid" : msgid
                # } )
                
            except (splunk.ResourceNotFound), e:
                logger.debug("404 resource not found.")
                msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', ('Resource not found: %s' % e.msg) )
                
                return self.redirect_to_url("/manager/%s" % namespace, _qs={
                    "msgid" : msgid
                } )

            breadcrumbs = self.generateBreadcrumbs(namespace, endpoint_path)

        if statusCode == 402:
                    
            #hows about some nagware to go along with this error message?
            if uiHelper.has_key("menu"):
                endpoint_label = uiHelper['menu']['label']
            elif uiHelper.get('breadcrumb_entityname'):
                endpoint_label = uiHelper['breadcrumb_entityname']
            elif uiHelper.get('breadcrumb_name'):
                endpoint_label = uiHelper['breadcrumb_name']
            elif uiHelper.has_key('header'):
                endpoint_label = uiHelper['header']
            endpoint_description = uiHelper.get("menu", {}).get("description")
            
            nagwareMessaging["label"] = endpoint_label
            nagwareMessaging["description"] = endpoint_description

            return self.render_admin_template('admin/402.html', {
                'namespace'             : namespace,
                'msgid'                 : msgid,
                'nagwareMessaging'      : nagwareMessaging,
                'breadcrumbs'           : breadcrumbs
            })      
                
        if noContainer:
            returnTemplate = 'admin/edit.html'
        else:
            returnTemplate = 'admin/index.html'
        
        return self.render_admin_template(returnTemplate, {
            'namespace'     : namespace,
            'breadcrumbs'    : breadcrumbs,
            'entities'      : entities,
            'entity'        : entity,
            'form_defaults' : form_defaults,
            'entity_name'   : entity_name,
            'isCloneAction' : isCloneAction,
            'uiHelper'      : DynamicUIHelper(uiHelper, __context={'form_defaults':form_defaults, 'entity_name':entity_name}) if uiHelper else uiHelper,
            'uiHelper_elements'      : uiHelper_elements,
            'showNewButton' : showNewButton,
            'endpoint_path' : endpoint_path,
            'endpoint_base' : endpoint_base,
            'error'         : error,
            'nagwareMessaging'         : nagwareMessaging,
            'eleOnly'       : eleOnly,
            'kwargs'        : kwargs,
            'appOptionList' : appOptionList,
            'pwnrOptionList': pwnrOptionList,
            'msgid'         : msgid
        })

    @route('/:namespace/*endpoint_base/:element=_element/:element_name', methods=['POST','GET'])
    @expose_page(must_login=True)
    def fetch_element(self, namespace, endpoint_base, element, element_name, form_defaults=None, element_overlay=None, entity_name=None, eai_attributes=None):
        """
        Fetch an individual form element for an enedpoint using the supplied form values 
        Optionally update the element's uiHelper configuration using element_overlay
        """
        uiHelper = self.fetchUIHelpers(search=endpoint_base, namespace=namespace)
        if len(uiHelper) == 0:
            logger.error('listProperties - unable to load the section form definition for endpoint=%s' % endpoint_base)
            raise cherrypy.NotFound("Endpoint/element not found")

        form_defaults = json.loads(form_defaults) if form_defaults else {}
        element_overlay = json.loads(element_overlay) if element_overlay else {}
        eai_attributes = json.loads(eai_attributes) if eai_attributes else {}

        uiHelper_flatten = {}
        self.flattenElements(uiHelper['elements'], uiHelper_flatten)
        # deep_update_dict modifies in place; deep copy in case this data ends up being cached in the future which would
        # probably lead to a hard to track down bug
        element = copy.deepcopy(uiHelper_flatten[element_name])
        util.deep_update_dict(element, element_overlay)

        # NOTE: this intentionally calls render_template(), not render_admin_template()
        return self.render_template('admin/element.html', {
            'form_defaults' : form_defaults,
            'eaiAttributes' : eai_attributes,
            'form_errors': {},
            'element': DynamicUIHelper(element, __context={'form_defaults':form_defaults, 'entity_name':entity_name})
        })


    @route('/:action=permissions/:namespace/*endpoint_path', methods=['POST', 'GET'])
    @expose_page(must_login=True)
    @set_cache_level('never')
    def permissions(self, namespace, action, endpoint_path, sharing='user', owner=None, perms_read=None, perms_write=None, **kwargs):
        '''
        Permission control for an endpoints ACL.
        
        Arg:
        namespace: The user namespace.
        action: The hard-coded route beginning pattern of 'permissions/'
        endpoint_base: A valid endpoint to perform ACL actions on (ie., /search/saved/foo%wbar)
        sharing: One of 'app' or 'global' or defaults to 'user'. Sharing is caring.
        owner: The owner of the object or defaults to None.
        perms_read: A string or list of read owners/roles, defaults to None.
        perms_write: A string or list of write owners/roles, defaults to None
        '''
        # app
        app = None
        # the cancel button url
        cancel_url = self.getCancelURL()
        # place holder for error messages
        error = None
        # place holder for general info messages
        info = None
        # acl entity
        entity = None
        # the permission entity name
        entity_name = 'acl'
        # active selected roles
        active_roles = []
        # modifiable setting
        modifiable = None
        # object label (eg., Saved Search, etc...)
        object_label = 'Object'
        # object label mapping (eg., /saved/searches/foobar is saved->Saved Search)
        object_labels = {
            'saved/eventtypes': _('saved eventtype'),
            'saved/searches': _('saved search'),
            'data/ui/views': _('views'),
            'commands/': _('search commands and scripts'),
        }

        # control if redirect on successful post operation
        redirect_success = True
        # admin dependencies
        uiHelper = {
            'header': object_label,
            'elements': []
        }

        segments = endpoint_path.split("/")
        entity_name = segments.pop()
        endpoint_base = "/".join(segments)

        # fully qualified action path
        action_path = ''

        # derive the object label
        for object_label_key in object_labels:
            if endpoint_base.find(object_label_key)>-1:
                object_label = object_labels[object_label_key]
                break

        post_error = None

        # handle POST methods
        if cherrypy.request.method == 'POST':
            # add set entity routine
            entityUpdate = None
            entityBoilerPlate = None
            try:
                entityBoilerPlate = en.Entity(endpoint_base, entity_name, namespace=namespace, owner=splunk.auth.getCurrentUser()['name'])
            except Exception, e:
                post_error = _('Splunk could not update permissions for resource %(resource)s %(error)s') % {'resource':endpoint_base, 'error':e}
            # if we have an update entity object proceed
            if entityBoilerPlate:
                # update entity boiler plate sharing properties
                entityBoilerPlate.properties['sharing'] = sharing
                # update entity boiler plate owner properties
                entityBoilerPlate.properties['owner'] = owner
                # update entity boiler plate read perms properties
                if perms_read:
                    # single read perms
                    if isinstance(perms_read, unicode):
                        entityBoilerPlate.properties['perms.read'] = perms_read
                    # multiple read perms
                    elif isinstance(perms_read, list):
                        entityBoilerPlate.properties['perms.read'] = ",".join(perms_read)
                # update entity boiler plate write perms properties
                if perms_write:
                    # single write perms
                    if isinstance(perms_write, unicode):
                        entityBoilerPlate.properties['perms.write'] = perms_write
                    # multiple write perms
                    elif isinstance(perms_write, list):
                        entityBoilerPlate.properties['perms.write'] = ",".join(perms_write)
                # update permissions for this resources ACL
                try:
                    result = en.setEntity(entityBoilerPlate, uri=_extractPathFromURI(kwargs['uri']) + '/acl')
                except Exception, e:
                    post_error = _('Splunk could not update permissions for resource %(resource)s %(error)s') % {'resource':endpoint_base, 'error':e}
                # log properties that could not be saved   
                if post_error:
                    logger.error('Splunk could not update ACL with the following params: %s' % entityBoilerPlate.properties)
            # set proper info message for successful update of permissions
            if post_error is None:
                return self.redirect_to_url(cancel_url)

        #gracefully handle entity retrieval
        try:
            entity_name = cpUnquoteEntity(entity_name)
            entity = en.getEntity(endpoint_base, entity_name, uri=_extractPathFromURI(kwargs.get('uri', None)), namespace=namespace, owner=auth.getCurrentUser()['name'], sessionKey=cherrypy.session['sessionKey'])
            entity_name = entity.name
        except Exception, e:
            error = _('Splunk could not retrieve permissions for resource %(resource)s %(error)s') % {'resource':endpoint_base, 'error':e}

        # if no error and entity acl data exists perform standard routines
        if error is None and entity['eai:acl']:
            # if acl and perms related data present:
            # 1) build an ordered list of active roles 
            # 2) assign perms_read and perms_write list

            if entity['eai:acl']['perms']:
                active_roles = self._permissionsGetRoles(entity['eai:acl']['perms'])
                perms_read = entity['eai:acl']['perms'].get('read', [])
                perms_write = entity['eai:acl']['perms'].get('write', [])
                # set perms to empty list if None exist
            else:
                perms_read = []
                perms_write = []
                # set acl modifiable state if present

            if entity['eai:acl']['modifiable']:
                try:
                    modifiable = normalizeBoolean(entity['eai:acl']['modifiable'], enableStrictMode=True)
                except Exception, e:
                    logger.warn("Splunk could not normalize acl modifiable value to Boolean %s" % e)
                sharing = entity['eai:acl']['sharing']
                owner = entity['eai:acl']['owner']
                app = entity['eai:acl']['app']

        # available sharing options
        sharing_options = []

        if entity['eai:acl'].get('sharing') == 'user' or entity['eai:acl'].get('can_share_user', None) == '1':
            sharing_options.append(('user', _('Keep private')))

        if entity['eai:acl'].get('sharing') == 'app' or entity['eai:acl'].get('can_share_app', None) == '1':
            app_label = _('This app only (%s)') % app if app else _('This app only')  
            sharing_options.append(('app', app_label))

        if entity['eai:acl'].get('sharing') == 'global' or entity['eai:acl'].get('can_share_global', None) == '1':
            sharing_options.append(('global', _('All apps')))
                
        # set appropriate modifiable message
        if modifiable is False:
            info = _('Permissions for this object are restricted. It is not modifiable.')
        # get roles
        role_entities = None
        roles = []
        try:
            role_entities = en.getEntities('authentication/roles', namespace=namespace, count=-1)
        except Exception, e:
            pass
        # create list of roles
        if role_entities:
            roles = [x for x in role_entities]

        breadcrumbs = self.generateBreadcrumbs(namespace, endpoint_base, entity_name, 'Permissions')
        isFree = splunk.util.normalizeBoolean(cherrypy.config["is_free_license"])
        # we are should not display the 402 page, instead hide the role sharing table and only show the global / this app only options.
        """
        nagwareMessaging = {}
        msgid = None
        if (isFree):
            endpoint_label = _("Permissions")        
            endpoint_description = _("Manage object permissions.")
            
            nagwareMessaging["label"] = endpoint_label
            nagwareMessaging["description"] = endpoint_description

            return self.render_admin_template('admin/402.html', {
                'namespace'             : namespace,
                'msgid'                 : msgid,
                'nagwareMessaging'      : nagwareMessaging,
                'breadcrumbs'           : breadcrumbs
            })
        """



        return self.render_admin_template('admin/permissions.html', {
            'namespace' : namespace,
            'breadcrumbs' : breadcrumbs,
            'error': error or post_error,
            'info': info,
            'entity': entity,
            'active_roles': active_roles,
            'modifiable': modifiable,
            'sharing': sharing,
            'owner': owner,
            'app': app,
            'perms_read': perms_read,
            'perms_write': perms_write,
            'action_path': action_path,
            'object_label': object_label,
            'sharing_options': sharing_options,
            'perms_modifiable': entity['eai:acl'].get('can_change_perms', None) == '1',
            'roles': roles,
            'cancel_url': self.make_url(cancel_url),
            # required admin template args
            'entity_name' : "Permissions",
            'endpoint_base' : endpoint_base,
            'form_defaults' : entity,
            'uiHelper' : uiHelper
        })


    def _permissionsGetRoles(self, perms):
        """
        Get an ordered list of roles from an ACL's perms. 
        
        Arg:
        perms: A dictionary of permissions associated to a list of roles.
        """
        unique_rolls = []
        for perm, roles in perms.items():
            unique_rolls.extend(roles)
        return sorted(set(unique_rolls))        


    @route('/:namespace/*endpoint_base/:multiaction=multidelete', methods=['POST'])
    @expose_page(must_login=True)
    def multiaction(self, namespace, endpoint_base, multiaction, **kwargs):
        

        control = kwargs.pop('ctrl')
        control_link = kwargs.pop('ctrl_link')
        control_name = kwargs.pop('ctrl_name')
        showAppContext = False
        if kwargs.has_key('showAppContext'):
            showAppContext = kwargs.pop('showAppContext')
        
        if kwargs.get('redirect_to') != None:
            redirect_to = kwargs.pop('redirect_to')
            logger.debug("will redirect to %s." % redirect_to)
        else:
            redirect_to = "/manager/%s/%s" % (namespace, endpoint_base)

        logger.debug("perparing to %s %s." % (CONTROL_ACTIONS[control], control_name) )
        try:
            en.controlEntity(control, control_link)
            if showAppContext:
                # used for messaging what app context the action was applied to
                control_parts = control_link.split("/")
                context_part = control_parts[3]
                msgStr = '%s %s from %s.' % (CONTROL_ACTIONS[control], control_name, context_part)
            else:
                msgStr = '%s %s.' % (CONTROL_ACTIONS[control], control_name)
            msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('info', msgStr)
            kwargs['msgid'] = msgid
        except Exception, e:
            # logger.debug(e)
            errMsg = ": %s" %e
            # logger.debug("errMsg: %s" % errMsg)
            try:
                errMsg = e.extendedMessages[0].get("text")
            except (splunk.ResourceNotFound), e:
                errMsg = " Resource not found: %s" % e
            except:
                errMsg = ""
            msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', 'Error occured attempting to %s %s: %s.' % (control, control_name, errMsg))
            kwargs['msgid'] = msgid

        kwargs.pop('splunk_session_id', None)
        
        return self.redirect_to_url(redirect_to, _qs=kwargs)
    
    
    @route('/:namespace/*endpoint_base/:entity_name', methods=['POST'])
    @expose_page(must_login=True)
    def editProperties(self, namespace, endpoint_base, entity_name, **kwargs):
        """
        provide a location for posting entity updates
        a form for editing the properties of a given endpoint and entity
        test for editing properties.
        """
        
        fetchStaticHelpers = True
        error = None
        form_errors = {}
        form_defaults = {}
        uiHelper = {}
        redirectBase = "/manager/%s" % namespace
        redirectTail = kwargs.get("spl-ctrl_redirectionPath", endpoint_base)
        redirectToUrl = "%s/%s" % (redirectBase, redirectTail)    
        redirectQSName = kwargs.get("spl-ctrl_redirectQSName", "redirecting")
        redirectQSVal = kwargs.get("spl-ctrl_redirectQSVal", "true")
        entity_name = urllib.unquote(urllib.unquote(entity_name))
        
        #logger.debug("endpoint_base: %s" % endpoint_base)
        #logger.debug("entity_name: %s" % entity_name)

        if endpoint_base.startswith("apps/local/") and entity_name == "setup":
            fetchStaticHelpers = False
            
        if fetchStaticHelpers:
            uiHelper = self.fetchUIHelpers(search=endpoint_base, namespace=namespace)
        else:
            # can't generate uiHelper without the entity, don't need it right here though.
            uiHelper = {}
        
        if len(uiHelper) == 0:
            logger.error('editProperties - unable to load the section form definition for endpoint=%s' % endpoint_base)

        
        showAppContext = uiHelper.get("showAppContext")
        editForm = formencode.Schema()
        editForm.allow_extra_fields = True

        entity_namespace = kwargs.pop('__ns', namespace)
        entity_owner     = kwargs.pop('__owner', '')
        entity_uri       = kwargs.pop('__uri', None)
        do_redirect      = kwargs.pop('__redirect', '0') == '1' and uiHelper.get('redirect')

        endpoint_base_saveto = kwargs.get('spl-ctrl_endpoint_override', endpoint_base)

        try:
            saved = False
            uiHelper_elements = {}
            self.flattenElements(uiHelper.get('elements', []), uiHelper_elements)
            form_data = editForm.to_python(kwargs)
            for key, value in form_data.items():
                # some values need to be post-processed prior to submission
                if uiHelper_elements.has_key(key):
                    submitValue = 'submitValueAdd' if entity_name==en.NEW_EAI_ENTITY_NAME else 'submitValueEdit'
                    p = uiHelper_elements[key].get(submitValue)
                    if p:
                        try:
                            nv = eval(p.strip())
                            if nv is None:
                                del form_data[key]
                            else:
                                form_data[key] = nv
                        except Exception, e:
                            logger.error('uiHelper %s operator failed for endpoint_base=%s entity_name=%s elementName=%s: %s' % (submitValue, endpoint_base, entity_name, key, str(e)))

                    if uiHelper_elements[key].get('type') == 'password' and form_data[key]!='' and form_data[key]!='********' and form_data[key] != form_data['spl-ctrl_%s-confirm'%key]:
                        return self.render_api({
                            'status': 'FIELD_ERRORS',
                            'fields': { key: 'Passwords do not match' },
                            }, 
                            set_mime='text/html' # jquery form plugin requires json to be return as text/html for its submission-by-iframe trick
                        )
                        

            # Attempt to save the entity
            saved, err = self.saveEntity(endpoint_base_saveto, entity_name, form_data, entity_namespace, entity_owner, entity_uri)
            if err:
                logger.debug("Error returned was: %s" % err)
            
            # for messaging
            savedName = form_data.get("name") if entity_name==en.NEW_EAI_ENTITY_NAME else entity_name
        
            if saved:
                if entity_name == en.NEW_EAI_ENTITY_NAME:
                    actionApplied = _('saved')
                else:
                    actionApplied = _('updated')
                    
                # logger.debug("err type: %s" % type(err))
                appendMsg = ""
                if (type(err) == type(dict())):
                    if len(err["messages"]) > 0:
                        appendMsg = err["messages"][0].get("text")
                        if not appendMsg.endswith("."):
                            appendMsg = appendMsg + "."
                    
                if showAppContext:
                    # TRANS: "Successfully saved  mysearch in Search"
                    feedbackMessage = _('Successfully %(verb)s "%(noun)s" in %(namespace)s. %(extra)s') % {'verb':actionApplied, 'noun':savedName, 'namespace':entity_namespace, 'extra': appendMsg}
                else:
                    # TRANS: "Succesfully updated settings"
                    feedbackMessage = _('Successfully %(verb)s "%(noun)s". %(extra)s') % {'verb':actionApplied, 'noun': savedName, 'extra': appendMsg}
                    
                if do_redirect:
                    msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('info', feedbackMessage)
                    redirect_target = self.make_url(redirectToUrl, _qs={ 'msgid':msgid, 'ns': entity_namespace, redirectQSName: redirectQSVal })
                else:
                    redirect_target = ''

                return self.render_api({
                    'status': 'OK',
                    'msg': feedbackMessage,
                    'redirect': redirect_target
                    }, 
                    set_mime='text/html' # jquery form plugin requires json to be return as text/html for its submission-by-iframe trick
                )

            else:
                if entity_name == en.NEW_EAI_ENTITY_NAME:
                    feedbackMessage = _('Encountered the following error while trying to save: %s.') % err
                else:
                    feedbackMessage = _('Encountered the following error while trying to update: %s.') % err

                logger.debug("feedbackMessage: %s." % feedbackMessage)
                return self.render_api({
                    'status': 'ERROR',
                    'msg': feedbackMessage,
                    }, 
                    set_mime='text/html' # jquery form plugin requires json to be return as text/html for its submission-by-iframe trick
                )
                
        except formencode.api.Invalid, e:
            logger.debug("Form data failed validation, returning errors.")
            if e.error_dict:
                # return a mapping of field names to error messages
                return self.render_api({
                    'status': 'FIELD_ERRORS',
                    'fields': e.error_dict,
                    'set_mime': 'text/html' # jquery form plugin requires json to be return as text/html for its submission-by-iframe trick
                    })
            else:
                error = e.msg
                return render_api({
                    'status': 'ERROR',
                    'msg': e.msg,
                    }, 
                    set_mime='text/html' # jquery form plugin requires json to be return as text/html for its submission-by-iframe trick
                )
        
    
    
    @route('/:namespace/:action=datainputstats')
    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def datainputstats(self, namespace, action, operation=None):
        
        monitors = en.getEntities('/data/inputs/monitor', namespace=namespace)
        scripts  = en.getEntities('/data/inputs/script', namespace=namespace)
        tcps = en.getEntities('/data/inputs/tcp/raw', namespace=namespace)
        cookedTcps = en.getEntities('/data/inputs/tcp/cooked', namespace=namespace)
        udps = en.getEntities('/data/inputs/udp', namespace=namespace)

        # provide overview data for data inputs, links to create new etc.
        indexData = {
            'monitorCount'      : len( monitors ),
            'scriptCount'       : len( scripts ),
            'tcpCount'          : len( tcps ),
            'cookedTcpCount'    : len( cookedTcps ),
            'udpCount'          : len( udps )
        }
        
        showWinInputs = True if sys.platform.startswith('win') else False

        if showWinInputs:
            wmis = en.getEntities('/admin/win-wmi-collections', namespace=namespace)
            els = en.getEntities('/admin/win-event-log-collections', namespace=namespace)
            regs = en.getEntities('/admin/win-regmon', namespace=namespace)
        
            winData = {
                'wmiCount'          : len( wmis ),
                'elCount'           : len( els ),
                'regCount'          : len( regs )
            }
            indexData.update(winData)
            
        return self.render_admin_template('admin/datainputstats.html', {
            'namespace'       : namespace,
            'breadcrumbs'     : self.generateBreadcrumbs(namespace, 'datainputstats'),
            'indexData'       : indexData, # dict of index stats.
            'winData'         : showWinInputs
        })

    @route('/:namespace/:action=tags')
    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def tags(self, namespace, action, operation=None):
        return self.render_admin_template('admin/tags.html', {
            'namespace'       : namespace,
            'breadcrumbs'     : self.generateBreadcrumbs(namespace, 'tags'),
        })

    @route('/:namespace/:action=transforms')
    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def transforms(self, namespace, action, operation=None):
        return self.render_admin_template('admin/transforms.html', {
            'namespace'       : namespace,
            'breadcrumbs'     : self.generateBreadcrumbs(namespace, 'transforms'),
        })

    @route('/:namespace/:action=props')
    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def props(self, namespace, action, operation=None):
        return self.render_admin_template('admin/props.html', {
            'namespace'       : namespace,
            'breadcrumbs'     : self.generateBreadcrumbs(namespace, 'props'),
        })

    @route('/:namespace/:action=lookups')
    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def lookups(self, namespace, action, operation=None):
        return self.render_admin_template('admin/lookups.html', {
            'namespace'       : namespace,
            'breadcrumbs'     : self.generateBreadcrumbs(namespace, 'lookups'),
        })

    @route('/:namespace/:action=forwardreceive')
    @expose_page(must_login=True, handle_api=True, methods=['GET'])
    def forwardreceive(self, namespace, action, operation=None, **kwargs):
        
        # for the new link to enable lightweight forwarder... first check that it exists
        lwfExists = False
        try:
            lwf = en.getEntity("apps/local", "SplunkLightForwarder")
            lwfExists = True
        except (splunk.ResourceNotFound), e:
            logger.warn("splunkLightForwarder was not found")
        
        msgid = kwargs.get("msgid")
        
        return self.render_admin_template('admin/forwardreceive.html', {
            'lwfExists'       : lwfExists,
            'breadcrumbs'     : self.generateBreadcrumbs(namespace, 'forwardreceive'),
            'namespace'       : namespace,
            'msgid'           : msgid
        })


    @route('/:namespace/:action=distsearch')
    @expose_page(must_login=True, handle_api=True, methods=['GET'])
    def distsearch(self, namespace, action, operation=None, **kwargs):
        msgid = kwargs.get("msgid")
   
        return self.render_admin_template('admin/distsearch.html', {
            'breadcrumbs'     : self.generateBreadcrumbs(namespace, 'distsearch'),
            'namespace'       : namespace,
            'msgid'           : msgid
        })

    
    @route('/:namespace/:action=authoverview')
    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def authoverview(self, namespace, action, operation=None, **kwargs):
        statusCode = 0
        nagwareMessaging = {}
        
        try:
            active_authmodule = en.getEntity("authentication/providers/services","active_authmodule")
            authmode = active_authmodule["active_authmodule"]
            logger.info("active authmode: %s." % authmode)
            
            rolemapping_endpoint = en.getEntities("admin/LDAP-groups")
            rolemapping_available = False
            if len(rolemapping_endpoint) > 0:
                rolemapping_available = True
        except (splunk.ResourceNotFound, splunk.AuthorizationFailed), e:
            error = "Fail: %s." % e.msg
            msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', error)
            return self.redirect_to_url("/manager/%s" % namespace, _qs={
                "msgid" : msgid
            } )
        except (splunk.RESTException), e:
            logger.debug("RESTException.")
            statusCode = e.statusCode
            msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', '%s' % e.msg[0].get('text') )
        
        breadcrumbs = self.generateBreadcrumbs(namespace, 'authoverview')
            
        if statusCode == 402:
        
            endpoint_label = _("Authentication Method")        
            endpoint_description = _("Specify authentication method (Splunk or LDAP).")
            
            nagwareMessaging["label"] = endpoint_label
            nagwareMessaging["description"] = endpoint_description

            return self.render_admin_template('admin/402.html', {
                'namespace'             : namespace,
                'msgid'                 : msgid,
                'nagwareMessaging'      : nagwareMessaging,
                'breadcrumbs'           : breadcrumbs
            })      
        
        
        return self.render_admin_template('admin/authoverview.html', {
            'namespace'             : namespace,
            'breadcrumbs'           : breadcrumbs,
            'rolemapping_available' : rolemapping_available,
            'authmode'              : authmode # currently active splunk auth provider.
        })

    @route('/:namespace/:action=licensestats')
    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def licensestats(self, namespace, action, operation=None, **kwargs):
        
        try:
            licenseData = en.getEntity("admin/license","license")
            serverInfo = en.getEntity("server", "info", owner="-")
        except (splunk.ResourceNotFound, splunk.AuthorizationFailed), e:
            error = "Fail: %s." % e.msg
            msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', error)
            return self.redirect_to_url("/manager/%s" % namespace, _qs={
                "msgid" : msgid
            } )
        isTrial = splunk.util.normalizeBoolean( serverInfo.get("isTrial") )
        licenseType = licenseData.get("licenseType")
        if licenseType == "free":
            callToAction = _("This version of Splunk is Free. Free for you to use as long as you like, within the terms of the license agreement. However, the paid version of Splunk has many more features, such as:<br/><ul><li>alerting based on searches and advanced thresholds.</li><li>authentication, optionally integrated with your enterprise AAA system.</li><li>roles and users for a customizable Splunk experience.</li><li>distributed search for massive scale and searching across operational silos.</li><li>daily volumes up to many terrabytes per day.</li></ul>If you want to request an Enterprise Trial license for this Splunk instance, <a target='_blank' href = http://splunk.com/r/signup>you can request one right now.</a>, or <a href=mailto:sales@splunk.com>contact Splunk sales</a> about upgrading. If you thought you had an Enterprise Trial, but are seeing this screen instead, likely your trial term has expired. You can <a target='_blank' href = http://splunk.com/r/request_trial>request another Trial</a>, or <a target='_blank' href = http://splunk.com/r/download>download a new copy</a> of Splunk and start again.<br><br>Regardless, we hope you like Splunk, and want to <a target ='_blank' href = http://splunk.com/r/feedback>hear any feedback</a> about the product.")
        elif licenseType == "pro":
            if isTrial:
                callToAction = _("This is a Splunk Enterprise Trial license. It gives you access, for a limited time, to all the features of Splunk. While you're using this trial, be sure to try our advanced alerting and user customization before your trial expires. If you would like to <a target='_blank' href = http://splunk.com/r/renew>extend your Trial</a> or <a href = mailto:sales@splunk.com>get a quote</a> we would be happy to assist you. Finally, you may <a href='../../licensing'>switch to a free license</a> at any time.")
            else:
                callToAction = _("You are using Splunk with an Enterprise license. This license is good for one Splunk instance. If you'd like to increase your daily indexing volume or get an additional license for another Splunk instance, please <a href='mailto:sales@splunk.com'>contact Splunk</a>.")

        return self.render_admin_template('admin/licensestats.html', {
            'namespace'         : namespace,
            'breadcrumbs'       : self.generateBreadcrumbs(namespace, 'admin/license'),
            'licenseData'       : licenseData,   # dict of license stats.
            'callToAction'      : callToAction
        })
    

    @expose_page(must_login=True, handle_api=True, methods=['GET', 'POST'])
    def control(self, operation=None):
        if cherrypy.request.method == 'POST':
            if operation == 'restart_server':
                # Fetch the Port and SSL enabled status so the JS knows where to poll for status
                ent = en.getEntity('server/settings', 'settings', sessionKey=cherrypy.session['sessionKey'])
                uri = '/services/server/control'
                postargs = {'_action':'restart'}
                try:
                    serverResponse, serverContent = rest.simpleRequest(uri, postargs=postargs)
                    response = {
                        'status': 'OK',
                        'start_time': int(cherrypy.config['start_time']),
                        'port': int(ent['httpport']),
                        'ssl': splunk.util.normalizeBoolean(ent['enableSplunkWebSSL'])
                    }
                except splunk.AuthenticationFailed:
                    response = { 'status': 'AUTH' }
                except splunk.AuthorizationFailed:
                    response = { 'status': 'PERMS' }
                return self.render_api(response)
            elif operation == 'resync_auth':
                uri = '/services/authentication/providers/services/_reload'
                postargs = {'name':'restart'}
                try:
                    serverResponse, serverContent = rest.simpleRequest(uri, postargs=postargs)
                except splunk.AuthenticationFailed:
                    return 'AUTH'
                except splunk.AuthorizationFailed:
                    return 'PERMS'
                if serverResponse.status in (201, 200):
                    return 'OK'
                return serverResponse
            else:
                return _('Invalid Operation')
        return _('Invalid Operation %s') % cherrypy.request.method

    @route('/:namespace/:action=control')
    @expose_page(must_login=True, handle_api=False, methods=['GET'])
    def controlpage(self, namespace, action, operation=None):
        
        serverControls = en.getEntities("server/control")
        restartLink = filter((lambda x: x[0] == 'restart'), serverControls.links)
        if len(restartLink)>0:
            displayRestartButton = True
        else:
            displayRestartButton = False

        
        return self.render_admin_template('admin/control.html', {
            'namespace'     : namespace,
            'displayRestartButton'  : displayRestartButton,
            'breadcrumbs'   : self.generateBreadcrumbs(namespace, 'control')
        })

    @route('/:namespace/:endpoint_base=widget-hiding')
    @expose_page(must_login=True, methods=['GET'])
    def listWidgetHidingEntities(self, namespace, endpoint_base, **kwargs):
        """
        Endpoint that demos the principal of widget actions triggering the display/hiding
        of other widgets
        """
        error = None
        msgid = kwargs.get('msgid', None)
        return self.render_admin_template('admin/index.html', {
            'namespace'     : namespace,
            'breadcrumbs'       : self.generateBreadcrumbs(namespace, 'widget-hiding'),
            'entities'      : ['Demo'],
            'entity'        : None,
            'uiHelper'      : {},
            'kwargs'      : {},
            'endpoint_path' : endpoint_base,
            'endpoint_base' : endpoint_base,
            'error'         : error,
            'msgid'         : msgid
        })

    @route('/:namespace/:endpoint_base=widget-hiding/:widget_name=Demo', methods=['GET','POST'])
    @expose_page(must_login=True)
    def listWidgetHidingProperties(self, namespace, endpoint_base, widget_name, **kwargs):
        error = None
        outputFormat = kwargs.get('outputFormat', 'HTML')
        form_errors = {}
        form_defaults = {}
        uiHelper = self.fetchUIHelpers(search=endpoint_base, namespace=namespace)

        uiHelper_elements = {}
        self.flattenElements(uiHelper['elements'], uiHelper_elements)
        
        return self.render_admin_template('admin/index.html', {
            'namespace'         : namespace,
            'endpoint_path'     : endpoint_base,
            'endpoint_base'     : endpoint_base, # url base from admin ie: saved-searches
            'uiHelper'          : uiHelper, # dictionary of values for markup help
            'uiHelper_elements' : uiHelper_elements, # dictionary of elements organized by element name
            'entity_name'       : widget_name,
            'form_errors'       : {}, # dict of form errors
            'form_defaults'     : {'dummy':'dummy'}, # dict of entity values.
            'error'             : error
        })

    @route('/:namespace/:endpoint_base=view-widgets')
    @expose_page(must_login=True, methods=['GET'])
    def listViewWidgetsEntities(self, namespace, endpoint_base, **kwargs):
        """
        Create a virtual endpoint called view-widgets that lists the unqiue
        widgets currently defined in uiHelpers
        """
        widgets = self.loadUniqueWidgets() # find unique widgets defined in uiHelpers
        error = None
        msgid = kwargs.get('msgid', None)
        return self.render_admin_template('admin/index.html', {
            'namespace'         : namespace,
            'entities'      : ['All Widgets'] + [name for name in widgets.keys() if name!='hidden'],
            'endpoint_path' : endpoint_base,
            'endpoint_base' : endpoint_base,
            'error'         : error,
            'msgid'         : msgid
        })
        
    @route('/:namespace/:endpoint_base=view-widgets/:widget_name')
    @expose_page(must_login=True, methods=['GET','POST'])
    def listWidgetProperties(self, namespace, endpoint_base, widget_name, **kwargs):
        """
        display examples of one or more widget types by creating a fake uiHelper
        The examples are taken from the live uiHelper config
        This should be good enough for our current widgets
        """
        widgets = self.loadUniqueWidgets()
        error = None
        outputFormat = kwargs.get('outputFormat', 'HTML')
        form_errors = {}
        form_defaults = {}
        
        widget_names = widgets.keys() if widget_name=='All Widgets' else [widget_name]
        uiHelper = {'header' : widget_name, 'elements' : []} # fake up a uiHelpers for this virtual endpoint
        for name in widget_names:
            if name not in widgets:
                error = _('The widget name is invalid.')
            else:
                element = widgets[name].copy()
                if element.has_key('label'):
                    element['label'] = '(%s) %s' % (name, element['label'])
                uiHelper['elements'].append(element)
        
        uiHelper_elements = {}
        self.flattenElements(uiHelper['elements'], uiHelper_elements)

        if outputFormat == "XML":
            # format of XML returns the raw feed
            
            transformer = kwargs.get('transformer', None)
            if not transformer == None:
                # insert the transformer xslt document into the xml.
                # validate the xsl is xsl file and path is available.
                pass
            pass
        else:
            return self.render_admin_template('admin/index.html', {
                'namespace'         : namespace,
                'endpoint_path'     : endpoint_base,
                'endpoint_base'     : endpoint_base, # url base from admin ie: saved-searches
                'uiHelper'          : uiHelper, # dictionary of values for markup help
                'uiHelper_elements' : uiHelper_elements, # dictionary of elements organized by element name
                'entity_name'       : widget_name,
                'form_errors'       : {}, # dict of form errors
                'form_defaults'     : {'dummy':'dummy'}, # dict of entity values.
                'error'             : error
            })

    def prepareRemoteQuery(self, category=None, subcategory=None, page=None, offset=None):
        query = {}
        query['splunk_version'] = '4.0.0'
        if page:
            query['page'] = page
        if offset:
            query['offset'] = offset
        if subcategory:
            category = subcategory
        if category is not None:
            query['category'] = category

        return query   

    def getRemoteCategories(self, category=None):
        cats = []

        entities = en.getEntities('apps/remote/categories', sessionKey=cherrypy.session['sessionKey'])
        for e in entities:
            entities[e]['title'] = e
            if entities[e]['encodedName'] == category:
                subcats = []
                
                logger.error(str(entities[e]['encodedName']))
                kwargs = {'splunk_version' : '4.0.0'}
                subentities = en.getEntities('apps/remote/categories/%s' % entities[e]['encodedName'], sessionKey=cherrypy.session['sessionKey'], **kwargs)
                for sube in subentities:
                    subentities[sube]['title'] = sube
                    
                    if eval(subentities[sube]['count']) > 0:
                        subcats.append(subentities[sube])
                entities[e]['children'] = subcats
            else: 
                entities[e]['children'] = None
            
            if eval(entities[e]['count']) > 0:
                cats.append(entities[e])
        return cats

    def getRemoteEntries(self, category=None, subcategory=None, page=None, count=None):
        entries = []
        offset = None
        
        if page and count:
            offset = (int(page)-1) * int(count)
        
        url = 'apps/remote/entries' 
        query = self.prepareRemoteQuery(category, subcategory, None, offset)
        entities = en.getEntities(url, sessionKey=cherrypy.session['sessionKey'], **query)
        
        # paging
        totalResults = int(entities.totalResults)
        itemsPerPage = int(entities.itemsPerPage)
        
        pages = 1
        pages = totalResults / itemsPerPage
        if (totalResults % itemsPerPage) > 0:
            pages += 1
        
        for e in entities:
            entities[e]['link'] = 'http://splunkbase.com/apps/%s' % urllib.quote(e)
            entities[e]['title'] = e
            entities[e]['titleEncoded'] = urllib.quote_plus(e)
            entities[e]['dateAdded'] = parseISO(entities[e]['dateAddonCreated']).strftime('%m/%d/%y %I:%M:%S')
            entities[e]['dateUpdated'] = parseISO(entities[e]['dateAddonUpdated']).strftime('%m/%d/%y %I:%M:%S')
            
            entries.append(entities[e])
        return (entries, pages)
    
    def installApp(self, namespace, appname):
        sessionKey = None
        if 'sbSessionKey' in cherrypy.session:
            sessionKey = cherrypy.session['sbSessionKey']
                
        if sessionKey == None:
            raise SplunkbaseException(_('not authenticated on Splunkbase'))
        
        url = 'apps/remote/entries/%s' % appname
        requestArgs = {'action': 'install', 'auth': sessionKey.replace(' ', '+')}
        
        response, content = rest.simpleRequest(url, postargs=requestArgs, sessionKey=cherrypy.session['sessionKey'])
        
        if response.status in [200,201]:
            return True
        else:
            raise SplunkbaseException('server returned error: ' + str(response.status))
            
     
    @route('/:namespace/:part1=apps/:part2=remote', methods=['GET','POST'])
    @expose_page(must_login=True, handle_api=True)
    def browser(self, namespace, part1, part2, encodedAppname=None, action='view', category=None, subcategory=None, page=None, count=10, splunk_session_id=None, message=None, msgid=None):

        baseUrl = '/manager/%s/apps' % namespace
        logger.warn(str(category) if category else 'None')       
        if action == 'install':
            try:
                self.installApp(namespace, encodedAppname)

                # refresh apps
                response, content = rest.simpleRequest('apps/local', getargs={'refresh' : 'true'}, sessionKey=cherrypy.session['sessionKey'])
            
                msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('warn', _('Successfully installed app: %s. You must restart for the app to be enabled.') % urllib.unquote_plus(encodedAppname))
                self.redirect_to_url('%s/local' % baseUrl, _qs={'msgid':msgid})
            except SplunkbaseException, e:
                msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', _('Failed to install app: %s, %s') % (urllib.unquote_plus(encodedAppname), e))
                self.redirect_to_url('%s/remote' % (baseUrl,), _qs={'msgid':msgid})
                
        templateArgs = {'cats' : [], 'apps' : [], 'namespace': namespace}
       
        templateArgs['breadcrumbs'] = self.generateBreadcrumbs(namespace, 'splunkbase')
        templateArgs['cats'] = self.getRemoteCategories(category)
        templateArgs['apps'], templateArgs['totalPages'] = self.getRemoteEntries(category=category, subcategory=subcategory, page=page, count=count)
        templateArgs['selectedCat'] = category if category else 'All'
        templateArgs['selectedSubcat'] = subcategory
        templateArgs['urlQuery'] = self.prepareRemoteQuery(category, subcategory, page)
        templateArgs['cpSessionKey'] = cherrypy.session.id
        templateArgs['curPage'] = int(page) if page else 1
        templateArgs['message'] = message
        templateArgs['baseUrl'] = baseUrl + '/remote'
        templateArgs['msgid'] = msgid
        
        return self.render_admin_template('/remoteapps/browser.html', templateArgs)

    @route('/:namespace/:part1=apps/:part2=remote/:part3=login', methods=['GET', 'POST'])
    @expose_page(must_login=True, handle_api=True)
    @lock_session
    def login(self, namespace, part1, part2, part3, returnurl=None, username=None, password=None, splunk_session_id=None):
        
        baseUrl = '/manager/%s/apps' % namespace
        
        if cherrypy.request.method == 'POST':
            try:
                import xml.etree.cElementTree as et
                
                response, content = rest.simpleRequest('/apps/remote/login', postargs={'username' : username, 'password' : password}, sessionKey=cherrypy.session['sessionKey'])
                if response.status not in [200, 201]:
                    return ''
                    
                root = et.fromstring(content)
                cherrypy.session['sbSessionKey'] = root.findtext('sessionKey')
            except Exception, e:
                return ''

        return cherrypy.session['sbSessionKey']

            
def unit_test():
    adm = AdminController()

if __name__ == '__main__':
    unit_test()
