import cherrypy, json, time, logging
from splunk.appserver.mrsparkle import *
from splunk.appserver.mrsparkle.lib import i18n


logger = logging.getLogger('splunk.appserver.controllers.config')

class ConfigController(BaseController):
    """/config"""

    def _get_active_config(self):
        return {
            'SEARCH_RESULTS_TIME_FORMAT': i18n.ISO8609_MICROTIME,
            'DISPATCH_TIME_FORMAT': cherrypy.config.get('DISPATCH_TIME_FORMAT'),
            'MRSPARKLE_ROOT_PATH': cherrypy.request.script_name,
            'MRSPARKLE_PORT_NUMBER': cherrypy.config.get('server.socket_port'),
            'SPLUNKD_SESSION_KEY': cherrypy.session['sessionKey'],
            'VERSION_LABEL': cherrypy.config.get('version_label', 'UNKNOWN_VERSION'),
            'BUILD_NUMBER': cherrypy.config.get('build_number', 0),
            'BUILD_PUSH_NUMBER': cherrypy.config.get('_push_version', 0),
            'LOCALE': i18n.current_lang_url_component(),
            'FLASH_MAJOR_VERSION': cherrypy.config.get('flash_major_version', 0),
            'FLASH_MINOR_VERSION': cherrypy.config.get('flash_minor_version', 0),
            'FLASH_REVISION_VERSION': cherrypy.config.get('flash_revision_version', 0),
            "JS_LOGGER_MODE": cherrypy.config.get('js_logger_mode', 0),
            "JS_LOGGER_MODE_SERVER_END_POINT": cherrypy.config.get('js_logger_mode_server_end_point', '/'),
            "JS_LOGGER_MODE_SERVER_POLL_BUFFER": cherrypy.config.get('js_logger_mode_server_poll_buffer', 100000),
            "JS_LOGGER_MODE_SERVER_MAX_BUFFER": cherrypy.config.get('js_logger_mode_server_max_buffer', 1000),
            'NOCACHE' : cherrypy.config.get('js_no_cache', False),
            'UI_UNIX_START_TIME': int(cherrypy.config.get('start_time', 0)),
            'DEFAULT_NAMESPACE': splunk.getDefault('namespace'),
            'SYSTEM_NAMESPACE': SYSTEM_NAMESPACE,
            'POLLER_TIMEOUT_INTERVAL': cherrypy.config.get('poller_timeout_interval', 600000),
            'SERVER_TIMEZONE_OFFSET': self.getServerTimezoneOffset(),
            'SPLUNKD_FREE_LICENSE': cherrypy.config.get('is_free_license')
        }

    def _get_app_config(self, app):
        try:
            rawConfig = splunk.bundle.getConf('app', namespace=app)
        except splunk.ResourceNotFound:
            return {}
        return {
            'APP_BUILD': rawConfig['install'].get('build', 0)
        }

    @route('/')
    @expose_page(must_login=False, methods='GET')
    def index(self, autoload=False, namespace=None, asDict=False):
        '''
        Returns the configuration information for the main Splunk frontend.
        The values returned from the endpoint are subject to the following:
        
        1) values are idempotent
        2) any time values are in ISO-8601 format
        3) values are typed appropriately
        
        On the JS side, these values are all inserted into a config
        dictionary that is accessible at:
        
            window.$C[<key_name>]
            
        These values should be treated as read-only.
        
        TODO: attach event handlers to value changes on JS side
        
        '''
        
        cherrypy.response.headers['content-type'] = MIME_JSON

        # unauthed calls get nothing
        if not cherrypy.session.get('sessionKey'):
            return '/* unauthorized */'
            
        args = self._get_active_config()

        if namespace:
            args.update(self._get_app_config(namespace))

        # for debug page
        if asDict:
            return args
            
            
        if autoload:
            output = 'window.$C = %s' % json.dumps(args)
            logger.debug('config values: %s' % args)
        else:
            output = json.dumps(args)


        if util.apply_etag(output):
            return None
        else:
            return output

    @route('/:var')
    @expose_page(methods='GET')
    def getvar(self, var, **kw):
        cfg = self._get_active_config()
        if not cfg.has_key(var):
            raise cherrypy.NotFound()
        return str(cfg[var])

    @expose_page(must_login=False, methods='GET')
    def UI_UNIX_START_TIME(self, **kw):
        """/config/UI_UNIX_START_TIME is required without auth by the server restart command"""
        if (kw.get('fmt')=='jsonp' and kw.get('callback')):
            return "%s({start_time: %s})" % (kw['callback'], self.getvar('UI_UNIX_START_TIME'))
        else:
            return self.getvar('UI_UNIX_START_TIME')
    
    """  returns the offset from GMT in seconds  """
    def getServerTimezoneOffset(self) :
        # Somewhat shockingly, this clunky if/else is the official way to get the actual timezone offset, 
        # ie a offset int that is accurate in both DST and non-DST times.
        if (time.localtime()[-1] == 1): 
            return time.altzone
        else:
            return time.timezone
        
