import cherrypy
from splunk.appserver.mrsparkle import *
from splunk.appserver.mrsparkle.controllers.config import ConfigController
from splunk.appserver.mrsparkle.lib.module import *
import splunk.appserver.mrsparkle.lib.util as util
import splunk.appserver.mrsparkle.lib.cached as cached
import logging
import splunk.util
import os, json
import splunk.rest
import splunk.entity as en
import splunk.search as se
import splunk.auth as au
import __main__

import view

logger = logging.getLogger('splunk.appserver.controllers.debug')

class DebugController(BaseController):
    """/debug"""

    #@route('/:p=status')
    #@expose_page(methods='GET')
    def status(self, **args):
        '''
        Provides a debug output page for appserver config
        '''
        hasReadPerms = self._hasReadPerms()
        
        # get overview items
        general = splunk.util.OrderedDict()
        general['Appserver boot path'] = getattr(__main__,'__file__', '<N/A>')
        general['Splunkd URI'] = splunk.mergeHostPath()
        general['Debug Mode'] = __debug__
        
        # get various dicts
        configController = ConfigController()
        uiConfig = configController.index(asDict=True)
        
        mm = moduleMapper
        moduleMap = mm.getInstalledModules()

        uiPanels = splunk.util.OrderedDict()
        uiPanels['config'] = uiConfig
        uiPanels['views'] = en.getEntities(view.VIEW_ENTITY_CLASS, namespace=splunk.getDefault('namespace'))
        uiPanels['modules'] = moduleMap
        uiPanels['cherrypy'] = cherrypy.config
        uiPanels['request'] = args
        uiPanels['wsgi'] = cherrypy.request.wsgi_environ

        splunkdPanels = splunk.util.OrderedDict()

        #code to display splunkd debug information as well 
        try:
            serverResponse, serverContent = splunk.rest.simpleRequest('/services/debug/status', sessionKey=cherrypy.session['sessionKey'])
            atomFeed = splunk.rest.format.parseFeedDocument(serverContent)
            atomFeed_prim = atomFeed.toPrimitive()
            general['Splunkd time'] = splunk.util.getISOTime(atomFeed.updated)
            general['Splunkd home'] = atomFeed_prim.get('SPLUNK_HOME', '&lt;unknown&gt;')

            for key in atomFeed_prim:
                splunkdPanels[key] = atomFeed_prim[key]

        except splunk.AuthenticationFailed, e:
            splunkdPanels['errors'] = 'The appserver is not authenticated with splunkd; retry login'
            
        except splunk.SplunkdConnectionException, e:
            splunkdPanels['errors'] = 'The appserver could not connect to the splunkd instance at: %s' % splunk.mergeHostPath()
            
        except Exception, e:
            splunkdPanels['errors'] = 'Unhandled exception: %s' % str(e)
            
        
        cherrypy.response.headers['content-type'] = MIME_HTML
        
        return self.render_template('debug/status.html', {
            'uiPanels': uiPanels, 
            'splunkdPanels': splunkdPanels,
            'appserverTime': splunk.util.getISOTime(),
            'general': general,
            'hasReadPerms': hasReadPerms
        })


    def _hasReadPerms(self):
        '''
        Use services/server/settings as a proxy for read permissions.
        '''
        
        # NOTE:Due SPL-21113 BETA: unify ACL actions to read/write we cannot use the settings endpoint defer to user admin for now.
        return True if 'admin' == au.getCurrentUser()['name'] else False

        entity = None
        try:
            entity = en.getEntity('/server/', 'settings', namespace=splunk.getDefault('namespace'))
        except Exception, e:
            return False
        if not entity['eai:acl']:
            return False
        if not entity['eai:acl']['perms']:
            return False
        if au.getCurrentUser()['name'] in entity['eai:acl']['perms'].get('read', []):
            return True
        else:
            return False
        
        
    @route('/:p=jsunits')
    def unittests(self, **kwargs):
        '''
        Returns a JSON array of all the HTML jsunit tests in mrsparkle
        
        Ex:
        
        [
            ["all", "/splunk/static/@40691/testing/tests/all.html"], 
            ["ascii_timeline", "/splunk/static/@40691/testing/tests/ascii_timeline.html"], 
            ["job", "/splunk/static/@40691/testing/tests/job.html"]
        ]
        '''
        
        relativePath = '/testing/tests'
        extension = '.html'
        
        testFileDir = util.make_absolute(cherrypy.config['static_dir'] + relativePath)
        logger.debug('Fetching HTML test files from: %s' % testFileDir)
        
        output = []
        for f in os.listdir(testFileDir):
            if f.endswith(extension):
                output.append([f[0:-len(extension)], self.make_url('/static' + relativePath + '/' + f)])
              
        cherrypy.response.headers['content-type'] = MIME_JSON
        return json.dumps(output)


    @route('/:p=reset')
    def reset(self, **kwargs):
        '''
        Resets the user space to a clean state; usually used for testingm
        '''
        has_perms = True if 'admin'==au.getCurrentUser()['name'] else False
        jobs_cancelled = []
        if has_perms and cherrypy.request.method=='POST':
            jobs = se.listJobs()
            for job in jobs:
                j = se.getJob(job['sid'])
                j.cancel()
                jobs_cancelled.append(job['sid'])
        return self.render_template('debug/reset.html', {
            'has_perms': has_perms,
            'method': cherrypy.request.method,
            'jobs_cancelled': jobs_cancelled
        })
    
    
    @route('/:p=echo') 
    def echo(self, **kw):
        '''echos incoming params'''
        cherrypy.response.headers['content-type'] = MIME_TEXT
        output = []
        for key, val in kw.items():
            output.append('Received key: %s, type = %s, value = %s' % (key, str(type(val)), str(val)))
        
        return '\n'.join(output)


    @expose_page()
    def refresh(self, entity=None, **kwargs):
        '''
Forces a refresh on splunkd resources

This method calls a splunkd refresh on all registered EAI handlers that
advertise a reload function.  Alternate entities can be specified by appending 
them via URI parameters.  For example,

    http://localhost:8000/debug/refresh?entity=admin/conf-times&entity=data/ui/manager
    
will request a refresh on only 'admin/conf-times' and 'data/ui/manager'.

1) not all splunkd endpoints support refreshing.
2) auth-services is excluded from the default set, as refreshing that system will
   logout the current user; use the 'entity' param to force it
        '''
        
        # get auto-list of refreshable EAI endpoints
        allEndpoints = en.getEntities('admin', namespace="search")
        eligibleEndpoints = {}
        
        for name in allEndpoints:
            for link in allEndpoints[name].links:
                if link[0] == '_reload':
                    logger.debug('FOUND reload for %s' % name)
                    eligibleEndpoints[name] = allEndpoints[name]
                    break

        if isinstance(entity, list):
            entityPaths = entity
        elif isinstance(entity, basestring):
            entityPaths = [entity]
        else:
            # seed manual endpoints
            entityPaths = [
                'admin/conf-times',
                'data/ui/manager',
                'data/ui/nav',
                'data/ui/views'
            ]
            
            # add capable endpoints
            for name in sorted(eligibleEndpoints.keys()):
                if name not in ['auth-services']: # refreshing auth causes logout
                    entityPaths.append('%s/%s' % (allEndpoints[name].path, allEndpoints[name].name))
            
            
        cherrypy.response.headers['content-type'] = MIME_TEXT

        output = ['Entity refresh control page']
        output.append('=' * len(output[0]))
        output.append("'''")
        output.append(self.refresh.__doc__.strip())
        output.append("'''")
        output.append('')

        # call refresh on each
        for path in entityPaths:
            try:
                en.refreshEntities(path, namespace='search')
                output.append(('Refreshing %s' % path).ljust(40, ' ') + 'OK')
            except Exception, e:
                logger.exception(e)
                msg = e
                if hasattr(e, 'extendedMessages') and e.extendedMessages:
                    msg = e.extendedMessages[0]['text']
                output.append(('Refreshing %s' % path).ljust(43, ' ') + e.__class__.__name__ + ' ' + unicode(msg))
        
        output.append('DONE')
        return '\n'.join(output)
        
        
    @expose_page()
    def memotest(self):

        logger.debug("TEST: first call to getEntities")
        cached.getEntities('saved/searches', 'search')
        logger.debug("TEST: second call to getEntities should be cached")
        cached.getEntities('saved/searches', 'search')
        logger.debug("TEST: third call to getEntities should be cached")
        cached.getEntities('saved/searches', search="search")
        logger.debug("TEST: fourth call to getEntities should be cached")
        cached.getEntities('saved/searches', search="search")
        logger.debug("TEST: fourth call to getEntities should be cached")
        cached.getEntities('saved/searches', search={1:2,3:4,5:6})
        logger.debug("TEST: fourth call to getEntities should be cached")
        cached.getEntities('saved/searches', search={1:2,3:4,5:6})
        
