import cherrypy, urllib, operator, os
from splunk.appserver.mrsparkle import *
import json
import splunk.util
import logging
import splunk.search, splunk.rest, splunk.auth
import splunk.entity as en
import splunk.appserver.mrsparkle.lib.module as module
from splunk.appserver.mrsparkle.lib.jsonresponse import JsonResponse
from splunk.appserver.mrsparkle.lib import i18n
import lxml.etree as et

logger = logging.getLogger('splunk.appserver.controllers.search')

#
# setup logger to obtain actual searches executed by instance; output is sent
# to searches.log file.  Each entry is of the format <username> <search>
#
querylogger = logging.getLogger('splunk.searchhelper')
querylogger.setLevel(logging.DEBUG)
querylogger.propagate = False
queryhandler = logging.FileHandler(os.path.join(os.environ['SPLUNK_HOME'], 'var', 'log', 'splunk', 'searches.log'))
queryhandler.setFormatter(logging.Formatter("%(asctime)s - %(message)s"))
querylogger.addHandler(queryhandler)

EXPORT_HARDLIMIT = 10000
ASSET_EVENTS = 'events'
ASSET_RESULTS = 'results'
ASSET_TIMELINE = 'timeline'
ASSET_SUMMARY = 'summary'
JOB_ACTIONS = ['cancel', 'pause', 'unpause', 'finalize', 'save', 'touch']
REQUIRED_DISPATCH_ARGUMENTS = ['search', 'status_buckets', 'earliest_time', 'latest_time']

class JobsController(BaseController):
    """
    /search/jobs
    
    Manages job control.
    
    Endpoints return a standard json envelope:
    { messages: [{type: 'ERROR', 'text': 'Abe ate the job.'}],
      data: None,
      success: True }
    """

    def __init__(self):
        super(JobsController, self).__init__()

        mm = module.moduleMapper
        self.moduleRoster = mm.getInstalledModules()

    #
    # handles job listing and dispatch
    #


    @route('/', methods='GET')
    @expose_page(handle_api=ONLY_API)
    def listJobs(self, restrictToSession=True, nocache=False, s=None, cachebuster=None):
        '''
        Returns a listing of jobs that the client needs to be aware of;
        listing is restricted by user session, and optionally filtered by
        a whitelist provided by the client
        '''
        
        cherrypy.response.headers['content-type'] = MIME_JSON
        resp = JsonResponse()
        
        # dump out if no jobs are specified
        if not s:
            resp.data = []
            return resp.toJson()
            
        # normalize a single string into a list
        if isinstance(s, basestring): s = [s]

        # loop over all all requested jobs and ask server for status
        listing = []
        for requestSID in s:
            try:
                job = splunk.search.getJob(requestSID)
                listing.append(job.toJsonable())
                
            except splunk.ResourceNotFound:
                listing.append({'sid': requestSID, '__notfound__': True})
                nocache = True # ensure we always bust the cache otherwise, multiple requests may not find out that the job doesn't exist
                resp.addError(_('Splunk could not find a job with sid=%s.') % requestSID)
                
            except Exception, e:
                logger.exception(e)
                resp.success = False
                resp.addError(str(e))
                return resp.toJson()
                
        # normalize the key data
        for item in listing:
            if not item.get('__notfound__'):
                for key in ('eventCount', 'resultCount', 'scanCount'):
                    item[key] = int(item[key])
                for key in ('isDone', 'eventIsStreaming', 'isFinalized', 'isPaused', 'isSaved', 'isSavedSearch'):
                    item[key] = splunk.util.normalizeBoolean(item[key])
        
        # do caching on hash
        if not splunk.util.normalizeBoolean(nocache):

            # generate a copy of the listing without job TTL information; 
            # this is so we can hash contents to send a 304
            staticlisting = listing[:]
            for x in staticlisting:
                if not item.get('__notfound__'):
                    del x['ttl']
                
            staticoutput = json.dumps(staticlisting)

            if util.apply_etag(staticoutput):
                return None

        resp.data = listing
        return resp.toJson()


    @route('/', methods='POST')
    @expose_page(handle_api=ONLY_API)
    def dispatchJob(self, **kwargs):
        '''
        Dispatches a new job
        '''
        if not set(kwargs.keys()) >= set(REQUIRED_DISPATCH_ARGUMENTS):
            raise cherrypy.HTTPError(status=400, message="Missing one or more of the required arguments: 'search', 'statusBucketCount', 'earliestTime', 'latestTime'.")

        # setup the dispatch args
        options = kwargs.copy()
        q = options['search']
        del options['search']

        if 'maxEvents' not in options:
            options['maxEvents'] = EXPORT_HARDLIMIT
        
        # ensure that owner and namespace contexts are passed
        if 'owner' not in options:
            options['owner'] = cherrypy.session['user'].get('name')
        if 'namespace' not in options:
            options['namespace'] = splunk.getDefault('namespace')
            logger.warn('search was dispatched without a namespace specified; defaulting to "%s"' % options['namespace'])

        # Add the default time format
        options['time_format'] = cherrypy.config.get('DISPATCH_TIME_FORMAT')
            
        cherrypy.response.headers['content-type'] = MIME_JSON
        resp = JsonResponse()

        try:
            try:
                userid = cherrypy.session['user']['name']
                querylogger.warn("%s\t%s" % (userid, q))
            except:
                pass
            
            logger.debug('q=%s' % q)
            logger.debug('options=%s' % options)
            job = splunk.search.dispatch(q, sessionKey=cherrypy.session['sessionKey'], **options)
            resp.data = job.id
        except splunk.SplunkdConnectionException, e:
            logger.exception(e)
            resp.success = False
            resp.addFatal(str(e))
        except Exception, e:
            logger.exception(e)
            resp.success = False
            resp.addError(str(e))
            
        logger.debug('dispatch returned %s' % resp)
        return resp.toJson()

        
    # It might be worth revisiting this endpoint since it returns a myriad
    # list of response types.  In its current configuration it can even
    # return JSON in two formats, one from the JsonResponse obj and another
    # directly from splunkd
    @route('/:sid/:asset', methods='GET')
    @expose_page(handle_api=ONLY_API)
    def getJobAsset(self, sid, asset, **kwargs):
        '''
        Returns specific asset for a given job
        '''
        
        resp = JsonResponse()

        # get the job
        try:
            job = splunk.search.getJob(sid, sessionKey=cherrypy.session['sessionKey'])
        except splunk.ResourceNotFound:
            cherrypy.response.headers['content-type'] = MIME_JSON
            cherrypy.response.status = 404
            resp.success = False
            resp.addError("job sid=%s not found" % sid)
            return resp.toJson()

        # set response type; default to XML output
        if 'outputMode' not in kwargs:
            kwargs['outputMode'] = 'xml'
        outputMode = kwargs['outputMode']
        if outputMode == 'json': ct = MIME_JSON
        elif outputMode == 'raw': ct = MIME_TEXT
        elif outputMode == 'csv': ct = MIME_CSV
        elif outputMode == 'xml': ct = MIME_XML
        cherrypy.response.headers['content-type'] = ct
        
        jobFeedRequestMethod = None

        # if we're exporting, set the correct headers, to get the brower to show a download
        # dialog. also hardlimit the export cap to 10,000 events.
        if 'isDownload' in kwargs:
            jobFeedRequestMethod = 'POST'
            if outputMode == 'raw': extension = 'txt'
            else: extension = outputMode

            if kwargs.has_key('filename') and len(kwargs["filename"]) > 0:
                if kwargs['filename'].find('.') > -1:
                    filename = kwargs['filename']
                else:
                    filename = "%s.%s" % ( kwargs['filename'], extension)
            else:
                filename = "%s.%s" % ( sid.replace('.','_'), extension)
            cherrypy.response.headers['content-type'] = 'application/force-download'  
            cherrypy.response.headers['content-disposition'] = 'attachment; filename="%s"' % filename

            if kwargs.has_key('count'):
                try:
                    if int(kwargs['count']) > EXPORT_HARDLIMIT:
                        logger.debug("Hardlimiting exported event count to %s" % EXPORT_HARDLIMIT)
                        kwargs['count'] = str(EXPORT_HARDLIMIT)
                except ValueError:
                    pass
                    
            # by default, exclude underscore fields except time and raw
            if 'field_list' not in kwargs:
                kwargs['field_list'] = [
                    f for f in job.results.fieldOrder 
                    if (f=='_time' or f=='_raw' or not f.startswith('_'))
                ]
            
        # set default time format
        if 'time_format' not in kwargs and 'timeFormat' not in kwargs:
            kwargs['time_format'] = cherrypy.config.get('DISPATCH_TIME_FORMAT')

        # handle errors
        if job.isFailed or len(job.messages) > 0:
            for severity in job.messages:
                for text in job.messages[severity]:
                    resp.addMessage(severity, text)
            if job.isFailed :
                resp.success = False
                if ct == MIME_XML:
                    logger.error('Job was marked as failed during an XML request to %s. Some messages have been suppressed.' % asset)
                else :
                    cherrypy.response.headers['content-type'] = MIME_JSON
                    return resp.toJson()
                    
            else : 
                resp.success = True
            
            
        # pass through the search options
        output = job.getFeed(asset, http_method=jobFeedRequestMethod, **kwargs)
        
        # TODO:
        # handle server-side XSL transforms
        moduleName = cherrypy.request.headers.get('X-Splunk-Module', None)
        if 'moduleName' in kwargs:
            moduleName = kwargs.get('moduleName')
            
        if moduleName and ('xsl' in kwargs) and output:
            
            #logger.debug('search api got xsl request: %s' % moduleName)
            
            # get XSL file
            xslFilePath = os.path.join(self.moduleRoster[moduleName]['path'], kwargs['xsl'])
            f = open(xslFilePath, 'r')
            xslt_doc = et.parse(f)
            f.close()
            
            # generate transformer
            transform = et.XSLT(xslt_doc)
            
            # transform the XML
            xmlDoc = et.fromstring(output)
            transformedOutput = transform(xmlDoc)
            
            cherrypy.response.headers['content-type'] = MIME_HTML

            html = et.tostring(transformedOutput)
            if not html:
                if job.isDone:
                    output = 'No results found'
                else:
                    output = 'Loading...'
            else:
                output = html
        
        # This handles the edge case when output returns no results but
        # a content-type of html is still expected, say by jQuery's $.ajax
        # method.  This could be avoided if the response returned a valid
        # xml document while maintaining a content-type of xml.  Currently
        # empty results are rendered as content-length 0 which jQuery fails
        # on parsing, as it expects xml.
        elif moduleName and ('xsl' in kwargs) and not output:
            logger.debug('Search api got xsl request, but no search results '+
                         'were returned. Setting content-type to html anyway')
            cherrypy.response.headers['content-type'] = MIME_HTML

            if job.isDone:
                output = 'No results found'
            else:
                output = 'Loading...'

        # otherwise, return raw contents
        if util.apply_etag(output):
            return None
        else:
            return output


    @route('/:sid/:ctl=control', methods='POST')
    @expose_page(handle_api=ONLY_API)
    def controlJob(self, sid, ctl, action):
        '''
        Executes control for a given job
        '''
        cherrypy.response.headers['content-type'] = MIME_JSON
        resp = JsonResponse()
        
        # SDK does not have any functionality to change ACLs, so i have 
        # had to add these custom actions for now. 
        # TODO - if/when SDK has support for changing ACL's,  rewrite this 
        # code to not use the Entity class.
        if (action=="makeWorldReadable" or action=="undoWorldReadable") :
            jobPath = "search/jobs/" + sid
            entityName = "acl"
            aclEntity = en.getEntity(jobPath, entityName, namespace="system", owner=splunk.auth.getCurrentUser()['name'])

            if (action == "makeWorldReadable") :    
                aclEntity.properties['perms.read'] = ["*"]
            else :
                aclEntity.properties['perms.read'] = []
            try:
                result = en.setEntity(aclEntity, uri=jobPath + "/acl")
            except Exception, e:
                logger.exception(e)
                resp.addError(_("Splunk could not update permissions for this job"))

            return resp.toJson()

        # get the job
        try:
            job = splunk.search.getJob(sid, sessionKey=cherrypy.session['sessionKey'])
        except splunk.ResourceNotFound:
            cherrypy.response.status = 404
            resp.success = False
            resp.addError(_("Splunk could not find the following job id=%s.") % sid)
            return resp.toJson()

        # check that action is supported
        try:
            actionMethod = getattr(job, action.lower())
        except:
            cherrypy.response.status = 400
            resp.success = False
            resp.addError(_("The following invalid action was passed: %s") % action.lower())
            return resp.toJson()

        wasSuccess = actionMethod()
        resp.data = "%s" % str(wasSuccess).lower()
        return resp.toJson()

    
    @route('/:ctl=control', methods='POST')
    @expose_page(handle_api=ONLY_API)
    def batchControl(self, ctl, sid=None, action=None, **kw):
        
        cherrypy.response.headers['content-type'] = MIME_JSON
        resp = JsonResponse()
        
        if sid == None or action == None or not action in JOB_ACTIONS:
            cherrypy.response.status = 400
            resp.success = False
            resp.addError(_('You must provide a job id(s) and a valid action.'))
            return resp.toJson()
        
        if not type(sid) == list:
            sid = [sid]
        
        resp.data = []
        action = action.lower()
        
        for searchId in sid:
            try:
                job = splunk.search.getJob(searchId, sessionKey=cherrypy.session['sessionKey'])
                actionMethod = getattr(job, action)
                response = actionMethod()
            except splunk.ResourceNotFound, e:
                resp.addError(_('Splunk could not find a job with a job id of %s.') % searchId, sid=searchId)
                response = False
            resp.data.append({'sid': searchId, 'action':action, 'response': response})                
            
        return resp.toJson()


class SearchController(BaseController):
    """
    /search
    """

    # delegate results and jobs to sub-controllres
    jobs = JobsController()

    @expose_page(handle_api=ONLY_API)
    def index(self, **kwargs):
        
        if cherrypy.request.is_api:

            return 'you got api access'
            serverResponse, serverContent = splunk.rest.simpleRequest(
                cherrypy.request.path, 
                cherrypy.session['sessionKey'], 
                #getargs=, 
                #postargs=postargs,
                method=cherrypy.request.method
            )

            for header in serverResponse : 
                cherrypy.response.headers[header] = serverResponse[header]

            return serverContent



        else:
            
            return 'This is the search endpoint'


    @route('/:p=typeahead')
    @expose_page(handle_api=ONLY_API)
    @conditional_etag()
    def typeahead(self, q, index='main', namespace=None, earliest_time=None, latest_time=None, count=None, max_time=None, servers=None, **kwargs):
        '''
        Provides typeahead services.  Currently configured to work with the stock jquery autocomplete plugin:
        http://bassistance.de/jquery-plugins/jquery-plugin-autocomplete/
        '''
        
        cherrypy.response.headers['content-type'] = MIME_JSON

        requestArgs = {
            'output_mode': 'json',
            'prefix': q,
            'count': count,
            'max_time': max_time
        }

        if earliest_time: requestArgs['earliest_time'] = earliest_time
        if latest_time: requestArgs['latest_time'] = latest_time
        
        if servers:
            logger.warn('typeahead server spec not implemented yet')

        uri = splunk.entity.buildEndpoint(
            'search/typeahead', 
            namespace=namespace, 
            owner=cherrypy.session['user'].get('name')
            )
        
        try:
            response, content = splunk.rest.simpleRequest(uri, getargs=requestArgs)
        except Exception, e:
            logger.error('Failed to get typeahead data: %s' % e)
            return json.dumps([{'data': ['Error in typeahead: %s' % e]}])

        # server says no data
        if response.status == 204:
            return '[]'
            
        # rewrite keys to suit UI; currently using jquery autocomplete plugin
        # TODO: update to reflect needs of client-side library
        try:
            output = json.loads(content)
            
            for item in output:
                item['count'] = i18n.format_number(int(item['count']))
                item['result'] = item['content']
                #del item['content']
                item['value'] = '%(content)s (%(count)s)' % item
                item['data'] = ['%(content)s (%(count)s)' % item]

        except Exception, e:
            logger.debug('Error in parsing typeahead data from splunkd: %s' % e)
            output = []

        return json.dumps(output)
