import cherrypy
from cherrypy import expose
from splunk.appserver.mrsparkle import *
from splunk.appserver.mrsparkle.lib import util
import splunk.auth
import logging

# pull in active controllers
from account import AccountController
from admin import AdminController
from config import ConfigController
from debug import DebugController
from lists import ListsController 
from messages import MessagesController
from module import ModuleController
from parser import ParserController
from prototype import PrototypeController
from proxy import ProxyController
from search import SearchController
from tags import TagsController
from utility import UtilityController
from view import ViewController
from searchhelper import SearchHelperController
from ifx import IFXController
from etb import ETBController
from viewmaster import ViewmasterController
from licensing import LicensingController
#from printer import PrintController

# this must be imported after the controllers.
from lib.module import moduleMapper

from splunk.entity import getEntities

logger = logging.getLogger('splunk.appserver.controllers.top')

class APIController(BaseController):
    pass

class TopController(BaseController):
    
    # set base endpoint controllers
    account = AccountController()
    # this change was for the URL structure. file is still admin.py
    manager = AdminController()
    api = APIController()
    app = ViewController()
    config = ConfigController()
    debug = DebugController()
    lists = ListsController()
    messages = MessagesController()
    module = ModuleController()
    parser = ParserController()
    prototype = PrototypeController()
    search = SearchController()
    tags = TagsController()
    splunkd = ProxyController()
    util = UtilityController()
    shelper = SearchHelperController()
    ifx = IFXController()
    etb = ETBController()
    viewmaster = ViewmasterController()
    licensing = LicensingController()
    #locals()['print'] = PrintController() # "print" is a reserved word
    
    @expose
    def admin(self):
        '''
        redirect to manager in case old admin url is hit.
        '''
        self.redirect_to_url('/manager')
    
    @expose
    def index(self):
        '''
        Serves the root of the webserver
        '''
        
        key = cherrypy.session.get('sessionKey')
        if key and splunk.auth.ping(sessionKey=key):
            
            # If the license is expired, redirect to the licensing endpoint.
            # Since we have no way of determining if the user has permissions to change
            # licenses, there is still the chance that a basic user could hit the root
            # endpoint and get redirected to licensing by hitting "/" with an expired license.
            if cherrypy.config['license_state'] == 'EXPIRED':
                return self.redirect_to_url('/licensing')
                
            return self.redirect_to_url('/app/%s' % splunk.auth.getUserPrefs('default_namespace'))
        else:
            return self.redirect_to_url('/account/login')
            
    @expose
    def login(self):
        """Legacy 3.x login url"""
        return self.redirect_to_url('/account/login')

    @expose
    def paths(self):
        """
        Generates an HTML page documenting accessible paths on this site
        and the methods responsible for generating them
        """
        mappings = util.urlmappings(self, cherrypy.request.script_name+'/', exclude=cherrypy.request.script_name+'/api') 
        mappings.sort(lambda a,b: cmp(a['path'], b['path']))
        paths = [ (i, data['path']) for (i, data) in enumerate(mappings) ]
        return self.render_template('top/paths.html', { 'pathnames' : paths, 'mappings' : mappings })

    @expose
    def modules(self):
        """
        Generates an HTML page documenting all registered modules
        """
        definitions = moduleMapper.getInstalledModules()
        names = definitions.keys()
        names.sort()
        return self.render_template('top/modules.html', { 'modules': definitions, 'names': names })

    @expose
    @conditional_etag()
    def help(self, **kwargs):
        """
        Redirects user to context-sensitive help
        """
        return self.render_template('top/help.html', kwargs)

    @expose_page()
    def _bump(self, **kwargs):
        """
        Bumps push_version so that clients are forced to reload static resources.
        Static resources are currently under /static/@12345.  If the bump number
        is non-zero, then the URI becomes /static/@12345.6, where '6' is the
        bump number.
        
        Usage:
        
            POST /splunk/_bump
        """
        
        if cherrypy.request.method == 'POST':
            self.incr_push_version()
            logger.info('appserver static bump number set to %s' % self.push_version())
            return "Version bumped to %i" % self.push_version()
        return """<html><body>Current version: %i<br>
            <form method=\"post\">
            <input type="hidden" name="splunk_session_id" value="%s">
            <input type=\"submit\" value=\"Bump version\">
            </form></body></html>""" % (self.push_version(), cherrypy.session.id)


# Copy TopControllers attributes into the APIController, except for api to avoid recursion
[ setattr(APIController, attr, TopController.__dict__[attr]) for attr in TopController.__dict__ if attr[:2]!='__' and attr!='api' ]
