import cherrypy, logging, datetime, md5
from splunk.appserver.mrsparkle import *
import splunk.util, json

logger = logging.getLogger('splunk.appserver.controllers.util')


class UtilityController(BaseController):
    """/util"""


    @route('/:m=md5')
    @expose_page(must_login=False)
    def md5(self, thing, salt=None, **args):
        '''
        Provides md5 hashing for strings.
        
        Usage:
        
            /util/md5?thing=helloworld ==> fc5e038d38a57032085441e7fe7010b0
            /util/md5?thing=helloworld&salt=23 ==> b6bc495e785755a8385a2f8b4ef12ec7
        '''
        
        if salt:
            return md5.new('%s%s' % (salt, thing)).hexdigest()
        else:
            return md5.new(str(thing)).hexdigest()
        
        
    @route('/:m=strftime')
    @expose_page(must_login=False)
    def strftime(self, ts=None, format=splunk.util.ISO_8601_STRFTIME, offset=None, **args):
        '''
        Provides datetime formatting services.
        
        Args:
            
            ts: A unix timestamp or ISO-8601 time string to process.  Default is now().
            format: A standard strftime time output format string.  Default is ISO-8601.
            offset: An ISO-8601 offset string, i.e. -0700, +12:00
            
        Usage: 
            
            /util/strftime ==> 2008-07-17T15:09:28-0700 (the current time in local timezone)
            /util/strftime?ts=1216332841 ==> 2008-07-17T15:14:01-0700 (passed time, formatted, in local timezone)
            /util/strftime?offset=-0300 ==> 2008-07-17T19:10:04-0300 (the current time in -0300 offset)
            /util/strftime?format=%m/%d/%y ==> 07/17/08 (the current time with formatting applied)
        '''
        
        cherrypy.response.headers['content-type'] = MIME_TEXT
        
        cacheable = True
        
        if offset:
            tz = splunk.util.TZInfo(splunk.util.parseISOOffset(offset), '')
        else:
            tz = splunk.util.TZInfo()
            
        if not ts:
            cacheable = False
            ts = datetime.datetime.now(tz)
            
        else:
            try:
                ts = float(ts)
                ts = datetime.datetime.fromtimestamp(ts, tz)
            except Exception, e:
                logger.warn('Unable to parse as unix time: %s' % e)
                
                try:
                    ts = splunk.util.parseISO(ts)
                except Exception, e:
                    logger.warn('Failed to parse ISO time: %s' % ts)
                    return 'Bad time: %s' % ts

        try:
            output = ts.strftime(format)
        except Exception, e:
            logger.exception(e)
            return 'Unparseable time format: %s' % e 
        
        # set caching rules if a specific time stamp is provided
        if cacheable and cherrypy.config.get('use_future_expires', False):
            cherrypy.response.headers["cache-control"] = 'max-age=31536000' # 1 year
            
        # output
        return output


    @route('/:m=strptime')
    @expose_page(must_login=False)
    def stpftime(self, time, time_format=None, tz=None, **args):

        pass

    @route('/:log/:js')
    @expose_page(must_login=False, methods='POST')
    def log_js(self, log, js, data='[]', **args):
        '''
        Provides javascript logging service.
        
        Args:
            
            data: A JSON encoded collection of JavaScript log messages to process.  Default is an empty collection.

        '''
        cherrypy.response.headers['content-type'] = MIME_JSON
        try:
            logs = json.loads(data)
        except Exception, e:
            logger.error("Could not parse json from javascript logger. Data set was:\n %s" % data)
            return '{"events_logged":"0"}'
        successCount = 0;
        for item in logs:
            event = []
            event.append('name=javascript')
            event.append('class=%s' % item.get('class','undefined'))
            event.append('%s' % item.get('message',''))
            eventFormat = ", ".join(event);
            try:
                getattr(logger, item.get('level', 'info'))(eventFormat)
                successCount = successCount + 1
            except Exception, e:
                logger.error("Could not log javascript event. Event was:\n%s" % eventFormat)
        return '{"events_logged":"%s"}' % successCount

