import os, copy, logging, traceback, time
import re
import urllib
import json
import lxml.etree as et
import splunk.bundle
import splunk.entity as en
import splunk.saved
import splunk.util
import splunk.search.Parser
import splunk.search.Transformer as xformer

from splunk.appserver.mrsparkle import *
from splunk.search.TransformerUtil import *

from splunk.appserver.mrsparkle.lib import module, util, viewconf, viewstate, jsonresponse, i18n, cached
from splunk import search

logger = logging.getLogger('splunk.appserver.controllers.view')

# define the splunkd entity path where views are stored
VIEW_ENTITY_CLASS = 'data/ui/views'

# define path to get current app view organization data
NAV_ENTITY_CLASS = 'data/ui/nav'
NAV_ENTITY_NAME = 'default'
NAV_ALT_ENTITY_NAME_S = 'default-%s'
NAV_CLASS_FREE = 'free'

# define the default view for saved searches.
DEFAULT_DISPLAYVIEW = 'flashtimeline'

class InvalidViewException(Exception):
    pass


class ViewController(BaseController):
    
    # /////////////////////////////////////////////////////////////////////////
    #  Supporting methods
    # /////////////////////////////////////////////////////////////////////////
    
    def __init__(self):
        '''
        Boot up the view controller
        
        The view system renders the views for the current user and namespace.
        Module definitions, including those registered inside apps, are cached
        inside this controller object (the long-lived instance held by
        cherrypy)
        '''
        
        super(ViewController,self).__init__()

        # load all UI modules
        mapper = module.moduleMapper
        self.moduleDefinitions = mapper.getInstalledModules()

        
    def getAppManifest(self):
        '''
        Returns a dict of all available apps to current user
        '''
        
        output = cached.getEntities('apps/local', search=['disabled=false','visible=true'], count=-1)
                
        return output
        
        
    def getViewManifest(self, namespace, currentViewName=None, viewstate_id=None, refresh=0):
        '''
        Returns a flat dict of all available views to current user and namespace.
        Iterate over this output to inspect the properties of the various views.
        '''
        
        output = {}
        
        # get available views to user/app
        try:
            allViews = en.getEntities(VIEW_ENTITY_CLASS, namespace=namespace, refresh=refresh, count=500)
        except splunk.ResourceNotFound:
            return output
            
        for item in allViews:
            try:
                nativeObject = viewconf.loads(allViews[item].get('eai:data'), item)
                output[item] = self._generateViewRoster(item, nativeObject)
            except Exception, e:
                logger.error('Error loading view "%s"' % item)
                logger.exception(e)
                if item == currentViewName:
                    raise


        #
        # handle sticky state params for current view
        # the load sequence (from lowest the highest priority) is:
        # 1) default defined in <module>.conf
        # 2) view configuration XML
        # 3) specified viewstate param set; _current if not specified
        #

        if currentViewName and currentViewName in output:

            currentUser = cherrypy.session['user'].get('name')
            
            try:
                persisted = viewstate.get(currentViewName, viewstate_id=viewstate_id, namespace=namespace, owner=currentUser)

            except splunk.ResourceNotFound:
                # just stop trying to load non-existent viewstates
                logger.warn('getViewManifest - unable to load requested viewstate id=%s' % viewstate_id)
                return output
            
            logger.debug('getViewManifest - loading overlay viewstate id=%s' % viewstate_id)

            rosters = output[currentViewName]['layoutRoster']
            
            keyedParams = output[currentViewName].get('keyedParamMap')
            
            for panelName in rosters:
                for currentModule in rosters[panelName]:
                    
                    moduleId = currentModule['id']
                    if moduleId in persisted.modules:
                        for paramName in persisted.modules[moduleId]:
                            logger.debug('PERSISTENCE - override module=%s paramName=%s: %s ==> %s' % (
                                moduleId,
                                paramName,
                                currentModule['params'].get(paramName, 'UNDEFINED'),
                                persisted.modules[moduleId][paramName]
                            ))
                            currentModule['params'][paramName] = persisted.modules[moduleId][paramName]

        return output
        

    
    def getAppNav(self, app, viewManifest):
        '''
        Retrieves the app nav hierarchy, and generates a tree of links
        
            1.  read in XML definition
            2.  replace run-time tokens, i.e. all unclassified views, recent 
                items, etc.
            3.  return hierarchical ordered list of labels and links to 
                views/searches/links
            
        '''

        navDefinition = None
        navAltClass = None

        # set alternate class
        if cherrypy.config.get('is_free_license'):
            navAltClass = NAV_CLASS_FREE

        # try alternate nav
        if navAltClass:
            try:
                navDefinition = en.getEntity(NAV_ENTITY_CLASS, NAV_ALT_ENTITY_NAME_S % NAV_CLASS_FREE, namespace=app)
            except splunk.ResourceNotFound:
                pass
            
        # if no alt, then proceed with default
        if not navDefinition:
            try:
                navDefinition = en.getEntity(NAV_ENTITY_CLASS, NAV_ENTITY_NAME, namespace=app)
            except splunk.ResourceNotFound:
                logger.warn('"%s" app does not have a navigation configuration file defined.' % app) # TK mgn 06/19/09
            except Exception, e:
                logger.exception(e)
                raise

        # parse the XML
        try:
            parser = et.XMLParser(remove_blank_text=True)
            navDefinition = et.XML(navDefinition['eai:data'], parser)
        except et.XMLSyntaxError, e:
            logger.error('Invalid app nav XML encountered: %s' % e)
            raise Exception, _('The following XML navigation configuration for this app is invalid: %s' % e)
        except Exception, e:
            logger.exception(e)
            raise

        output = []

        # if application has nav defined
        if navDefinition:

            # empty nav means don't do anything; omitted nav is treated down below
            if len(navDefinition) == 0:
                return output

            self._replaceNavTokens(navDefinition, viewManifest, app)
            output = self._decorateNavItems(navDefinition, viewManifest, app)

            # check for the default view; if no default set, pick the first
            # view listed in nav; if none, try to get first in manifest            
            defaultNodes = navDefinition.xpath('//view[@default]')
            for node in defaultNodes:
                if splunk.util.normalizeBoolean(node.get('default')):
                    defaultView = node.get('name')
                    break
            else:
                fallbackNodes = navDefinition.xpath('//view[@name]')
                for node in fallbackNodes:
                    defaultView = node.get('name')
                    break
                else:
                    defaultView = sorted(viewManifest.keys())[0]

            
        # otherwise dump all views into a generic menu
        else:
            logger.warn('Unable to process navigation configuration for app "%s"; using defaults.' % app) # TK mgn 06/19/09
            DEFAULT_VIEW_COLLECTION = 'Default Views'
            output.append({
                'label': DEFAULT_VIEW_COLLECTION, 
                'submenu': [
                    {'label': _(viewManifest[name]['label']), 'uri': name} 
                    for name 
                    in sorted(viewManifest.keys())
                ]
            })
            
            try:
                defaultView = sorted(viewManifest.keys())[0]
            except:
                logger.warn('Unable to determine a default view for app "%s"' % app)
                defaultView = None

        return output, defaultView
        

    def _replaceNavTokens(self, navDefinition, viewManifest, app):
        '''
        Inserts the proper view and saved search items as required by the XML
        nodes placed into the nav XML data.  Modified the 'navDefinition' lxml
        node in-place.

        The XML nodes currently recognized are:
            <view source="unclassified" />
            <view source="all" />
        '''

        try:
            searches = en.getEntities('saved/searches', namespace=app, search='disabled=0', count=500)

        except splunk.ResourceNotFound:
            logger.warn('Unable to retrieve current saved searches')
            searches = {}            
        
        # get a list of explicitly marked views and saved searches
        markedViews = []
        for node in navDefinition.xpath('//view[@name]'):
            markedViews.append(node.get('name'))
        
        markedSaved = []
        for node in navDefinition.xpath('//saved[@name]'):
            if searches.has_key(node.get('name')):
                markedSaved.append(node.get('name'))
            else:
                node.getparent().remove(node)
        #
        # handle views
        # identify the <view source="" /> nodes and fill in with views
        #
        for node in navDefinition.xpath('//view[@source]'):
            source = node.get('source')
            match  = node.get('match', '').lower()

            if source == 'all':
                for viewName in sorted(viewManifest.keys()):
                    if match and viewName.lower().find(match) == -1:
                        continue
                    linkNode = et.Element('view')
                    linkNode.set('name', viewName)
                    node.addprevious(linkNode)
            
            elif source == 'unclassified':
                for viewName in sorted(viewManifest.keys()):
                    if (viewName in markedViews) or (match and viewName.lower().find(match) == -1):
                        continue
                    linkNode = et.Element('view')
                    linkNode.set('name', viewName)
                    node.addprevious(linkNode)
 		    if match:
                       markedViews.append(viewName) 
            
            else:
                logger.warn('Unable to process view item; unknown source: %s' %  source)
                
            node.getparent().remove(node)
            
        #    
        # handle saved searches
        # identify the <saved source="" /> nodes and fill in with the proper
        # saved search items; allow matching on name substring
        #
        for node in navDefinition.xpath('//saved[@source]'):
            source = node.get('source', '').lower()
            match = node.get('match', '').lower()

            if source == 'all':
                keys = splunk.util.objUnicode(searches.keys())
                for savedName in sorted(keys, key=unicode.lower):
                    if match and savedName.lower().find(match) == -1:
                        continue
                    savedNode = et.Element('saved')
                    savedNode.set('name', savedName)
                    savedNode.set('sharing', searches[savedName].get('eai:acl',{}).get('sharing'))
                    if node.get('view'):
                        savedNode.set('view', node.get('view'))
                    node.addprevious(savedNode)

            elif source == 'unclassified':
                keys = splunk.util.objUnicode(searches.keys())
                for savedName in sorted(keys, key=unicode.lower):
                    if savedName not in markedSaved:
                        if match and savedName.lower().find(match) == -1:
                            continue
                        savedNode = et.Element('saved')
                        savedNode.set('name', savedName)
                        savedNode.set('sharing', searches[savedName].get('eai:acl',{}).get('sharing'))
                        if node.get('view'):
                            savedNode.set('view', node.get('view'))
                        node.addprevious(savedNode)
                        if match:
                            markedSaved.append(savedName)

            else:
                logger.warn('Unable to process saved search item; unknown source: %s' % source)
                
            node.getparent().remove(node)
        
        
        
    def _decorateNavItems(self, branch, viewManifest, app):
        '''
        Rewrites the incoming nav definition by decorating view names with
        proper links, and saved searches as views with search name specified.
        This recursive method is used by getAppNav().
        
        Input Example:
            <nav>
                <collection label="Dashboards">
                    <a href="http://google.com">Google</a>
                </collection>
                <collection label="Views">
                    <view source="all" />
                </collection>
                <collection label="Saved Searches" sort="alpha">
                    <collection label="Recent Searches">
                        <saved source="recent" />
                    </collection>
                    <saved name="All firewall errors" />
                    <divider />
                </collection>
            </nav>
        
        Output Example:
        
        
        '''
        
        output = []
        for node in branch:
            
            # update the view nodes with the proper links and labels
            if node.tag == 'view':
                viewData = viewManifest.get(node.get('name'))
                if viewData:
                    if viewData['isVisible']:
                        output.append({
                            'label': _(viewData.get('label')), 
                            'uri': self.make_url('/app/%s/%s' % (app, node.get('name', '')))
                        })
                else:
                    logger.warn(_('An unknown view name \"%(view)s\" is referenced in the navigation definition for \"%(app)s\".') % {'view': node.get('name'), 'app': app})
                    
            # update saved searches and point them to the saved search redirector
            elif node.tag == 'saved':
                if node.get('view'):
                    uri = self.make_url(
                        '/app/%s/%s' % (app, node.get('view')), 
                        {'s': node.get('name')}
                    )
                else:
                    uri = self.make_url(
                        '/app/%s/@go' % app, 
                        {'s': node.get('name')}
                    )
                output.append({
                    'label': node.get('name'), 
                    'uri': uri,
                    'sharing': node.get('sharing', None)
                })
                
            elif node.tag == 'a':
                uri = node.get('href')
                output.append({
                    'label': _(node.text), 
                    'uri': self.make_url(uri) if uri.startswith('/') else uri
                })

            elif node.tag == 'divider':
                output.append({
                    'label': '------', 
                    'uri': '#', 
                    'divider': 'actionsMenuDivider'
                })
                
            elif node.tag == 'collection':
                subcollection = {
                    'label': _(node.get('label')), 
                    'submenu': self._decorateNavItems(node, viewManifest, app)
                }
                
                # only show submenu if it contains something
                if len(subcollection['submenu']) > 0:
                    output.append(subcollection)

        return output
        
        
    def getAppConfig(self, appName, currentViewName=None, build_nav=True):
        '''
        Returns a dict of properties for appName.
        '''

        # determine viewstate set
        viewstate_id = None
        if 'vs' in cherrypy.request.params:
            viewstate_id = cherrypy.request.params['vs']
        
        elif 's' in cherrypy.request.params:
            try:
                ss = en.getEntity('saved/searches', cherrypy.request.params['s'], namespace=appName)
                viewstate_id = ss.get('vsid', None)
                logger.debug('getAppConfig - saved search has viewstate=%s' % viewstate_id)
            except splunk.ResourceNotFound:
                viewstate_id = None

        elif 'sid' in cherrypy.request.params:
            savedSearchObject = splunk.saved.getSavedSearchFromSID(cherrypy.request.params.get('sid'))
            if savedSearchObject and len(savedSearchObject.get('vsid', '')) > 0:
                viewstate_id = savedSearchObject['vsid']

        # set default output
        output = {
            'is_visible': False,
            'label': appName,
            'nav': {},
            
            # TODO: this should be in a method dedicated to the
            # current view, not the app config
            'viewstate_id': viewstate_id,
            
            'available_views': self.getViewManifest(appName, currentViewName, viewstate_id=viewstate_id)
        }

        # verify presence of app config
        try:
            rawConfig = splunk.bundle.getConf('app', namespace=appName)
        except splunk.ResourceNotFound:
            logger.warn(_('Splunk cannot load app "%s" because it could not find a related app.conf file.') % appName)
            return output

        # check pre-reqs
        if 'ui' not in rawConfig:
            logger.warn(_('Splunk found an app.conf for for app "%s", but no UI stanza was present.') % appName)
            return output

        # get app nav
        if build_nav:
            output['nav'], tmp_dv = self.getAppNav(appName, output['available_views'])
        else:
            output['nav'] = {}
        
        # pick up the remaining app config items
        if rawConfig['ui'].get('is_visible'):
            output['is_visible'] = splunk.util.normalizeBoolean(rawConfig['ui'].get('is_visible'))
        if rawConfig['ui'].get('label', appName):
            output['label'] = rawConfig['ui'].get('label')
        
        return output
        
        
    def _generateViewRoster(self, viewName, viewConfig):
        '''
        Returns a dict of properties for viewName, to be used by all the view
        components.

            layoutRoster: a dict of modules, keyed by their layoutPanel
                          assigment.
            activeModules: a dict of all modules requested by viewName
        
        SAMPLE OUTPUT:

            {
                layoutRoster: ...
                activeModules: ...
            }
        '''
        logger.debug("template for view " + viewName + " is " + viewConfig.get('template', "undefined"))

        output = {
            'layoutRoster': {}, 
            'activeModules': {}, 
            'label': viewConfig.get('label', viewName),
            'isVisible'  : viewConfig.get('isVisible', True),
            'displayView': viewConfig.get('displayView', viewName),
            'refresh'    : viewConfig.get('refresh', None),
            'template'   : viewConfig.get('template', 'search.html'),
            'onunloadCancelJobs': viewConfig.get('onunloadCancelJobs', True),
            'autoCancelInterval': viewConfig.get('autoCancelInterval'),
            'stylesheet': viewConfig.get('stylesheet', None),
            'objectMode': viewConfig.get('objectMode')
        }
        
        localUsageCounter = {}

        # loop over all the modules requested in view config
        for module in viewConfig['modules']:

            # process the module config through generator, and collate them
            # into the layout panels, and identify as active module
            for item in self._moduleGenerator(module, panelName=viewName, usageCounter=localUsageCounter):
                if not output['layoutRoster'].get(item.get('layoutPanel')):
                    output['layoutRoster'][item.get('layoutPanel')] = []
                output['layoutRoster'][item.get('layoutPanel')].append(item)
                if 'inheritance' in item:
                    for inheritedClass in item['inheritance']:
                        output['activeModules'][inheritedClass] = 1
                if 'include' in item: 
                    for includedClass in item['include']:
                        output['activeModules'][includedClass] = 1


        output['activeModules'] = self._sortModuleNames(output['activeModules'])
            
        return output
        
        
    def _sortModuleNames(self, moduleList):
        '''
        Returns a module list, sorted by inheritance order.  Called from
        _generateViewRoster().
        '''
        
        output = []
        for moduleName in self.moduleDefinitions:
            if moduleName in moduleList:
                output.append(moduleName)
                
        return output
            
            
    def _moduleGenerator(self, moduleConfig, position=0, depth=0, parentLayout=None, parentId=None, panelName=None, usageCounter=None, isHidden=False):
        '''
        Yields every module that is listed in moduleConfig.  The dict yielded 
        for each module looks like:
        
            {
                'parentmodule': '#AsciiTimeline_0_0',
                'inheritance': [
                    'Splunk.Module',
                    'Splunk.Module.SearchModule',
                    'Splunk.Module.SimpleFieldViewer'
                ],
                'className': 'SimpleFieldViewer',
                'layoutPanel': 'fullWidthControls',
                'id': 'SimpleFieldViewer_1_1',
                'templatePath': '../../modules/field_viewer/field_viewer.html',

                'params': {
                    'paramName1': {
                        'default': 'defaultValue1',
                        'required': True,
                        'values': ['acceptableValue1','acceptableValue2','acceptableValue3']
                    }
                },
                'stickyParams': ['key1','key2','key3'],
                'persistableParams': ['key1','key2','key3'],
                
                // custom param fields...
                'count': 10,
                'field': 'twikiuser'
            }
        
        'id' is of the form: <module_name>_<module_usage_count>_<tree_depth>_<sibling_position>
        'templatePath' is the relative URL (mako seems to need it) of the template
        'parentmodule' is the DOM id of parent; not present if a top level module
        
        '''
        
        # first lookup the module definition
        # TODO: the module prefix is patched in here
        moduleDefinition = self.moduleDefinitions.get("Splunk.Module." + moduleConfig['className'], {})
        
        # generate module config for current module
        output = moduleConfig.copy()

        # add the module persistence information
        output['stickyParams'] = moduleDefinition.get('stickyParams', [])
        output['persistableParams'] = moduleDefinition.get('persistableParams', [])

            
        output.setdefault('params', {})

        # merge default params supplied by the module's manifest file with the view's params
        if moduleDefinition:

            # check that required params defined in .conf are present
            for param_name, param in moduleDefinition['params'].items():
                
                if not output['params'].has_key(param_name):
                    if param['default'] is not None:
                        output['params'][param_name] = param['default']
                    elif param['required']:
                        lib.message.send_client_message(
                            'error', 
                            _("Misconfigured view '%(panel)s' - Parameter '%(param)s' is required for module %(module)s") % 
                            {
                                'panel':panelName, 
                                'param':param_name, 
                                'module':moduleConfig['className']
                            }
                        )

                if param['values'] and param['required'] and output['params'].get(param_name) not in param['values']:
                    lib.message.send_client_message(
                        'error', 
                        _("Misconfigured view '%(panel)s' - Parameter '%(param)s' must be set to one of %(values)s for module %(module)s") % 
                        {
                            'panel':panelName, 
                            'param':param_name, 
                            'values': param['values'],
                            'module': moduleConfig['className']
                        }
                    )

            # check if the module config defines wildcarded parameters
            wildcardPrefixes = [x[0:-1] for x in moduleDefinition['params'] if x.endswith('*')]
            
            # the reverse check - enforce that all configured params are documented in the conf.
            for param_name in output['params'] : 
                
                # ensure that params that match wildcards are allowed through
                if any(map(param_name.startswith, wildcardPrefixes)):
                    continue
                
                if not moduleDefinition['params'].has_key(param_name) and param_name not in {"group":1, "groupLabel": 1, "autoRun":1, "altTitle":1} :
                    lib.message.send_client_message(
                        'error', 
                        _("Misconfigured view '%(panel)s' - Unknown parameter '%(param)s' is defined for module %(module)s. Make sure the parameter is specified in %(module)s.conf.") % 
                        {
                            'panel':panelName, 
                            'param':param_name, 
                            'module':moduleConfig['className']
                        }
                    )

                if moduleDefinition['params'].has_key(param_name) and moduleDefinition['params'][param_name]['translate']:
                    translate = moduleDefinition['params'][param_name]['translate']
                    if translate == 'string':
                        output['params'][param_name] = _(output['params'][param_name])
                    else:
                        i18n.translate_view_params(translate, output['params'][param_name])
                        

        # assign a unique DOM ID to this module
        usageCounter.setdefault(moduleConfig['className'], -1)
        usageCounter[moduleConfig['className']] += 1
        output['id'] = '%s_%s_%s_%s' % (
            moduleConfig['className'], 
            usageCounter[moduleConfig['className']], 
            depth, 
            position)

        output['isHidden'] = isHidden
        
        output['layoutPanel'] = moduleConfig.get('layoutPanel') or parentLayout

        if not output['layoutPanel']:
            raise Exception, _('Splunk cannot load the specified view because a layoutPanel is not defined for module %s.') % output['id']

        if parentId:
            output['parentmodule'] = '#%s' % parentId
        if 'html' in moduleDefinition:
            output["templatePath"] = '=' + moduleDefinition['html'] # use '=' prefix to denote absolute template path
        if 'inheritance' in moduleDefinition:
            output['inheritance'] = moduleDefinition['inheritance']
        if 'include' in moduleDefinition:
            output['include'] = moduleDefinition['include']
        if 'children' in output: del output['children']
        yield output

        # process children
        if 'children' in moduleConfig:
            for i, child in enumerate(moduleConfig['children']):
                hideChild = True if (isHidden or splunk.util.normalizeBoolean(output['params'].get('hideChildrenOnLoad'))) else False
                childOutput = self._moduleGenerator(child, i, depth+1, output['layoutPanel'], output['id'], panelName=panelName, usageCounter=usageCounter, isHidden=hideChild)
                if isinstance(childOutput, dict):
                    yield childOutput
                else:
                    for x in childOutput: yield x


    def filePathToUrl(self, filePath, pivotPath='modules'):
        '''
        Converts a filesystem path to a relative URI path, on the assumption that
        both paths contain a common pivot path segment
        '''

        parts = filePath.split(os.sep)
        try: pivot = parts.index(pivotPath)
        except: return False
        return '/' + os.path.join(*parts[pivot:]).replace('\\', '/')
    
    def _getTermsForSavedSearch(self, savedSearchName, namespace) :
        savedSearch = en.getEntity("saved/searches", savedSearchName, namespace=namespace)
        return savedSearch.get('qualifiedSearch', False)

    
        
    # /////////////////////////////////////////////////////////////////////////
    #  Main route handlers
    # /////////////////////////////////////////////////////////////////////////
        
    @route('/:app')
    @expose_page()
    def appDispatcher(self, app):
        '''
        Redirect user to the default view, as specified in the nav XML
        '''
        
        apps = self.getAppManifest()
        #validate app exists
        if apps.has_key(app):
            # locate default view
            views = self.getViewManifest(namespace=app, refresh=1)
            nav, defaultView = self.getAppNav(app, views)
            return self.redirect_to_url('/app/%s/%s' % (app, defaultView))
        else:
            cherrypy.response.status = 404
            return self.render_template('view/app_error.html', {'app':app, 'apps':apps})

    
    @route('/:app/:p=@go')
    @expose_page()
    def redirect(self, app, **kwargs):
        '''
        Dispatches generic object requests to the appropriate view.  Currently
        used for redirecting saved searches and sids

        TODO: this needs to be app configurable
        '''
        view = None

        def getViewForSavedSearch(savedSearchName):
            try:
                saved = en.getEntity('saved/searches', savedSearchName, namespace=app)
                return saved.get('displayview', saved.get('view' , DEFAULT_DISPLAYVIEW))
            except splunk.ResourceNotFound, e:
                raise cherrypy.HTTPError(404, _('The following requested saved search is unknown: "%s". As a result, Splunk is unable to redirect to a view.') % savedSearchName)

        def redirect(app, view, options):
            return self.redirect_to_url('/app/%s/%s' % (app, view), _qs=options)

        if 's' in kwargs:
            view = getViewForSavedSearch(kwargs['s'])

            # removes the sid explicitly if one is set, this allows @go to follow the regular view.py behavior
            if 'sid' in kwargs:
                del kwargs['sid']

        elif 'sid' in kwargs:
            try:
                job = splunk.search.getJob(kwargs['sid'])
            except splunk.ResourceNotFound, e:
                raise cherrypy.HTTPError(404, _('Could not find a job with an sid of "%(sid)s".  Cannot redirect to a view.') % {'sid': kwargs['sid']})
           
            if job.isSavedSearch and job.label:
                view = getViewForSavedSearch(job.label)

            else:
                try:
                    request = job.request
                    view = request.get('ui_dispatch_view', False)
                except AttributeError:
                    pass

                if not view:
                    view = DEFAULT_DISPLAYVIEW
                    
        if view:
            if 'p' in kwargs:
                del kwargs['p']

            if kwargs.get('media') == 'print' and view == 'report_builder_display':
                view = 'report_builder_print'

            redirect(app, view, kwargs)

        raise cherrypy.HTTPError(400, _('The requested object is unknown. You must provide an sid or saved search as a query param. e.g. ?sid=<sid> or ?s=<saved search name>.'))


    def buildViewTemplate(self, app, view_id, action=None, q=None, sid=None, s=None, earliest=None, latest=None, remote_server_list=None, render_invisible=False, build_nav=True):
        """
        Build the template args required to render a view
        The render() handler below calls this, as do other controllers that need to embed a view 
        into their own templates (see the AdminController for example)
        """

        # assert on view name; may only be alphanumeric, dot, dash, or underscore
        view_id_checker = re.compile('^[\w\.\-\_]+$')
        if not view_id_checker.match(view_id):
            raise cherrypy.HTTPError(400, _('Invalid view name requested: "%s". View names may only contain alphanumeric characters.') % view_id)
        
        
        # get list of all UI apps
        appList = self.getAppManifest()
        
            # get current app configuration
        appConfig = self.getAppConfig(app, view_id, build_nav)
        
        if not appConfig['is_visible'] and not render_invisible:
            raise cherrypy.HTTPError(404, _('App "%s" does not support UI access.  See its app.conf for more information.') % appConfig['label'])
        
        # get all views for current user in app context
        availableViews = appConfig['available_views']
        if len(availableViews) == 0:
            raise cherrypy.HTTPError(404, _('App "%s" does not have any available views.') % app)
        
        # get template layout for current config
        currentViewConfig = availableViews.get(view_id)
        if not currentViewConfig:
            raise cherrypy.HTTPError(404, _('Splunk cannot find the  "%s" view.') % view_id)

        # get asset rosters
        moduleMap = currentViewConfig.get('layoutRoster', {})
        cssList = []
        jsList = []
        for moduleName in currentViewConfig.get('activeModules', []):
            moddef = self.moduleDefinitions[moduleName]
            if 'css' in moddef: cssList.append(self.filePathToUrl(moddef['css']))
            if 'js' in moddef: jsList.append(self.filePathToUrl(moddef['js']))

        # compile link list for all available views
        viewList = [{'label': availableViews[k].get('label', k), 'uri': self.make_url('/app/%s/%s' % (app,k))} for k in availableViews]
        viewList.sort(lambda x,y: cmp(x['label'], y['label']))
        
        # gather application specific static content
        # allowed css is 'application.css', then any css declared in viewConfig
        customCssList = []
        printCssList = []
        
        allowedCss = [currentViewConfig.get('stylesheet')]
        # manually exclude core views from getting skins
        if view_id not in ('_admin', 'job_management'):
            allowedCss.append('application.css')

        allowedPrintCss = ['print.css']
        from lib.apps import local_apps
        if local_apps[app].has_key('static'):
            appstatics = local_apps[app]['static'].get('css', [])
            for filename in appstatics:
                if filename in allowedCss:
                    customCssList.append("/static/app/%s/%s" % (app,filename))
                elif filename in allowedPrintCss:
                    printCssList.append("/static/app/%s/%s" % (app,filename))

        # translate the view label
        label = currentViewConfig.get('label')
        if label:
            label = _(label)
        else:
            label = '(%s)' % view_id
        
        # Safely get the displayView. Somehow this gets set to None elsewhere
        # which causes some problems further up in the stack
        displayView = currentViewConfig.get('displayView')
        if displayView == None:
            displayView = view_id
        
        # assemble the template vars
        templateArgs = {
        
            # define standard params
            'APP': {'id': app, 'label': appConfig['label']},
            'VIEW': {'id': view_id,
                     'label': label,
                     'displayView': displayView,
                     'refresh': currentViewConfig.get('refresh', 10),
                     'viewstateId': appConfig['viewstate_id'] or '', # TODO: relocate me
                     'onunloadCancelJobs': currentViewConfig.get('onunloadCancelJobs'),
                     'autoCancelInterval': currentViewConfig.get('autoCancelInterval'),
                     'template': currentViewConfig.get('template', []),
                     'objectMode': currentViewConfig.get('objectMode')
                    },
            'DATA': {'mode': None},
            'toBeResurrected': None,
            
            # define HTML asset params
            'navConfig': appConfig['nav'],
            'appList': appList,
            'viewList': viewList,
            'modules':  moduleMap,
            "cssFiles" : cssList,
            "customCssFiles" : customCssList,
            "printCssFiles" : printCssList,
            "jsFiles"  : jsList
        }

        # PROCESSING ARGUMENTS FOR PERMALINK.
        
        # 0. Verify that between s, q, and sid,  only one is actually specified. 
        # NOTE - when we have the ability for messaging here to send warnings down to the user, we should just 
        # pick the first one specified, between s, q, sid.   And if a second or third is specified,  
        # then maybe check whether it's all still consistent,   but write some stuff that tells the user either 
        # that the args were inconsistent,  or that some of the args were totally ignored. 
        if s and (q or sid):
            raise Exception, _("If you specify s (saved search name), you cannot also specify q or sid (job id).")
        if sid and (q or s):
            raise Exception, _("If you specify sid (job id), you cannot also specify q or s (saved search name).")
        if q and (s or sid):
            raise Exception, _("If you specify q, you cannot also specify s (saved search name) or sid (job id).")
        
        # 1. Saved search name, aka 's'.
        if (s) :
            try:
                # We use the SDK to get a savedSearch object
                savedSearchObject = search.getSavedSearch(s, namespace=app, owner=cherrypy.session['user'].get('name'))
            except splunk.ResourceNotFound: 
                lib.message.send_client_message('error', _('The saved search "%(savedSearchName)s" could not be found.') % {'savedSearchName': s})
            else:               
                # ensure that the saved search has a viewstate object to save stuff to
                # -- if saved search is owned by another user and is publically read-only,
                #    the auto viewstate will fail
                # TODO #1 it seems odd to do this here rather than in resurrectFromSavedSearch
                #      for instance HiddenSavedSearch will now not get the same treatment. 
                #      it's probably OK, but seems inconsistent.
                if savedSearchObject.get('vsid') == None:
                    newVsid = viewstate.generateViewstateId(make_universal=True)
                    self.setViewstate(app, view_id, newVsid, _is_shared=True, _is_autogen=True)
                    savedSearchObject['vsid'] = newVsid
                    logger.info('Saved search "%s" has no viewstate; auto generating vsid=%s' % (s, newVsid))
                    try:
                        en.setEntity(savedSearchObject)
                    except splunk.AuthorizationFailed:
                        logger.warn('loaded a saved search without a viewstate; current user not authorized to generate viewstate')
                    except splunk.ResourceNotFound, e:
                        logger.warn('There was an error generating the view state object. The view will likely continue to work, but will not be able to persist its state.')
                    except Exception, e:
                        logger.exception(e)
                        
                context = util.resurrectFromSavedSearch(
                    savedSearchObject=savedSearchObject,
                    hostPath=self.splunkd_urlhost,
                    namespace=app,
                    owner=cherrypy.session['user'].get('name')
                )
                
                # TODO #2.  And obviously by pulling it up we would delete this line
                context["vsid"] = savedSearchObject.get('vsid')

                templateArgs['DATA'] = {'mode': 'saved', 'name': s, 'vsid': context["vsid"]}
                templateArgs["toBeResurrected"] = json.dumps(context)

        # 2. search Id.        
        if (sid) :
            try :
                job = splunk.search.getJob(sid=sid,sessionKey=cherrypy.session['sessionKey'])
                jsonableJob = job.toJsonable(timeFormat='unix')
                
                # See -- SPL-24020  Saving a report using relative time ranges such as "Previous business week" gives me epoch earliest and lastest times in the save dialog    
                # to fix SPL-24020 it now only uses the earliestTime, latestTime from the job when the job was dispatched by the scheduler. 
                #    Note: earliest/latest from the job are always absolute epochTime.
                #    in all other cases it will use the earliest, latest args from the client, which may be relative time terms. 
                #    see SPL-24020 for further comments.
                earliestTime = None
                latestTime   = None
                if ("delegate" in jsonableJob and jsonableJob["delegate"]) :
                    earliestTime = jsonableJob["earliestTime"]
                    latestTime   = jsonableJob["latestTime"]
                elif ("request" in jsonableJob and jsonableJob["request"]) :
                    earliestTime = jsonableJob["request"].get("earliest_time", None)
                    latestTime   = jsonableJob["request"].get("latest_time", None)
                else :
                    logger.error("Found job with no delegate that also had no request parameter. Unable to resurrect earliest and latest times.")
                    
                
                context = util.resurrectSearch(
                    hostPath = self.splunkd_urlhost,
                    q = job.search, 
                    earliest = earliestTime, 
                    latest   = latestTime, 
                    remote_server_list = None, # job.searchProviders,
                    namespace = app,
                    owner=cherrypy.session['user'].get('name')
                )
                
                # try to get the relevant viewstate; first check the URI, if not
                # passed, then check if job was dispatched from a saved search;
                # if so, try to pull the persisted viewstate id
                context["job"] = jsonableJob
                savedSearchName = None
                if 'vs' in cherrypy.request.params:
                    context["vsid"] = cherrypy.request.params['vs']
                else:
                    savedSearchObject = splunk.saved.getSavedSearchFromSID(sid)
                    if savedSearchObject:
                        savedSearchName = savedSearchObject.name
                        if len(savedSearchObject.get('vsid', '')) > 0:
                            context['vsid'] = savedSearchObject['vsid']
                        

                templateArgs['DATA'] = {'mode': 'sid', 'name': savedSearchName, 'sid': sid, 'vsid': context.get('vsid')}
                templateArgs["toBeResurrected"] = json.dumps(context)


            except splunk.ResourceNotFound : 
                lib.message.send_client_message('error', _('Splunk cannot find a job with an sid = %(sid)s. It may have expired.') % {'sid': sid})
            
            
        # 3) straight up search language string + any manually passed in getargs.
        elif (q):
            logger.debug("DECOMPOSITION > %s" % q)
            context = util.resurrectSearch(
                hostPath = self.splunkd_urlhost,
                q = q, 
                earliest = earliest, 
                latest = latest, 
                remote_server_list = remote_server_list,
                namespace = app,
                owner = cherrypy.session['user'].get('name')
            )
            if 'vs' in cherrypy.request.params:
                context["vsid"] = cherrypy.request.params['vs']

            templateArgs["toBeResurrected"] = json.dumps(context)
            logger.debug("DECOMPOSITION < %s" % templateArgs["toBeResurrected"])
            
        return templateArgs   
                


    @route('/:app/:view_id')
    @expose_page(handle_api=True)
    def render(self, app, view_id, action=None, q=None, sid=None, s=None, earliest=None, latest=None, remote_server_list=None, **kw):
        '''
        Handle main view requests
        '''

        #
        # DEBUG
        #

        if kw.get('showsource'):
            templateArgs = {'viewConfig': viewconf.loads(en.getEntity(VIEW_ENTITY_CLASS, view_id, namespace=app).get('eai:data'), view_id)}
            return self.render_template('/view/pyramid.html', templateArgs)
            
        if kw.get('showtree'):
            viewConfig = viewconf.loads(en.getEntity(VIEW_ENTITY_CLASS, view_id, namespace=app).get('eai:data'), view_id)
            
            def convertModuleToJit(module):
                newModule = {
                    'id': '',
                    'name': '',
                    'children': []
                }
                
                if 'className' in module:
                    newModule['name'] = module['className']
                    if module['className'] in seenModules:
                        seenModules[module['className']] += 1
                    else:
                        seenModules[module['className']] = 1
                    newModule['id'] = '_'.join([module['className'], str(seenModules[module['className']])])

                if 'children' in module:
                    for childModule in module['children']:
                        newModule['children'].append(convertModuleToJit(childModule))
                return newModule

            
            seenModules = {}
            moduleTree = [{'id': 'root', 'name': 'root', 'children': []}]
            if 'modules' in viewConfig.keys():
                for module in viewConfig['modules']:
                    moduleTree[0]['children'].append(convertModuleToJit(module))
                    
            templateArgs = {'moduleTree': moduleTree}
            return self.render_template('/view/tree2.html', templateArgs)

        time1 = time.time()
        templateArgs = self.buildViewTemplate(app, view_id, action, q, sid, s, earliest, latest, remote_server_list)
        time2 = time.time()
        output = self.render_template('/view/' + templateArgs['VIEW']['template'], templateArgs)
        time3 = time.time()
        logger.info('PERF - viewTime=%ss templateTime=%ss' % (round(time2-time1, 4), round(time3-time2, 4)))
        if cherrypy.request.is_api:
            # set non json serializable types to None revisit if required.
            for i in templateArgs['appList']:
                templateArgs['appList'] = None
            templateArgs['make_url'] = None
            templateArgs['make_route'] = None
            templateArgs['attributes'] = None
            templateArgs['controller'] = None
            return self.render_api(templateArgs)
        else:
            cherrypy.response.headers['content-type'] = MIME_HTML
            return output
    

    @route('/:app/:view_id/:viewstate_id', methods='GET')
    @expose_page(handle_api=True, methods='GET')
    def getViewstate(self, app, view_id, viewstate_id):
        '''
        Returns a JSON structure representing all of the module params for
        view_id.
        '''

        try:
            viewManifest = self.getViewManifest(namespace=app, currentViewName=view_id, viewstate_id=viewstate_id)
        except splunk.ResourceNotFound:
            raise cherrypy.HTTPError(status=404, message=(_('Viewstate not found; view=%s viewstate=%s') % (view_id, viewstate_id)))

        
        currentView = viewManifest[view_id]

        output = {}

        for panelName in currentView['layoutRoster']:
            for currentModule in currentView['layoutRoster'][panelName]:
                output[currentModule['id']] = currentModule.get('params', {})


        cherrypy.response.headers['content-type'] = MIME_JSON
        return json.dumps(output)
        
        
        
    @route('/:app/:view_id/:viewstate_id')
    @expose_page(handle_api=True, methods=['GET','POST'])
    def setViewstate(self, app, view_id, viewstate_id, _is_shared=False, _is_autogen=False, **form_args):
        '''
        Persists module params to the specific viewstate object.  Writes are
        done in an overlay fashion; unspecified params will not be overwritten.
        Parameters are accepted in the following format:

            <module_DOM_id>.<param_name>=<param_value>

        Ex:

            SearchBar_0_0_0.useTypeahead=true

        Only string values are accepted; complex data types are not persistable.
        '''

        cherrypy.response.headers['content-type'] = MIME_JSON
        output = jsonresponse.JsonResponse()

        # determine the desired viewstate
        altView, vsid = viewstate.parseViewstateHash(viewstate_id)
        if altView == None:
            altView = view_id

        # scaffold object mapper
        vs = viewstate.Viewstate()
        vs.namespace = app
        vs.view = altView
        vs.id = vsid
        vs.owner = cherrypy.session['user'].get('name')

        # add in stub property
        if _is_autogen and len(form_args) == 0:
            form_args['is.autogen'] = 1

        if len(form_args) == 0:
            logger.warn('setViewstate - no parameters received; nothing persisted')
            output.success = False
            output.addError(_('No parameters received; aborting'))
            return output.toJson()
            
        # insert all passed params
        for key in form_args:
            parts = key.split('.', 2)
            if len(parts) != 2:
                logger.warn('setViewstate - invalid viewstate param name: %s; aborting' % key)
                output.success = False
                output.addError(_('Invalid viewstate param name: %s; aborting') % key)
                return output.toJson()

            vs.modules.setdefault(parts[0], {})
            vs.modules[parts[0]][parts[1]] = form_args[key]
            logger.debug('setViewstate - setting module=%s param=%s value=%s' % (parts[0], parts[1], form_args[key]))

        # commit
        try:
            viewstate.commit(vs)
        except Exception, e:
            logger.exception(e)
            output.success = False
            output.addError(str(e))
            return output.toJson()


        # set sharing bit, if requested; this will fail for non-power users
        if splunk.util.normalizeBoolean(_is_shared):
            try:
                viewstate.setSharing(vs, 'app')
            except Exception, e:
                logger.exception(e)
                output.success = False
                output.addError(str(e))
                return output.toJson()

        return output.toJson()
