import cherrypy, json, math, re
from splunk.appserver.mrsparkle import *
import splunk.appserver.mrsparkle.lib.jsonresponse as jsonresponse
import lxml.etree as et
import splunk.entity as en
import lib.view.dashboard, lib.view.form, lib.view.panel

import logging
logger = logging.getLogger('splunk.appserver.controllers.viewmaster')


VIEW_ENTITY_CLASS = 'data/ui/views'

# define default number of panels to place in a row on dashboards
DEFAULT_DASHBOARD_ROW_SIZE = 2


class ViewmasterController(BaseController):
    '''
    Provides widget-based controls for view management
    '''
    
    # /////////////////////////////////////////////////////////////////////////
    #  Helpers
    # /////////////////////////////////////////////////////////////////////////

    def getViewObject(self, view_id, namespace, outputEntity=False):
        '''
        Returns a native ViewObject version of the requested view_id
        '''
        
        viewEntity = en.getEntity(VIEW_ENTITY_CLASS, view_id, namespace=namespace)
        
        parser = et.XMLParser(remove_blank_text=True, remove_comments=True, remove_pis=True)
        root = et.XML(viewEntity[en.EAI_DATA_KEY], parser)

        if root.tag == 'dashboard':
            objectClass = lib.view.dashboard.SimpleDashboard()
        elif root.tag == 'form':
            objectClass = lib.view.form.SimpleForm()
        else:
            raise NotImplementedError, 'Cannot manage view object of type: %s' % root.tag

        objectClass.fromXml(root)
        
        if outputEntity:
            return objectClass, viewEntity
        else:
            return objectClass
        

    def viewObjectToXmlString(self, viewObject):
        '''
        Serializes a view object to XML string, ready to be committed to
        splunkd
        '''
        
        objectXml = viewObject.toXml()
        objectXml.insert(0, 
            et.Comment('\nNOTE: This file was automatically generated by Splunk.  Use caution when editing manually.\n')
        )
        return et.tostring(objectXml, xml_declaration=True, encoding='utf-8', pretty_print=True)
        
        
        
    # /////////////////////////////////////////////////////////////////////////
    #  view container management
    #
    #  These endpoints provide create/edit/delete functionality for the
    #  view objects, via the client-side viewmaster
    # /////////////////////////////////////////////////////////////////////////

    @route('/:namespace')
    @expose_page(methods=['GET','POST'])
    @set_cache_level('never')
    def createContainer(self, namespace, view_id='', view_label='', container_type='dashboard'):
        '''
        Handles dashboard creation
        
        GET /<namespace>
            ==> HTML template form to input create data
            
        POST /<namespace>
            ==> Saves the HTML input values from GET
            
            == returns JSON response
        '''

        #
        # serve template
        #
        
        if cherrypy.request.method == 'GET':
            return self.render_template('viewmaster/create_dashboard.html', {'namespace': namespace})
        
            
        #
        # handle create view
        #
        
        cherrypy.response.headers['content-type'] = MIME_JSON
        output = jsonresponse.JsonResponse()

        # clean inputs
        view_id = re.sub(r'[^\w]', '', view_id)
        view_label = view_label.strip() or view_id
        
        if view_id == '':
            output.success = False
            output.addError(_('Dashboard ID cannot be empty'))
            return output.toJson()
            
        # ensure that the view doesn't already exist
        try:
            view = self.getViewObject(view_id, namespace=namespace)
            output.success = False
            output.addError(
                _('Cannot create new %(container_type)s: %(view_id)s already exists') \
                % {'container_type': container_type, 'view_id': view_id})
            return output.toJson()

        except splunk.ResourceNotFound:
            pass
        
        try:    
            # generate empty dashboard
            view = lib.view.dashboard.SimpleDashboard()
            view.label = view_label
        
            # insert into entity wrapper
            viewEntity = en.getEntity(VIEW_ENTITY_CLASS, en.NEW_EAI_ENTITY_NAME, namespace=namespace)
            viewEntity['name'] = view_id
            viewEntity[en.EAI_DATA_KEY] = self.viewObjectToXmlString(view)
        
            en.setEntity(viewEntity)
            
            output.data = {'view_id': view_id, 'view_label': view_label}
        
            output.addInfo(_('Successfully created new %(container_type)s: %(view_id)s') \
                % {'container_type': container_type, 'view_id': view_id})
            logger.info('Created new %s: namespace=%s id=%s label=%s' % (container_type, namespace, view_id, view_label))

        except Exception, e:
            logger.exception(e)
            output.success = False
            output.addError(_('Unable to create dashboard: %s') % e)
            
        return output.toJson(indent=2)


        
    @route('/:namespace/:view_id')
    @expose_page(methods=['GET'], handle_api=True)
    @set_cache_level('never')
    def getContainer(self, namespace, view_id, **unused):
        '''
        Renders the dashboard edit page
        
        GET /<namespace>/<view_id>
            ==> HTML form page with dashboard edit form (labels, panels) and
                master panel edit form (hidden)
            
        GET /api/<namespace>/<view_id>
            ==> JSON structure of dashboard config (unused?)
        '''
        
        # serve data feed
        output = jsonresponse.JsonResponse()

        try:
            view = self.getViewObject(view_id, namespace=namespace)
            output.data = view.toJsonable()
        except splunk.ResourceNotFound:
            cherrypy.response.status = 404
            output.success = False
            output.addError(_('View %s was not found') % view_id)
            return output.toJson(indent=2)
        
        # serve template page
        if cherrypy.request.is_api:
            cherrypy.response.headers['content-type'] = MIME_JSON
            return output.toJson(indent=2)
        else:
            cherrypy.response.headers['content-type'] = MIME_HTML
            return self.render_template('viewmaster/edit_dashboard.html', 
                {
                    'namespace': namespace,
                    'view_id': view_id,
                    'view_object': view,
                    'APP': {'id': namespace}
                })



    @route('/:namespace/:view_id')
    @expose_page(methods=['POST'])
    @set_cache_level('never')
    def setContainer(self, namespace, view_id, action, view_json=None):
        '''
        Provides support to modify dashboard configs
        
        POST /<namespace>/<view_id>
            &action=delete
                --> deletes the current view
            &action=edit
                --> updates the current view config (view JSON object)
                
            ==> returns a JSON response
        '''

        cherrypy.response.headers['content-type'] = MIME_JSON
        output = jsonresponse.JsonResponse()

        try:
            
            if action == 'delete':
                en.deleteEntity(VIEW_ENTITY_CLASS, view_id, namespace=namespace, owner=cherrypy.session['user'].get('name'))
                output.addInfo(_('Successfully deleted %s') % view_id)

            elif action == 'edit':

                # get most recent copy of view
                view = self.getViewObject(view_id, namespace=namespace)
                
                # convert incoming JSON to native struct; clean strings, 
                view_json = json.loads(view_json)

                if view_json.get('label'):
                    view.label = view_json['label'].strip()
                if view_json.get('refresh'):
                    view.refresh = int(view_json['refresh'].strip())
                    
                # handle panel reordering
                newPanelDefinition = []
                for row in view_json['new_panel_sequence']:
                    if len(row) > 0:
                        newRow = []
                        for seq in row:
                            newRow.append(view.getPanelBySequence(seq))
                        newPanelDefinition.append(newRow)
                view.rows = newPanelDefinition
                
                # ensure that the row grouping array is synced
                if len(view.rowGrouping) < len(view.rows):
                    view.rowGrouping.extend([None] * (len(view.rows) - len(view.rowGrouping)))

                # commit
                viewEntity = en.getEntity(VIEW_ENTITY_CLASS, view_id, namespace=namespace)
                viewEntity[en.EAI_DATA_KEY] = self.viewObjectToXmlString(view)
                en.setEntity(viewEntity)

                output.addInfo(_('Successfully updated %s') % view_id)

            else:
                output.success = False
                output.addError(_('Unrecognized dashboard action: %s; cannot process') % action)
                logger.error('Unrecognized dashboard action: %s; cannot process' % action)

        except splunk.ResourceNotFound:
            cherrypy.response.status = 404
            output.addWarn(_('"%s" was not found; no action taken') % view_id)

        except Exception, e:
            output.success = False
            output.addError(_('Unable to update view %s: %s') % (view_id, e))
            logger.exception(e)

        return output.toJson(indent=2)


        
    # /////////////////////////////////////////////////////////////////////////
    #  view container panel management
    #
    #  These endpoints provide create/edit/delete functionality for the
    #  panels contained in views; these are HTML generating
    # /////////////////////////////////////////////////////////////////////////

    @route('/:namespace/:view_id/:panel_type')
    @expose_page(methods=['POST'], handle_api=True)
    @set_cache_level('never')
    def createPanel(self, namespace, view_id, panel_type, panel_class, **panel_definition):
        '''
        Create a new panel to add to an existing dashboard
        
        POST /<namespace>/<view_id>/<panel_type=panel>
            &panel_class={table | chart | html | event | list}
            &<panel_property>=<property_value>
                --> creates a new panel
                
            the <panel_property> is a direct map to the panel XML data
            
            ==> returns a JSON response
        '''
        
        cherrypy.response.headers['content-type'] = MIME_JSON
        output = jsonresponse.JsonResponse()

        try:
            
            if panel_type != 'panel':
                raise ValueError, 'Only panel type "panel" is currently supported'
                
            viewObject, viewEntity = self.getViewObject(view_id, namespace=namespace, outputEntity=True)

            # calculate desired number of panels that constitute a complete row
            rowCount = len(viewObject.rows)
            if rowCount == 0:
                expectedRowSize = DEFAULT_DASHBOARD_ROW_SIZE
            else:
                expectedRowSize = int(math.floor(sum(map(len, viewObject.rows)) / rowCount))
                
            # if last row still has room, don't add row; otherwise create new row
            if rowCount > 0 and len(viewObject.rows[-1]) < expectedRowSize:
                pass
            else:
                viewObject.rows.append([])
                viewObject.rowGrouping.append(None)

            # generate new panel object
            panelObject = lib.view.panel.createPanel(panel_class)
            panelObject.fromJsonable(panel_definition)

            # insert into view object into last row
            viewObject.rows[-1].append(panelObject)
            
            # update the EAI wrapper and commit at container level
            viewEntity[en.EAI_DATA_KEY] = self.viewObjectToXmlString(viewObject)
            en.setEntity(viewEntity)
            
        except Exception, e:
            logger.exception(e)
            output.success = False
            output.addError(_('Unable to add panel: %s') % e)
            
        return output.toJson(indent=2)

    
    @route('/:namespace/:view_id/:panel_type/:panel_sequence')
    @expose_page(methods=['GET'], handle_api=True)
    @set_cache_level('never')
    def getPanel(self, namespace, view_id, panel_type, panel_sequence, **unused):
        '''
        Returns a dashboard panel config
        
        GET /<namespace>/<view_id>/panel/<panel_sequence>
            --> panel config for panel at <panel_sequence>
            
            ==> returns a JSON response
        '''
        
        output = jsonresponse.JsonResponse()

        try:
            viewObject = self.getViewObject(view_id, namespace=namespace)
            i,j = viewObject.getPanelPositionBySequence(panel_sequence)
            panel = viewObject.rows[i][j]
            output.data = panel.toJsonable()
            output.data['panel_sequence'] = panel_sequence

        except IndexError, e:
            cherrypy.response.status = 404
            output.success = False
            output.addError(_('Requested panel %s does not exist') % panel_sequence)
            
        except Exception, e:
            logger.exception(e)
            output.success = False
            output.addError(_('Unable to get panel at sequence %s: %s') % (panel_sequence, e))


        if not cherrypy.request.is_api:
            cherrypy.response.headers['content-type'] = MIME_HTML
            return self.render_template('viewmaster/edit_panel.html', 
                {
                    'namespace': namespace,
                    'view_id': view_id,
                    'view_object': viewObject,
                    'panel_sequence': panel_sequence,
                    'APP': {'id': namespace}
                })

        else:
            cherrypy.response.headers['content-type'] = MIME_JSON
            return output.toJson(indent=2)
                
        
        
    @route('/:namespace/:view_id/:panel_type/:panel_sequence')
    @expose_page(methods=['POST'], handle_api=True)
    @set_cache_level('never')
    def setPanel(self, namespace, view_id, panel_type, panel_sequence, panel_class=None, action='edit', **panel_definition):
        '''
        Provides management for view panel objects
        
        The HTTP signature for this method expects standard form params to match
        the property names used in the panel objects.  All form params are
        inserted into a dict and passed into the panel object for processing
        
        POST /<namespace>/<view_id>/panel/<panel_sequence>
            &action=edit
            --> updates the existing panel at <panel_sequence>
            &action=delete
            --> deletes the panel at <panel_sequence>
            
            ==> returns JSON response
        '''

        cherrypy.response.headers['content-type'] = MIME_JSON
        output = jsonresponse.JsonResponse()

        try:
            
            # get the container and panel info
            viewObject = self.getViewObject(view_id, namespace=namespace)
            i,j = viewObject.getPanelPositionBySequence(panel_sequence)
            
            # process actions
            if action == 'edit':

                if panel_class == None:
                    raise ValueError, 'Cannot set panel with unknown class; panel_definition=%s' % panel_definition

                panel = viewObject.rows[i][j]
                
                # for now, remove any time terms from saved search mode
                # TODO: revisit when we verify support for this
                if panel_definition.get('searchMode') == 'saved':
                    panel_definition['searchEarliestTime'] = None
                    panel_definition['searchLatestTime'] = None

                # if the class changes, make a new object
                if panel_class != panel.matchTagName:
                    newPanel = lib.view.panel.createPanel(panel_class)
                    newPanel.fromJsonable(panel_definition)
                    viewObject.rows[i][j] = newPanel
                else:
                    panel.fromJsonable(panel_definition)
                    
        
            elif action == 'delete':
                viewObject.rows[i].pop(j)
                output.data = {'panel_sequence': panel_sequence}
                
            else:
                raise ValueError, 'Unknown action requested: %s' % action

            # generate EAI wrapper and commit at container level
            viewEntity = en.getEntity(VIEW_ENTITY_CLASS, view_id, namespace=namespace)
            viewEntity[en.EAI_DATA_KEY] = self.viewObjectToXmlString(viewObject)
            en.setEntity(viewEntity)
            
            output.addInfo(_('Successfully updated %s' % view_id))

        except Exception, e:
            logger.exception(e)
            output.success = False
            output.addError(_('Unable to update panel at sequence %s: %s') % (panel_sequence, e))

        return output.toJson(indent=2)
    
