"""
An override of cherrypy.tools.staticdir that allows for a callback resolver to be triggered
prior to serving a static file.  The resolver can return a local pathname to serve instead of
allowing staticdir to default to serving files out of its configured directory.

Ideally this would be implemented as a completely separate tool (ie. not clobber cherrypy.tools.staticdir)
but unfortunately cherrypy's default dispatcher makes assumptions about how static files are served
at the moment; hopefully it'll be fixed in the future.
"""

import cherrypy
import urllib, re, os.path
import i18n
import logging
from lib.util import make_url, use_future_expires

logger = logging.getLogger('splunk.appserver.lib.staticdir')

def custom_staticdir(section, dir, root="", match="", content_types=None, index="", generate_indexes=False, resolver=None, strip_version=False, default_ext=None):
    """
    Backwards compatible, i18n enabled, staticdir tool replacement.
    If a callable called resolver is supplied then it is called prior to the original staticdir tool being fired.
    If the resolver returns a pathname, then we attempt to static serve that file.
    If the resolver returns False, or serve_file raises a NotFound exception, then we fallback
    to the original staticdir behaviour.
    """
    # first path segment should be locale
    if section == 'global':
        section = "/" 

    branch_offset = cherrypy.request.path_info.find(section)
    branch = cherrypy.request.path_info[branch_offset + len(section):]
    branch = urllib.unquote(branch.lstrip(r"\/"))

    if strip_version and len(branch) and branch[0]=='@':
        version = branch.split('/', 1)[0]
        branch = branch[len(version):].lstrip('/')
        use_future_expires()
    
    if resolver:
        # call the user's resolver callback
        filename = resolver(section, branch, dir)
        if filename:
            # try to find localized versions of the requested file
            fnlist = i18n.path_to_i18n_paths(filename)
            for fn in fnlist:
                if fn.endswith('.css'):
                    if os.path.exists(fn):
                        serve_static_css(fn)
                        return True
                else:
                    try:
                        cherrypy.lib.static.serve_file(fn)
                        return True
                    except cherrypy.NotFound:
                        # fall back to the original staticdir handler if we're out of filenames to try
                        pass

    # duplicate the original staticdir functionality but add i18n filename matching
    if match and not re.search(match, cherrypy.request.path_info):
        return False

    # Allow the use of '~' to refer to a user's home directory.
    dir = os.path.expanduser(dir)

    # If dir is relative, make absolute using "root".
    if not os.path.isabs(dir):
        if not root:
            msg = _("Static dir requires an absolute dir (or root).")
            raise ValueError(msg)
        dir = os.path.join(root, dir)

    # If branch is "", filename will end in a slash
    filename = os.path.join(dir, branch)

    # There's a chance that the branch pulled from the URL might
    # have ".." or similar uplevel attacks in it. Check that the final
    # filename is a child of dir.
    if not os.path.normpath(filename).startswith(os.path.normpath(dir)):
        raise cherrypy.HTTPError(403, "%s does not start with %s" % (filename, dir)) # Forbidden

    # get filenames based on user's current locale
    fnlist = []
    filenames = [filename]
    if default_ext:
        filenames.append("%s.%s" % (filename, default_ext))
    for try_filename in filenames:
        fnlist += i18n.path_to_i18n_paths(try_filename) # get filenames

    # see if we can serve any of the local language filenames
    for fn in fnlist:
        if fn.endswith('.css'):
            if os.path.exists(fn):
                serve_static_css(fn)
                return True
        else:
            handled = cherrypy.lib.static._attempt(fn, content_types)
            if handled:
                return handled

    # didn't find a file if we reach here; perhaps it's a directory
    if not handled:
        # Check for an index file if a folder was requested.
        if index:
            handled = cherrypy.lib.static._attempt(os.path.join(filename, index), content_types)
            if handled:
                cherrypy.request.is_index = filename[-1] in (r"\/")
        if not handled and generate_indexes:
            if callable(generate_indexes):
                handled = cherrypy.lib.static.generate_indexes(filename, branch)
            else:
                handled = cherrypy.lib.static.render_index(filename, branch)
    return handled


def serve_static_css(fn):
    """
    Translate urls defined in static css files into local urls
    taking locale and root prefix into account
    """
    # make sure we keep cached copies for each locale/version tuple
    lang = i18n.current_lang_url_component()
    cache = serve_static_css.cache.setdefault(lang, {})
    cherrypy.response.headers['Content-Type'] = 'text/css'

    if fn in cache and cache[fn][0]==os.path.getmtime(fn):
        css =  cache[fn][1]
    else:
        f = file(fn)   
        css = re.sub(r'([\s:]+url\([\'"]?)([^\)\'"]+)', lambda m: m.group(1)+make_url(m.group(2)), f.read(250000))
        cache[fn] = (os.path.getmtime(fn), css)
        f.close()
    cherrypy.response.body = css 
    cherrypy.response.headers['Content-Length'] = len(css)
serve_static_css.cache = {}


# clobber Cherrypy's default staticdir handler :-/
# this thread might help explain why: http://osdir.com/ml/python.cherrypy/2006-10/msg00021.html
cherrypy.tools.staticdir = cherrypy._cptools.HandlerTool(custom_staticdir, name='staticdir')
