import urllib
import cherrypy
import splunk
import logging
import os
import json

from splunk.appserver.mrsparkle.lib import util, startup, message
from routes import RequestRefused
from decorator import decorator
from splunk.appserver.mrsparkle.list_helpers.formattermapper import FormatterMapper
from cherrypy.lib import http

logger = logging.getLogger('splunk.appserver.mrsparkle.lib.decorators')

ONLY_API=1

def chain_decorators(fn, *declist):
    """
    Called from a decorator to chain other decorators together
    eg. chain_decorators(fn, require_login(), cherrypy.expose)
    """
    for dec in declist[::-1]: # wrap them in the same order as they'd be wrapped if they were used normally
        fn = dec(fn)

    @decorator
    def rundecs(_fn, *a, **kw):
        return fn(*a, **kw)

    return rundecs

def expose_page(must_login=True, handle_api=False, methods=None, verify_session=True):
    """
    Use this instead of cherrypy.expose
    Ensures that user's are logged in to view the page by default
    
    set handle_api=True to have requests beginning with /api sent to the handler 
    as well as non-api requests
    set handle_api=ONLY_API to have it only accept api requests
    (check cherrypy.request.is_api to see whether this is an api request if set to True)

    set methods to a list of method names to accept for this handler (default=any)
    """
    @decorator
    def check(fn, self, *a, **kw):
        is_api = util.is_api()
        if not handle_api and is_api:
            raise RequestRefused(404)
        if handle_api is ONLY_API and not is_api:
            raise RequestRefused(404)
        _methods = methods
        if _methods:
            if isinstance(_methods, basestring):
                _methods = [ _methods ]
            if cherrypy.request.method not in _methods:
                raise RequestRefused(405)
        
        # verify that version info is good; do it here so that any URI access
        # will trigger the check
        startup.initVersionInfo()

        # CSRF protection
        # Disable in tests by setting cherrypy.config.update({'environment': 'test_suite'})
        if verify_session and cherrypy.request.method == 'POST' and not cherrypy.config.get('environment') == 'test_suite':
            session_id = None
            if util.is_xhr():
                session_id = cherrypy.request.headers.get('X-Splunk-Session')
            else:
                session_id = cherrypy.request.params.get('splunk_session_id')

            if not util.is_valid_cherrypy_session_id(session_id):
                error_message = _('Splunk cannot authenticate the request. CSRF validation failed.')
                logger.warn('Splunk cannot authenticate the request. CSRF validation failed.')
                raise cherrypy.HTTPError(401, error_message)

        return fn(self, *a, **kw)

    def dec(fn):
        if must_login:
            return chain_decorators(fn, check, require_login(), ExceptionHandler(), cherrypy.expose)(fn)
        else:
            return chain_decorators(fn, check, ExceptionHandler(), cherrypy.expose)(fn)
            
    return dec
        

def lock_session(fn):
    """
    Use this if your handler will make changes to cherrypy.session
    It causes CherryPy to acquire an exclusive lock on the session for the 
    duration of the request ensuring there aren't any race conditions with
    other requests that are also accessing session data.
    """
    fn.lock_session = True
    return fn

def require_login():
    """
    If for some reason you're not using the expose_page decorator
    you can use this to require a user to be logged in instead.
    use expose_page though. really.
    """
    @decorator
    def check_login(fn, self, *a, **kw):
        session_key = cherrypy.session.get('sessionKey', None)
        is_api = util.is_api()

        if not session_key:
            logger.info('No sessionKey session variable set')
            if is_api or util.is_xhr():
                raise cherrypy.HTTPError(401)
            else:
                self.redirect_to_url('/account/login', _qs=[ ('return_to', util.current_url_path()) ] )
            
        try:
            return fn(self, *a, **kw)
        except splunk.AuthenticationFailed:
            logger.info('sessionKey rejected by splunkd')
            cherrypy.session.delete()
            if is_api or util.is_xhr():
                raise cherrypy.HTTPError(401)
            else:
                self.redirect_to_url('/account/login', _qs=[ ('return_to', util.current_url_path()) ] )

    return check_login


def ExceptionHandler():
    """
    Handles exceptions returned by simpleRequest
    """
    @decorator
    def handle_exceptions(fn, self, *a, **kw):
        from controllers.admin import AdminController

        try:
            return fn(self, *a, **kw)
            
        except splunk.AuthenticationFailed:
            # redirect to the login page if auth fails
            cherrypy.session['sessionKey'] = None
            self.redirect_to_url('/account/login', _qs=[ ('return_to', util.current_url_path()) ] )
            
        except splunk.AuthorizationFailed, e:
            if isinstance(self, AdminController):
                return self.render_admin_template('admin/error.html', {'namespace' : 'search', 'excp_msg': e, 'excp_details' : 'None'})   
            else:
                raise
                
        except splunk.SplunkdConnectionException, e:
            logger.exception(e)
            raise cherrypy.HTTPError(503, _('The splunkd daemon is no longer responding and may have crashed.  The UI is fine, so check splunkd.log or look for any crash dumps.'))
            
        except splunk.BadRequest, e:
            logger.exception(e)
            if e.msg == "Couldn't parse xml reply":
                raise cherrypy.HTTPError(500, _('The splunkd python dispatcher was unable to properly process script output.'))
            else:
                raise
                
    return handle_exceptions


def conditional_etag():
    '''
    DEPRECATED.  Use @set_cache_level('etag') instead.
    
    Similar to the util.apply_etag(content) method this wraps the 
    entire response with predefined 304 behavior.
    '''
    @decorator
    def apply_etag(fn, self, *a, **kw):
        response = fn(self, *a, **kw)
        if (util.apply_etag(response)):
            return None
        else:
            return response
        
    return apply_etag

def set_cache_level(cache_level):
    '''
    This is a convience wrapper for util.set_cache_level, providing
    backwards compatibility with the original set_cache_level decorator.
    
    The body of this method was moved to the util module so that individual
    modules could dictate whether or not they need to cache their getResults
    responses.
    '''
    @decorator
    def apply_cache_headers(fn, self, *a, **kw):
        response = fn(self, *a, **kw)
        return util.set_cache_level(cache_level, response)
    return apply_cache_headers

def normalize_list_params():
    '''
    Requires the underlying class implements the default list params:
        COUNT
        OFFSET
        SORT_KEY
        SORT_DIR
    '''
    
    @decorator
    def apply_normalized_list_params(fn, self, *a, **kw):
        params = {
            'count': kw.get('count', self.COUNT),
            'offset': kw.get('offset', self.OFFSET),
            'sort_key': kw.get('sort_key', self.SORT_KEY),
            'sort_dir': kw.get('sort_dir', self.SORT_DIR)
        }
        
        try:
            params['count'] = int(params['count'])
        except ValueError:
            params['count'] = int(self.COUNT)
            
        try:
            params['offset'] = int(params['offset'])
        except ValueError:
            params['offset'] = int(self.OFFSET)
        
        kw.update(params)

        return fn(self, *a, **kw)
    
    return apply_normalized_list_params

def format_list_template():

    def decode_fields(fields):
        if type(fields) != list:
            fields = [fields]
    
        decoded_fields = []
        for field in fields:
            try:
                decoded_fields.append(json.loads(field))
            except ValueError, e:
                decoded_fields.append(field)
        return decoded_fields
    
    @decorator
    def response_template(fn, self, *a, **kw):

        # Retrieve the list of dictionaries
        kw['list_data'] = fn(self, *a, **kw)

        # protect against dir traversal 
        # We might want to consider more strict filtering, but shouldn't
        # have any issues with double encoding or unicode filtering based
        # on quick testing.
        output_mode = 'lists/%s.html' % kw.get('output_mode', 'li').split('/').pop()
        
        fields = kw.get('fields')
        staticFields = kw.get('staticFields')
       
        # TODO: make this more sane using a cherrypy tool that translates
        # field[0][label]=foo,field[0][value]=bar to:
        # field = {'0': {'label':'foo', 'value':'bar'}}
        if fields:
            kw['fields'] = decode_fields(fields)

        if staticFields:
            kw['staticFields'] = decode_fields(staticFields)

        # Some list templates require delimiters.
        # This may not be the ideal place to put this, but where else?
        if not kw.get('delimiter'):
            kw['delimiter'] = ''

        return self.render_template(output_mode, kw)

    return response_template

def memoize():
    '''
    Simple memoization decorator preserving signature.
    
    Used to speed up your programs by avoiding repeated calculation of results for previously-processed inputs.
    
    Note: You can not memoize mutable arguments.
    
    w00t!
    '''
    
    def getattr_(obj, name, default_thunk):
        "Similar to .setdefault in dictionaries."
        try:
            return getattr(obj, name)
        except AttributeError:
            default = default_thunk()
            setattr(obj, name, default)
            return default
    
    @decorator
    def memo(func, *args, **kwargs):
        dic = getattr_(func, "memo_dic", dict)
        key = (args, tuple(kwargs.iteritems()))
        if key in dic:
            logger.info("Memoized")
            return dic[key]
        else:
            logger.info("Not memoized")
            result = func(*args, **kwargs)
            dic[key] = result
            return result
        
    return memo
