import os, sys
import util
import splunk.appserver.mrsparkle.controllers.module
import controllers.module
import splunk.util
import logging
from ConfigParser import *


from lib.apps import local_apps

logger = logging.getLogger('splunk.appserver.lib.module')

MASTER_EXTENSION = '.js'
MANIFEST_EXTENSION = '.conf'

UNDEFINED_MODULE_ASSERTION_ERROR = "%s is not a defined module classname. This is probably either a typo in the first line of a module's js file, or an old out-of-date module file hanging around in the modules directory."

class ModuleMapperException(Exception): 
    pass

class ModuleMapper(object):

    def __init__(self):
        self.installedModules = None
        self.installedModules = self.getInstalledModules()
    
    def getModuleList(self, root):
        '''
        Generates a list of module detail by crawling the entire modules directory
        for JS files.
        '''
        moduleList = []
        moduleHash = {}
        _duplicateDefender = set()
        _seen_filenames = set()

        def scan_file(dirpath, name):
            if name.endswith(MANIFEST_EXTENSION) and os.path.exists(os.path.join(dirpath, name[:-len(MANIFEST_EXTENSION)] + MASTER_EXTENSION)):
                cfg_pathname = os.path.join(dirpath, name)
                if cfg_pathname in _seen_filenames:
                    return
                _seen_filenames.add(cfg_pathname)

                config = ConfigParser()
                config.readfp(open(cfg_pathname))
                

                # assert that conf file has base classes defined
                
                try:
                    className = config.get('module', 'className')
                    superClass = config.get('module', 'superClass') if config.has_option('module', 'superClass') else None
                except NoSectionError:
                    logger.error("Manifest file %s does not contain a module section" % cfg_pathname)
                    return
                except NoOptionError:
                    logger.error("Manifest file %s contains an incomplete module section" % cfg_pathname)
                    return

                if className == superClass:
                    raise ModuleMapperException('%s defines className == superClass !!' % cfg_pathname)


                # assemble the parameter configuration info

                params = {}
                stickyParams = []
                persistableParams = []
                for section in config.sections():
                    if section.startswith('param:'):
                        pname = section[6:]
                        if config.has_option(section, 'default') and config.has_option(section, 'required') and config.getboolean(section, 'required'):
                            raise ModuleMapperException(
                                'Cannot use required=True with a default value in Manifest file %s, parameter %s' 
                                % (cfg_pathname, pname))

                        params[pname] = {
                            'required': config.getboolean(section, 'required') if config.has_option(section, 'required') else False,
                            'default': config.get(section, 'default') if config.has_option(section, 'default') else None,
                            'values' : [val.strip() for val in config.get(section, 'values').split(',')] if config.has_option(section, 'values') else None,
                            'label': config.get(section, 'label') if config.has_option(section, 'label') else None,
                            'translate': config.get(section, 'translate') if config.has_option(section, 'translate') else None
                        }

                        # add params to persistence lists
                        if config.has_option(section, 'sticky') and config.getboolean(section, 'sticky'):
                            stickyParams.append(pname)
                        if config.has_option(section, 'persistable') and config.getboolean(section, 'persistable'):
                            persistableParams.append(pname)
                        
                
                # check dupes
                
                if (className in _duplicateDefender):
                    logger.error("ASSERT - duplicate definition of %s in %s/%s" % (className, dirpath, name)) 
                    return
                _duplicateDefender.add(className)


                # enable support for including other module configuration
                
                if config.has_option('module', 'include'):
                    include = [ mod_name.strip() for mod_name in config.get('module', 'include').split(',') ]
                else:
                    include = []


                # assemble final module definition dict

                mod = {
                    'class': className,
                    'appName': '',
                    'superClass': superClass,
                    'path': dirpath,
                    'filePrefix': name[:-len(MANIFEST_EXTENSION)],
                    'params': params,
                    'stickyParams': stickyParams,
                    'persistableParams': persistableParams,
                    'include': include,
                    'description': config.get('module', 'description') if config.has_option('module', 'description') else None
                }
                moduleList.append(mod)
                moduleHash[className] = mod


        # generate a dict of all the accessible Splunk.Modules
        for dirpath, subdirs, filenames in os.walk(root):
            for name in filenames:
                scan_file(dirpath, name)
        

        # Scan installed applications for modules that they might define
        for (app_name, module_name, module_path) in local_apps.getAllModules():
            for filename in os.listdir(module_path):
                if filename.endswith('.py'):
                    # load module's python handlers
                    modname = filename[:-3]
                    if modname not in sys.modules.keys():
                        sys.path.insert(0, module_path)
                        mod = __import__(modname)
                        # find all the module classes that subclass module.ModuleHandler()
                        for c in util.get_module_classes(mod):
                            # push module into controllers.module namespace.. 
                            # not sure i really like this, but does keep it super easy for module writers
                            if issubclass(c, controllers.module.ModuleHandler):
                                logger.debug('module loader - loading python module handler: %s' % c.__name__)
                                # XXX Normalize this asap!
                                try:
                                    setattr(splunk.appserver.mrsparkle.controllers.module, c.__name__, c) 
                                except AttributeError:
                                    pass
                                setattr(controllers.module, c.__name__, c) 
                        del sys.path[0]
                else:
                    scan_file(module_path, filename)

        # modules inherit their parameters from their superClass
        processed = set()
        def update_params(mod):
            if mod['class'] in processed:
                return
            processed.add(mod['class'])
            if not (mod['superClass'] and moduleHash.has_key(mod['superClass'])):
                return
            supermod = moduleHash[mod['superClass']]
            if mod['superClass'] not in processed:
                update_params(supermod)
            for pname, param in supermod['params'].items():
                if not mod['params'].has_key(pname):
                    mod['params'][pname] = param.copy()
        for mod in moduleList:
            update_params(mod)

        if len(moduleList) == 0:
            logger.warn('getModuleList - did not find any modules to load from: %s' % root)
            
        return moduleList
        
                   
    def sortModuleList(self, moduleList):
        '''
        Sorts a module list (in-place) according to inheritance order
        '''
        
        # create link list
        links = dict([(x['class'], x['superClass']) for x in moduleList])
        
        def getNumberOfGenerations(name, count=0):
            # check for simple but common configuration error, for instance where 
            # an old version of a module file refers to a nonexistent class.
            if (name not in links):
                logger.warn(UNDEFINED_MODULE_ASSERTION_ERROR % name)
                return -1

            superClass = links[name]
            
            if superClass == None: return count
            return getNumberOfGenerations(superClass, count + 1)

        # count the number of parents until a class hits top
        numberOfGenerationsMap = {}
        
        for item in moduleList:
            numberOfGenerationsMap[item['class']] = getNumberOfGenerations(item['class'])
        
        # remove the ones that were flagged earlier as having -1 ancestral generations.
        startingLen = len(moduleList) - 1
        for i, x in enumerate(reversed(moduleList)):
            if numberOfGenerationsMap.get(x['class'], 0) < 0:
                moduleList.pop(startingLen - i)

        # sort the moduleList by the numberOfGenerationsMap of each class
        def comp(a, b):
            return cmp(numberOfGenerationsMap[a['class']], numberOfGenerationsMap[b['class']])
        moduleList.sort(comp)
        
                    
    def updateModuleAssets(self, moduleList):
        '''
        Updates a module list (in-place) with asset flags.  Any JS file with
        matching files will result in a new key being added to the module list.
        Ex: if a matching HTML file is found, a 'html': True key will be added.
        '''
        
        for module in moduleList:
            for name in os.listdir(module['path']):
                if name.startswith(module['filePrefix'] + '.'):
                    parts = name.split('.')
                    if len(parts) > 1:
                        module[parts[-1]] = os.path.join(module['path'], module['filePrefix'] + '.' + parts[-1])

    def updateModuleAppName(self, moduleList):
        """
        Look at the modulelist's path attribute to get what application it came from.
        """
        
        for module in moduleList:
            appnameShard = module['path'].split(os.sep).index('modules')-2
            appname = module['path'].split(os.sep)[appnameShard]
            if appname == 'splunk':
                module['appName'] = splunk.getDefault('namespace')
            else:
                module['appName'] = appname
                      
    def updateModuleInheritance(self, moduleList):
        '''
        Updates a module list (in-place) with the chain of inheritance as the
        'inheritance' property
        '''
        
        # generate an ordered dict version of the list for easy access
        workingList = splunk.util.OrderedDict()
        for x in moduleList:
            workingList[x['class']] = x.copy()
            
        # flag all of the modules that are active in this view
        def mwalk(name):
            if name: workingList[name]['isActive'] = True
            else: return
            mwalk(workingList[name]['superClass'])

        # inspect each module and generate its inheritance chain
        for moduleName in workingList:

            # reset
            for x in workingList:
                workingList[x]['isActive'] = False

            mwalk(workingList[moduleName]['class'])
            
            workingList[moduleName]['inheritance'] = []
            for x in workingList:
                if workingList[x]['isActive']:
                    workingList[moduleName]['inheritance'].append(x)
                    
            for x in moduleList:
                x['inheritance'] = workingList[x['class']].get('inheritance',[])

                        
    def extractClassName(self, jsFilePath) : 
        """
        Extracts from the first line of the js file, the classname and any 
        parent class relationships.
        """            

        className = None
        superClass = None
        handle = open(jsFilePath, 'r')
        try : 
            for line in handle:
                firstLine = line.strip('/\r\n').split(" extends ")
                className   = firstLine[0].strip()
                if (len(firstLine) > 1) :
                    superClass = firstLine[1].strip()
                break        
        except Exception, e:
            logger.exception(e)
        finally:
            if handle: handle.close()

        return className, superClass


    def getInstalledModules(self):
        '''
        Returns a list of module information that will be used 
        to construct the various view templates.
        Top-level keys are module names like "Splunk.Module.Topnav".
        and those top level values are each their own dictionary.
        
        Within those second level dictionares, the most important keys are 
        'path', and 'fileprefix', giving clients the path to the files.
        -- and then individual keys like 'html','css','js', that are 
        bools, each indicating whether the module specifies a file 
        for that specific resource.
        
        Example:
        
        {
            'Splunk.Module.FlashTimeline': {
                'filePrefix': 'flash_timeline',
                'inheritance': [
                    'Splunk.Module',
                    'Splunk.Module.SearchModule',
                    'Splunk.Module.FlashWrapper',
                    'Splunk.Module.FlashTimeline'
                ],
                'js': '/Users/johnvey/build/current/share/splunk/search_mrsparkle/modules/flash_timeline/flash_timeline.js',
                'html': '/Users/johnvey/build/current/share/splunk/search_mrsparkle/modules/flash_timeline/flash_timeline.html',
                'superClass': 'Splunk.Module.FlashWrapper',
                'path': '/Users/johnvey/build/current/share/splunk/search_mrsparkle/modules/flash_timeline',
                'class': 'Splunk.Module.FlashTimeline',
                'css': '/Users/johnvey/build/current/share/splunk/search_mrsparkle/modules/flash_timeline/flash_timeline.css'
            }
            
            ...
            
        }
        '''

        if self.installedModules is not None:
            return self.installedModules

        # TODO: pull out path into global config area
        # read in the system modules
        root = util.make_absolute("share/splunk/search_mrsparkle/modules")

        mods = self.getModuleList(root)
        self.sortModuleList(mods)
        self.updateModuleAssets(mods)
        self.updateModuleInheritance(mods)
        self.updateModuleAppName(mods)
        
        # generate an ordered dict version of the list for easy access
        output = splunk.util.OrderedDict()
        for x in mods:
            output[x['class']] = x
        
        # memoize the list of installed modules
        self.installedModules = output

        return output

### make ModuleMapper a singleton
moduleMapper = ModuleMapper()
