"""
These classes supply custom session lock behaviour

Normally cherrypy acquires an exclusive lock on a session for the duration of a request,
which essentially means requests are handled sequentially for a given client even
if the request handler doesn't make any changes to the session data.

This update causes cherrypy to acquire a shared lock instead by default.  A handler
can acquire an exclusive lock instead by using the @lock_session decorator.
"""

import cherrypy
import cherrypy._cptools
from cherrypy.lib import http
from lib import portalocker
import os.path, time

import logging

logger = logging.getLogger('splunk.appserver.lib.sessions')


class SessionException(Exception): pass

class SessionTool(cherrypy._cptools.SessionTool):
    def _lock_session(self):
        if not hasattr(cherrypy.request.handler, 'callable') or not getattr(cherrypy.request.handler.callable, 'lock_session', None):
            # default to acquiring a shared lock for sessions unless the handler explicitly requests otherwise with the @lock_session decorator
            try:
                cherrypy.serving.session.acquire_lock(read_lock=True)
                return
            except TypeError:
                pass
        cherrypy.serving.session.acquire_lock()

cherrypy.tools.sessions = SessionTool()


class FileSession(cherrypy.lib.sessions.FileSession):

    def __init__(self, *a, **kw):
        self._changed = False
        self.locktype = None
        super(FileSession, self).__init__(*a, **kw)

    def __get_changed(self):
        return self._changed

    def __set_changed(self, val):
        if val:
            self._ensure_ex_lock()
        self._changed = val

    changed = property(__get_changed, __set_changed)

    def __setitem__(self, key, value):
        self.changed = True
        super(FileSession, self).__setitem__(key, value)

    def __delitem__(self, key):
        self.changed = True
        super(FileSession, self).__setitem__(key, value)

    def _ensure_ex_lock(self):
        if self.locktype == portalocker.LOCK_EX:
            return
        if not self.locked:
            return self.acquire_lock(read_lock=False)
        logger.warn('Attempted to update session data while holding a shared lock - Consider using @lock_sesion - request_path=%s' % (cherrypy.request.path_info))
        self.escalate_lock()

    def _touch_session(self):
        try:
            os.utime(self._get_file_path(), None)
        except OSError:
            pass # file doesn't exist which is normal if it's already expired and hasn't been updated

    def save(self):
        if self._changed and self.loaded:
            self._save(self.timeout) # specify relative minutes for timeout instead of absolute time
        elif self.loaded:
            self._touch_session()

    def load(self):
        data = self._load()
        if data is None:
            self._data = {}
        else:
            mtime = os.path.getmtime(self._get_file_path())
            if time.time() - mtime > data[1]*60:
                self._data = {} # expired!
                self._changed = True
                logger.debug('Session expired mtime=%d time=%s path=%s' % (mtime, time.time(), self._get_file_path()))
            else:
                self._data = data[0]
        self.loaded = True

        # Stick the clean_thread in the class, not the instance.
        # The instances are created and destroyed per-request.
        cls = self.__class__
        if self.clean_freq and not cls.clean_thread:
            # clean_up is in instancemethod and not a classmethod,
            # so that tool config can be accessed inside the method.
            t = cherrypy.process.plugins.Monitor(
                cherrypy.engine, self.clean_up, self.clean_freq * 60) 
            t.subscribe()
            cls.clean_thread = t 
            t.start()

    def clean_up(self):
        """Clean up expired sessions."""
        now = time.time()
        suffix_len = len(self.LOCK_SUFFIX)
        # Iterate over all session files in self.storage_path
        lockfiles = set()
        active_sessions = set()
        for fname in os.listdir(self.storage_path):
            if fname.startswith(self.SESSION_PREFIX):
                path = os.path.join(self.storage_path, fname)
                if not fname.endswith(self.LOCK_SUFFIX):
                    # We have a session file: lock and load it and check
                    #   if it's expired. If it fails, nevermind.
                    active_sessions.add(path)
                    self.acquire_lock(path)
                    try:
                        contents = self._load(path)
                        # _load returns None on IOError
                        if contents is not None:
                            mtime = os.path.getmtime(path)
                            data, expiration_time = contents
                            if now - mtime > expiration_time*60:
                                # Session expired: deleting it
                                os.unlink(path)
                                self.release_lock(path)
                                active_sessions.discard(path)
                    except:
                        active_sessions.discard(path) # make sure lock file gets deleted
                    finally:
                        self.release_lock(path)
                else:
                    session_path = path[:-suffix_len]
                    lockfiles.add(session_path)

        # find entries in lockfiles that aren't in session_files are delete them
        for session_path in lockfiles.difference(active_sessions):
            self.delete_lock(session_path)
        


    def acquire_lock(self, path=None, read_lock=False):
        """Acquire an exclusive or read lock on the currently-loaded session data."""
        if self.locked:
            return
        if path is None:
            path = self._get_file_path()
        path += self.LOCK_SUFFIX
        lockfile = open(path, 'a+')

        self.locktype = locktype = portalocker.LOCK_SH if read_lock else portalocker.LOCK_EX
        portalocker.lock(lockfile, locktype)

        self.locked = True
        self.lockfile = lockfile

    def release_lock(self, path=None):
        if not self.locked:
            return
        if path is None:
            path = self._get_file_path()
        path += self.LOCK_SUFFIX
        portalocker.unlock(self.lockfile)
        self.locked = False
        self.lockfile = None

    def delete_lock(self, path):
        path += self.LOCK_SUFFIX
        os.unlink(path)

    def escalate_lock(self):
        """
        Convert a read lock into a write lock - will reload the session first
        to avoid any possible race conditions
        """
        if not self.locked:
            return self.acquire_lock(read_lock=False)
        if self.locktype == portalocker.LOCK_EX:
            return
        self.release_lock()
        self.acquire_lock(read_lock=False)
        self.load()


# Replace Cherrypy's version of this so that cookies alwasy last for
# 24 hours by default - Setting them to 1 hour causes problems for IE
# users who's clocks are off from the server
def set_response_cookie(path=None, path_header=None, name='session_id',
                        timeout=60, domain=None, secure=False):
    """Set a response cookie for the client.

    path: the 'path' value to stick in the response cookie metadata.
    path_header: if 'path' is None (the default), then the response
        cookie 'path' will be pulled from request.headers[path_header].
    name: the name of the cookie.
    timeout: the expiration timeout for the cookie.
    domain: the cookie domain.
    secure: if False (the default) the cookie 'secure' value will not
        be set. If True, the cookie 'secure' value will be set (to 1).
    """
    # Set response cookie
    cookie = cherrypy.response.cookie
    cookie[name] = cherrypy.serving.session.id
    cookie[name]['path'] = (path or cherrypy.request.headers.get(path_header)
                            or '/')

    # We'd like to use the "max-age" param as indicated in
    # http://www.faqs.org/rfcs/rfc2109.html but IE doesn't
    # save it to disk and the session is lost if people close
    # the browser. So we have to use the old "expires" ... sigh ...
##    cookie[name]['max-age'] = timeout * 60
    if timeout:
        #cookie[name]['expires'] = http.HTTPDate(time.time() + (timeout * 60))
        timeout = timeout * 60 # convert to seconds
        if timeout < 86400:
            timeout = 86400 # Cookies last at least 24 hours
        cookie[name]['expires'] = http.HTTPDate(time.time() + timeout)
    if domain is not None:
        cookie[name]['domain'] = domain
    if secure:
        cookie[name]['secure'] = 1


cherrypy.lib.sessions.FileSession = FileSession
cherrypy.lib.sessions.set_response_cookie = set_response_cookie
