import splunk.bundle
import splunk.util as util
import cached

import logging
logger = logging.getLogger('splunk.appserver.mrsparkle.lib.times')


def getTimeRanges(namespace=None):
    '''
    Returns a list of splunk search time identifiers that are defined
    in the time.conf files, ordered ascending by the 'order' key in each 
    time stanza.
    
    Sample time.conf stanza:
    
        [last_7d]
        label = Last 7 Days
        earliest_time = -7d
        latest_time = false
        order = 30

        [previous2_m]
        label = Last Month
        earliest_time = -2m@m
        latest_time = -1m@m
        order = 40

    Sample return structure:
    
        [
            {
                'key' = 'last_7d',
                'label' = 'Last 7 Days',
                'earliest_time' = '-7d',
                'latest_time' = False,
                'order' = '30'
            },

            {
                'key' = 'previous2_m',
                'label' = 'Last Month',
                'earliest_time' = '-2m@m',
                'latest_time' = '-1m@m',
                'order' = '40'
            }
        ]
        
    '''
    
    #stanzas = cached.getEntities('admin/conf-times', namespace=namespace)
    stanzas = splunk.bundle.getConf('times', namespace=namespace)

    # the final flat ordered array we will output.
    orderedStanzas = []

    # two dicts that we use only to check for configuration error cases. 
    # if you use sub_menu = foo  in a stanza in times.conf (to place that stanza within the sub_menu with label=foo)
    # you MUST have a stanza with label=foo,  and is_sub_menu=True.
    # if you dont have this, right we just log an error at runtime.
    subMenusDefinedByChildren = {}
    subMenusPresent = {}

    for s in stanzas:
        if s == 'default':
            continue
            
        item = {
            'key': s,
            'label': stanzas[s]['label'],    
            'order': int(stanzas[s].get('order', 99999))
        }
        
        # stanzas that are just sub_menu containers.  
        # we only care so we can check for misconfigurations.
        if ('is_sub_menu' in stanzas[s] and util.normalizeBoolean(stanzas[s]['is_sub_menu'])):
            # were still outputting a flat list at the end of the day. 
            # the caller will use this flag to build the hierarchy appropriately.
            item['is_sub_menu'] = True
            subMenusPresent[stanzas[s]['label']] = True
            
        # items that are meant to be INSIDE a submenu. If omitted it the item will go into the main menu.
        if (stanzas[s].get('sub_menu')) :
            item['sub_menu'] = stanzas[s]['sub_menu']
            subMenusDefinedByChildren[stanzas[s]['sub_menu']] = True
            
        # header_label is optional. If omitted it will use the 'header' as the label. 
        if (stanzas[s].get('header_label')) :
            item['header_label'] = stanzas[s]['header_label']
        
        # only add time bounds if evaluates to something true
        for p in ('earliest_time', 'latest_time'):
            if util.normalizeBoolean(stanzas[s].get(p)):
                item[p] = stanzas[s][p]
            else:
                item[p] = False
                
        orderedStanzas.append(item)
    
    if (subMenusPresent.keys() != subMenusDefinedByChildren.keys()) :
        logger.error("Configuration error in times.conf.  For each sub_menu key (%s) there must be an existing stanza (%s) and vice versa." % (subMenusDefinedByChildren.keys(),  subMenusPresent.keys()))
    
    # even though the sub_menu = <sub_menu_name> items will be 
    # sorted in alongside the main-level items, and other sub-menu items, 
    # the client code will pull out all sub_menu items and attach them to the 
    # is_sub_menu items. 
    # The thinking is that some clients (ResultsHeader.html) just want a flat list of timeranges and iterating over a tree structure would suck. 
    # other clients that need the tree (TimeRangePicker.html) will have to build it themselves using sub_menu and is_sub_menu
    orderedStanzas.sort(cmp=lambda a,b: cmp(a['order'], b['order']))
    return orderedStanzas
    
