import lxml.etree as et
import base
import splunk.util
import xml.sax.saxutils as su

import logging
logger = logging.getLogger('splunk.appserver.lib.view.panel')


def createPanel(name):
    '''
    Factory method for creating an appropriate panel object based upon the
    name.  Returns an instance of a BasePanel subclass, or throws a
    NotImplementedError if no suitable mapper is found.
    
    This method works by inspecting all objects that sublcass BasePanel and
    attempting to match their matchTagName class attribute.
    '''
    
    if not name:
        raise ValueError, 'Cannot create panel from nothing'
        
    for key, obj in globals().items():
        try:
            if issubclass(obj, BasePanel) and name == obj.matchTagName:
                return obj()
        except:
            pass
    raise NotImplementedError, 'Cannot find object mapper for panel type: %s' % name



class BasePanel(base.ViewObject):
    '''
    Represents a view result display object.
    '''
    
    # define common XML node -> object property mappings
    commonNodeMap = [
        ('title', 'title'),
        ('earliestTime', 'searchEarliestTime'),
        ('latestTime', 'searchLatestTime')
    ]
    
    # define search mode XML node -> object property mappings
    searchModeNodeMap = [
        ('searchString', base.STRING_SEARCH_MODE),
        ('searchName', base.SAVED_SEARCH_MODE),
        ('searchTemplate', base.TEMPLATE_SEARCH_MODE),
        ('searchPostProcess', base.POST_SEARCH_MODE)
    ]
    
    
    
    def __init__(self):
        
        # init standard panel params
        self.title = None
        self.layoutPanel = 'results'
        self.autoRun = True
        self.options = {}
        
        # init standard search configuration params
        self.searchMode = None
        self.searchCommand = None
        self.searchEarliestTime = None
        self.searchLatestTime = None
        self.searchFieldList = []

        
    def fromXml(self, lxmlNode):
        '''
        Extracts common panel attributes from the XML
        '''
        
        # import common nodes
        for nodeName, memberName in self.commonNodeMap:
            val = lxmlNode.findtext(nodeName)
            if val:
                setattr(self, memberName, val)
        
        # option params get their own container        
        for node in lxmlNode.findall('option'):
            self.options[node.get('name')] = node.text
        
        # handle different search modes
        for pair in self.searchModeNodeMap:
            if lxmlNode.find(pair[0]) != None:
                self.searchMode = pair[1]
                self.searchCommand = lxmlNode.findtext(pair[0])
                break

        # handle field lists
        if lxmlNode.find('fields') != None:
            self.searchFieldList = splunk.util.stringToFieldList(lxmlNode.findtext('fields'))
                      
          
    def toXml(self):
        '''
        Serializes the current panel object into its XML form.  Returns an
        lxml node.
        '''
        
        if not self.matchTagName:
            raise NotImplementedError, 'Unable to convert panel object to XML; panel.title=%s panel.searchMode=%s' % (self.title, self.searchMode)
            
        root = et.Element(self.matchTagName)

        for pair in self.searchModeNodeMap:
            if self.searchMode == pair[1]:
                et.SubElement(root, pair[0]).text = self.searchCommand
                break

        for pair in self.commonNodeMap:
            if getattr(self, pair[1]) != None:
                et.SubElement(root, pair[0]).text = getattr(self, pair[1])
            
        if self.searchFieldList:
            et.SubElement(root, 'fields').text = splunk.util.fieldListToString(self.searchFieldList)

        for option in sorted(self.options):
            elOption = et.SubElement(root, 'option')
            elOption.set('name', option)
            elOption.text = self.options[option]
            
        
        return root
          
        
    def toObject(self):
        '''
        Returns the expanded module object hierarchy needed to render this
        panel into a view
        '''
        raise NotImplementedError, 'The %s class has not properly implemented toObject()' % self.__class__
       
       
    def decorateSearchAndOptions(self, child):
        
        if not isinstance(child, dict):
            raise TypeError, 'Expected dict structure for param "child"'
        
        # process subpanels that aren't search generating    
        if self.searchMode == None:

            # add pager if asked
            if splunk.util.normalizeBoolean(self.options.get('showPager')):
                output = self._generatePagerModule()
                output['children'].append(child)
            else:
                output = child
            
            if self.title:
                output['params']['group'] = self.title
                output['params']['groupLabel'] = _(self.title)
            if self.layoutPanel:
                output['layoutPanel'] = self.layoutPanel
            return output
            

        # continue on with search generators
        
        searchParams = {
            'autoRun': self.autoRun
        }
        viewstateParams = {}

        if self.title:
            searchParams['group'] = self.title or ''
            searchParams['groupLabel'] = _(self.title) if self.title else '' 
            
        if self.searchMode in (base.STRING_SEARCH_MODE, base.TEMPLATE_SEARCH_MODE) :
            searchModuleName = 'HiddenSearch'
            searchParams['search'] = self.searchCommand
            searchParams['earliest'] = self.searchEarliestTime
            searchParams['latest'] = self.searchLatestTime
            
        elif self.searchMode == base.SAVED_SEARCH_MODE:
            searchModuleName = 'HiddenSavedSearch'
            searchParams['savedSearch'] = self.searchCommand
            viewstateParams['savedSearch'] = self.searchCommand

        elif self.searchMode == base.POST_SEARCH_MODE:
            searchModuleName = 'HiddenPostProcess'
            searchParams['search'] = self.searchCommand
            
        else:
            raise AttributeError, 'Unable to process panel because of unknown search mode: %s' % self.searchMode


        # construct base panel tree
        output = {
            'className': searchModuleName,
            'layoutPanel': self.layoutPanel,
            'params': searchParams,
            'children': [
                {
                    'className': 'ViewstateAdapter',
                    'params': viewstateParams,
                    'children': [
                        {
                            'className': 'HiddenFieldPicker',
                            'params': {
                                'fields': splunk.util.fieldListToString(self.searchFieldList) or '',
                                'strictMode': True
                            },
                            'children': [
                                {
                                    'className': 'JobProgressIndicator',
                                    'children': []
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        insertionPoint = output['children'][-1]['children'][-1]['children']
        
        # add pager if asked
        if splunk.util.normalizeBoolean(self.options.get('showPager')):
            insertionPoint.append(self._generatePagerModule())
        insertionPoint = insertionPoint[-1]['children']
        
        # append the data panel
        insertionPoint.append(child)
            
        return output
            
        
    def _generatePagerModule(self):
        '''
        Returns a dict instance of a pager module
        '''
        
        return {
            'className': 'Paginator',
            'params': {
                'count': self.options.get('count', 10),
                'entityName': 'results' # TODO: this needs an 'auto' mode
            },
            'children': []
        }
        
        
    jsonablePropertyList = ['title', 'searchMode', 'searchCommand', 'searchEarliestTime', 'searchLatestTime', 'options']
        
    def toJsonable(self):
        '''
        Returns a native structure that represents the current panel.  Used
        with the UI viewmaster framework.
        '''
        
        output = {
            'type': self.matchTagName
        }
        for k in self.jsonablePropertyList:
            output[k] = getattr(self, k)
        
        return output
        
        
    def fromJsonable(self, primitive):
        '''
        Parses an object primitive and populates the correct members
        '''
        
        for k in self.jsonablePropertyList:
            if primitive.get(k) != None:
                setattr(self, k, primitive.get(k))
        
        
    
class Chart(BasePanel):
    '''
    Represents a standard chart display of a search object
    '''
    
    # define the XML tag
    matchTagName = 'chart'
    
    
    def toObject(self):
        
        chartObjectParams = {
            'width': '100%'
        }
        if 'height' in self.options:
            chartObjectParams['height'] = self.options['height']

        chartFormatParams = {}
        for k in self.options:
            if k.startswith('charting.'):
                chartFormatParams[k] = self.options[k]
        
        output = {
            'className': 'HiddenChartFormatter',
            'params': chartFormatParams,
            'children': [
                {
                    'className': 'FlashChart',
                    'params': chartObjectParams,
                    'children': [
                        {
                            'className': 'ViewRedirectorLink',
                            'params': {
                                'viewTarget': 'flashtimeline'
                            }
                        }
                    ]
                }
            ]
        }
        output = self.decorateSearchAndOptions(output)
        
        return output
        
        
        
class Table(BasePanel):
    '''
    Represents a basic tabular output of search object
    '''        

    matchTagName = 'table'

            
    def toObject(self):

        resultsTableParams = {
            'allowTransformedFieldSelect': True
        }
        for k in ['count', 'displayRowNumbers']:
            if k in self.options:
                resultsTableParams[k] = self.options[k]
        
        output = {
            'className': 'SimpleResultsTable',
            'params': resultsTableParams,
            'children': [
                {
                    'className': 'ViewRedirectorLink',
                    'params': {
                        'viewTarget': 'flashtimeline'
                    }
                }
            ]
        }
        output = self.decorateSearchAndOptions(output)
        
        return output


        
class Event(BasePanel):
    '''
    Represents a raw event renderer
    '''        

    matchTagName = 'event'


    def toObject(self):

        eventViewerParams = {
        }
        for k in ['count', 'displayRowNumbers', 'entityName', 'segmentation', 'maxLines']:
            if k in self.options:
                eventViewerParams[k] = self.options[k]

        output = {
            #'className': 'HiddenFieldPicker',
            #'params': {
            #    'fields': splunk.util.fieldListToString(self.searchFieldList) or '',
            #    'strictMode': False
            #},
            #'children': [
            #    {
                    'className': 'EventsViewer',
                    'params': eventViewerParams
            #    }
            #]
        }
        output = self.decorateSearchAndOptions(output)

        return output


        
class List(BasePanel):
    '''
    Represents a basic list
    '''        

    matchTagName = 'list'


    def toObject(self):

        listerParamList = ['initialSortDir', 'labelFieldSearch', 'valueField', 'labelField', 'labelFieldTarget', 'initialSort']

        listerParams = {}
        for k in listerParamList:
            if k in self.options:
                listerParams[k] = self.options[k]
        
        output = {
            'className': 'LinkList',
            'params': listerParams
        }
        output = self.decorateSearchAndOptions(output)
        
        return output



class Single(BasePanel):
    '''
    Represents a single value panel.
    '''        

    matchTagName = 'single'


    def toObject(self):

        output = {
            'className': 'SingleValue',
            'params': self.options
        }
        output = self.decorateSearchAndOptions(output)

        return output

  
  
class Html(BasePanel):
    '''
    Represents a basic HTML content
    '''        

    matchTagName = 'html'


    def fromXml(self, lxmlNode):
        '''
        Override default parser and just get the text
        '''
        flatString = et.tostring(lxmlNode, method='html')
        flatString = flatString.replace('<html>', '').replace('</html>', '')
        self.options['rawcontent'] = flatString.strip('\r\n\t ')
        
        
    def toXml(self):
        
        xml = '<%s>%s</%s>' % (self.matchTagName, self.options['rawcontent'], self.matchTagName)
        return et.fromstring(xml)
        
        
    def toObject(self):
        
        output = {
            'className': 'StaticContentSample',
            'params': {'text': self.options['rawcontent']}
        }
        output = self.decorateSearchAndOptions(output)

        return output
              
