# Copyright (C) 2005-2010 Splunk Inc.  All Rights Reserved.  Version 4.0
import logging as logger
import os, manage_search, shutil, socket, subprocess, sys, xml.dom.minidom
import build_info, index, migration
import splunk.clilib.cli_common as comm
import control_exceptions as cex
import bundle_paths

ARG_DRYRUN    = "dry-run"

PARM_TRUE     = "true"

DEFAULT_SERVER_NAME  = "server_name_in_splunkd.xml"

# TODO move this to common
PATH_AUTH_DIR        = os.path.join(comm.splunk_home, "etc", "auth")
PATH_DB_MANIPULATOR  = os.path.join(comm.splunk_home, "bin", "dbmanipulator.py")
PATH_FTR_FILE        = os.path.join(comm.splunk_home, "ftr")
PATH_LICENSE_FILE    = os.path.join(comm.splunk_home, "license-eula.txt")
PATH_SELINUX_CONFIG  = os.path.join(os.path.sep, "etc", "selinux", "config")
PATH_GEN_AUDIT_KEYS  = os.path.join(comm.splunk_home, "bin", "genAuditKeys.py")
PATH_AUDIT_KEY_DIR   = os.path.join(comm.splunk_home, "etc", "auth", "audit")
PATH_AUDIT_PRIV_KEY  = os.path.join(PATH_AUDIT_KEY_DIR, "private.pem")
PATH_AUDIT_PUB_KEY   = os.path.join(PATH_AUDIT_KEY_DIR, "public.pem")

PATH_FIRST_INST_LOG  = os.path.join(comm.splunk_home, "var", "log", "splunk", "first_install.log")

ENV_IGNORE_SELINUX   = "SPLUNK_IGNORE_SELINUX"

#                         0-------------------------------------------------------------------------------80
PROMPT_MIGRATION     = """--------------------------------------------------------------------------------"""  + "\n" \
                                                                                                               + "\n" \
                       """Splunk has detected an older version of Splunk installed on this machine. To"""      + "\n" \
                       """finish upgrading to the new version, Splunk's installer will automatically"""        + "\n" \
                       """update and alter your current configuration files. Deprecated configuration"""       + "\n" \
                       """files will be renamed with a .deprecated extension."""                               + "\n" \
                                                                                                               + "\n" \
                       """You can choose to preview the changes that will be made to your configuration"""     + "\n" \
                       """files before proceeding with the migration and upgrade:"""                           + "\n" \
                                                                                                               + "\n" \
                       """If you want to migrate and upgrade without previewing the changes that will be"""    + "\n" \
                       """made to your existing configuration files, choose 'y'."""                            + "\n" \
                       """If you want to see what changes will be made before you proceed with the"""          + "\n" \
                       """upgrade, choose 'n'. """                                                             + "\n" \
                                                                                                               + "\n" \
                       """Perform migration and upgrade without previewing configuration changes? [y/n] """


###
###  things called by bin/splunk, etc.
###


def setDefaultHostname(argsDict, fromCLI):
  # validate the input
  paramReq = ("hostname",)
  paramOpt = ()
  comm.validateArgs(paramReq, paramOpt, argsDict)

  name = argsDict["hostname"]
  path = bundle_paths.make_path("inputs.conf")

  if name.count(" ") + name.count(":") > 0:
    raise cex.ArgError, "Name '%s' is invalid.  Names cannot contain spaces or colons." % name

  # read in current inputs.conf
  props = comm.readConfFile(path)

  # manip the [default] stanza
  if not 'default' in props.keys():
    props['default'] = {}

  props['default']['host'] = name

  # write out new inputs.conf
  comm.writeConfFile(path, props)


def checkXmlFiles(args, fromCLI):
  """
  Gathers all .xml files in path, recursively, and runs them through the
  minidom parser.
  """

  print "\tChecking configuration... ",; sys.stdout.flush()


  paramReq = ()
  paramOpt = ()
  comm.validateArgs(paramReq, paramOpt, args)

  retDict  = {}

  ################
  ################ BEGIN PART 1
  ################

  xmlErrs  = 0
  xmlFiles = []
  # path to search.
  dirPath  = os.path.join(comm.splunk_home, "etc")
  # gather .xml files into xmlFiles.
  xmlFiles = comm.findFiles(dirPath, "\\.xml$", caseSens = False, minBytes = 10)

  # run the cmd on appropriately sized subsets of xml files.
  for file in xmlFiles:
    try:
      xml.dom.minidom.parse(file)
    except:
      value = sys.exc_info()[1]
      logger.error("Error while parsing '%s':\n %s\n", file, value)
      xmlErrs += 1

  ### all done, report/prompt now.

  if xmlErrs > 0:
    if not comm.prompt_user("\nThere were problems with the configuration files.\nWould you like to ignore these errors? [y/n]:"):
      raise cex.ParsingError, "Parsing error in configuration files."

  print "Done." 
  return retDict


def firstTimeRun(args, fromCLI):
  """
  All of our first time run checks that used to happen in the former bin/splunk shell script.
  Does any number of things, such as config migration, directory validation, and so on.  For
  the most up to date info, read the code.  It tends to be fairly well documented.
  """

  paramReq = ()
  paramOpt = (ARG_DRYRUN,)
  comm.validateArgs(paramReq, paramOpt, args)

  isDryRun   = None # we check for this later.
  if ARG_DRYRUN in args:
    isDryRun = (PARM_TRUE == args[ARG_DRYRUN].lower())
  retDict  = {}

  # ...arg parsing done now.

  ##### prompt user for license acceptance as necessary
  if not comm.acceptLic:
    promptLicense()


  # none of the changes that are made in this function are subjected to isDryRun.
  # these things just have to be done - they're not considered to be migration.


  ##### if user doesn't have a splunkd.xml, put our default in its place.
  if not os.path.exists(migration.PATH_SPLUNKD_XML):
    comm.copyItem(migration.PATH_SPLUNKD_XML_DEF, migration.PATH_SPLUNKD_XML)
  ##### if user doesn't have a ldap.conf, put our default in its place.
  if not os.path.exists(migration.PATH_LDAP_CONF):
    comm.copyItem(migration.PATH_LDAP_CONF_DEF, migration.PATH_LDAP_CONF)
  if not os.path.exists(PATH_AUDIT_PRIV_KEY) and not os.path.exists(PATH_AUDIT_PUB_KEY):
    kScript, kPriv, kPub, kDir = PATH_GEN_AUDIT_KEYS, PATH_AUDIT_PRIV_KEY, PATH_AUDIT_PUB_KEY, PATH_AUDIT_KEY_DIR
    retCode = comm.runAndLog(["python", kScript, "-p", kPriv, "-k", kPub, "-d", kDir])
    if 0 != retCode:
      raise cex.FilePath, "Could not create audit keys (returned %d)." % retCode

  try:    
    keyScript = comm.getConfKeyValue("distsearch", "tokenExchKeys", "genKeyScript" );
    pubFilename = comm.getConfKeyValue("distsearch", "tokenExchKeys", "publicKey" );
    privateFilename = comm.getConfKeyValue("distsearch", "tokenExchKeys", "privateKey" );
    certDir = comm.getConfKeyValue("distsearch", "tokenExchKeys", "certDir" )
    certDir = os.path.expandvars( certDir )
    privateFilename = os.path.join( certDir,privateFilename )
    pubFilename = os.path.join( certDir, pubFilename )
    if not ( os.path.exists( os.path.join( certDir,privateFilename ) ) or os.path.exists( os.path.join( certDir, pubFilename ) ) ):
      cmdList = ["python", os.path.expandvars(keyScript), "-p", privateFilename, "-k", pubFilename,"-d", certDir ]
      success = comm.runAndLog( cmdList ) == 0
      if not success:
        logger.warn("Unable to generate distributed search keys."); #TK mgn 06/19/09
        raise cex.FilePath, "Unable to generate distributed search keys." #TK mgn 06/19/09
  except:
    logger.warn("Unable to generate distributed search keys."); #TK mgn 06/19/09
    raise 

  ##### set the default input hostname and dist search/etc servername if necessary
  try:
    newServerName = comm.getConfKeyValue("server", "general", "serverName")
  except:
    newServerName = ""
  if (0 == len(comm.grep("<serverName>", migration.PATH_SPLUNKD_XML))) and (newServerName == build_info.defaultServerName):
    logger.info("\nThis appears to be your first time running this version of Splunk.")
    open(PATH_FIRST_INST_LOG, "w").write(open(migration.PATH_VERSION_FILE, "r").read()) # dump version file only once, ever.
    hostname = socket.gethostname()
    setDefaultHostname({"hostname" : hostname}, fromCLI = True)
    manage_search.setLocalName({"name" : hostname}, fromCLI = True)
  ##### if the servername has already been set, we *have* to be in an upgrade situation.
  else:
    logger.info("\nThis appears to be an upgrade of Splunk.")
    ##### find out if we need to prompt people to ask if they want a migration preview.
    if None == isDryRun:
      isDryRun = False # assume 'no'
      # if this is being run from a terminal, go ahead and prompt the user.
      if comm.isTerminal: # TODO: check if this works as expected.  ie, does it skip the prompting in ssh?
        isDryRun = not comm.prompt_user("\n" + PROMPT_MIGRATION, checkValidResponse = True)

    ##### now do the actual migration (or fake it, if the user wants).
    #     upon faking, this function will throw an exception.
    migration.autoMigrate(isDryRun)


      

  ##### validate database directories...
  validateDirectories()


  ##### FTR succeeded, remove FTR marker.  johnvey is eggless. :(
  if os.path.exists(PATH_FTR_FILE):
    os.remove(PATH_FTR_FILE)

  # --- done w/ FTR, now i can has bucket?? ---
  return retDict


def preFlightChecks(args, fromCLI = False):
  paramReq = ()
  paramOpt = ()
  comm.validateArgs(paramReq, paramOpt, args)
  #
  checkDb({}, fromCLI)
  checkPerms()
  index.importAllFlatFiles({}, fromCLI)
  checkSearchthing()
  checkSELinux()


def checkDb(args, fromCLI = False):
  # looks like the 2 blocks here can prob be consolidated.  investigate (TODO)
  """
  replaces check_db.
  """
  paramReq = ()
  paramOpt = ()
  comm.validateArgs(paramReq, paramOpt, args)

  #
  print "\tChecking index directory... ",; sys.stdout.flush()
  
  if not os.path.exists(comm.splunk_db):
    logger.info("WARNING: Index directory '%s' does not exist." % comm.splunk_db)
    if not comm.prompt_user("Create a new index at '%s'? [y/n]: " % comm.splunk_db, checkValidResponse = True):
      raise cex.FilePath("Index directory does not exist, cannot continue.")
    validateDirectories()
  print "Done."
  validateDirectories()
  

###
###  things we call internally.
###

def callDbManipValidate():
  retCode = comm.runAndLog(["python", PATH_DB_MANIPULATOR, "--validate"])
  if 0 != retCode:
    raise cex.FilePath, "Could not validate indexes, will not continue (returned %d)." % retCode


def checkPerms():
  """
  Ensures that we can write to the dirs we need to write to.
  """
  testFilename = "permsTest"
  testDirs = (
    comm.splunk_home,
    comm.splunk_db
  )
  for oneDir in testDirs:
    testPath = os.path.join(oneDir, testFilename)
    try:
      comm.touch(testPath)
    except IOError:
      raise cex.FilePath, "Splunk is unable to write to the directory '%s' and therefore will not run.  Please check for appropriate permissions on this directory and its contents as necessary." % oneDir
    os.remove(testPath)


def checkSearchthing():
  """
  Performs a locking test on the splunk_db dir.
  """
  retCode = comm.runAndLog(["locktest"], logStdout = False)
  if 0 != retCode:
    raise cex.FilePath, "Locking test failed on filesystem in path '%s' with code '%d'.  Please file a case online at http://www.splunk.com/page/submit_issue" % (comm.splunk_db, retCode)

def promptLicense():
  if not comm.isWindows: # windows will probably display its license in the installer.
    if os.path.exists(PATH_LICENSE_FILE):
      if sys.stdin.isatty():
        subprocess.call(["more", PATH_LICENSE_FILE]) # TODO: test this on dirs w/ spaces.  prob works fine.
      else:
        logger.info(open(PATH_LICENSE_FILE, 'r').read())
    else: # don't barf in dev envs
      comm.out("Could not find license file.")
    if not comm.prompt_user("Do you agree with this license? [y/n]: ", checkValidResponse = True):
      raise cex.InputError("License refused - exiting.")


def checkSELinux():
  if ENV_IGNORE_SELINUX in os.environ:
    comm.out("Skipping SELinux check (to enable this check, unset the '%s' environment variable)." % ENV_IGNORE_SELINUX)
  else:
    if os.path.exists(PATH_SELINUX_CONFIG):
      comm.out("\tChecking for SELinux.")
      if "enforcing" in comm.sed("^SELINUX=(.*)", "\\1", PATH_SELINUX_CONFIG):
        raise cex.ArgError, "Splunk will not run with SELinux enabled.\nIf you have adjusted Splunk's security level with chcon, you can bypass this check by setting the '%s' environment variable." % ENV_IGNORE_SELINUX

def validateDirectories():
  """
  Ensures that various important paths exist and are directories.
  """
  splunkHomeDirs = (
    os.path.join("var", "lib"),
    os.path.join("var", "log", "splunk"),
    os.path.join("var", "run", "splunk"),
    os.path.join("var", "run", "splunk", "upload"),
    os.path.join("var", "spool", "splunk"),
    os.path.join("var", "spool", "dirmoncache")
  )

  splunkDbDirs = (
    "authDb",
    "hashDb"
  )

  dirsToCheck      = [os.path.join(comm.splunk_home, x) for x in splunkHomeDirs]
  dirsToCheck.extend([os.path.join(comm.splunk_db, x)   for x in splunkDbDirs  ])

  for oneDir in dirsToCheck:
    if not os.path.exists(oneDir):
      logger.info("\tCreating: %s" % oneDir)
      os.makedirs(oneDir)
    else:
      if not os.path.isdir(oneDir):
        raise cex.FilePath, "The path '%s' should be a directory, but is not.  Please file a case online at http://www.splunk.com/page/submit_issue"

  callDbManipValidate()
