# Copyright (C) 2005-2010 Splunk Inc.  All Rights Reserved.  Version 4.0
import logging as logger
import splunk.clilib.cli_common as comm
from control_exceptions import *
import glob, os, sys
import bundle_paths

if comm.isWindows:
	import splunk.windows as win

dbmanipulator_path = os.path.join(comm.splunk_home, "bin", "dbmanipulator.py")
fishbucket_path    = os.path.join(comm.splunk_db, "fishbucket")
PROMPT_CONTINUE    = "Are you sure you want to continue [y/N]? "

def dummy():
  pass

def skipConfirm(args):
  skip_confirmation = False
  if "force" in args.keys():
    # this condition should be pretty unnecessary...
    if isinstance(args["force"], str):
      if args["force"].lower() == "-f":
        skip_confirmation = True
      else:
        raise ArgError, "The value '%s' is not a valid force argument." % args["force"]
    else:
      if args["force"] == True: # don't one-line this - may not necessarily be a bool
        skip_confirmation = True
  return skip_confirmation


def cleanFishbucket():
  if os.path.exists(fishbucket_path):
    map(os.remove, comm.findFiles(fishbucket_path, ".*")) # .* == all files.


def cleanAll(args, fromCLI):
  """
  Removes ALL non-essential data - ie. metadata, indexes, users, saved queue data...
  """
  paramsReq = ()
  paramsOpt = ("force",)
  comm.validateArgs(paramsReq, paramsOpt, args)

  if comm.isLocalSplunkdUp():
    raise ServerState, "In order to clean, Splunkd must not be running."

  confirmed = True

  if fromCLI:
    skip_confirmation = skipConfirm(args)
    if not skip_confirmation:
        logger.info("This action will permanently erase all indexed event data, user information, global data, and internal logs; it cannot be undone.")
        confirmed = comm.prompt_user(PROMPT_CONTINUE)
                      
  if confirmed:
    logger.info("Cleaning all data...")
    for file in ("passwd", "unknown-props.xml", "unknown-types.xml"):
      fullPath = os.path.join(comm.splunk_home, "etc", file)
      if os.path.exists(fullPath):
        comm.remove_file(fullPath)

    for direc in (os.path.join("spool", "splunk"),
                  os.path.join("spool", "dirmoncache"),
                  os.path.join("spool", "splunkcopy"),
                  os.path.join("spool", "splunksym"),
                  os.path.join("log", "splunk"),
                  os.path.join("lib", "splunk", "bundles"),
                  os.path.join("lib", "splunk", "queues"),
                  os.path.join("lib", "splunk", "persistentstorage"),
                  os.path.join("run", "splunk", "dispatch"),
		  os.path.join("run", "splunk", "srtemp")):

      fullPath = os.path.join(comm.splunk_home, "var", direc)
      comm.remove_dirs(fullPath)
    # don't delete subdirs in the following dir - only files (recursively).
    spoolPath = os.path.join(comm.splunk_home, "var", "run", "splunk")
    map(os.remove, comm.findFiles(spoolPath, ".*")) # .* == all files.

    cleanUsers(                    {}, fromCLI = False)
    cleanMetadata({"indexes" : False}, fromCLI = False)

    cleanFishbucket()

    retCode = comm.runAndLog(["python", dbmanipulator_path, "--clean"])
    if 0 != retCode:
      raise FilePath, "Clean did not finish successfully."



def cleanIndex(args, fromCLI):
  """
  Removes all data in a given index, or all indexes if none is specified.
  """
  paramsReq = ()
  paramsOpt = ("index", "force",)
  comm.validateArgs(paramsReq, paramsOpt, args)

  if comm.isLocalSplunkdUp():
    raise ServerState, "In order to clean, Splunkd must not be running."

  confirmed  = True
  cmd        = ["python", dbmanipulator_path, "--clean"]
  cleanfish  = True
  if "index" in args.keys():
    cleanfish = ( args["index"] == "_thefishbucket" )
    cmd.append(args["index"])
    confirmMsg = "This action will permanently erase all events from the index '%s'; it cannot be undone." % args["index"]
  else:
    confirmMsg = "This action will permanently erase all events from ALL indexes; it cannot be undone."

  if fromCLI:
    skip_confirmation = skipConfirm(args)
    if not skip_confirmation:
      logger.info(confirmMsg)
      confirmed = comm.prompt_user(PROMPT_CONTINUE)

  if confirmed:

    #SPL-18401
    for direc in (os.path.join("lib", "splunk", "persistentstorage"),):
      fullPath = os.path.join(comm.splunk_home, "var", direc)
      comm.remove_dirs(fullPath)

    # the fishbucket is no longer a real index, but let's mimic old behavior and clean it here.
    if cleanfish:
      cleanFishbucket()

    ret = comm.runAndLog(cmd)
    if 0 != ret:
      raise FilePath, "There was an error while cleaning (code: %s)." % ret
    logger.info("Cleaned successfully.")



def cleanUsers(args, fromCLI):
  """
  Removes all users and related data.
  """
  paramsReq = ()
  paramsOpt = ("force",)
  comm.validateArgs(paramsReq, paramsOpt, args)

  if comm.isLocalSplunkdUp():
    raise ServerState, "In order to clean, Splunkd must not be running."

  confirmed    = True
  usersDB      = os.path.join(comm.splunk_home, "etc", "passwd")

  if fromCLI:
    skip_confirmation = skipConfirm(args)
    if not skip_confirmation:
      logger.info("This action will permanently erase all user accounts, Saved Splunks, and Live Splunks; it cannot be undone")
      confirmed = comm.prompt_user(PROMPT_CONTINUE)
  
  if confirmed:
    if fromCLI:
      logger.info("Removing user info.")

    if os.path.exists(usersDB):
      comm.remove_file(usersDB)



def cleanMetadata(args, fromCLI):
  """
  Removes all metadata (event type ids, event type tags, custom sourcetype names...)
  """
  paramsReq = ()
  paramsOpt = ("force", "indexes")
  comm.validateArgs(paramsReq, paramsOpt, args)

  if comm.isLocalSplunkdUp():
    raise ServerState, "In order to clean, Splunkd must not be running."

  dontCleanIndexes = ("indexes" in args and str(args["indexes"]).lower() == "false")

  confirmed        = True

  if fromCLI:
    skip_confirmation = skipConfirm(args)
    if not skip_confirmation:
      logger.info("This action will permanently erase all global event type ids, event type tags, and custom source type names; it cannot be undone.")
      confirmed = comm.prompt_user(PROMPT_CONTINUE)

  if confirmed:
    if fromCLI:
      logger.info("Removing global data ...")
    if not dontCleanIndexes:
      cleanFishbucket()
    comm.remove_dirs(os.path.join(comm.splunk_db, "hashDb"))
    # remove all the *.conf and *-model.xml files as well
    bundle, file = ("learned", "*.conf")
    learned = (glob.glob(bundle_paths.make_bundle_path(bundle, file)) +
               glob.glob(bundle_paths.make_legacy_bundle_path(bundle, file)));
    for learnedFile in learned:
      comm.remove_file(learnedFile)

