import os, sys, platform, shutil, tarfile
import stat
import logging as logger

SPLUNK_HOME   = os.environ['SPLUNK_HOME']
RESULTS_DIR   = os.path.join(SPLUNK_HOME, 'var', 'run', 'splunk', 'diag-temp','splunk-diag')
RESULTS_TGZ   = os.path.join(SPLUNK_HOME, os.path.basename(RESULTS_DIR) + ".tar.gz")
MSINFO_FILE   = 'msinfo-sum.txt'
SYSINFO_FILE  = 'systeminfo.txt'
COMPOSITE_XML = os.path.join(SPLUNK_HOME, 'var', 'run', 'splunk', 'composite.xml')

system_info = None # systeminfo.txt file obj; if you want to use reporting
                   # functions externally, set this as desired.

def splunkVersion():
    """ Use splunk version to figure the version"""

    system_info.write('\n\n********** Splunk Version **********\n\n')
    sver = os.popen('splunk version')
    system_info.write(sver.read())


# this uses python's uname function to get info in a cross-platform way.
def systemUname():
    """ Python uname output """ 

    system_info.write('\n\n********** Uname **********\n\n')
    suname = platform.uname()
    system_info.write(str(suname))
    #SPL-18413
    system_info.write("\n")
    splunkdpath = os.path.join(SPLUNK_HOME, 'bin', 'splunkd') 
    if suname[0] == 'Windows':
        splunkdpath += ".exe"
    arch = str(platform.architecture(splunkdpath)) 
    system_info.write(arch)


def networkConfig():
    """ Network configuration  """ 

    system_info.write('\n\n********** Network Config  **********\n\n')
    # we call different utilities for windows and "unix".
    if os.name == "posix":
        # if running as a non-root user, you may not have ifconfig in your path.
        # we'll attempt to guess where it is, and if we can't find it, just
        # assume that it is somewhere in your path.
        ifconfigCmd = '/sbin/ifconfig'
        if not os.path.exists(ifconfigCmd):
            ifconfigCmd = 'ifconfig'
        system_info.write(os.popen('%s -a' % ifconfigCmd).read())
    else:
        system_info.write(os.popen('ipconfig /all').read())


def networkStat():
    """ Network Status """ 

    system_info.write('\n\n********** Network Status **********\n\n')
    # just like with ifconfig, we attempt to guess the path of netstat.
    # if we can't find it, we leave it up to it being in your path.
    # also, if we're on windows, just go with the path-less command.
    netstatCmd = '/bin/netstat'
    if not os.name == "posix" or not os.path.exists(netstatCmd):
        netstatCmd = 'netstat'
    system_info.write(os.popen("%s -a -n" % netstatCmd).read())


def systemResources():
    """ System Memory """ 

    # on windows, we use msinfo to get all the relevant output.
    if os.name == "posix":
        system_info.write('\n\n********** System Ulimit **********\n\n')
        system_info.write(os.popen('ulimit -a').read())
        system_info.write('\n\n********** System Memory **********\n\n')
        suname = platform.uname()
        #SPL-17593
        if suname[0] == 'SunOS':
            system_info.write(os.popen('vmstat').read())
        #SPL-16915
        elif suname[0] == 'Darwin':
            system_info.write(os.popen('vm_stat').read())
        else:
            system_info.write(os.popen('free').read())
        system_info.write('\n\n********** DF output **********\n\n')
        system_info.write(os.popen('df').read())
    else:
        os.popen('start /wait msinfo32.exe /report "%s" /categories +SystemSummary' % MSINFO_FILE)
        shutil.move(MSINFO_FILE, RESULTS_DIR)

def splunkDBPaths():
    # using btool, get the merged indexes.conf listing.
    btoolLines = os.popen('btool indexes list')

    # now extract any lines that have Path in them.
    pathLines  = filter(lambda x: x.count("Path") > 0, btoolLines)

    # separate each line on the = (key=val), take the val, and strip surrounding whitespace.
    paths = map(lambda x: x.split("=")[1].strip(), pathLines)

    def normalize_path(btool_path):
        # SPL-25568 first try to hack around use of '/' on Windows.. soo ugly
        # 'safe' because / is not legal path character on win32.  
        # bundle layer should normalize this to universal seperator or
        # always platform specific.
        if os.name == "nt":
            if (not '\\' in btool_path) and ('/' in btool_path):
                btool_path = btool_path.replace('/','\\')

        # if it's absolute, there's no processing to do here.
        # (also fixes SPL-26129)
        if os.path.isabs(btool_path):
            return btool_path

        # btool normally gives us $VARS in the path (how it is in file)
        # This is clumsy but workable on unix, but nonfunctional on windows
        # so substitute the full path in there manually.  
        # Start by splitting the very first section of the path, and checking to
        # see if it's a unix variable.
        try:
            pathBegin, pathEnd = btool_path.split(os.path.sep, 1)
        except ValueError:
            err_mesg = "No directory seperator found in index path: %s\n" % btool_path
            sys.stderr.write(err_mesg)
            system_info.write(err_mesg)
            return None

        if pathBegin.startswith('$'):
            # if it does start with $, remove the $ and lookup the remainder in our env vars,
            # replacing the variable with the result.
            try:
                pathBegin = os.environ[pathBegin[1:]]
                # rebuild the path
                return os.path.join(pathBegin, pathEnd)
            except KeyError, key_e:
                # only log error, let dir or ls show the problem as well
                err_mesg = "No env var available from index path: %s\n" % key_e.args[0]
                sys.stderr.write(err_mesg)
                system_info.write(err_mesg)

        return btool_path


    paths = map(normalize_path, paths)
    paths = filter(None, paths) # remove chaff from normalize_paths
    return paths

def splunkDBListing():
    """ Index Listing Output""" 
    dbPaths = splunkDBPaths()

    system_info.write('\n\n********** find **********\n\n')
    for dbPath in dbPaths:
        if os.name == "posix":
            # get a recursive listing of that path.
            # SPL-19438 -- show if dir is symlink
            system_info.write("\nls -ld \"%s\" \n" % dbPath)
            system_info.write(os.popen("ls -ld \"%s\"\n" % dbPath).read())
            system_info.write("\nls -lR \"%s\"\n" % dbPath)
            system_info.write(os.popen("ls -lR \"%s\"" % dbPath).read())
        else:
            system_info.write(os.popen("dir /s/a \"%s\"" % dbPath).read())

def get_recursive_dir_info(dir_path, dest_file):
    """ do a recursive listing of the given dir and write the results on the given file """
    if not os.path.exists(os.path.dirname(dest_file)):
        os.makedirs(os.path.dirname(dest_file)) 
    f = open(dest_file, 'w') 
    try:
       if os.name == "posix":
          f.write("ls -lR \"%s\"\n" % dir_path)
          f.write(os.popen("ls -lR \"%s\"" % dir_path).read())
       else:
          f.write("dir /s/a \"%s\"\n" % dir_path)
          f.write(os.popen("dir /s/a \"%s\"" % dir_path).read())
    finally: 
       f.close()

def get_worddata_filenames(path):
    "return relative paths for all .data files under a given path (hopefully a SPLUNK_DB or index)"
    filenames = []

    for dir, subdirs, files in os.walk(path):
        data_files = filter(lambda s: s.endswith(".data"), files)
        relative_dir = dir[len(path)+1:]
        # prepend the directory
        paths = map( lambda fname: os.path.join(relative_dir, fname), data_files)
        filenames.extend(paths)
    return filenames

def copy_file_with_parent(src_file, dest_file):
    "copy a file while creating any parent directories for the target"
    # ensure directories exist
    dest_dir = os.path.dirname(dest_file)
    if not os.path.isdir(dest_dir):
        os.makedirs(os.path.dirname(dest_file))

    shutil.copy(src_file, dest_file)


def copy_worddata(target_path):
    "copy all .data files in all indices to the results dir"

    worddata_filenames = []

    for dbPath in splunkDBPaths():
        rel_paths = get_worddata_filenames(dbPath)
        # make absolute paths
        abs_paths = map(lambda f: os.path.join(dbPath, f), rel_paths)
        worddata_filenames.extend(abs_paths)

    for src_file in worddata_filenames:
        dest_file = src_file

        # trim SPLUNK_HOME from target directory
        if dest_file.startswith(SPLUNK_HOME):
                dest_file = dest_file[len(SPLUNK_HOME):]

        # windows tomfoolery....
        # make an informative path to store the worddata file
        if os.name != 'posix':
            unc, dest_file = os.path.splitunc(dest_file)
            # yank initial backslashes off UNC expression for constructing
            # relative path later
            unc_dirnames = unc.lstrip('\\')
        else:
            unc_dirnames = ''
        # capture drive letter if exists
        drive , dest_file = os.path.splitdrive(dest_file)
        if drive:
            # letter only, since windows is unahppy with colons in dirnames
            drive = drive[0]

        # make relative, for join to work
        if dest_file[0] == os.sep:
            dest_file = dest_file[1:]

        dest_file = os.path.join(target_path, unc_dirnames, drive, dest_file)

        copy_file_with_parent(src_file, dest_file)

def copy_dispatch_dir(target_path):
     dispatch_dir = os.path.join(SPLUNK_HOME, "var", "run", "splunk", "dispatch") 
    
     # collect info iff dispatch dir is present
     if not os.path.exists(dispatch_dir):    
         return
 
     get_recursive_dir_info(dispatch_dir,  os.path.join(target_path, "dir_info.txt")) 
     for job in os.listdir(dispatch_dir):
         job_dir = os.path.join(dispatch_dir, job)
         if not os.path.isdir(job_dir):
            continue
         
         get_recursive_dir_info(job_dir, os.path.join(target_path, job, "dir_info.txt"))
         # copy only files in the job's dir, skip all the .gz stuff
         for f in os.listdir(job_dir):
            src_file = os.path.join(job_dir, f)
            if f.endswith('.gz') or f.startswith('results') or os.path.isdir(src_file):
               continue
                   
            copy_file_with_parent(src_file, os.path.join(target_path, job, f)) 
     

def deleteTree(path):
    """ Delete a directory, if it exists. """
    def handle_readonly_filedir_errors(func_called, path, exc):
        "try to make this more like rm -rf"
        error_type = exc[1]
        full_perms = stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO

        parent = os.path.dirname(path)
        os.chmod(parent, full_perms)
        os.chmod(path, full_perms)
        if func_called == os.rmdir:
            os.rmdir(path)
        elif func_called == os.remove:
            os.remove(path)
        else:
            raise
    if os.path.isdir(RESULTS_DIR):
        shutil.rmtree(RESULTS_DIR, onerror=handle_readonly_filedir_errors)

def main():
    try:
        logger.info("Ensuring clean temp dir...")
        # delete any old results tarball, and create a fresh results directory
        deleteTree(RESULTS_DIR)
    except Exception, e:
        #logger.error(e)
        logger.error("""Couldn't initally clean diag temp directory.  
To work around move %s aside, or delete it.
A listing of the entire contents of this directory will help us identify the problem.""" % RESULTS_DIR)
        raise

    os.makedirs(RESULTS_DIR)
    if os.path.exists(RESULTS_TGZ):
        os.unlink(RESULTS_TGZ)
    
    # make sure it's a supported os.
    if not os.name in ("posix", "nt"):
        logger.error("Unsupported OS (%s)." % os.name)
        sys.exit(1)

    logger.info("Starting splunk diag...")

    try:
        try:
            global system_info
            system_info = open(os.path.join(RESULTS_DIR, SYSINFO_FILE), 'w')
        except IOError:
            logger.error("Exiting: Cannot create system info file.  Permissions may be wrong.")
            sys.exit(1)

        #get the splunk version
        splunkVersion()
        
        #uname
        systemUname()
        
        #ifconfig
        networkConfig()
        
        #netstat
        networkStat()
        
        #ulimit
        systemResources()

        #ls
        splunkDBListing()
    finally:
        system_info.close()
    
    #copy etc and log into results too
    logger.info("Copying Splunk configuration files...")
    shutil.copytree(os.path.join(SPLUNK_HOME, "etc"), os.path.join(RESULTS_DIR, "etc"), symlinks = True)

    # quick hack for SPL-26622  -- will be adding filters for later SPLs
    secret_filename = os.path.join(RESULTS_DIR, "etc", "auth", "splunk.secret")
    os.unlink(secret_filename)

    logger.info("Copying Splunk log files...")
    shutil.copytree(os.path.join(SPLUNK_HOME, "var", "log", "splunk"), \
                    os.path.join(RESULTS_DIR, "log"), symlinks = True)

    # get worddata "metadata" files
    logger.info("Copying index worddata files...")
    copy_worddata(RESULTS_DIR)

    if not os.path.exists(COMPOSITE_XML):
        logger.warn("Unable to find composite.xml file, product has likely not been started.")
    else:
        shutil.copy(COMPOSITE_XML, RESULTS_DIR)

    logger.info("Copying Splunk log files in the dispatch dir...")
    copy_dispatch_dir(os.path.join(RESULTS_DIR, "dispatch"))

    #create a tar.gz file out of the results dir
    logger.info("Creating archive file...")
    destFile = tarfile.open(RESULTS_TGZ, 'w:gz')
    destFile.add(RESULTS_DIR, os.path.basename(RESULTS_DIR))
    destFile.close()

    try:
        #clean up the results dir
        logger.info("Cleaning up...")
        deleteTree(RESULTS_DIR)
    finally:
        logger.info("Splunk diagnosis file created: %s" % RESULTS_TGZ)


def pclMain(args, fromCLI):
    return main()

if __name__ == "__main__":
    # get logging behavior similar to normal splunk invocation
    # logging apis suck, renaming them sucks more
    sh = logger.StreamHandler()
    sh.setFormatter(logger.Formatter("%(message)s"))
    l = logger.getLogger()
    l.setLevel(logger.INFO)
    l.addHandler(sh)
    main()
