import logging as logger
import os, stat, subprocess, sys
if sys.platform == "win32":
  import win32api
  import winerror
  import win32service
import splunk.clilib.cli_common as comm
import splunk.clilib.control_exceptions as cex


# NOTE: since we use os.popen all over this script, it is not safe to call these
# functions from SplunkWeb/twistd.  if we want to expose this feature in the UI,
# we'll need to change things around to not collide with twistd.

from splunk.clilib import build_info
SVC_SPLUNKD   = build_info.SVC_SPLUNKD
SVC_SPLUNKWEB = build_info.SVC_SPLUNKWEB

# User under which services should run.  Should be specified in the format
# 'domain\user'.  Set to None to run as "LOCAL SYSTEM".
SVC_USER      = None

WEB_SVC_PATH  = os.path.join(comm.splunk_home, "bin", "SplunkWebService.py")

WIN_ON_BOOT   = "AUTO_START"


#####################################################################
# Entrypoints (enable/disable/status)                               #
#####################################################################


def add_args(listOrStr):
  isStr = (str == type(listOrStr))
  if (comm.acceptLic):
    listOrStr += (isStr and (" " + "--accept-license") or (["--accept-license"]))
  if (comm.answerYes):
    listOrStr += (isStr and (" " + "--answer-yes")     or (["--answer-yes"]))
  if (comm.isNoPrompt):
    listOrStr += (isStr and (" " + "--no-prompt")      or (["--no-prompt"]))
  return listOrStr

def en_bootStartLoop(args, fromCLI):
  """
  This cmd is used by windows - it's easier to do it here than in installshield.
  """
  ranOnce = False
  isGreatSuccess = False
  while not isGreatSuccess:
    if ranOnce:
      comm.prompt_user("\nPress [Enter] to try again...")
    else:
      ranOnce = True
    cmdList = ["splunk", "enable", "boot-start"]
    cmdList = add_args(cmdList)
    proc = subprocess.Popen(cmdList)
    proc.wait()
    isGreatSuccess = (proc.returncode == 0)

def en_bootStart(args, fromCLI):
  paramsReq = ()
  paramsOpt = (PARM_OS, PARM_RCDIR, PARM_USER, PARM_VARIANT)
  comm.validateArgs(paramsReq, paramsOpt, args)

  osName, user, variant, rcdir, script = getSpecs(args)

  # do first time run if necessary, so it doesn't end up happening at boot time.
  cmdStr = "splunk ftr"
  cmdStr = add_args(cmdStr)
  os.system(sucmd(user, cmdStr))

  if os.path.exists(os.path.join(comm.splunk_home, "ftr")):
    raise cex.ArgError, "First-time-run has not finished.  Ignore this error when previewing migration - exiting."

  setupBootStart(osName, user, variant, rcdir, script) # script gets enabled in here, too.

  show_bootStart(args, fromCLI)


def dis_bootStart(args, fromCLI):
  paramsReq = ()
  paramsOpt = (PARM_OS, PARM_RCDIR, PARM_USER, PARM_VARIANT)
  comm.validateArgs(paramsReq, paramsOpt, args)

  osName, user, variant, rcdir, script = getSpecs(args)

  disableInit(osName, variant, rcdir, script)
  if UNAME_WIN32 == osName:
    # remove the windows service from the configuration
    os.system("splunkd uninstall")
    os.system("python \"%s\" remove" % WEB_SVC_PATH)
  elif UNAME_AIX == osName:
    for service in [ "splunkd", "splunkweb" ]:
      os.system("/usr/bin/rmssys -s %s >/dev/null 2>/dev/null" % service)
  else:
    if os.path.exists(script):
      os.remove(script)
    if UNAME_OSX == osName:
      if os.path.exists(rcdir + "/StartupParameters.plist"):
        os.remove(rcdir + "/StartupParameters.plist")
      try:
        os.rmdir(rcdir)
      except:
        pass

  if fromCLI:
    logger.info("Disabled.")


def show_bootStart(args, fromCLI):
  paramsReq = ()
  paramsOpt = (PARM_OS, PARM_RCDIR, PARM_USER, PARM_VARIANT)
  comm.validateArgs(paramsReq, paramsOpt, args)

  osName, user, variant, rcdir, script = getSpecs(args)

  retDict = {}

  if UNAME_WIN32 == osName:
    # see if the service exists first.
    retDict["scriptExists"] = True
    for svc in (SVC_SPLUNKD, SVC_SPLUNKWEB):
      if svc_isInstalled(svc) == False:
        retDict["scriptExists"] = False
	break
    # if there's no services, they sure can't be enabled. :)
    if not retDict["scriptExists"]:
      retDict["initEnabled"] = False
    # otherwise make sure they're enabled, just like above.
    else:
      retDict["initEnabled"] = True
      for svc in (SVC_SPLUNKD, SVC_SPLUNKWEB):
        if svc_isAutostart(svc) == False:
          retDict["initEnabled"] = False
	  break
  else:
    if UNAME_AIX == osName:
      chken = os.popen("/usr/bin/lssrc -g splunk >/dev/null")
      retDict["scriptExists"] = (chken.close() is None)
    else:
      retDict["scriptExists"] = os.path.exists(script)
    retDict["initEnabled"]  = checkInit(osName, variant, rcdir, script)

  if fromCLI:
    if UNAME_WIN32 == osName:
      if retDict["scriptExists"]:
        logger.info("Windows services installed.")
      else:
        logger.info("Windows services not installed.")
      desc_script = "Windows services are"
    elif UNAME_AIX == osName:
      if retDict["scriptExists"]:
        logger.info("SRC subsystem group installed.")
      else:
        logger.info("SRC subsystem group not installed.")
      desc_script = "SRC subsystem group is"
    else:
      if retDict["scriptExists"]:
        logger.info("Init script installed at %s." % script)
      else:
        logger.info("Init script is not installed (checked: %s)." % script)
      desc_script = "Init script is"

    if retDict["initEnabled"]:
      logger.info("%s configured to run at boot." % desc_script)
    else:
      logger.info("%s not configured to run at boot." % desc_script)


#####################################################################
# Constants                                                         #
#####################################################################


# parameters

PARM_OS       = "os"
PARM_RCDIR    = "rc-dir"
PARM_USER     = "user"
PARM_VARIANT  = "variant"

# some consts we use to compare against os.uname() output.
# used to figure out which OS we're on.

UNAME_LINUX   = "linux"
UNAME_FBSD    = "freebsd"
UNAME_SOL_OLD = "sunos"
UNAME_OSX     = "darwin"
UNAME_AIX     = "aix"
UNAME_WIN32   = "win32"

# where in the sysv init process do we start Splunk...

START_ORDER   = 80
STOP_ORDER    = 100 - START_ORDER

# some distro labels.  mostly used to keep track of things...

DISTRO_DUNNO  = "unknown"
DISTRO_RH     = "redhat"
DISTRO_DEB    = "debian"
DISTRO_SLACK  = "slackware"
DISTRO_SUSE   = "suse"

ERR_NOERR     = "0"
ERR_USAGE     = "1"

MODE_CHECK    = "check"
MODE_DIS      = "disable"
MODE_EN       = "enable"

# mapping of distributions and which files can be used to identify them.
linuxFileMap = [
  [ DISTRO_RH,    "/etc/redhat-release" ],
  [ DISTRO_RH,    "/etc/fedora-release" ],
  [ DISTRO_DEB,   "/etc/debian_version" ],
  [ DISTRO_SLACK, "/etc/slackware-version" ],
  [ DISTRO_SUSE,  "/etc/SuSE-release" ],
]

# our list of "functions" files to look for (and possibly include).
FUNC_FILE_PATHS = ("/etc/init.d/functions", "/etc/rc.d/init.d/functions")

# no need for this thing to change, just plop it down later.
CODE_CASE="""
case "$1" in
  start)
    splunk_start
    ;;
  stop)
    splunk_stop
    ;;
  restart)
    splunk_restart
    ;;
  status)
    splunk_status
    ;;
  *)
    echo "Usage: $0 {start|stop|restart|status}"
    exit %s
esac
""" % ERR_USAGE




#####################################################################
# Distribution-specific metadata                                    #
#####################################################################


# many distros/*nixes tend to have configuration utilities (such as chkconfig).
# for each distro that we know of, add the config metadata here so we can pick
# it out later.
METADATA = {}

# starting with the # makes this fit in the comment header a bit cleaner...
METADATA[DISTRO_RH]="""#
# chkconfig: 2345 90 60
# description: Splunk indexer service"""

METADATA[DISTRO_SLACK]="""#
# please ensure that /etc/rc.d/rc.sysvinit is executable."""

METADATA[DISTRO_SUSE]="""#
### BEGIN INIT INFO
# Provides: splunkd
# Required-Start: $local_fs $remote_fs
# Should-Start: splunkd
# Required-Stop: $local_fs $remote_fs
# Default-Start: 3 5
# Default-Stop: 0 1 2 6
# Short-Description: Start splunk
# Description: Splunk indexer service
### END INIT INFO"""



#####################################################################
# Manage the various init script systems.                           #
#####################################################################

# returns the runlevels that we should be started and stopped in
def runlevels(ostype):
  if UNAME_SOL_OLD == ostype:
     return (["3"], ["0", "1", "2", "S"])
  return (["3", "4", "5"], ["0", "1", "6"])

def makeInitdSymlink(dest):
  if not os.path.exists(dest):
    os.symlink("../init.d/splunk", dest)

def editConfFile(fname, var, newval):
  contents = ""
  modified = False
  found = False
  hand = open(fname)
  while True:
    line = hand.readline()
    if not line: break
    if line.strip().startswith(var):
      found = True
      if not line.strip().startswith(var + newval):
        modified = True
	line = var + newval + "\n"
    contents += line
  hand.close()
  if not found:
    contents += var + newval + "\n"
    modified = True
  if modified:
    hand = open(fname, "w")
    hand.write(contents)
    hand.close()

def removeConfFileLine(fname, var):
  contents = ""
  modified = False
  hand = open(fname)
  while True:
    line = hand.readline()
    if not line: break
    if line.strip().startswith(var):
      modified = True
    else:
      contents += line
  hand.close()
  if modified:
    hand = open(fname, "w")
    hand.write(contents)
    hand.close()

# enable, disable, and check status of splunk via chkconfig.
def enableChkconfig(ostype, rcDir):
  logger.info(os.popen("chkconfig splunk on").read())
def disableChkconfig(ostype, rcDir):
  logger.info(os.popen("chkconfig splunk off").read())
def isEnabledChkconfig(ostype, rcDir):
  # if for some reason the init script doesn't exist when we check this, don't spit out an error.
  pieces = os.popen("chkconfig --list splunk 2>/dev/null").read().split(" ")
  if len(pieces) > 1:
    onesThatSayOn = filter(lambda x: x.endswith(":on"), pieces[1:])
    return (onesThatSayOn > 0)
  return False

# there's no way to check the status with update-rc.d.  we'll use the symlink check.
def enableUpdateRcD(ostype, rcDir):
  logger.info(os.popen("update-rc.d splunk defaults").read())
def disableUpdateRcD(ostype, rcDir):
  logger.info(os.popen("update-rc.d -f splunk remove").read())

# manually create symlinks for the various run levels...
# we'll use this for unknown distros (and slackware :))
def enableSysVSymlinks(ostype, rcDir):
  START_RUN_LEVELS, STOP_RUN_LEVELS = runlevels(ostype)
  for level in START_RUN_LEVELS:
    sympath = "%s/rc%s.d" % (rcDir, level)
    if not os.path.exists(sympath):
      os.makedirs(sympath)
    makeInitdSymlink("%s/S%dsplunk" % (sympath, START_ORDER))
  for level in STOP_RUN_LEVELS:
    sympath = "%s/rc%s.d" % (rcDir, level)
    if not os.path.exists(sympath):
      os.makedirs(sympath)
    makeInitdSymlink("%s/K%dsplunk" % (sympath, STOP_ORDER))

# manually delete symlinks for the various run levels...
# we'll use this for unknown distros (and slackware :))
def disableSysVSymlinks(ostype, rcDir):
  START_RUN_LEVELS, STOP_RUN_LEVELS = runlevels(ostype)
  for level in START_RUN_LEVELS:
    sympath = "%s/rc%s.d" % (rcDir, level)
    os.remove("%s/S%dsplunk" % (sympath, START_ORDER))
  for level in STOP_RUN_LEVELS:
    sympath = "%s/rc%s.d" % (rcDir, level)
    os.remove("%s/K%dsplunk" % (sympath, STOP_ORDER))

# manually check and see if there's a splunk start symlink in at least one runlevel.
# we need this for slackware naturally, but also debian's update-rc.d won't tell us, so we use it there as well.
def isEnabledSysVSymlinks(ostype, rcDir):
  START_RUN_LEVELS, STOP_RUN_LEVELS = runlevels(ostype)
  foundStart = False
  for level in START_RUN_LEVELS:
    # explore /etc/rc.d/rc3.d or whatever.
    symPath = "%s/rc%s.d" % (rcDir, level)
    for foo, bar, files in os.walk(symPath):
      # match things like "S80splunk".
      matches = filter(lambda x: x.startswith("S") and x.endswith("splunk"), files)
      if len(matches) > 0:
        foundStart = True
        break
  return foundStart



#
# setup a dictionary of which functions to use for which distros.
#

INITMGMT = {}
INITMGMT[DISTRO_DEB]   = {MODE_CHECK : isEnabledSysVSymlinks, MODE_DIS : disableUpdateRcD,    MODE_EN : enableUpdateRcD}
INITMGMT[DISTRO_RH]    = {MODE_CHECK : isEnabledChkconfig,    MODE_DIS : disableChkconfig,    MODE_EN : enableChkconfig}
INITMGMT[DISTRO_SLACK] = {MODE_CHECK : isEnabledSysVSymlinks, MODE_DIS : disableSysVSymlinks, MODE_EN : enableSysVSymlinks}
INITMGMT[DISTRO_SUSE]  = {MODE_CHECK : isEnabledChkconfig,    MODE_DIS : disableChkconfig,    MODE_EN : enableChkconfig}
# don't know the distro? go with generic sysvinit style symlinks - no chkconfig, no update-rc.d, etc.
INITMGMT[DISTRO_DUNNO] = {MODE_CHECK : isEnabledSysVSymlinks, MODE_DIS : disableSysVSymlinks, MODE_EN : enableSysVSymlinks}

#
# our generic entrypoints for enabling/disabling the init scripts on the various distros.
#

def enableInit(ostype, variant, rcdir, script):
  if UNAME_LINUX == ostype:
    (INITMGMT[variant][MODE_EN])(ostype, rcdir) # we'll always have a valid variant - unknown == DISTRO_DUNNO.
    if DISTRO_SLACK == variant:
      # if we're going as far as to chkconfig people's systems, might as well make sure this file is proper, too.
      setFileExecutable(rcdir + "/rc.d/rc.sysvinit")
  elif UNAME_SOL_OLD == ostype:
    enableSysVSymlinks(ostype, rcdir)
  elif UNAME_FBSD == ostype:
    editConfFile("/etc/rc.conf", "splunk_enable=", "\"YES\"")
  elif UNAME_OSX == ostype:
    editConfFile("/etc/hostconfig", "SPLUNK=", "-YES-")
  elif UNAME_AIX == ostype:
    editConfFile("/etc/inittab", "splunk:", "2:once:/usr/bin/startsrc -g splunk > /dev/console 2>&1")
  elif UNAME_WIN32 == ostype:
    # set the service as start on boot.
    for oneSvc in (SVC_SPLUNKD, SVC_SPLUNKWEB):
      svc_setAutostart(oneSvc)
  else:
    raise cex.ArgError, ("This operation not supported on OS '%s'." % ostype)


def disableInit(ostype, variant, rcdir, script):
  if UNAME_WIN32 == ostype:
    # set the service to NOT start on boot (only manually).
    for oneSvc in (SVC_SPLUNKD, SVC_SPLUNKWEB):
      svc_setAutostart(oneSvc, False)
  elif UNAME_LINUX == ostype:
    (INITMGMT[variant][MODE_DIS])(ostype, rcdir)
  elif UNAME_SOL_OLD == ostype:
    disableSysVSymlinks(ostype, rcdir)
  elif UNAME_FBSD == ostype:
    removeConfFileLine("/etc/rc.conf", "splunk_enable=")
  elif UNAME_OSX == ostype:
    removeConfFileLine("/etc/hostconfig", "SPLUNK=")
  elif UNAME_AIX == ostype:
    removeConfFileLine("/etc/inittab", "splunk:")
  else:
    raise cex.ArgError, ("This operation not supported on OS '%s'." % ostype)


def checkInit(ostype, variant, rcdir, script):
  if UNAME_LINUX == ostype:
    return (INITMGMT[variant][MODE_CHECK])(ostype, rcdir)
  elif UNAME_SOL_OLD == ostype:
    return isEnabledSysVSymlinks(ostype, rcdir)
  elif UNAME_FBSD == ostype:
    chken = os.popen(". /etc/rc.subr && { load_rc_config splunk ; checkyesno `set_rcvar splunk`; } >/dev/null 2>/dev/null")
    return chken.close() is None
  elif UNAME_OSX == ostype:
    chken = os.popen(". /etc/hostconfig && echo $SPLUNK")
    result = False
    while True:
      line = chken.readline()
      if not line: break
      if line.startswith("-YES-"):
        result = True
	break
    chken.close()
    return result
  elif UNAME_AIX == ostype:
    found = False
    hand = open("/etc/inittab")
    while True:
      line = hand.readline()
      if not line: break
      line = line.strip()
      if line.startswith("splunk:") and line.split(":")[2] == "once":
        found = True
        break
    hand.close()
    return found
  else:
    raise cex.ArgError, ("This operation not supported on OS '%s'." % ostype)

#####################################################################
# How we figure out our Linux distribution.                         #
#####################################################################

def default_os_name():
  if sys.platform == "win32":
    return "win32"
  return os.uname()[0].lower()

def default_user():
  if sys.platform == "win32":
    return "Administrator"
  return "root"

# how we determine which linux distribution we're on.
def guessDistro(osName):
  if UNAME_WIN32 == osName:
    versinfo = win32api.GetVersionEx()
    if versinfo[0] == 5:
      if versinfo[1] == 0:
        result = "Windows 2000"
      elif versinfo[1] == 1:
        result = "Windows XP"
      else:
        result = "Windows Server 2003"
    else:
      result = "Windows Vista" # Or 2008
    if versinfo[4] is not None and versinfo[4] != "":
      result += " " + versinfo[4]
    return result
  if UNAME_LINUX == osName:
    for distro, path in linuxFileMap:
      if os.path.exists(path):
        return distro
  elif UNAME_FBSD == osName:
    return os.uname()[2]
  elif UNAME_SOL_OLD == osName:
    return os.uname()[2].split(".")[1]
  elif UNAME_OSX == osName:
    uname_ver = os.uname()[2].split(".")
    return "10." + str(int(uname_ver[0]) - 4) + "." + uname_ver[1]
  elif UNAME_AIX == osName:
    return os.uname()[3] + "." + os.uname()[2]
  else:
    raise cex.ArgError, ("OS '%s' not supported just yet." % osName)
  return DISTRO_DUNNO


# figure out which dir the rc file should live in
def guessRCDir(osName, variant):
  if UNAME_OSX == osName:
    return "/System/Library/StartupItems/Splunk"
  if UNAME_FBSD == osName:
    return "/etc/rc.d"
  return "/etc"

# common arg handling, and figuring out things about the system.
def getSpecs(args):
  if (PARM_VARIANT in args) and (not PARM_OS in args):
    raise cex.ArgError, "Parameter '%s' must be specified when '%s' is specified." % (PARM_OS, PARM_VARIANT)

  # these are kinda in order.. some depend on others.
  osName  = (PARM_OS      in args) and args[PARM_OS]      or default_os_name()
  user    = (PARM_USER    in args) and args[PARM_USER]    or default_user()
  variant = (PARM_VARIANT in args) and args[PARM_VARIANT] or guessDistro(osName)
  rcdir   = (PARM_RCDIR   in args) and args[PARM_RCDIR]   or guessRCDir(osName, variant)
  if UNAME_OSX == osName:
    script  = rcdir + "/Splunk"
  elif UNAME_FBSD == osName:
    script  = rcdir + "/splunk"
  else:
    script  = rcdir + "/init.d/splunk"
  return (osName, user, variant, rcdir, script)


#####################################################################
# Random helper functions...                                        #
#####################################################################

def makeFuncSourceLine():
  for onePath in FUNC_FILE_PATHS:
    if os.path.exists(onePath):
      return ". %s" % onePath
  return ""

def sucmd(user, cmd):
  if sys.platform == "win32" or "root" == user:
    return cmd
  return "su %s -c \"%s\"" % (user, cmd.replace('"', '\\"'))

def makeCtlFunc(name, action, display, user):
  fullAction = sucmd(user, "\"$SPLUNK_HOME/bin/splunk\" %s" % action)
  return (("%s() {\n"
        + "  echo %s\n"
        + "  %s\n"
        + "  RETVAL=$?\n"
        + "}\n") % (name, display, fullAction))

def setFileExecutable(script):
  #                fetch current modes...    and add 0100 (u+x).
  os.chmod(script, os.stat(script).st_mode | stat.S_IXUSR)



#####################################################################
# Create SysV init scripts... currently adapted to various Linux    #
# distributions.  Should end up having code for FBSD/etc.           #
#####################################################################

def createSysVScript(outFile, ostype, user, variant):
  def writeln(s):
    outFile.write(str(s) + "\n")
  # write out any metadata the the OS configuration tools might want to read.
  if variant in METADATA:
    writeln(METADATA[variant])
  writeln("#")

  # setup a few vars.
  writeln("SPLUNK_HOME=\"%s\"" % comm.splunk_home)
  writeln("RETVAL=%s" % ERR_NOERR)
  # load pretty output functions - one of the init scripts was doing this
  writeln("")
  writeln(makeFuncSourceLine())
  writeln("")
  # create functions to do the actual calls.
  writeln(makeCtlFunc("splunk_start",     "start",     "Starting Splunk...",    user))
  writeln(makeCtlFunc("splunk_stop",      "stop",      "Stopping Splunk...",    user))
  writeln(makeCtlFunc("splunk_restart",   "restart",   "Restarting Splunk...",  user))
  writeln(makeCtlFunc("splunk_status",    "status",    "Splunk status:",        user))
  # add main case statement.
  writeln(CODE_CASE)
  writeln("")
  # buh bye.
  writeln("exit $RETVAL\n")

def createOsxScript(outFile, user, variant):
  def writeln(s):
    outFile.write(str(s) + "\n")
  writeln("")
  writeln(". /etc/rc.common")
  writeln("")
  writeln("SPLUNK_HOME=\"%s\"" % comm.splunk_home)
  writeln("export SPLUNK_HOME")
  writeln("PATH=\"$SPLUNK_HOME/bin:$PATH\"")
  writeln("export PATH")
  writeln("")
  for ln in [
    "StartService ()",
    "{",
    "\tif [ \"${SPLUNK:=-NO-}\" = \"-YES-\" ]; then",
    "\t\techo \"Starting Splunk server\"",
    "\t\t" + sucmd(user, "splunk start"),
    "\tfi",
    "}",
    "",
    "StopService ()",
    "{",
    "\techo \"Stopping Splunk server\"",
    "\t" + sucmd(user, "splunk stop"),
    "}",
    "",
    "RestartService ()",
    "{",
    "\tif [ \"${SPLUNK:=-NO-}\" = \"-YES-\" ]; then",
    "\t\techo \"Restarting Splunk server\"",
    "\t\t" + sucmd(user, "splunk restart"),
    "\telse",
    "\t\tStopService",
    "\tfi",
    "}",
    "",
  ]:
    writeln(ln)
  writeln("RunService \"$1\"")

def createScript(ostype, user, variant, rcdir, script):
  outFile = open(script, 'w')
  def writeln(s):
    outFile.write(str(s) + "\n")
  # header.
  writeln("#!/bin/sh")
  if UNAME_FBSD == ostype:
    writeln("")
    writeln("# PROVIDE: splunkd")
    writeln("# REQUIRE: LOGIN")
    writeln("# KEYWORD: shutdown")
    writeln("")
  else:
    writeln("#")
  writeln("# %s" % script)
  writeln("# init script for Splunk.")
  writeln("# generated by 'splunk enable boot-start'.")
  if UNAME_OSX == ostype:
    createOsxScript(outFile, user, variant)
  elif UNAME_FBSD == ostype:
    writeln("")
    writeln(". /etc/rc.subr")
    writeln("")
    writeln("name=splunk")
    writeln("extra_commands=\"status\"")
    writeln("rcvar=`set_rcvar`")
    writeln("eval \"${rcvar}=\\${${rcvar}:-'NO'}\"")
    writeln("")
    writeln("SPLUNK_HOME=${splunk_home:-\"%s\"}" % comm.splunk_home)
    writeln("export SPLUNK_HOME")
    writeln("")
    for cmd in [ "start", "stop", "restart", "status" ]:
      writeln("splunk_%s()" % cmd)
      writeln("{")
      writeln("\t%s \"$@\"" % (sucmd(user, "\"$SPLUNK_HOME/bin/splunk\" %s" % cmd)))
      writeln("}")
      writeln("%s_cmd=splunk_%s" % (cmd, cmd))
      writeln("")
    writeln("load_rc_config $name")
    writeln("run_rc_command \"$@\"")
  else:
    createSysVScript(outFile, ostype, user, variant)
  outFile.close()
  setFileExecutable(script)

#####################################################################
# Defines which operations to perform per OS...                     #
#####################################################################

def setupBootStart(ostype, user, variant, rcdir, script):
  # Create any needed files
  if ostype in [ UNAME_LINUX, UNAME_FBSD, UNAME_SOL_OLD ]:
    createScript(ostype, user, variant, rcdir, script)
  elif UNAME_OSX == ostype:
    if not os.path.isdir(rcdir):
      os.mkdir(rcdir)
    createScript(ostype, user, variant, rcdir, script)
    plistfile = open(rcdir + "/StartupParameters.plist", 'w')
    plistfile.write(
      "{\n" +
      "  Description   = \"Splunk server\";\n" +
      "  Provides      = (\"Splunk server\");\n" +
      "  Uses          = (\"Disks\", \"NFS\");\n" +
      "}\n")
    plistfile.close()
  elif UNAME_AIX == ostype:
    for service in [ "splunkd", "splunkweb" ]:
      os.system("/usr/bin/rmssys -s %s >/dev/null 2>/dev/null" % service)
      killsig = 15
      if service == "splunkd":
        killsig = 2
      os.system("/usr/bin/mkssys -G splunk -s %s -p \"%s/bin/splunk\" -u %s -a _internal_exec_%s -S -n %d -f 9" % (service, comm.splunk_home, user, service, killsig))
  elif UNAME_WIN32 == ostype:
    # create the windows services for splunkd and splunkweb
    os.system("splunkd install")

    os.system("pythonservice -register")
    os.system("python \"%s\" install" % WEB_SVC_PATH)

    # set username and password for service if needed
    username = os.environ.get('SPLUNK_USER', SVC_USER)
    if username is not None:
      password = os.environ.get('SPLUNK_PASS', None)
      domain = os.environ.get('SPLUNK_DOMAIN', None)
      if domain is not None:
        # if SPLUNK_DOMAIN is not provided, assume that it is
        username = "%s\\%s" % (domain, username)
      for svc in [SVC_SPLUNKD, SVC_SPLUNKWEB]:
        if not svc_setUserPass(svc, username, password):
          raise cex.ArgError("Unable to set username/password for service %s, user %s. Make sure that the service is installed and the credentials are correct." % (svc, username))
  else:
    raise cex.ArgError("Unknown operating system: %s." % ostype)
  enableInit(ostype, variant, rcdir, script)


#####################################################################
# Utility functions to support manipulating Windows services        #
#####################################################################

# returns a tuple of two handles on, or None on error
def svc_openService(svcName):

	scm = None
	svc = None
	try:
		scm = win32service.OpenSCManager(None, None, win32service.SC_MANAGER_ALL_ACCESS)
		if scm is None:
			raise
		svc = win32service.OpenService(scm, svcName, win32service.SC_MANAGER_ALL_ACCESS);
		if svc is None:
			raise
	except:
		if svc is not None:
			win32service.CloseServiceHandle(svc)
		if scm is not None:
			win32service.CloseServiceHandle(scm)

		# error opening service
		return None

	return (scm, svc)

def svc_closeService(handles):
	if handles is None:
		return

	(scm, svc) = handles
	win32service.CloseServiceHandle(svc)
	win32service.CloseServiceHandle(scm)

def svc_changeConfig(handles, startType, username = None, password = None):
	(scm, svc) = handles
	try:
		win32service.ChangeServiceConfig(
			svc,
			win32service.SERVICE_NO_CHANGE,
			startType,
			win32service.SERVICE_NO_CHANGE,
			None,
			None,
			0,
			None,
			username,
			password,
			None,
		)
	except:
		return False

	return True

def svc_setAutostart(svcName, auto = True):
	handles = svc_openService(svcName)
	if handles is None:
		return False

	st = win32service.SERVICE_AUTO_START
	if auto == False:
		st = win32service.SERVICE_DEMAND_START
	retval = svc_changeConfig(handles, startType = st)

	svc_closeService(handles)
	return retval

# returns True, False, or None on error
def svc_isAutostart(svcName):
	handles = svc_openService(svcName)
	if handles is None:
		return None

	svcConf = win32service.QueryServiceConfig(handles[1])
	retval = False
	if svcConf[1] == win32service.SERVICE_AUTO_START:
		retval = True

	svc_closeService(handles)
	return retval

# returns True, False, or None on error
def svc_isInstalled(svcName):
	handles = svc_openService(svcName)
	if handles is None:
		if win32api.GetLastError() == winerror.ERROR_SERVICE_DOES_NOT_EXIST:
			# no such service
			return False
		# error
		return None

	# service installed
	svc_closeService(handles)
	return True

def svc_setUserPass(svcName, user, passw):
	handles = svc_openService(svcName)
	if handles is None:
		# error
		return False

	retval = svc_changeConfig(
		handles,
		startType = win32service.SERVICE_NO_CHANGE,
		username = user,
		password = passw)

	svc_closeService(handles)
	return retval

# test code
def svc_test():
	cmd = sys.argv[1]
	if cmd == 'isInst':
		svc = sys.argv[2]
		print "%s is installed: %s" % (svc, svc_isInstalled(svc))

	elif cmd == 'isAuto':
		svc = sys.argv[2]
		print "%s is auto: %s" % (svc, svc_isAutostart(svc))

	elif cmd == 'setAuto':
		svc = sys.argv[2]
		yesno = sys.argv[3]

		if yesno == "yes":
			svc_setAutostart(svc)
		elif yesno == "no":
			svc_setAutostart(svc, False)
		else:
			print "Bad arguments"

