# Copyright (C) 2005-2010 Splunk Inc.  All Rights Reserved.  Version 4.0
import logging as logger
import splunk
import splunk.clilib.cli_common as comm
import control_exceptions as cex
import xml.etree.cElementTree as et
import xml.dom.minidom
import copy, filecmp, os, re, shutil, subprocess, sys, time
import dbmanipulator_lib
import splunk.clilib.bundle_paths as bundle_paths
import urllib


ARG_DRYRUN           = 'dry-run'
ARG_NOWAIT           = 'no-wait'

PERMS_OWNER_RW_ONLY = '600'

EMPTY_LINE     = "\n"

BACKUP_EXT     = "-migrate.bak"

STANZA_MIG_HISTORY   = 'history'

CONF_COMMANDS        = "commands"
CONF_SAVEDSEARCHES   = "savedsearches"
CONF_PROPS           = "props"

EXT_MIGRATE          = "migratePreview"
EXT_DEPRECATED       = "deprecated"
SUF_MIGRATE          = "." + EXT_MIGRATE

KEY_LAST_MIG_VERSION = "lastVersionMigratedTo"

PATH_APPS_DIR         = os.path.join(comm.splunk_home, "etc", "apps")

PATH_LOG_DIR          = os.path.join(comm.splunk_home, "var", "log", "splunk")
PATH_VERSION_FILE     = os.path.join(comm.splunk_home, "etc", "splunk.version")

PATH_LICENSE_ACTIVE   = os.path.join(comm.splunk_home, "etc", "splunk.license")
PATH_LICENSE_FREE     = os.path.join(comm.splunk_home, "etc", "splunk-enttrial.license")
PATH_LICENSE_USER     = os.path.join(comm.splunk_home, "etc", "splunk-user.license")

PATH_AUTHEN_CONF_OLD  = bundle_paths.make_path("auth.conf")
PATH_AUTHEN_CONF_BAK  = PATH_AUTHEN_CONF_OLD + BACKUP_EXT
PATH_AUTHEN_CONF_NEW  = bundle_paths.make_path("authentication.conf")

PATH_INPUTS_CONF      = bundle_paths.make_path("inputs.conf")
PATH_INDEXES_CONF     = bundle_paths.make_path("indexes.conf")

PATH_SPLUNKD_XML     = os.path.join(comm.splunk_home, "etc", "myinstall", "splunkd.xml")
PATH_SPLUNKD_XML_BAK = os.path.join(comm.splunk_home, "etc", "myinstall", "splunkd.xml-migrate.bak")
PATH_SPLUNKD_XML_DEF = os.path.join(comm.splunk_home, "etc", "myinstall", "splunkd.xml.cfg-default")

PATH_LDAP_CONF       = os.path.join(comm.splunk_home, "etc", "openldap", "ldap.conf")
PATH_LDAP_CONF_DEF   = os.path.join(comm.splunk_home, "etc", "openldap", "ldap.conf.default")

PATH_BATCHCONF_XML   = os.path.join(comm.splunk_home, "etc", "modules", "input", "batchfile", "config.xml")
PATH_BATCHCONF_DIS   = PATH_BATCHCONF_XML + ".upgraded"

PATH_TXN_TYPES_CONF   = bundle_paths.make_path("transactiontypes.conf")
PATH_SAVSRCH_CONF     = bundle_paths.make_path("savedsearches.conf")
PATH_EVTTYPE_CONF     = bundle_paths.make_path("eventtypes.conf")
PATH_TAGS_CONF        = bundle_paths.make_path("tags.conf")
PATH_PROPS_CONF       = bundle_paths.make_path("props.conf")
PATH_SECRET_FILE      = os.path.join(comm.splunk_home, "etc", "splunk.secret")

PATH_PASSWD_FILE     = os.path.join(comm.splunk_home, "etc", "passwd")

PATH_OLD_DEF_BUNDLES = os.path.join(comm.splunk_home, "etc", "bundles", "default")

PATH_AUTHORIZE_CONF  = bundle_paths.make_path("authorize.conf")
PATH_MIGRATION_CONF  = bundle_paths.make_path("migration.conf")
PATH_SERVER_CONF     = bundle_paths.make_path("server.conf")
PATH_WEB_CONF        = bundle_paths.make_path("web.conf")
PATH_LOCALMETA_CONF  = os.path.join( bundle_paths.etc(), "system", "metadata", "local.meta" )
PATH_WMI_CONF        = bundle_paths.make_path("wmi.conf")

PATH_SEARCH_LOCALMETA_CONF  = os.path.join( bundle_paths.etc(), "apps", "search", "metadata", "local.meta" )
PATH_SEARCH_SAVSRCH_CONF    = os.path.join( bundle_paths.etc(), "apps", "search", "local", "savedsearches.conf" )

PATH_ETC_USERS              = os.path.join(comm.splunk_home, "etc", "users")

#                         0-------------------------------------------------------------------------------80
POST_DRY_RUN_MESSAGE = """Stopping because you requested a preview of the configuration files to be"""         + "\n" \
                       """migrated.  Migration and upgrade will not continue until the configuration files"""  + "\n" \
                       """have actually been migrated."""                                                      + "\n" \
                                                                                                               + "\n" \
                       """To proceed, run 'splunk start' again and choose 'y'."""

#                         0-------------------------------------------------------------------------------80
TIMEZONE_WARNING_MSG =                                                                                           "\n" \
                       """WARNING: If you have configured timestamp offsets using pre-Splunk 3.2 POSIX"""      + "\n" \
                       """instructions, you must reconfigure them using the information on this page:"""       + "\n" \
                       """http://www.splunk.com/doc/latest/admin/ConfigureTimeZoneOffsets.  If you do not"""   + "\n" \
                       """do this, your timestamp information will be incorrect."""

#                         0-------------------------------------------------------------------------------80
UNMIGRATABLE_PROMPT  =                                                                                           "\n" \
                       """The above configuration files cannot be migrated automatically.  They will need"""   + "\n" \
                       """to be migrated manually as documented at:"""                                         + "\n" \
                       """ http://splunk.com/base/Documentation/latest/Installation/Stepsformanualmigration""" + "\n" \
                                                                                                               + "\n" \
                       """You can choose to disregard this warning and continue with automatic migration"""    + "\n" \
                       """of the remaining configuration files, but this may result in certain features"""     + "\n" \
                       """being disabled or not working properly until the above files are migrated"""         + "\n" \
                       """properly."""                                                                         + "\n" \
                                                                                                               + "\n" \
                       """Would you like to ignore this warning and continue with migration? [y/n] """


#                         0-------------------------------------------------------------------------------80
UNMIGRATABLE_MSG     = """Stopping as requested."""                                                            + "\n" \
                                                                                                               + "\n" \
                       """To proceed, please complete the above steps and run 'splunk start' again."""



######################################## UTILS ######################################## 

def upgradingFromBeforeSplunk4():
  """
  Returns boolean after checking for serverName node in splunkd.xml.
  This check is ONLY valid when run before splunkd.xml is migrated 4.0.0.
  """
  return (0 != len(comm.grep("<serverName>", PATH_SPLUNKD_XML)))

def findPreviewFiles():
  return comm.findFiles(comm.splunk_home, "\\.%s$" % EXT_MIGRATE)

def getPreviewName(filename):
  return filename + SUF_MIGRATE

def getUnpreviewName(filename):
  if filename.endswith(SUF_MIGRATE):
    filename = filename[:-1 * len(SUF_MIGRATE)]
  return filename

def chooseFile(path, isDryRun, useNewPaths):
  # when we're in dry-run and migrating from the old bundles dir structure
  # to the new one, we haven't actually moved the files - so allow preview
  # to happen based on the old, still valid, file locations.
  if not useNewPaths:
    path = path.replace(bundle_paths.get_system_bundle_path(),
        bundle_paths.get_legacy_base_path(), 1)

  doritosRule = path
  if isDryRun:
    doritosRule = getPreviewName(doritosRule)
    if os.path.exists(path):
      shutil.copy(path, doritosRule)
  return doritosRule


def setMigHistory(key, val, isDryRun):
  if isDryRun:
    return
  migSettings = {}
  if os.path.exists(PATH_MIGRATION_CONF):
    migSettings = comm.readConfFile(PATH_MIGRATION_CONF)
  if not STANZA_MIG_HISTORY in migSettings:
    migSettings[STANZA_MIG_HISTORY] = {}
  migSettings[STANZA_MIG_HISTORY][key] = val
  comm.writeConfFile(PATH_MIGRATION_CONF, migSettings)

def getMigHistory(key):
  val = None
  if os.path.exists(PATH_MIGRATION_CONF):
    migSettings = comm.readConfFile(PATH_MIGRATION_CONF)
    if STANZA_MIG_HISTORY in migSettings:
      if key in migSettings[STANZA_MIG_HISTORY]:
        val = migSettings[STANZA_MIG_HISTORY][key]
  return val


####################################### //UTILS ####################################### 

def checkCommandsConfig(confName):
  """
  Compare etc/searchscripts with the merged commands.conf, which now must
  specify all such scripts.
  
  If a given script does not have a corresponding FILENAME=<name> entry in
  some stanza of commands.conf, show a warning.

  Since some genius shoved a .conf file in that dir, we'll be sure to only
  inspect files ending in .pl or .py, which is what commands.conf.spec says
  we support.

  This function makes no changes to your filesystem.
  """
  logger.info("\nChecking script configuration...")
  keyFilename = 'filename'
  keyFileType = 'type'
  scriptsDir  = 'searchscripts'
  fileTypeMap = { 'perl' : '.pl', 'python' : '.py' }

  # collect all known script names from commands.conf into a list.
  commandsDict = comm.getMergedConf(confName)
  # this is the list of valid script filenames.
  commandList  = [kvPair[keyFilename] for kvPair in commandsDict.values() if kvPair.has_key(keyFilename)]
  commandList = []
  for cmdName, cmdSetts in commandsDict.items():
    # if the filename is specified in the stanza settings, go with that.
    if cmdSetts.has_key(keyFilename):
      commandList.append(cmdSetts[keyFilename])
    # otherwise we have to figure it out.
    else:
      # figure out whether this is a python or perl file.
      fileType = (keyFileType in cmdSetts and cmdSetts[keyFileType] or 'python')
      if not fileType in fileTypeMap:
        logger.warn("File type '%s' in stanza '%s' is unrecognized.  This command may not work." % (fileType, cmdName))
        continue # skip this command, we dunno what it is.
      # add the stanza name + appropriate extension as a command.
      commandList.append(cmdName + fileTypeMap[fileType])
  logger.debug("The following commands are specified in commands.conf:  %s" % str.join(", ", commandList))

  # find all .py or .pl files in etc/searchscripts.
  scriptFiles = comm.findFiles(os.path.join(bundle_paths.etc(), scriptsDir), r'\.(py|pl)', caseSens = False)
  
  # out of that script list, pick out all those that are not in the .conf file, discounting the path.
  badScripts = filter(lambda x: os.path.split(x)[1] not in commandList, scriptFiles)

  if len(badScripts) > 0:
    logger.warn("The following scripts will not work without being configured in commands.conf:"
        "\n\t%s" % str.join("\n\t", badScripts))


def checkSavedSearches(confName):
  """
  Iterates all saved searches (after bundles have been merged)
  and warns about searches that are now unsafe.

  This function makes no changes to your filesystem.

  Thanks to Steveyz for scripting this up.
  """
  logger.info("\nChecking saved search compatibility...")
  KEY_QUERY = "query"
  savedSearches = comm.getMergedConf(confName)
  for ssName, ssSetts in savedSearches.items():
    if not KEY_QUERY in ssSetts:
      continue

    searchString = ssSetts[KEY_QUERY]
    warning = ""
    
    searchModifiers = ("maxresults", "maxtime", "savedsplunk", "savedsearch", "eventtype", "tag",
        "typetag", "eventtypetag", "hosttag", "sourcetype", "source", "host", "index", "readlevel",
        "startdaysago", "startminutesago", "starthoursago", "startmonthsago", "enddaysago",
        "endminutesago", "endhoursago", "endmonthsago", "searchtimespanhours", "searchtimespanminutes",
        "searchtimespandays", "searchtimespanmonths","starttime", "timeformat", "endtime", "timeformat",
        "starttimeu", "endtimeu", "daysago", "minutesago", "hoursago", "monthsago")
    specialExceptions = ("group",) # group = SPL-12442.
    
    firstTerm = searchString.split(" ")[0]
    if firstTerm in ("admin", "remote"):
      warning = "A search starting with '%s' will no longer invoke the %s operator.  To invoke the '%s'" \
          " operator, prepend '|' to your search, e.g. '| %s'." % (firstTerm, firstTerm, firstTerm, searchString)
    elif firstTerm == "search":
      warning = "A leading 'search' term is no longer ignored.  Your query will search for the literal term 'search'"
    else:
      # find maxresults/maxtime, which are deprecated
      maxrestimeRE = re.compile(r"($|.*\s+)(maxresults|maxtime)(\:\:|==?)\d+(\s+.*|$)")
      if maxrestimeRE.match(searchString):
        warning = "The maxresults and maxtime search modifiers have been deprecated and will be ignored."
      # this is trickier, find foo=bar terms in the search clause before the first pipe
      # not the most robust parsing, but simple and covers 99% of cases, should be good enough
      searchString = searchString.split("|")[0]
      srcfieldRE = re.compile(r"^.*?(\w+)=(.*)$")
      while True:
        sfmatch = srcfieldRE.match(searchString)
        if sfmatch == None:
          break
        fieldname = sfmatch.group(1)
        searchString = sfmatch.group(2)
        if (not fieldname in searchModifiers) and (not fieldname in specialExceptions):
          if len(warning) > 0:
            warning += "\n\n"
          warning += ("If '%s' is an indexed field with values not found in the raw event, you must indicate that" \
              " this is an indexed field in fields.conf or use '::' instead of '=' to separate the field name" \
              " from the value.") % fieldname
    
    if len(warning) > 0:
      logger.warn("WARNING for saved search '%s': %s" % (ssName, warning))


def checkTimezones(confName, isDryRun):
  """
  This is the check for SPL-13537, the issue with the previously-wrong timezone offset rules.

  The logic here is:
    - for each stanza from the post-merge props.conf,
    - if the stanza has a TZ key, check the value.
    - if there's a slash in the value, truncate the string just before it.
    - if the value has a 1-9 digit (not 0) at this point, show the warning.

  Upon showing the warning, when we're not in dry-run, a migration.conf file will be left behind.
  This will be checked the next time FTR goes through.  If we have a special flag in the file,
  we won't show the warning again - so people will only ever see it once per instance.
  """
  WARNING_VERSION = '1'
  KEY_TIMEZONE    = 'TZ'
  KEY_TZWARNING   = 'checkedOrDisplayedTimezoneWarning'

  # check and see if we've already shown this warning.
  migSettings = {}
  if os.path.exists(PATH_MIGRATION_CONF):
    migSettings = comm.readConfFile(PATH_MIGRATION_CONF)
    if STANZA_MIG_HISTORY in migSettings:
      if KEY_TZWARNING in migSettings[STANZA_MIG_HISTORY]:
        if migSettings[STANZA_MIG_HISTORY][KEY_TZWARNING] >= WARNING_VERSION:
          return

  logger.info("\nChecking for possible timezone configuration errors...")
  iTawtISawAPuddyTat = False
  propsStanzas = comm.getMergedConf(confName)
  # for each [stanza] in props...
  for stanza, settings in propsStanzas.items():
    # no TZ key?  skip it.
    if not KEY_TIMEZONE in settings:
      continue
    tzVal = settings[KEY_TIMEZONE]
    # strip any slashes and anything after them.
    firstSlashPos = tzVal.find('/')
    if firstSlashPos >= 0:
      tzVal = tzVal[:firstSlashPos]
    # position of first non-zero digit in this TZ value, if any.
    for oneChar in tzVal:
      if oneChar.isdigit() and ('0' != oneChar):
        iTawtISawAPuddyTat = True
        # don't bother checking anymore.
        break

  # do the warning if necessary.
  if iTawtISawAPuddyTat:
    logger.warn(TIMEZONE_WARNING_MSG)

  # if this isn't a dry run, make a note that we've done this warning check.
  if not isDryRun:
    if not STANZA_MIG_HISTORY in migSettings:
      migSettings[STANZA_MIG_HISTORY] = {}
    migSettings[STANZA_MIG_HISTORY][KEY_TZWARNING] = WARNING_VERSION
    comm.writeConfFile(PATH_MIGRATION_CONF, migSettings)


def migrateSplunkdXml_3_0_0(path):
  """
  Migrate splunkd.xml to 3.0.0.
  This function WILL OVERWRITE YOUR FILE.
  The caller should create a backup first.
  """
  XML_PIPELINE   = "pipeline"
  XML_PROCESSOR  = "processor"
  PIPE_NAME      = "name"
  PROC_NAME      = "name"
  PROC_PLUGIN    = "plugin"
  INDEXQ_NAME    = "indexIn"
  BLOCK_MGMTSVR  = "managementServer"
  BLOCK_PRELOAD  = "preloaderConfig"
  THRUP_PLUGIN   = "thruputprocessor"
  THRUP_OLDNAME  = "thruput"
  THRUP_NEWNAME  = "index_thruput"
  UPGPIPE_NAME   = "upgraderPipe"
  TCPOUT_NAME    = "tcp-output-generic-processor"

  PROC_TCPO="""
                            <processor name="tcp-output-generic-processor" plugin="tcpoutputprocessor">
                               <config>
                               </config>
                            </processor>"""

  oldDom = xml.dom.minidom.parse(path)
  pipelines = oldDom.getElementsByTagName(XML_PIPELINE)
  processors = oldDom.getElementsByTagName(XML_PROCESSOR)
 
  # remove managementServer block, if any.
  for mgmt in oldDom.getElementsByTagName(BLOCK_MGMTSVR):
    mgmt.parentNode.removeChild(mgmt)

  # remove preloaderConfig block, if any.
  for prel in oldDom.getElementsByTagName(BLOCK_PRELOAD):
    prel.parentNode.removeChild(prel)


  # add the tcpout processor, if it's missing.
  if 0 == len(comm.grep(TCPOUT_NAME, path)):
    # start by finding the index input queue...
    indexQ = None
    for proc in processors:
      if proc.getAttribute(PROC_NAME) == INDEXQ_NAME:
        indexQ = proc
        break
    if None == indexQ:
      raise cex.ParsingError, "Could not find indexer input queue, Please file a case online at http://www.splunk.com/page/submit_issue"
    # then get the next node after that...
    afterIndexQ = indexQ.nextSibling
    if None == afterIndexQ:
      raise cex.ParsingError, "No node after index queue. Please file a case online at http://www.splunk.com/page/submit_issue"
    # and add the tcp out node just before that.  ain't minidom a stinker?
    tcpOutNode = xml.dom.minidom.parseString(PROC_TCPO).documentElement
    afterIndexQ.parentNode.insertBefore(oldDom.createTextNode(EMPTY_LINE + "\t\t\t"), afterIndexQ)
    afterIndexQ.parentNode.insertBefore(tcpOutNode, afterIndexQ)

  # if we have a "thruput" processor, rename it to "index_thruput".
  for proc in processors:
    if THRUP_PLUGIN == proc.getAttribute(PROC_PLUGIN):
      if THRUP_OLDNAME == proc.getAttribute(PROC_NAME):
        proc.setAttribute(PROC_NAME, THRUP_NEWNAME)

  # remove the upgrader pipe if it's still there.
  for pipe in pipelines:
    if UPGPIPE_NAME == pipe.getAttribute(PIPE_NAME):
      pipe.parentNode.removeChild(pipe)

  # file is done, write it out
  newFile = open(path, "w")
  newFile.write(oldDom.toxml())
  newFile.close()


def migrateSplunkdXml_3_2_0(path):
  """
  Migrate splunkd.xml to 3.2.0.
  This function WILL OVERWRITE YOUR FILE.
  The caller should create a backup first.
  """
  XML_PROCESSOR  = "processor"
  PROC_NAME      = "name"
  TCPOUT_NAME    = "tcp-output-generic-processor"
  STREAMOUT_NAME = "stream-output-processor"
  INDEXER_NAME   = "indexer"
  SIGNING_PLUGIN = "signingprocessor"



  PROC_STREAMOUT="""
                        <processor name="stream-output-processor" plugin="streamoutputprocessor">
                           <config>
                           </config>
                        </processor>"""

  PROC_SIGNING="""
                        <processor name="signing" plugin="signingprocessor">
                            <config>
                            </config>
                        </processor>"""


  oldDom = xml.dom.minidom.parse(path)
  processors = oldDom.getElementsByTagName(XML_PROCESSOR)

  # add the stream output processor if missing.
  if 0 == len(comm.grep(STREAMOUT_NAME, path)):
    # start by finding the tcp output processor...
    tcpOutProc = None
    for proc in processors:
      if TCPOUT_NAME == proc.getAttribute(PROC_NAME):
        tcpOutProc = proc
        break
    if None == tcpOutProc:
      raise cex.ParsingError, "Could not find tcp output block, Please file a case online at http://www.splunk.com/page/submit_issue"
    # then get the next node after that...
    afterTcpOut = tcpOutProc.nextSibling
    if None == afterTcpOut:
      raise cex.ParsingError, "No node after index queue. Please file a case online at http://www.splunk.com/page/submit_issue"
    # and add the streamout node just before that.  eyeroll.
    streamOutNode = xml.dom.minidom.parseString(PROC_STREAMOUT).documentElement
    afterTcpOut.parentNode.insertBefore(oldDom.createTextNode(EMPTY_LINE + "\t\t\t"), afterTcpOut)
    afterTcpOut.parentNode.insertBefore(streamOutNode, afterTcpOut)



  # add the signing processor if missing.
  if 0 == len(comm.grep(SIGNING_PLUGIN, path)):
    # first find the indexer processor...
    indexProc = None
    for proc in processors:
      if INDEXER_NAME == proc.getAttribute(PROC_NAME):
        indexProc = proc
        break
    if None == indexProc:
      raise cex.ParsingError, "Could not find indexer block. Please file a case online at http://www.splunk.com/page/submit_issue"
    # then add the signing proc node just before that.
    signProcNode = xml.dom.minidom.parseString(PROC_SIGNING).documentElement
    indexProc.parentNode.insertBefore(oldDom.createTextNode(EMPTY_LINE + "\t\t\t"), indexProc)
    indexProc.parentNode.insertBefore(signProcNode, indexProc)
    indexProc.parentNode.insertBefore(oldDom.createTextNode(EMPTY_LINE + "\t\t\t"), indexProc)


  # file is done, write it out
  newFile = open(path, "w")
  newFile.write(oldDom.toxml())
  newFile.close()


def migSplunkdXml_4_0_0(srcXml, destConf):
  """
  Migrates splunkd.xml values for server name and disk usage processor settings
  to server.conf.
  Overwrites existing splunkd.xml and server.conf.
  """

  STANZA_GENERAL = "general"
  KEY_SRV_NAME   = "serverName"

  # from...
  DISK_USAGE_NAME = "diskusage"
  FREE_SPACE_TAG  = "minFreeSpace"
  POLL_FREQ_TAG   = "pollingFrequency"
  # to...
  KEY_DISK_USAGE  = "diskUsage"

  srvName, minFreeSpace, pollFreq = None, None, None

  srcDom = xml.dom.minidom.parse(srcXml)

  # find serverName node and get value.
  nodeList = srcDom.getElementsByTagName(KEY_SRV_NAME)
  if 0 != len(nodeList):
    srvNameNode = nodeList[0]

    srvName = srvNameNode.firstChild.nodeValue
    if 0 == len(srvName):
      raise cex.ParsingError, "Found splunkd.xml serverName node, but could not extract value.  Please file a case online at http://www.splunk.com/page/submit_issue"
    # remove node from dom.
    srvNameNode.parentNode.removeChild(srvNameNode)

  # find diskusage processor node and get value(s).
  diskUsageNode = None
  nodeList = srcDom.getElementsByTagName("processor")
  for node in nodeList:
    if node.hasAttribute("name") and DISK_USAGE_NAME == node.getAttribute("name"):
      diskUsageNode = node
      break
  if None != diskUsageNode:
    # get minFreeSpace value if avail.
    nodeList = diskUsageNode.getElementsByTagName(FREE_SPACE_TAG)
    if 0 != len(nodeList):
      node = nodeList[0]
      minFreeSpace = node.firstChild.nodeValue
      if 0 == len(minFreeSpace):
        raise cex.ParsingError, "Found splunkd.xml minFreeSpace node, but could not extract value. Please file a case online at http://www.splunk.com/page/submit_issue"
    # get pollingFrequency if avail.
    nodeList = diskUsageNode.getElementsByTagName(POLL_FREQ_TAG)
    if 0 != len(nodeList):
      node = nodeList[0]
      pollFreq = node.firstChild.nodeValue
      if 0 == len(pollFreq):
        raise cex.ParsingError, "Found splunkd.xml pollingFrequency node, but could not extract value. Please file a case online at http://www.splunk.com/page/submit_issue"
    diskUsageNode.parentNode.removeChild(diskUsageNode)


  if srvName == None and minFreeSpace == None and pollFreq == None:
    return # nothing changed, nothing to write out.

  # before editing xml file, load conf if it exists and save an updated version.
  destDict = {}
  if os.path.exists(destConf):
    destDict = comm.readConfFile(destConf)
  if srvName != None:
    if not STANZA_GENERAL in destDict:
      destDict[STANZA_GENERAL] = {}
    destDict[STANZA_GENERAL][KEY_SRV_NAME] = srvName
  if minFreeSpace != None:
    if not KEY_DISK_USAGE in destDict:
      destDict[KEY_DISK_USAGE] = {}
    destDict[KEY_DISK_USAGE][FREE_SPACE_TAG] = minFreeSpace
  if pollFreq != None:
    if not KEY_DISK_USAGE in destDict:
      destDict[KEY_DISK_USAGE] = {}
    destDict[KEY_DISK_USAGE][POLL_FREQ_TAG] = pollFreq
  # save local server.conf.
  comm.writeConfFile(destConf, destDict)

  # write out updated xml file.
  open(srcXml, "w").write(srcDom.toxml())


def migrateSavedSearches(path):
  """
  Change isGlobal in savedsearches.conf.  Use case insensitive because there was
  some confusion with this variable early on, we may have lowercase and camelcase.
    isGlobal = 1   ->   role = Everybody
    isGlobal = 0   ->   role = 
    isGlobal =     ->   role = 
  """
  if os.path.exists(path):
    change = lambda x, y: comm.sed(x, y, path, inPlace = True)
    change("is[Gg]lobal\\s*=\\s*1", "role = Everybody")
    change("is[Gg]lobal\\s*=\\s*0", "role = ")
    change("is[Gg]lobal\\s*=\\s*$", "role = \n")
    if len(comm.grep("^[^#]*is[Gg]lobal", path)) > 0:
      logger.warn("*** After saved search migration, still have isGlobal in %s.  This should not happen. Please file a case online at http://www.splunk.com/page/submit_issue" % path)



def migrateWebSSL(srvConfPath, webConfPath):
  """
  Changes from the old web/ssl variable in the server.conf to the new one in web.conf.
  """
  oldParm    = "enableSplunkSearchSSL"
  newParm    = "enableSplunkWebSSL"
  kvSep      = "="
  settStanza = "[settings]"
  # local file doesn't exist?  done.
  if not os.path.exists(srvConfPath):
    return
  # old string not in local file?  done.
  if 0 == len(comm.grep(oldParm, srvConfPath)):
    return

  # fyi, this is a little more complicated than it should be because we're preserving comments.

  # get all existing lines in server.conf.
  srvFile     = open(srvConfPath, 'r')
  oldSrvLines = srvFile.readlines()
  srvFile.close()
  # remove lines w/ old setting and store setting.
  newSrvLines = []
  oldSetting  = None
  for line in oldSrvLines:
    # if this line has no =, just add it to the new lines.
    if 0 == line.count(kvSep):
      newSrvLines.append(line)
    # otherwise, dig deeper.
    else:
      key, val = line.split(kvSep, 1)
      key = key.strip()
      # if this is the SSL key, save the value.
      if oldParm == key:
        oldSetting = val
      # otherwise add it to the lines to be written out again.
      else:
        newSrvLines.append(line)
  
  # write out the new server.conf.
  srvFile = open(srvConfPath, 'w+') # w+ = write/truncate.
  srvFile.write(str.join("", newSrvLines))
  srvFile.close()

  webConfLines = []
  # gather old web config lines, if it exists.
  if os.path.exists(webConfPath):
    webConfLines = open(webConfPath, 'r').readlines()
  settStanzaInd = None
  # find [settings] if possible.
  for i in range(0, len(webConfLines)):
    if settStanza == webConfLines[i].strip():
      settStanzaInd = i
      break
  # wasn't there?  add it as index 0.
  if None == settStanzaInd:
    webConfLines.append(settStanza + "\n")
    settStanzaInd = 0
  # insert the setting.
  webConfLines.insert(settStanzaInd + 1, "%s =%s" % (newParm, oldSetting))

  # write out new web.conf.
  newWebFile = open(webConfPath, 'w+') # w+ = write/truncate.
  newWebFile.write(str.join("", webConfLines))
  newWebFile.close()



def removeDirMon(isDryRun):
  """
  disable the old batch module or we'll be in pain.
  """
  if os.path.exists(PATH_BATCHCONF_XML):
    if not isDryRun:
      logger.info("Upgrading batch file input module...")
    comm.moveItem(PATH_BATCHCONF_XML, PATH_BATCHCONF_DIS, isDryRun)


def removeDeprecated(isDryRun):
  """
  move now-unused or potentially harmful files out of the way.
  """
  logger.notice("\nHandling deprecated files...")

  bundleFiles = ("access_control.conf", "auth.conf", "metaevents.conf")
  # some files are moved elsewhere.  this one is done in rory's auth migration.
  # but, we still want to delete the example/spec files.
  exclude     = "auth.conf"
  bundlePath  = os.path.join(comm.splunk_home, "etc", "bundles")

  for filename in bundleFiles:
    # good idea to do case-insensitive, for windows in the future.
    #                                           match on foo, foo.example, foo.spec.
    fileList = comm.findFiles(bundlePath, "\\b%s(|.example|.spec)$" % filename, caseSens = False)
    for oneFile in fileList:
      if not oneFile.endswith(exclude):
        filename   = os.path.basename(oneFile)
        parentPath = os.path.dirname(oneFile)
        parentDir  = os.path.basename(parentPath)
        depDir     = os.path.join(parentPath, "deprecated")

        # apparently we install files in 3.2 that are pre-deprecated.  a winrar is us..
        if 'deprecated' != parentDir:
          if not isDryRun and not os.path.exists(depDir):
            os.mkdir(depDir)
          # if this file lives in our bundles/README or bundles/DEFAULT dir, move it to a "deprecated" subdir.
          if parentDir in ('default', 'README'):
            comm.moveItem(oneFile, os.path.join(depDir, filename + "." + EXT_DEPRECATED), isDryRun)
          # otherwise just rename it.
          else:
            comm.moveItem(oneFile, oneFile + "." + EXT_DEPRECATED, isDryRun)


def migrateAuthSecret(isDryRun):
  """
  For upgrades, move the existing splunk.secret in place.
  """
  if os.path.exists(PATH_SECRET_FILE):
    comm.moveItem(PATH_SECRET_FILE, PATH_AUTH_DIR, isDryRun)


def migrateAuth_3_2_0(newConfPath, isDryRun):
  """
  Migrate the old auth.conf -> authentication.conf for 3.2.0.
  """
  ROLE_ADMIN   = "Admin"
  ROLE_POWER   = "Power"
  ROLE_USER    = "User"
  STANZA_AUTH  = "auth"

  # if we don't have local/auth.conf, there's nothing to do.
  if not os.path.exists(PATH_AUTHEN_CONF_OLD):
    return

  # copy auth.conf to the new path.  this is what we'll edit.
  shutil.copy(PATH_AUTHEN_CONF_OLD, newConfPath)
  # rename auth.conf to auth.conf-migrate.bak or whatever.
  comm.moveItem(PATH_AUTHEN_CONF_OLD, PATH_AUTHEN_CONF_BAK, isDryRun)

  stanzaDict = comm.readConfFile(newConfPath)

  roleMap = {}
  # find Admin/Power/User settings as necessary.  save values and remove
  # from settings dictionary.
  for stanzaName, stanzaSetts in stanzaDict.items():
    for key, val in stanzaSetts.items():
      if   ROLE_ADMIN == key:
        roleMap[ROLE_ADMIN] = val
        del stanzaSetts[key]
      elif ROLE_POWER == key:
        roleMap[ROLE_POWER] = val
        del stanzaSetts[key]
      elif ROLE_USER  == key:
        roleMap[ROLE_USER]  = val
        del stanzaSetts[key]
    
  if 0 != len(roleMap):
    stanzaDict["roleMap"] = roleMap
  
  if STANZA_AUTH in stanzaDict:
    stanzaDict["authentication"] = stanzaDict.pop(STANZA_AUTH)

  # done, write out the new authentication.conf.
  comm.writeConfFile(newConfPath, stanzaDict)

def migrateDistSearchXML(xmlPath, confPath, isDryRun):
  # <config>
  BLOCK_CONFIG  = "config"
  # these live in <config>.
  BLOCK_BLIST   = "blacklist"
  BLOCK_SERVERS = "servers"

  TAG_SERVER    = "server"
  TAG_URL       = "url"

  ATTR_SERVNAME = "name"
  ATTR_SRVURL   = "url"

  KEY_SERVERS   = "servers"
  KEY_BLISTNAME = "blacklistNames"
  KEY_BLISTURL  = "blacklistURLs"

  # load the XML and make sure we have <config>.
  distDom = xml.dom.minidom.parse(xmlPath)
  try:
    configBlock = distDom.getElementsByTagName(BLOCK_CONFIG)[0]
  except IndexError:
    raise cex.ParsingError, "Bad distributed search config file. Please file a case online at http://www.splunk.com/page/submit_issue"

  # only migrate if we have xml nodes in <config></config> other than whitespace and comment nodes.
  if 0 == len(filter(lambda x: x.nodeType not in (x.TEXT_NODE, x.COMMENT_NODE), configBlock.childNodes)):
    return # had 0 nodes left after filtering those useless ones out.


  # make a backup of the distributed search config.xml.  not in dry run though, looks kinda silly
  # to say we'd make a backup of the "preview" file.
  if not isDryRun:
    comm.copyItem(xmlPath, xmlPath + BACKUP_EXT, isDryRun)



  # this is where we'll collect all the settings from the xml file.
  settsDict = {}

  # iterate almost all xml tags in <config> and store them in our dict.
  for child in configBlock.childNodes:
    valueFound = False
    # if it's not an element, screw it.  prob whitespace/etc.
    if child.ELEMENT_NODE == child.nodeType:
      # skip <servers> and <blacklist>, we handle those specially.
      if not child.nodeName in (BLOCK_BLIST, BLOCK_SERVERS):
        # could be empty tags, check for that.
        if child.hasChildNodes():
          valueFound = True
          # add value in <tags>value</tags>.
          settsDict[child.nodeName] = child.firstChild.nodeValue
        # otherwise just add it with an empty value.
        else:
          settsDict[child.nodeName] = ""

  # now process the nested server nodes.
  serverList = []
  for servBlock in configBlock.getElementsByTagName(BLOCK_SERVERS):
    for urlTag in servBlock.getElementsByTagName(TAG_URL):
      if urlTag.hasChildNodes():
        serverList.append(urlTag.firstChild.data)
  # add any collected servers to the settings as a CSV list.
  if len(serverList) > 0:
    settsDict[KEY_SERVERS] = str.join(",", serverList)

  # process the nested blacklist nodes.
  blacklistNames, blacklistURLs = [], []
  for blistBlock in configBlock.getElementsByTagName(BLOCK_BLIST):
    for servNode in blistBlock.getElementsByTagName(TAG_SERVER):
      if servNode.hasAttribute(ATTR_SERVNAME):
        # if there's a name=foo attribute, store it.
        blacklistNames.append(servNode.getAttribute(ATTR_SERVNAME))
      if servNode.hasAttribute(ATTR_SRVURL):
        # also store any url=foo attribute.
        blacklistURLs.append(servNode.getAttribute(ATTR_SRVURL))

  # add the settings we just gathered to the dictionary.
  if len(blacklistNames) > 0:
    settsDict[KEY_BLISTNAME] = str.join(",", blacklistNames)
  if len(blacklistURLs)  > 0:
    settsDict[KEY_BLISTURL]  = str.join(",", blacklistURLs)

  # create the file-level dictionary that we'll write out.
  #             stanza               k/v pairs
  confDict  = {"distributedSearch" : settsDict}

  # write out the new .conf file.
  comm.writeConfFile(confPath, confDict)

  # now that all that was successful, add an empty <config> block,
  # delete the old one, and write out the new xml file.
  configBlock.parentNode.appendChild(distDom.createElement("config"))
  configBlock.parentNode.removeChild(configBlock)
  open(xmlPath, "w").write(distDom.toxml())

  # done.


def migratePasswd(path):
  """
  Check and make sure that all users are in the "Everybody" role.
  """
  if not os.path.exists(path):
    return

  fieldDelim = ":"
  roleDelim  = ";"
  roleIndex  = 5   # counting from 0.
  roleAll    = "Everybody"
  changed    = False
  # get all passwd lines.
  passwdLines = open(path, 'r').readlines()
  # use indices so we can edit the array.
  for i in range(len(passwdLines)):
    # pick out the roles field, the last one in each line.
    splitted = passwdLines[i].split(fieldDelim)

    # if we don't have 6 fields, maybe it's a commnent? skip it.
    if len(splitted) != (roleIndex + 1):
      continue

    # if Everybody isn't in the list of roles, add it.
    roles = [x.strip() for x in splitted[roleIndex].split(roleDelim)]
    if not roleAll in roles:
      changed = True
      roles.append(roleAll)
      # toss empty fields.
      roles = filter(lambda x: len(x) > 0, roles)
      # update the roles field in this line.
      splitted[roleIndex] = str.join(roleDelim,  roles)
      # update this line in the file array.
      passwdLines[i]      = str.join(fieldDelim, splitted) + "\n"
  # if changes were made, write the passwd file out agian.
  if changed:
    open(path, 'w').writelines(passwdLines)


def fixPasswdPermissions(path, isDryRun):
  """
  In some versions prior to 3.2.0, passwd files were created world-readable.
  On Unix systems, just ensure the perms are set properly.
  """

  # can't do anything about this on windows, and free doesn't have a passwd file.
  if comm.isWindows or not os.path.exists(path):
    return
  if isDryRun:
    logger.info("Would ensure '%s' is mode %s." % (path, PERMS_OWNER_RW_ONLY))
  else:
    if 0 != subprocess.call(['chmod', PERMS_OWNER_RW_ONLY, path]):
      raise cex.ArgError, "Was unable to change permissions of '%s'.  Unwise to continue without figuring out why."


def migQueryToSearch_3_3_0(pathSaved, pathEvttype):
  """
  The language police has mandated that the 'query' key change to 'search' for
  saved searches and event types.  This is critical.
  """
  for path in (pathSaved, pathEvttype):
    if os.path.exists(path):
      comm.sed(r'^(\s*)query(\s*=)', r'\1search\2', path, inPlace = True)


def migCapabilities_3_3_0(authorizeConf):
    """
    Migrate capabilities that have been renamed.
    """
    if not os.path.exists(authorizeConf):
        return
    for substr, replacement in (('edit_exec', 'edit_scripted'),
                                ('edit_watch', 'edit_batch'),
                                ('edit_tail', 'edit_monitor')):
        comm.sed(r'^(\s*)%s(\s*=)' % substr,
                 r'\1%s\2' % replacement,
                 authorizeConf,
                 inPlace = True)

def is_int(s):
        try:
                int(s)
                return True
        except ValueError:
                return False

def migEtcUsers_4_0_5(users_dir, isDryRun):

        reserved_dir = os.path.join(users_dir, '_reserved')
        # no user dirs to migrate
        if not os.path.exists(reserved_dir) or not os.path.isdir(reserved_dir):
           return 

        # matches user.original_user_md5 dir names
        regex = re.compile('^(.*)\\..*')
        dirs  = os.listdir(reserved_dir) 
       
        for d in dirs:
          user_dir = os.path.join(reserved_dir, d)
          match    = regex.match(d) 
          if not os.path.isdir(user_dir) or match == None:
             continue;
      
          new_user_dir = os.path.join(users_dir, match.group(1))

          if os.path.exists(new_user_dir):
              logger.warn("\nCannot automatically migrate user dir " + str(user_dir) + " to " + str(new_user_dir) + ". Destination dir already exists" )
	      continue
 
          comm.moveItem(user_dir, new_user_dir, isDryRun)    
     
        users_ini = os.path.join(users_dir, "users.ini")
        if os.path.exists( users_ini ):
           comm.moveItem(users_ini, os.path.join(users_dir, "users.ini.pre405"), isDryRun)  
         	
        comm.removeItem(reserved_dir, isDryRun)

def migSavedsearches_4_0_0(src_savedsearch_conf_path, dst_savedsearch_conf_path, local_meta_path, isDryRun):
        ###
        # see SPL-20459 - savedsearch.conf migration
        ###
        if not os.path.exists( src_savedsearch_conf_path ):
                return

        if not os.path.exists( os.path.dirname(dst_savedsearch_conf_path) ):
               os.makedirs ( os.path.dirname(dst_savedsearch_conf_path) )

        inPlace = src_savedsearch_conf_path == dst_savedsearch_conf_path

        old_backup_ext = ".old"

        # make backup of original if not dryrun
        comm.copyItem( src_savedsearch_conf_path, src_savedsearch_conf_path + old_backup_ext, isDryRun )

        stanzas = comm.readConfFile( src_savedsearch_conf_path )
        form_search_matcher = re.compile( "\$.*\$" );

        meta_stanza = comm.readConfFile( local_meta_path )

        id2user = comm.id2userMapFromPasswdFile( os.path.join( bundle_paths.etc(), "passwd" ) )

        for stanza,settings in stanzas.items():

                if settings.has_key( 'search' ) \
                        and form_search_matcher.search( settings['search'] ):
                                settings['disabled'] = '1'

                meta_stanza_key = 'savedsearches/' + urllib.quote(stanza)
                tmp = { }

                if settings.has_key( 'role' ):
                        lowercase_role = settings['role'].lower()
                        if lowercase_role == "everybody":
                                lowercase_role = "*"
                        access_val = 'read : [ ' + lowercase_role  + ' ]'
                        tmp[ 'access' ] = access_val

                if settings.has_key( 'userid' ):
                        if id2user.has_key( settings['userid'] ):
                                tmp[ 'owner' ] = id2user[ settings['userid'] ]
                        else:
                                tmp[ 'owner' ] = settings['userid']
                                logger.warn('could not find user in splunk etc/passwd corresponding to id='
                                        + settings['userid'] 
                                        + ' from savedsearch: ' 
                                        + stanza
                                        + ' ... simply using as is for meta') 
                tmp['export'] = 'system'
                if len(tmp) != 0 and not stanza == "default":
                        meta_stanza[ meta_stanza_key ] = tmp

        if inPlace:
            comm.writeConfFile(dst_savedsearch_conf_path, stanzas)
        else:
             dst_stanzas = comm.readConfFile( dst_savedsearch_conf_path )
             if not dst_stanzas is None and len(dst_stanzas) > 0:
                stanzas.update(dst_stanzas)
             dst_stanzas = stanzas

             comm.writeConfFile(dst_savedsearch_conf_path, dst_stanzas)

   	     f = open(src_savedsearch_conf_path, 'w')
             f.write("###############\n")
             f.write("# contents migrated to: " + str(dst_savedsearch_conf_path) + "\n")
             f.write("###############\n")
             f.close()

        # comment out default view states
        comm.sed( r'(^\s*viewstate.*$)', r'# \1 --commented out for migration-- ', dst_savedsearch_conf_path, inPlace = True )

        comm.writeConfFile(local_meta_path, meta_stanza )

def migTranactionTypes_4_0_0(txn_conf_path, isDryRun):

        if not os.path.exists( txn_conf_path ):
                return

        old_backup_ext = ".old"

        # make backup of original if not dryrun
        comm.copyItem( txn_conf_path, txn_conf_path + old_backup_ext, isDryRun )

        # comment out deprecated options
        comm.sed( r'(^\s*(?:aliases|pattern|match|exclusive|maxrepeats).*$)', r'# \1 --commented out by migration, this option has been deprecated -- ', txn_conf_path, inPlace = True )

        # migrate the usage of ` in starts/endswith
        comm.sed( r'(^\s*(?:startswith|endswith))\s*=\s*"([^"]+)"\s*$', r'\1 = eval( \2 )\n', txn_conf_path, inPlace = True )
        comm.sed( r"'([^']+)'", r'"\1"', txn_conf_path, inPlace = True )
        comm.sed( r"`([^`]+)`", r'"\1"', txn_conf_path, inPlace = True )
        
# Sourcetype aliasing used to be done with tags - migrate all tags on sourcetype to rename in props.conf
def migSourcetypeAliases_4_0_0(tags_conf_path, props_conf_path, isDryRun):
  if not os.path.exists( tags_conf_path ):
    return

  # check that a sourcetype stanza exists in tags.conf
  tag_stanzas = comm.readConfFile( tags_conf_path )
  if not "sourcetype" in tag_stanzas:
    return

  # move every enabled tag of sourcetype to props.conf
  renamePairs = {}
  for k, v in tag_stanzas["sourcetype"].items():
    if v == "enabled":
      toks = k.split("::")
      if len(toks) != 3 or toks[0] != "tag":
        print "Failed to parse", k, "as a tag"
      elif toks[1] in renamePairs:
        print "Multiple aliases found for sourcetype", toks[1],"- only renaming to",renamePairs[toks[1]]
      else:
        renamePairs[toks[1]] = toks[2]

  if len(renamePairs) == 0:
    return

  # make backup of original if not dryrun
  comm.copyItem( tags_conf_path, tags_conf_path + ".old", isDryRun )
  if os.path.exists( props_conf_path ):
    comm.copyItem( props_conf_path, props_conf_path + ".old", isDryRun )

  del tag_stanzas["sourcetype"]
  props_stanzas = comm.readConfFile( props_conf_path )

  for oldname, newname in renamePairs.items():
    if oldname in props_stanzas:
      if "rename" in props_stanzas[oldname]:
        print "Sourcetype",oldname,"already has a rename specified. Ignoring old alias",newname
      else:
        props_stanzas[oldname]["rename"] = newname
    else:
      props_stanzas[oldname] = {"rename":newname}

  # Write the new stanzas out to conf files
  comm.writeConfFile(tags_conf_path, tag_stanzas)
  comm.writeConfFile(props_conf_path, props_stanzas)

def migIndexes_4_0_0(indexes_conf_path, isDryRun):
        ###
        # see SPL-20802 - indexes.conf migration
        ###
        if not os.path.exists( indexes_conf_path ):
                return

        old_backup_ext = ".old"

        # make backup of original if not dryrun
        comm.copyItem( indexes_conf_path, indexes_conf_path + old_backup_ext, isDryRun )

        stanzas = comm.readConfFile( indexes_conf_path )

        keys_to_remove = ['_actions'
                        ,'maxTerms'
                        ,'maxTermChars'
                        ,'maxPostings'
                        ,'maxValues'
                        ,'waitForOptimize'
                        ,'indexThreads'
                        ,'maxMemMB']

        for stanza,settings in stanzas.items():
                #
                # we guess that an index is high-volume in 3.X if indexThreads are
                # set to > 0 for this stanza
                #
                is_high_volume_index = False
                if stanza != 'default'                                  \
                        and settings.has_key( 'indexThreads' )          \
                        and is_int( settings['indexThreads'] )          \
                        and int( settings['indexThreads'] ) > 0:
                                is_high_volume_index = True

                for key in keys_to_remove:
                        if settings.has_key( key ):
                                del settings[key]
                
                #
                # if we have a high-volume index, we better set 
                # the corresponding high-volume attributes
                #
                if is_high_volume_index:
                        settings['maxHotBuckets'] = '10'
                        settings['maxHotIdleSecs'] = '86400'
                        settings['maxMemMB'] = '20'

        comm.writeConfFile(indexes_conf_path, stanzas)

# Migrates any enabled (and potentially modified) WMI inputs that were shipped
# under etc\system\default\wmi.conf.
def migWmi_4_0_0(migWmiConf, dryRun):
    # 3.4.x's etc\system\default\wmi.conf in a dict
    wmi_conf_default = {
        "WMI:LocalApplication" : { "interval" : "10", "event_log_file" : "Application", "disabled" : "1" },
        "WMI:LocalSystem" : { "interval" : "10", "event_log_file" : "System", "disabled" : "1" },
        "WMI:LocalSecurity" : { "interval" : "10", "event_log_file" : "Security", "disabled" : "1" },
        "WMI:CPUTime" : { "interval" : "5", "wql" : "SELECT PercentProcessorTime FROM Win32_PerfFormattedData_PerfOS_Processor", "disabled" : "1" },
        "WMI:Memory" : { "interval" : "5", "wql" : "SELECT CommittedBytes, AvailableMBytes, PagesPerSec FROM Win32_PerfFormattedData_PerfOS_Memory", "disabled" : "1" },
        "WMI:LocalDisk" : { "interval" : "5", "wql" : "SELECT PercentDiskTime, AvgDiskQueueLength FROM Win32_PerfFormattedData_PerfDisk_PhysicalDisk", "disabled" : "1" },
        "WMI:FreeDiskSpace" : { "interval" : "5", "wql" : "SELECT FreeMegabytes FROM Win32_PerfFormattedData_PerfDisk_LogicalDisk", "disabled" : "1" },
        }

    if not os.path.exists( migWmiConf ):
        return

    old_backup_ext = ".old"

    # make backup of original if not dryrun
    comm.copyItem( migWmiConf, migWmiConf + old_backup_ext, dryRun )

    # this makes sure that all required properties are available in the migrated file
    stanzas = comm.readConfFile( migWmiConf )
    for stanza in stanzas.keys():
        if stanza in wmi_conf_default:
            for k in wmi_conf_default[stanza]:
                if k not in stanzas[stanza]:
                    stanzas[stanza][k] = wmi_conf_default[stanza][k]

    comm.writeConfFile(migWmiConf, stanzas)

######################
# Begin eventtype tag migration from eventtypes.conf to tags.conf
######################
def getCurrentTaggedEventtypes( typeConfFile ):
    conf = comm.readConfFile( typeConfFile )
    eventtypeTags = {}
    for k,v in conf.items():
        eventtype = k
        if "tags" in v and len(v["tags"]) > 0:
            eventtypeTags[ eventtype ] = v["tags"].split()
    return eventtypeTags

def getCurrentEventtypeTags( tagConfFile ):
    conf = comm.readConfFile( tagConfFile )
    eventtypeTags = {}
    for k, v in conf.items():
        if k == "eventtype":
            for g,k in v.items():
                vars = g.split("::")
                if( len(vars) != 3 ):
                    print "Failed to parse ",vars,"as a tag"                    
                type = vars[1]
                tag = vars[2]
                if type not in eventtypeTags:
                    eventtypeTags[type] = [ tag ]
                else:
                    eventtypeTags[type].append( tag )
    return eventtypeTags, conf

def mergeEventtypeTagsIntoTags( incomingTags, tagsConfFile, isDryRun=False ):
    if not incomingTags:
        ##Nothing to do
        return
    currentTags,conf = getCurrentEventtypeTags(tagsConfFile)    
    if "eventtype" not in conf:
        conf["eventtype"] = {}
    for type in incomingTags.keys():
        tags = incomingTags[ type ]
        if type in currentTags:
            for ctag in currentTags[type]:
                if ctag in tags:
                    tags.remove( ctag )
        for tag in tags:
            conf["eventtype"]["tag::" + type + "::" + tag ] = "enabled"
        if tags:
            print "Migrated the following tags from eventtype=",type,str(tags)
    if not isDryRun:
        comm.writeConfFile( tagsConfFile, conf )


def migEventTypeTags_4_0_0( typeConfFile, tagConfFile, isDryRun):
    mergeEventtypeTagsIntoTags( getCurrentTaggedEventtypes( typeConfFile ), tagConfFile, isDryRun )
##############
# End type tag migration
###############
    
def migInputs_3_3_0(path, isDryRun):
  """
  Convert deprecated inputs to new ones:
  - if there's a tail://, convert it to monitor://.
  - if there was a batch://...
    - if it had move_poly = sinkhole, leave it untouched!
    - otherwise, comment that line and change stanza to monitor://.

  In an attempt to preserve comments, some trickery is employed.  Instead of
  reading via readConfFile and writing out the new settings via
  writeConfFilePath, which will toss all the comments, we use the former to
  load all the settings, and then do a bunch of sed() calls to make changes.
  """
  if not os.path.exists(path):
    return

  tailInputs          = []
  batchInputsNormal   = [] # move_policy != sinkhole, or not specified.
  batchPrefix         = 'batch://'
  monitorPrefix       = 'monitor://'
  tailPrefix          = 'tail://'
  movePolicyKey       = 'movePolicy'
  sinkholePolicy      = 'sinkhole'

  # start by reading the file into our dicts, to easily find all tail stanzas,
  # all batch stanzas that should change to monitor, and all batch stanzas that
  # are sinkholes and should remain untouched.
  localInputs = comm.readConfFile(path)
  for input, settings in localInputs.items():
    # tail:// inputs...
    if input.startswith(tailPrefix):
      tailInputs.append(input)
    # batch:// inputs...
    elif input.startswith(batchPrefix):
      # if it's not a sinkhole, hold on to it.
      if (not movePolicyKey in settings) or (settings[movePolicyKey] != sinkholePolicy):
        batchInputsNormal.append(input)

  # now that we know which stanzas should change to what, we can just do a bunch
  # of regex replacements on the conf file.
  
  # start by converting the tail and batch (non-sinkhole) stanzas...
  for inputList, inputPrefix in (
      (tailInputs,        tailPrefix),
      (batchInputsNormal, batchPrefix)
      ):
    for oneInput in inputList:
      # strip out the prefix and get the path.
      inputPath = oneInput.replace(inputPrefix, '', 1)
      # pick out everything surrounding input prefix in the stanza.
      searchRE  = r"^(\s*\[)%s(%s\]\s*)$" % (inputPrefix, re.escape(inputPath))
      # replace the stanza with tail:// or batch:// changed to monitor://.
      replaceRE = r"\1%s\2" % monitorPrefix
      # actually run the regex.
      comm.sed(searchRE, replaceRE, path, inPlace = True)

  # since move_policy settings that are 'passive_copy' and 'passive_symlink'
  # are no longer any use to us, just comment them out.  'sinkhole' remains.
  policySearch  = r'^(\s*move_policy\s*=\s*passive_.*)'
  policyReplace = r'# DEPRECATED (SPLUNK MIGRATED): \1'
  comm.sed(policySearch, policyReplace, path, inPlace = True)


def migWinSavedSearches(args, fromCLI):
  """
  CLI wrapper for windows saved search migration.
  """
  paramsReq = ()
  paramsOpt = (ARG_DRYRUN, ARG_NOWAIT,)
  comm.validateArgs(paramsReq, paramsOpt, args)

  delay, isDryRun = 5, False
  if ARG_NOWAIT in args and comm.getBoolValue(ARG_NOWAIT, args[ARG_NOWAIT]):
    delay = 0
  if ARG_DRYRUN in args and comm.getBoolValue(ARG_DRYRUN, args[ARG_DRYRUN]):
    isDryRun = True

  logger.info('This command will update your saved searches, but you must review its changes' + '\n' +
                'once it is finished.'                                                          + '\n' +
                                                                                                  '\n' +
                'Will begin in %d seconds (press Ctrl-C to cancel).' % delay                    + '\n')

  time.sleep(delay)

  if not os.path.exists(PATH_SAVSRCH_CONF):
    raise cex.FilePath, "Nothing to migrate - path doesn't exist (%s)." % PATH_SAVSRCH_CONF

  if not isDryRun:
    bakPath = PATH_SAVSRCH_CONF + BACKUP_EXT
    comm.copyItem(PATH_SAVSRCH_CONF, bakPath, isDryRun)

  path = chooseFile(PATH_SAVSRCH_CONF, isDryRun, useNewPaths=True)
  migWinSavedSearchesWorker(path, isDryRun)

  logger.info("Updated: %s." % path)


def migWinSavedSearchesWorker(path, isDryRun):
  """
  This guy does the work of actually migrating saved searches based on the old Windows
  field names.  It should only be run after 3.3.0 migration has finished, otherwise it
  will report errors.

  This SHOULD NOT be run as a part of autoMigrate(), as its results may not always be
  completely correct, due to the nature of the conversion.
  """
  if not os.path.exists(path):
    raise cex.FilePath, "Nothing to migrate - path doesn't exist (%s)." % path

  confDict = comm.readConfFile(path)


  def makeOldName(name):
    return str(name) + "-old"

  regexes = (
      (re.compile(r'\bevtlog_category\b'),  'CategoryString'),
      (re.compile(r'\bevtlog_id\b'),        'EventCode'),
      (re.compile(r'\bevtlog_severity\b'),  'Type'),
      (re.compile(r'\bevtlog_account\b'),   'User'),
      (re.compile(r'\bevtlog_domain\b'),    'ComputerName'),
      (re.compile(r'\bevtlog_sid\b'),       'Sid'),
      (re.compile(r'\bevtlog_sid_type\b'),  'SidType'))
  newSSs = {}
  for oneSSName, oneSSSetts in confDict.items():
    if "search" in oneSSSetts:
      copied = False
      for regex, replacement in regexes:
        if not copied:
          # should we backup?
          if regex.search(oneSSSetts["search"]) != None:
            copied = True
            newSSs[makeOldName(oneSSName)] = copy.deepcopy(oneSSSetts)
        confDict[oneSSName]["search"] = regex.sub(replacement, oneSSSetts["search"])
  for name, setts in newSSs.items():
    confDict[name] = setts

  comm.writeConfFile(path, confDict)


def migSSLConf_4_0_0(path):
  """
  Clear up ambiguities in server.conf:
    keyfile         -> sslKeysfile
    keyfilePassword -> sslKeysfilePassword
  """
  STANZA_SSL       = "sslConfig"
  KEY_KEYFILE_OLD  = "keyfile"
  KEY_KEYFILE_NEW  = "sslKeysfile"
  KEY_PASSWORD_OLD = "keyfilePassword"
  KEY_PASSWORD_NEW = "sslKeysfilePassword"

  if not os.path.exists(path):
    return

  changed = False

  settsDict = comm.readConfFile(path)
  if STANZA_SSL in settsDict and KEY_KEYFILE_OLD  in settsDict[STANZA_SSL]:
    changed = True
    settsDict[STANZA_SSL][KEY_KEYFILE_NEW]  = settsDict[STANZA_SSL].pop(KEY_KEYFILE_OLD)
  if STANZA_SSL in settsDict and KEY_PASSWORD_OLD in settsDict[STANZA_SSL]:
    changed = True
    settsDict[STANZA_SSL][KEY_PASSWORD_NEW] = settsDict[STANZA_SSL].pop(KEY_PASSWORD_OLD)

  if not changed:
    return

  comm.writeConfFile(path, settsDict)


def migAuthConf_4_0_0(path):
  """
  Lowercase splunk roles in LDAP role mappings, since our roles are all
  lowercased now.
  """
  STANZA_ROLEMAP = "roleMap"

  if not os.path.exists(path):
    return

  confDict = comm.readConfFile(path)

  if not STANZA_ROLEMAP in confDict:
    return

  changed = False
  for key in confDict[STANZA_ROLEMAP].keys():
    lowercased = key.lower()
    if key != lowercased:
      changed = True
      confDict[STANZA_ROLEMAP][lowercased] = confDict[STANZA_ROLEMAP].pop(key)

  if not changed:
    return

  comm.writeConfFile(path, confDict)


def migPAM_Scripted_4_1(path, isDryRun):
  """
  The old pamScripted.py example auth script was heinously insecure, so check
  to see if they are using it with some grepping around. If they are, throw
  up a big warning message and move the script so they don't use it.
  """
  if not os.path.exists(path):
    return

  confDict = comm.readConfFile(path)

  # First ensure we have the required keys
  authen = 'authentication'
  authType = 'authType'
  authSettings = 'authSettings'
  if authen not in confDict or authType not in confDict[authen] or authSettings not in confDict[authen]:
    return

  # Only migrate if we're using Scripted and the settings exist
  if confDict[authen][authType] != 'Scripted' or confDict[authen][authSettings] not in confDict:
    return

  stanza = confDict[authen][authSettings]
  settings = confDict[stanza]

  # Look for the path of the script, making sure it exists
  if 'scriptPath' not in settings:
    return

  # The script path line contains the call to python, so grab everything after that and try that path
  # Clearly this won't ALWAYS work, a path could have 'python ' in it. But you can't simply split on spaces
  script = settings['scriptPath']
  pyloc = script.find('python ')
  if pyloc == -1:
    return

  script = script[pyloc + len('python '):]
  script = os.path.expandvars(script.strip())

  if not os.path.exists(script):
    return

  # Read at most 20k bytes of the file (the initial sample script is ~6 KB)
  FH = open(script)
  contents = FH.read(20 * 1024)
  FH.close()

  # Search for some old indicators of the bad script. Print a big warning message
  if contents.find('shell=True') != -1 and contents.find('pamauth " + str(infoIn[') != -1:
    logger.warn('********************************   WARNING   ********************************')
    logger.warn('The sample PAM authentication script being used is deprecated and poses a security risk.')
    if not isDryRun:
      logger.warn('The script will be disabled by moving it to ' + script + '.bak')
      logger.warn('Please copy the new version of pamScripted.py to ' + script + ' and edit it with your PAM settings.')
    logger.warn('******************************** END WARNING ********************************')
    comm.moveItem(script, script + '.bak', isDryRun)


def showPreviewFiles(previewFileList):
  """
  if this is a dry run, let's make sure we display all the files that would have been changed.
  we will not display preview files that were unchanged from the original versions.
  """
  toRemove     = [] 
  # get rid of preview files that have not changed from the original version (if it exists).
  for prevFile in previewFileList:
    unprevFile = getUnpreviewName(prevFile)
    # for some of these preview files, a direct counterpart doesn't exist.  they don't need dupe checking anyway.
    if os.path.exists(unprevFile):
      fileIsSame = filecmp.cmp(prevFile, unprevFile, shallow = False)
      # if the file hasn't changed, add it to our trim list.
      if fileIsSame:
        toRemove.append(prevFile)
  # for every file that we saved, delete it and remove it from the list.
  for dupeFile in toRemove:
    # remove the .migratePreview file - it's the same as the existing file (no change).
    os.unlink(dupeFile)
    # and take it out of the list we display.
    previewFileList.remove(dupeFile)
  # now do the notifications.
  logger.notice("\nThe following files would be created or modified (without the %s extension):\n%s"
                % (SUF_MIGRATE, str.join("\n", [("  " + x) for x in previewFileList])))


def warnUnmigratableFiles(isDryRun):
  """
  there are a few config files that we don't have any migration code for, at
  least in 4.0.  make sure users are aware of this if it affects them.
  """

  if not upgradingFromBeforeSplunk4():
    return

  badPaths_4_0_0 = (bundle_paths.make_path("deployment.conf"),
                    bundle_paths.make_path("distsearch.conf"),
                    bundle_paths.make_path("outputs.conf"))

  hasUnmigratableFiles = False
  for badPath in badPaths_4_0_0:
    if os.path.exists(badPath):
      if not hasUnmigratableFiles: # first occurrence
        logger.warn("\nCannot automatically migrate:")
      hasUnmigratableFiles = True
      logger.warn("  %s" % badPath)
  if hasUnmigratableFiles:
    if isDryRun:
      logger.warn(UNMIGRATABLE_PROMPT)
    else:
      if not comm.prompt_user(UNMIGRATABLE_PROMPT, checkValidResponse = True):
        raise cex.SuccessException, UNMIGRATABLE_MSG


def replaceLicense(isDryRun):
  """
  the 3.x license is incompatible with 4.0.  if we haven't seen a 4.0 
  installation before, copy over our 4.0 splunk-free license to get the user
  started.  if a splunk-user.license exists, copy that over instead of the free
  license, so users have some reasonable way to automate this.

  we decide to replace the license only upon seeing <serverName> in splunkd.xml,
  as this indicates that we've come from a splunk version that's never been
  upgraded to 4.x.
  """

  if not upgradingFromBeforeSplunk4():
    return

  # backup old license.
  if os.path.exists(PATH_LICENSE_ACTIVE):
    comm.moveItem(PATH_LICENSE_ACTIVE, PATH_LICENSE_ACTIVE + BACKUP_EXT, isDryRun)

  # copy new license in place.
  licPath = os.path.exists(PATH_LICENSE_USER) and PATH_LICENSE_USER or PATH_LICENSE_FREE
  comm.copyItem(licPath, PATH_LICENSE_ACTIVE, isDryRun) 


def replaceSplunkdXml_4_0_7(srcPath, dstPath, isDryRun):
  """
  Clobber the user's existing splunkd.xml with our freshly shipped one.  As of
  4.0.0, there is no user-configurable data in splunkd.xml.  migSplunkdXml_4_0_0
  will have been called already, so any user settings are in server.conf at this
  point.
  """
  if not os.path.exists(srcPath):
    raise cex.FilePath("Cannot copy '%s' to '%s' - if this is a development build, you may"
        " want to:\n\tcp '%s' '%s'" % (srcPath, dstPath, dstPath, srcPath))
  comm.moveItem(srcPath, dstPath, isDryRun)


def warnOutdatedApps():
  """
  Upon initial upgrade from pre-4.0 to 4.x, warn about apps that don't appear to
  be compatible with 4.x.  We determine this by looking for the metadata
  directory in the app, which is essentially required for 4.x apps.
  Also special-case a warning for the CM and PCI apps, which are only for 3.x.
  """
  if not upgradingFromBeforeSplunk4():
    return # have already warned once before.
  knownIncompatibleApps = ("change_management", "pci") 

  # check for apps that we KNOW aren't yet 4.x.
  foundIncompatibleApps = []
  lowerAppList = [x.lower() for x in os.listdir(PATH_APPS_DIR)] # lowercase to ease comparisons.
  for app in knownIncompatibleApps:
    if app in lowerAppList:
      foundIncompatibleApps.append(app)
  if len(foundIncompatibleApps) > 0:
    logger.warn("")
    logger.warn("The following 3.x Apps in your Splunk deployment are incompatible with Splunk 4")
    logger.warn("and must be replaced:")
    for app in foundIncompatibleApps:
      logger.warn("  %s" % app)
    logger.warn("Updated versions of these Apps are not yet available. Please file a case online at http://www.splunk.com/page/submit_issue to")
    logger.warn("learn when updated versions will be released.")
  
  # now find the rest, by checking for metadata dirs.
  appsWithoutMetadata = []
  for app in os.listdir(PATH_APPS_DIR):
    if not app.lower() in knownIncompatibleApps: # skip dupes of above.
      metadataPath = os.path.join(PATH_APPS_DIR, app, "metadata")
      if not os.path.exists(metadataPath):
        appsWithoutMetadata.append(app)
  if len(appsWithoutMetadata) > 0:
    logger.warn("")
    logger.warn("The following 3.x Apps in your Splunk deployment must be updated before they will")
    logger.warn("work properly with Splunk 4:")
    for app in appsWithoutMetadata:
      logger.warn("  %s" % app)
    logger.warn("Visit the following URL and review the Splunk Developer Manual to learn how to")
    logger.warn("convert your 3.x Apps:")
    logger.warn("  http://www.splunk.com/base/Documentation/latest/Developer/Migrate3x")

  logger.info("")


def autoMigrate(dryRun = False):
  """
  The central migration control function.  This should automagically migrate an old Splunk's configs
  to an uptodate set of configuration files.
  """

  if comm.isLocalSplunkdUp():
    raise cex.ServerState, "In order to migrate, Splunkd must not be running."

  #
  # setup logging.  all migration stuff will go to file (as well as stdout/stderr).
  # at the end of this function, we'll remove the file-logging of all logger messages.
  # otherwise all subsequent pcl messages would get logged to file as well.
  #

  logFilename = "migration.log.%s" % time.strftime("%Y-%m-%d.%H-%M-%S")
  logFilePath = os.path.join(PATH_LOG_DIR, logFilename)
  migLogFile  = open(logFilePath, 'w')
  migLogFileHandler = comm.newLogHandler(migLogFile, comm.debugMode, normal = False, filter = False)
  logger.getLogger().addHandler(migLogFileHandler)

  logger.info("\n-- Migration information is being logged to '%s' --" % logFilePath)
  logger.info("\nMigrating to:")
  logger.info("\n%s", open(PATH_VERSION_FILE, "r").read()) # dump contents of version file.

  if dryRun:
    logger.notice("\n********** BEGIN PREVIEW OF CONFIGURATION FILE MIGRATION **********")

  #
  # if this is false, that means we need to move our bundles to the new
  # location.  this file gets created by one of our migration calls.
  # for fresh installs post-move, check for bundles/default/ as well...
  #
  isAfterBundleMove = os.path.exists(PATH_MIGRATION_CONF) or not os.path.exists(PATH_OLD_DEF_BUNDLES)

  # start by ensuring that everything we want to migrate is in etc/system/...
  if not isAfterBundleMove:
    # while we're in here, delete the old 'default' bundle, or there's gonna
    # be some confused customers.  migrate_bundles will to the right thing,
    # ie delete, with the default bundle when asked.  all other bundles will
    # remain.
    for bundle in ('local', 'default'):
      if os.path.exists(bundle_paths.make_legacy_bundle_install_path(bundle)):
        bundle_paths.migrate_bundles({'dry-run' : dryRun, 'name' : bundle}, fromCLI = True)
  if not dryRun:
    isAfterBundleMove = True


  # these rely on splunkd.xml - be sure to run this before that's upgraded for 4.0.0!
  replaceLicense(dryRun)
  warnOutdatedApps()
  warnUnmigratableFiles(dryRun)

  #
  #
  # choose which files to edit (based on whether we're previewing migration or not).
  #
  #

  migAuthConf      = chooseFile(PATH_AUTHEN_CONF_NEW,  dryRun, isAfterBundleMove)
  migInputsConf    = chooseFile(PATH_INPUTS_CONF,      dryRun, isAfterBundleMove)
  migSplunkdXml    = chooseFile(PATH_SPLUNKD_XML,      dryRun, isAfterBundleMove)
  migPasswdFile    = chooseFile(PATH_PASSWD_FILE,      dryRun, isAfterBundleMove)
  migSavedConf     = chooseFile(PATH_SAVSRCH_CONF,     dryRun, isAfterBundleMove)
  migEvttypeConf   = chooseFile(PATH_EVTTYPE_CONF,     dryRun, isAfterBundleMove)
  migSrvConf       = chooseFile(PATH_SERVER_CONF,      dryRun, isAfterBundleMove)
  migWebConf       = chooseFile(PATH_WEB_CONF,         dryRun, isAfterBundleMove)
  migIndexesConf   = chooseFile(PATH_INDEXES_CONF,     dryRun, isAfterBundleMove)
  migTagsConf      = chooseFile(PATH_TAGS_CONF,        dryRun, isAfterBundleMove)
  migPropsConf     = chooseFile(PATH_PROPS_CONF,       dryRun, isAfterBundleMove)

  migTxnTypesConf  = chooseFile(PATH_TXN_TYPES_CONF,   dryRun, isAfterBundleMove)
  migLocalMeta     = chooseFile(PATH_LOCALMETA_CONF,   dryRun, isAfterBundleMove)
  

  migWmiConf          = chooseFile(PATH_WMI_CONF,                dryRun, isAfterBundleMove)
  migSearchLocalMeta  = chooseFile(PATH_SEARCH_LOCALMETA_CONF,   dryRun, isAfterBundleMove)
  migSearchSavedConf  = chooseFile(PATH_SEARCH_SAVSRCH_CONF,     dryRun, isAfterBundleMove)

  #
  #
  # do the actual migration steps...
  # this section should basically be in order from oldest migration functions
  # to newest migration functions (for the most part...).
  #
  #

  # make a backup of splunkd.xml.
  comm.copyItem(PATH_SPLUNKD_XML, PATH_SPLUNKD_XML_BAK, dryRun)

  # make our various changes to splunkd.xml.
  migrateSplunkdXml_3_0_0(migSplunkdXml)
  migrateSplunkdXml_3_2_0(migSplunkdXml)

  migrateAuthSecret(dryRun)
  migratePasswd(migPasswdFile)
  removeDirMon(dryRun)

  checkSavedSearches(CONF_SAVEDSEARCHES)
  checkTimezones(CONF_PROPS, dryRun)
  migrateAuth_3_2_0(migAuthConf, dryRun)
  migrateSavedSearches(migSavedConf)
  migrateWebSSL(migSrvConf, migWebConf)
  removeDeprecated(dryRun)

  # we run this even though we've previously moved everything out of
  # bundles/local, as it can still produce valid warnings depending upon what
  # the user has done.
  bundle_paths.warn_about_legacy_bundles(dryRun)

  # continue on with migration functions that don't depend on the old dir structure.
  checkCommandsConfig(CONF_COMMANDS)
  fixPasswdPermissions(PATH_PASSWD_FILE, dryRun)
  migInputs_3_3_0(migInputsConf, dryRun)
  migQueryToSearch_3_3_0(migSavedConf, migEvttypeConf)
  migCapabilities_3_3_0(PATH_AUTHORIZE_CONF)


  migIndexes_4_0_0(migIndexesConf, dryRun)
  migEventTypeTags_4_0_0(migEvttypeConf, migTagsConf, dryRun)
  migSplunkdXml_4_0_0(migSplunkdXml, migSrvConf)
  migSSLConf_4_0_0(migSrvConf)
  migSavedsearches_4_0_0(migSavedConf, migSearchSavedConf, migSearchLocalMeta, dryRun)
  migAuthConf_4_0_0(migAuthConf)
  migTranactionTypes_4_0_0(migTxnTypesConf, dryRun)
  migSourcetypeAliases_4_0_0(migTagsConf, migPropsConf, dryRun)
  migWmi_4_0_0(migWmiConf, dryRun)

  migEtcUsers_4_0_5(PATH_ETC_USERS, dryRun)
  replaceSplunkdXml_4_0_7(PATH_SPLUNKD_XML_DEF, PATH_SPLUNKD_XML, dryRun)

  migPAM_Scripted_4_1(migAuthConf, dryRun)

  #
  #
  # finished w/ migration (or dry run).
  #
  #

  # store a marker of what version we've upgraded to.
  setMigHistory(KEY_LAST_MIG_VERSION, splunk.getReleaseVersion(), dryRun)

  # find all files in splunkhome that we may have created during migration preview/dry run.
  previewFiles = findPreviewFiles()
  if dryRun:
    showPreviewFiles(previewFiles)
    logger.notice("********** END PREVIEW OF CONFIGURATION FILE MIGRATION **********\n")
    raise cex.SuccessException, POST_DRY_RUN_MESSAGE
  # if it's not a dry run, go through and try to delete all our .migratePreview files from any prior runs!
  # this is so they don't get detected in subsequent upgrades, and accidentally get displayed.
  else:
    for macLikeFile in previewFiles:
      os.unlink(macLikeFile) # delete useless files.
  
  #
  # revert logging.  we don't want all subsequent logging calls to keep going to the mig log file.
  # in the case of an exception above here, we'll be exiting anyway - so no worries about this
  # not getting called in those cases.
  # 

  logger.getLogger().removeHandler(migLogFileHandler)
