# Copyright (C) 2005-2010 Splunk Inc.  All Rights Reserved.  Version 4.0
import logging as logger
import os, sys
import shutil

import cli_common as comm
import control_exceptions as cex
import xml.etree.cElementTree as et
import bundle
import bundle_paths

ARG_PATH = "path"
ARG_NAME = "name"
ARG_UPGRADE = "upgrade"
ARG_ADDON_NAME = "addon-name"
ARG_INCLUDE = "include"
ARG_QUIET = "noprompt"
ARG_BUILDDIR = "builddir"
ARG_DEBUG = "debug"
KEY_BUNDLES = 'bundles'
KEY_BUNDLES_DISABLED = 'disabled-bundles'
DEFAULT_REPLACE_TOKEN = "$$"

BUNDLE_EXTENSIONS = ('.conf', '.bundle')

# Bundles that should not be shown to users
BUNDLES_IGNORE = ('default', 'local', 'README')

# Paths to bundle directories
PATH_BUNDLES = bundle_paths.get_base_path()
PATH_BUNDLES_DISABLED = os.path.join(comm.splunk_home, 'etc', 'disabled-bundles')

PATH_BUILDDIR = "bundles-build"
PATH_BUILDPATH = os.path.join(comm.splunk_home, 'etc', PATH_BUILDDIR) 

# strings
MESSAGE_USING_DEFAULT = "No name supplied, defaulting to local bundle"
NAME_PROMPT = "What would you like to name the addon?"
MESSAGE_VALID_ERROR =  "ERROR - you must supply a valid entry"
PROMPT_REPLACE_VALUES = "Do you want to have values replaced at installation time?"
MESSAGE_REPLACE_VALUES= "If you want to have the package command do variable replacement for you you need to add $$ around the values to replace in your config files before continuing on with this script."
PROMPT_HOW_MUCH_TO_PACKAGE = "Do you want to package all, some, or none of the files in source directory?"
MESSAGE_USING_NAME = "Using name '%s' for addon package"
MESSAGE_NO_DEFAULT_ERROR = "ERROR, you need to supply a default or argument in quiet mode"

#----------------------------------------------------
def printLine(output, extraline=False, args=()):
    extraText = extraline and "\n" or ""
    if args and len(args) > 0:
      logger.info((output % args) + extraText)
    else:
      logger.info(output + extraText)


#----------------------------------------------------------------
class AddonPackager:

  def __init__(self, bundle, args, fromCLI):
    self.bundleName = bundle 
    self.args = args
    self.fromCLI = fromCLI
    self.replaceToken = DEFAUL_REPLACE_TOKEN

    # -- See if any of the args was passed in. 
    # -- this command can be driven via args only.
    if args.has_key(ARG_ADDON_NAME):
        self.newPackageName = args[ARG_ADDON_NAME]
    else:
        self.newPackageName = ""

    if args.has_key(ARG_QUIET):
        self.quiet = True
    else:
        self.quiet = False

    if args.has_key(ARG_INCLUDE):
        self.include = args[ARG_INCLUDE]
    else:
        self.include = ""

    if args.has_key(ARG_BUILDDIR):
        self.builddir = args[ARG_BUILDDIR]
    else:
        self.builddir = PATH_BUILDPATH

    self.fullBuildPath = os.path.join(self.builddir, self.newPackageName)

    if args.has_key(ARG_DEBUG):
        self.debug = True
    else:
        self.debug = True

  #------------------------------------------------------------------
  def debugOut(self, message):
    if self.debug:
        printLine("DEBUG - " + message)

  #-----------------------------------------------------------------
  def getFullBundlePath(self):
    return os.path.join(PATH_BUNDLES, self.bundleName)


  #-----------------------------------------------------------------
  def getFullBuildPath(self):
    return self.fullBuildPath



  #----------------------------------------------------------------
  def promptForValue(self, prompt, values, default, nullOK, blankLine):
    fullprompt = prompt

    if self.quiet:
        if default:
            return default
        else:
            printLine(MESSAGE_NO_DEFAULT_ERROR, extraline=True)
            sys.exit()

    if values and len(values) > 0:   
        first = True
        for i in values:
            if not first:
                fullprompt += "|"
            else:
                fullprompt += " "
                first = False
            if i == default:
                fullprompt += '['+i+']'
            else:
                fullprompt += i

    fullprompt += " : "
    
    
    result = raw_input(fullprompt).strip()
    if result == "" and not nullOK:
      if default and not default =="":
          result = default
      else:
          printLine(MESSAGE_VALID_ERROR)
          result = self.promptForValue(prompt, values, default, nullOK, blankLine)
    else:
        if values and len(values) > 0:
            if result in values:
                return result
            else:
                for i in values:
                    if i.startswith(result):
                        return result
                printLine(MESSAGE_VALID_ERROR)
                result = self.promptForValue(prompt, values, default, nullOK, blankLine)


    if blankLine:
      logger.info("")
    return result


  #----------------------------------------------------------------
  def getPackageName(self):
    if not self.bundleName == 'local':
      newname = self.bundleName
    else:
      newname = ""

    return self.promptForValue(NAME_PROMPT, None, newname, False, True)

  
  #----------------------------------------------------------------
  def preReplaceValues(self):
    printLine(MESSAGE_REPLACE_VALUES, extraline=True)


  #----------------------------------------------------------------
  def ensureBuildDir(self):
    self.debugOut("Checking build dir..."+self.builddir)
    if not os.path.exists(self.builddir):
        self.debugOut("BuildDir not found, attempting to create dir "+self.builddir)
        try:
            os.mkdir(self.builddir)
            self.debugOut("BuildDir created successfully")

        except Exception, e:
            raise cex.FileAccess, 'Error while trying to create builddir' % (self.builddir)

  #---------------------------------------------------------------
  def backupPriorBuild(self):
    if os.path.isdir(self.fullBuildPath):
        self.debugOut("Backing up prior package")
        backupdir = self.fullBuildPath+".bu"
        if os.path.isdir(backupdir):
            shutil.rmtree(backupdir)
        shutil.copytree(self.fullBuildPath, backupdir)
        shutil.rmtree(self.fullBuildPath)

  #---------------------------------------------------------------
  def packageAll(self):
    logger.info("packaging up all files")

    self.ensureBuildDir()
    self.backupPriorBuild()

    # copy bundle directory
    fullpath = self.getFullBundlePath() 
    if os.path.isdir(fullpath):
        try:
            self.debugOut("About to copy " + fullpath + " to " + self.fullBuildPath)
            shutil.copytree(fullpath, self.fullBuildPath)
        except:
            raise cex.FilePath, "Unable to copy bundle at %s" % fullpath


  #---------------------------------------------------------------
  def packageSome(self):
    logger.info("packaging up some files")

  #---------------------------------------------------------------
  def packageNone(self):
    logger.info("why are we doing none?")


  #---------------------------------------------------------------
  def packageFiles(self):
    
    # See how much to package... All, Some, None

    if self.include == "":
        self.include = self.promptForValue(PROMPT_HOW_MUCH_TO_PACKAGE, ["all","some", "none"], "all", False, True)

    if self.include == "all":
        self.packageAll() 
    elif self.include =="some":
        self.packageSome()
    else:
        self.packageNone()

  #---------------------------------------------------------------
  def hasReplaceToken(f):
    self.debugOut("checking to see if file %s has token...", (f))
    try:
        fh= open(f)
        for line in fh.readlines():
            
            mt = re.search(self.relaceToken, line)
            
            if mt and mt != "":
                self.debugOut("found token", (mt))
                return true

    except Exception, e:
        self.printLine('Error while try to test file %s for replace tokens' % (f))
    
    return false


  #----------------------------------------------------------------
  def replaceOptions(self):
    self.debugOut("About to replace values in " + self.getFullBuildPath())
    
    # Check to see if any of the files have the replace tags
    for root, dirs, files in os.walk( self.getFullBuildPath()):
        for file in files:
            logger.info("File = " + file)



  #----------------------------------------------------------------
  def packageAddon(self):

    # Get the name the users wants to call the new package
    if  self.newPackageName == "":
        self.newPackageName = self.getPackageName()
    printLine(MESSAGE_USING_NAME, extraline=True, args=(self.newPackageName))
        

    # see if the users want to replace values
    if not self.quiet and self.promptForValue(PROMPT_REPLACE_VALUES, ["yes","no"],"yes",False,True):
        self.preReplaceValues()
   
    logger.info("")
    self.packageFiles()
    found = self.replaceOptions()
    

    #  ----  End PackageBundle  ------------------------------------------------------



#----------------------------------------------------------------
def processCommand(args, fromCLI):
  paramsReq = ()
  paramsOpt = (ARG_NAME, ARG_ADDON_NAME, ARG_QUIET, ARG_INCLUDE, ARG_DEBUG)
  comm.validateArgs(paramsReq, paramsOpt, args)

  retDict = {'restartRequired': True}
  if args.has_key(ARG_NAME):
    name = args[ARG_NAME]
  else:
    name = ""


  # Check to see if name was supplied, otherwise default to local
  if name == "":
    printLine(MESSAGE_USING_DEFAULT, extraline=True) 
    name = "local"
  else:
    # make sure that its a valid bundle name
    bundles = bundle.getBundleTuples(PATH_BUNDLES)
    logger.info("EEEE %s" % bundles)
    if name not in bundles.keys():
      raise cex.ArgError, '%s is not currently installed' % name
  packager = AddonPackager(name, args, fromCLI);
  packager.packageAddon()
  return "package successful"    
