import sys, os, getopt, traceback, shutil
import splunk.clilib.cli_common as comm
import control_exceptions as cex
import os, fnmatch
import apps, module

SPLUNK_HOME = os.getenv('SPLUNK_HOME')
if not SPLUNK_HOME:
    raise ConfigError('Environment variable SPLUNK_HOME must be set')

# How verbose we are
QUIET = 0
ERROR = 1
INFO = 2
DEBUG = 3

currentLevel = INFO
SPLUNK_BIN = os.path.join(SPLUNK_HOME, 'bin', 'splunk')
SPLUNK_MODULES = os.path.join(SPLUNK_HOME, 'etc', 'modules')

if sys.platform == "win32":
    IS_WINDOWS = True
else:
    IS_WINDOWS = False


CONFIGURATION_OPTIONS = (
    ('modules', '', 'which modules to disable'),
    ('apps', '', "which apps to enable"),
    ('files', {}, 'module configs to swap around'),
)

class Installer(object):
    action = ""
    uninstall = False
    apppath = ""
    confStanzas = ""

    #--------------------------------------------------------------
    def __init__(self, action, appname):
        # uninstall is just convienience to not test string all the time
        self.action = action
        self.appname = appname
        self.uninstall = (self.action == "uninstall")

        # action can be install, uninistall or list
        if  self.action == None or self.action == "":
            self.usage()
            raise cex.ConfigError, "must provide a valid install action: install, uninstall or list"
        
        if self.appname == None or self.appname == "":
            self.usage()
            raise cex.ConfigError, "must provide a valid appname to install"

        # cache some frequently used paths
        self.apppath = os.path.join(SPLUNK_HOME,"etc","apps",self.appname)
   
        # read the configs in to member var.
        self.readConfig()
 
    #--------------------------------------------------------------
    def usage(self):
        self.feedback(INFO, "updater takes one of the following arguments [ install, uninstall, update ]");



    #--------------------------------------------------------------
    def fixSlashes(self, mod):
        if IS_WINDOWS:
            mod = mod.replace('/', '\\')
        return mod


    #--------------------------------------------------------------
    def locate(self, pattern, root=os.curdir):
        for path, dirs, files in os.walk(os.path.abspath(root)):
            for filename in fnmatch.filter(dirs, pattern):
                yield os.path.join(path, filename)

    #--------------------------------------------------------------
    def readConfig(self):
        fullpath = os.path.join(self.apppath,"default","setup.conf")
        self.confStanzas = comm.readConfFile(fullpath)

    #--------------------------------------------------------------
    # wrapper for feedback to uesr
    def feedback( self, level, message):
        if ( currentLevel >= level ):
            print message

    #--------------------------------------------------------------
    def externcall(self, cmd):
        out = os.popen(cmd)
        for line in out.readlines():
            print line

    #--------------------------------------------------------------
    def enableApp(self, app, action):
        if self.uninstall:
            if  action == "disabled":
                action = "enabled"
            else:
                action = "disabled"
        try:
            if action == "disabled":
                self.feedback( INFO, "> Disabling app " + app )
                apps.localAppDisable({'name':app}, False)
            else:
                self.feedback( INFO, "> Enabling app " + app )
                apps.localAppEnable({'name':app}, False)    
        except Exception, e:
            pass

    #--------------------------------------------------------------
    def enableModule(self, mod, action):
        if self.uninstall:
            if action == "disabled":
                action = "enabled"
            else:
                action = "disabled"
        mod = self.fixSlashes(mod)
        try:
            if action == "disabled":
                self.feedback(INFO, "> Disabling module " + mod)
                module.localModuleDisable({'module':mod}, False)
            else:
                self.feedback(INFO, "> Enabling module " + mod)
                module.localModuleEnable({'module':mod}, False)
        except Exception, e:
            pass
    #--------------------------------------------------------------
    def switchConfigs(self, mod, conf):
        mod = self.fixSlashes(mod)
        modPath = os.path.join(SPLUNK_MODULES, mod)
        newFile = os.path.join(self.apppath, self.fixSlashes(conf))  
        self.feedback(INFO, "> Switching  module configs " + mod + " with " + conf)
            
        try:
            if (self.uninstall):
                #only  if disabled exists otherwise then we cannot be uninstalling...

                if (os.path.exists( os.path.join( modPath, "config.xml.disabled"))):

                    self.feedback(DEBUG, "> found disabled version of config " + modPath)
                    self.feedback(DEBUG, "> will delete current config.xml")
                    os.remove( os.path.join(modPath, "config.xml"))

                    # put back disabled
                    module.localModuleEnable({'module':mod}, False)
            else:
                # backup to disabled
                module.localModuleDisable({'module':mod}, False)
                self.feedback(DEBUG, "> About to copy in local config to conf.xml ")
                origFile = os.path.join( modPath, 'config.xml')
                comm.ensureDeletable(newFile)
                shutil.copyfile(newFile, origFile)
        except Exception, e:        
            pass

    
    #--------------------------------------------------------------
    # does all the work here....
    def install(self):
        self.feedback( DEBUG, "About to install - uninstall = " + str(self.uninstall))

        stanzaValue = self.confStanzas.get('apps', {})
        for optionName, optionValue in stanzaValue.items():
            self.enableApp( optionName, optionValue.lower() )

        stanzaValue = self.confStanzas.get('modules', {})
        for optionName, optionValue in stanzaValue.items():
            self.enableModule( optionName, optionValue.lower())

        stanzaValue = self.confStanzas.get('files', {})
        for optionName, optionValue in stanzaValue.items():
            self.switchConfigs(optionName, optionValue)

        return

    #--------------------------------------------------------------
    # we were succesful, anything we want to do at end?
    def success(self):
        self.feedback( INFO, "Installation Successful")

    #--------------------------------------------------------------
    # we failed! call uninstall and exit
    def failure(self):
        self.feedback( ERROR, "Installation Failed - reverting changes..")
        #uninstall.uninstall()
        os._exit(99)


#--------------------------------------------------------------
#  for testing or running by hand.
#--------------------------------------------------------------
if __name__ == "__main__":
    try:
      optlist, args = getopt.getopt(sys.argv[1:], '?',['install','uninstall', 'update'])

      action = ""
      appname = ""

      if args[0] == 'install':
        action = "install"
      elif args[0] == 'update':
        action = "update"
      elif args[0] == 'uninstall':
        action = "uninstall"
      elif args[0] == 'info':
        action = "info"

      appname = args[1]
      if appname == None or appname == "":
        self.feedback(ERROR, "no path to app provided")
    except getopt.error, val:
        print "must supply action and appnme"

    i = Installer(action, appname)
    i.install()
