#Copyright (C) 2005-2010 Splunk Inc. All Rights Reserved. This work contains trade
#secrets and confidential material of Splunk Inc., and its use or disclosure in
#whole or in part without the express written permission of Splunk Inc. is prohibited.


import time, re
import logging

logger = logging.getLogger('splunk.mining.FieldLearning') 

MAX_ITERATIONS = 1
MAX_REGEX_BEFORE_FAILURE = 100
MIN_EXAMPLE_MATCH = 5

# if regex of actual value is > 15 characters, loose up regex.
# revents overly complicated regex when they get unnecessary complex,
# and also prevents false extractions when regex is simple
MAX_VALUE_REGEX = 15

# terms that you do not want to use as a prefix in regex
BAD_MEAT = set(["sun", "mon", "tue", "tues", "wed", "thurs", "fri", "sat", "sunday", "monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec", "january", "february", "march", "april", "may", "june", "july", "august", "september", "october", "november", "december", "2003", "2004", "2005", "2006", "2007", "2008", "2009", "am", "pm", "ut", "utc", "gmt", "cet", "cest", "cetdst", "met", "mest", "metdst", "mez", "mesz", "eet", "eest", "eetdst", "wet", "west", "wetdst", "msk", "msd", "ist", "jst", "kst", "hkt", "ast", "adt", "est", "edt", "cst", "cdt", "mst", "mdt", "pst", "pdt", "cast", "cadt", "east", "eadt", "wast", "wadt"])


# characters used to jump to value (e.g. jump 3 tabs)
GUIDEPOST_CHARACTERS = "\"/\t()[]{}*+^$!-\\?!@#%+=:<>,?;" + "' &.~"

# same but value AFTER value, so willing to accept more unreliable characters . and space.
# unwilling to rely on spaces to jump to a value (because english text length can arbitrarily vary),
# but are willing to have space end a value (e.g "[blah blah blah] 7.0 blah" -> ] (\d+.\d+) ")
VALUEPOST_CHARACTERS = "&. " + GUIDEPOST_CHARACTERS

# don't allow \w+{6}
MAXMATCHCOUNT = 5
# don't allow jump to char if more than 30 chars away.  defeats much of purpose of dist
MAXDIST = 30


def cleanStanzaName(name):
    replacements = """ []$%"&<>()*=\\/"'"""
    for r in replacements:
        name = name.replace(r, '_')
    oldname = None
    while (oldname != name):
        oldname = name
        name = name.replace('__', '_')
    return name



def ruleCMP(x, y):
    return int(1000 * (y.getScore() - x.getScore()))

def learn(events, examples, counterexamples):

    rules = None
    newTerms = None
    newExamples = list(examples)
    # for each learning cycle
    for i in range(0, MAX_ITERATIONS):
        logger.info("GENERATING RULES FROM %s EVENTS" % len(events))
        rules = _generateRules(events, newExamples)
        newTerms = _validateRules(events, newExamples, counterexamples, rules)
        if len(newTerms) == 0:
            break
        newExamples.extend(newTerms)

    terms = set()
    for rule in rules:
        terms.update(rule.getExamplesCount().keys())
        terms.update(rule.getLearnedCount().keys())
        
    logger.debug("%s rules" % len(rules))
    orderedTerms = list(terms)
    orderedTerms.sort()
    logger.debug("Terms Learned: %s" % orderedTerms)

    rules.sort(ruleCMP)

    if len(rules) > MAX_REGEX_BEFORE_FAILURE:
        raise Exception('Too many rules were generated. Try providing values that are more unique to your field.')
    regexes = []
    extractions = {}
    for pos, rule in enumerate(rules):
        regexes.append(rule.getWholePattern())
        knownCounts = rule.getExamplesCount()
        learnedCounts = rule.getLearnedCount()
        terms = knownCounts.keys() + learnedCounts.keys()
        for term in terms:
            myrules = extractions.get(term, [])
            myrules.append(pos)
            extractions[term] = myrules
        
    return regexes, extractions



def removeDoomedRules(rules, doomed):
    for d in doomed:
        for i in range(0, len(rules)):
            if d == rules[i].getWholePattern():
                del rules[i]
                break

def patternComplexity(pattern):
    return pattern.count('(') + pattern.count('[') + pattern.count('\\')
    
    

RULE_TYPES = ["forward", "backward", "forward-literal", "forward-delimiter", "forward-delimiter-simple"]
              

def _generateRules(events, extractions):
    rules = {}
    linecount = 0
    for event in events:
        linecount += 1
        for ruletype in RULE_TYPES:
            myrules = _generateEventRules(events, event, extractions, ruletype)            
            for rule in myrules:
                rulestr = str(rule)
                #print "RULE %s BAD %s" %(rulestr, rule.getBad())
                if rulestr in rules:
                    rule = rules[rulestr]
                else:
                    rules[rulestr] = rule
                    #print "TYPE: %s \n\tRULE: %s \n\tEVENT: %s" % (ruletype, rulestr, event)

                rule.incMatchCount()
                
    return rules.values()

def removeBadRules(rules, badrules):
    for br in badrules:
        rules.remove(br)
    badrules.clear()


def _validateRules(events, examples, counterexamples, rules):
    
    totalExtractions = 0
    knownmatches = 0
    newExtractions = set()
    lineCount = 0
    badrules = set()

    # for each rule
    for rule in rules:
        pattern = rule._wholePattern
        # if the rule is ambiguous, remove it
        # for example: [a-z]+([a-z]+) which can get any arbitrary split
        # print "PATTERN:", pattern
        if re.search("(.+)\(\\1\)", pattern) != None:
            #print "THROWING OUT BAD RULE:", pattern
            badrules.add(rule)
            logger.debug("Removing ambiguously rule: %s" % pattern)

    # REMOVE RULES THAT HAD AMBIGUOUS REGEX
    removeBadRules(rules, badrules)

    for event in events:
        lineCount += 1
        for rule in rules:
            if rule in badrules:
                continue
            extractions = rule.findExtractions(event)
            if len(extractions) > 0:
                for extraction in extractions:
                    #print "EXTRACTED:", extractions, "RULE:", rule
                    rule.addExtraction(extraction)
                    if extraction in counterexamples:
                        badrules.add(rule)
                        logger.debug("Removing rule that learned counter example: %s Counterexample: %s" % (rule._wholePattern, extraction))
                        continue
                    totalExtractions += 1
                    if extraction in examples:
                        knownmatches += 1
                    else:
                        if not extraction in newExtractions:
                            newExtractions.add(extraction)

    logger.debug("%s known matches out of %s total matches." % (knownmatches, totalExtractions))
    logger.debug("New values: %s" % list(newExtractions))

    removeBadRules(rules, badrules)

    return newExtractions
    

def _generateEventRules(events, event, extractions, ruletype):
    rules = []
    for extraction in extractions:
        start = 0
        while True:
            start = event.find(extraction, start)
            if start < 0:
                break
            rule = PositionalRule(events, extraction, start, event, extractions, ruletype)
            if not rule._bad:
                #print "RULE: %s \t\tTYPE: %s\t\t%s" % (rule, ruletype, event)
                rules.append(rule)
            else:
                pass #print "BAD: %s" % ruletype
            start += 1
    return rules

def needsEsc(ch):
    return ".()[]{}*+^$!-\?|".find(ch) >= 0

def safeRegexLiteral(literal):
    safe = ""
    for ch in literal:
        if needsEsc(ch):
            ch = "\\" + ch
        safe += ch
    return safe

def getBestSkipPunc(text, isPrefix):
    punct = GUIDEPOST_CHARACTERS

    bestpos = -1
    bestscore = 999999999
    l = len(text) 
    for p in punct:
        thiscount = text.count(p)
        if thiscount == 0 or thiscount > MAXMATCHCOUNT:
            continue
        if isPrefix:
            thispos = text.rfind(p)
        else:
            thispos = text.find(p)
        if thispos >= 0:
            if isPrefix:
                dist = l - thispos
            else:
                dist = thispos+1
            if dist > MAXDIST:
                continue
            thisscore = dist * thiscount * thiscount
            #print "PUNCT: %s SCORE: %s DIST: %s COUNT %s" % (p, thisscore, dist, text.count(p))
            if thisscore < bestscore:
                bestscore = thisscore
                bestpos = thispos
    #print "WINNER: %s" % text[bestpos]
    return bestpos

def generateSimpleRegex(text, isPrefix, loosenCount = False):
    regex = ""
    inf = 99999999999.0 # infinity not supported in windows.  # float('inf')

    bestpos = getBestSkipPunc(text, isPrefix)
    
    if bestpos >= 0:
	ch = text[bestpos]
	count = text.count(ch)
	chStr = ""
	if ch == '\t':
	    chStr = "\\t"
	elif needsEsc(ch):
	    chStr = "\\" + ch
	else: 
	    chStr = ch
        countstr = ""
        if count > 1:
            if loosenCount:
                countstr = "+"
            else:
                countstr = "{%s}" % str(count)
        if isPrefix:
            if countstr == "":
                regex = "[^%s]*%s" % (chStr, chStr)
            else:
                regex = "(?:[^%s]*%s)%s" % (chStr, chStr, countstr)
        else:
            if countstr == "":
                regex = "%s" % chStr
            else:
                regex = "(?:%s[^%s]*)%s" % (chStr, chStr, countstr)

        if isPrefix:
            afterregex = splitText(text[bestpos+1:])
            regex = regex + afterregex
        else:
            beforeregex = splitText(text[:bestpos])
            regex = beforeregex + regex
    else:
        regex = splitText(text)
    return regex

def generateSearchRegex(text):
    regex = ""
    inf = 99999999999.0 # infinity not supported in windows.  # float('inf')

    bestpos = getBestSkipPunc(text, True)
    if bestpos >= 0:
	ch = text[bestpos]
	count = text.count(ch)
	chStr = ""
	if ch == '\t':
	    chStr = "\\t"
	elif needsEsc(ch):
	    chStr = "\\" + ch
	else: 
	    chStr = ch
        regex = "%s" % chStr
        if count > 1:
            regex = "(?:[^%s]*%s)%s" % (chStr, chStr, "{%s}" % str(count))

        afterregex = splitText(text[bestpos+1:])
        regex = regex + afterregex
    else:
        regex = splitText(text)
    regex = regex.replace(' ', '\\s')
    return regex


def fixHex(text):
    return re.sub("(?<![0-9a-z])(?:(?:(?:\d+[a-f]+)|(?:[a-f]+\d+))[0-9a-f]*)(?![0-9a-z])", "HEX", text)


def fixIdentifiers(types):
    oldtypes = None
    while oldtypes != types:
        oldtypes = types
        types = types.replace("[a-z]+\\d+[a-z]+", "[a-z]\\w+")
        types = types.replace("[a-z]+\\d+[a-z]", "[a-z]\\w+")
        types = types.replace("\\d+[a-z]+\\d+", "\\d+\\w+")
        types = types.replace("\\w+\\d+", "\\w+")
        types = types.replace("\\w+[a-z]+", "\\w+")
        types = types.replace("\\w+[a-z]", "\\w+")
        types = types.replace("\\w+\\w+", "\\w+")
        types = types.replace("\\w+\-\\w+", "[a-z_-]+")
        types = types.replace("\\w+_\\w+", "[a-z_-]+")
        types = types.replace("\\w+_", "\\w+")
        types = types.replace("_\\w+", "\\w+")
        types = types.replace("\\s+\-\\d", "\\s+[-+]\\d")        
        types = types.replace("\\s+\+\\d", "\\s+[-+]\\d")        
        # print "types: \n\t%s\n\t%s" % (oldtypes, types)
    return types
    
    
def splitText(text):
    types = []
    token = ""
    lastchtype = None
    lastallowplus = False
    text = text.lower()
    text = fixHex(text)
    for ch in text:
        chtype = ch
        allowplus = False
        if ch.isalpha():
            if ch.isupper():
                chtype = '[a-f0-9]'
            else:
                chtype = '\\w' 
            allowplus = True
        elif ch.isdigit():
            chtype = "\d"
            allowplus = True
        elif ch == '\t':
            chtype = "\\t"
        elif ch.isspace():
            chtype = "\s"
            allowplus = True          
        elif needsEsc(ch):
            chtype = "\\" + chtype # + "+"
        if lastchtype != None and chtype != lastchtype:
            if lastallowplus:
                lastchtype += "+"
            types.append(lastchtype)
            token = ""
        if chtype == lastchtype and not allowplus:
            types.append(lastchtype)
        token += ch
        lastchtype = chtype
        lastallowplus = allowplus

    if lastchtype != None:
        types.append(lastchtype)
    if lastallowplus:
        types.append("+")
    types = "".join(types)
    
    return types

def safeDelAttr(obj, attr):
    if getattr(obj, attr, False):
        delattr(obj, attr)

class PositionalRule:

    def __init__(self, filetype, extraction, startpos, event, examples, ruletype):

        self._examplesCount = {}
        self._filetype = filetype
        self._ruletype = ruletype
        self._bad = False        
        self.generatePattern(event, extraction, startpos)
        self.setExamples(examples)
        self._learnedExtractionsCount = {}
        self._score = None
        self._matchCount = 0
        self._bias = 1.0
        if ruletype == "forward-literal":
            self._bias = 1.3 # bonus for literal rule
        elif ruletype.startswith('forward-delimiter'):
            self._bias = 0.7 # penalty for less reliable rule
        elif ruletype.startswith('backward'):
            self._bias = 0.7 # penalty for less reliable rule
        
    def __str__(self):
        return "type: %s regex: %s" % (self._ruletype, str("".join(self._wholePattern)))
    def __hash__(self):
        return hash(str(self))
    def incMatchCount(self):
        self._matchCount += 1
    def getMatchCount(self):
        return self._matchCount

    def getScore(self):
        if self._score == None:
            self._score = self.calcScore()
        return self._score
    
    def calcScore(self):

        exampleVarietyPerc = float(sum([ 1 for v in self.getExamplesCount().values() if v > 0]))  / len(self.getExamplesCount())# number of examples this rule extracts
        learnedCount = len(self.getLearnedCount())                # learned more terms
        exampleCount = sum(self.getExamplesCount().values())      # number of examples matched
        regexSize = len(self.getWholePattern())                   # approximate measure of regex complexity
        center = 7
        goodCount = float(max(1, center - abs(learnedCount-center)))
        
        # score = (1. + learnedCount) * exampleCount / regexSize # favor rules that learned more terms, matched more examples, and are smaller regex
        # quality = exampleCount / (learnedCount + 1.)
        #score = quality * exampleCount / regexSize # favor rules that learned more terms, matched more examples, and are smaller regex

        score = (10000.0*exampleVarietyPerc) + (200.0*goodCount) + (200.0/regexSize) + 10*exampleCount
        score *= self._bias
        # print "%s bias: %s exampleVarietyPerc: %s learnedCount: %s exampleCount: %s regexsize: %s goodCount: %s rule: %s" % (int(score), self._bias, exampleVarietyPerc, learnedCount, exampleCount, regexSize, goodCount, self)
        # logger.error("%s bias: %s exampleVarietyPerc: %s learnedCount: %s exampleCount: %s regexsize: %s goodCount: %s rule: %s" % (int(score), self._bias, exampleVarietyPerc, learnedCount, exampleCount, regexSize, goodCount, self)        )
        # logger.debug("%s RULE: %s " % (score, str(self)))
        return score
    
    def getBad(self):
        return self._bad


    #RULE_TYPES = ["forward", "backward", "forward-prefix", "forward-delimiter", "forward-delimiter-simple"]


    def generatePattern(self, event, extraction, startpos):
        end = startpos + len(extraction)
        prefix = event[0:startpos]
        suffix = event[end:]
        #print "EXTRACTION:", extraction
        #print "OLD PREFIX: %s" % prefix[0:500]
        #print "OLD SUFFIX: %s" % suffix[0:500]
        SOL = prefix.rfind('\n')
        EOL = suffix.find('\n')
        multiline = ""
        if SOL >= 0 or EOL >= 0:
        #if event.find('\n') >= 0:
            multiline = "m"
            if SOL >= 0: prefix = prefix[SOL+1:]
            if EOL >= 0: suffix = suffix[:EOL]
            # print "NEW PREFIX: %s" % prefix
            # print "NEW SUFFIX: %s" % suffix        
        
        good = False
        ruletype = self._ruletype
        if ruletype == "forward":
            good = self.generateForwardRule(event, extraction, multiline, end, prefix, suffix)
        elif ruletype == "backward":
            good = self.generateBackwardRule(event, extraction, multiline, end, prefix, suffix)
        elif ruletype == "forward-literal":
            good = self.generateForwardLiteralRule(event, extraction, multiline, end, prefix, suffix)
        elif ruletype == "forward-delimiter":
            good = self.generateForwardDelimiterRule(event, extraction, multiline, end, prefix, suffix)
        elif ruletype == "forward-delimiter-simple":
            good = self.generateForwardDelimiterSimpleRule(event, extraction, multiline, end, prefix, suffix)
        else:
            raise  Exception("Unknown ruletype: %s" % ruletype)

        self._bad = not good

        
    def generateForwardDelimiterSimpleRule(self, event, extraction, multiline, end, prefix, suffix):
        return self.generateForwardDelimiterRule(event, extraction, multiline, end, prefix, suffix, True)

    def generateForwardDelimiterRule(self, event, extraction, multiline, end, prefix, suffix, simpleValue=False):

        if suffix == '' or suffix == '\n': suffix = "$" 
        if prefix == '': prefix = '^'
        suffix = suffix[0]
        prefix = prefix[-1]
        if suffix in extraction or prefix in extraction:
            return False

        suffixVal = safeRegexLiteral(suffix).replace("\$", '')
        prefixVal = safeRegexLiteral(prefix).replace('\^', '')

##         if simpleValue:
##             valueRegex = "[^%s%s]*" % (prefixVal, suffixVal)
##         else:
##             valueRegex = fixIdentifiers(splitText(extraction))

        valueRegex, oppositeRegex = self.getValueRegex(extraction, True, suffix)
            
        pattern = "(?i)%s(?P<FIELDNAME>%s)%s" % (prefixVal, valueRegex, suffixVal)
        self._wholePattern = pattern
        self._regex = re.compile(pattern)
        return True


    def generateForwardRule(self, event, extraction, multiline, end, prefix, suffix):
        oppositeChar = ""
        if suffix != "": oppositeChar = suffix[0]
        valueRegex, oppositeRegex = self.getValueRegex(extraction, True, suffix)
        if oppositeRegex == "": oppositeRegex = self.simpleOppositeRegex(oppositeChar, extraction)
        prefixRegex = self.getPrefixRegex(prefix, False)
        if prefixRegex == None:
            return False
        pattern = "(?i%s)^%s(?P<FIELDNAME>%s)%s" % (multiline, prefixRegex, valueRegex, oppositeRegex)
        self.cleanupPattern(pattern)
        return True
        
    def generateForwardLiteralRule(self, event, extraction, multiline, end, prefix, suffix):
        oppositeChar = ""
        if suffix != "": oppositeChar = suffix[0]
        valueRegex, oppositeRegex = self.getValueRegex(extraction, True, suffix)
        if oppositeRegex == "": oppositeRegex = self.simpleOppositeRegex(oppositeChar, extraction)        
        prefixRegex = self.getPrefixRegex(prefix, True)        
        if prefixRegex == None:
            return False
        pattern = "(?i)%s(?P<FIELDNAME>%s)%s" % (prefixRegex, valueRegex, oppositeRegex)
        self.cleanupPattern(pattern)
        return True
        
    def generateBackwardRule(self, event, extraction, multiline, end, prefix, suffix):
        oppositeChar = ""
        if prefix != "": oppositeChar = prefix[-1]        
        valueRegex, oppositeRegex = self.getValueRegex(extraction, False, suffix)
        suffixRegex = generateSimpleRegex(suffix, False)
        #print "SUFFIX:", suffix, "REGEX:", suffixRegex
        pattern = "(?i)%s(?P<FIELDNAME>%s)%s" % (oppositeRegex, valueRegex, suffixRegex)
        self.cleanupPattern(pattern)
        return True
    
    def getValueRegex(self, extraction, forward, suffix, max_value_regex=0): #MAX_VALUE_REGEX):

        # make value extractions
        valueRegex = splitText(extraction)

        oppositeRegex = "" 
        simplifiedValueRegex = False

        # if forward rule, and there is not suffix (we're at the end
        # of the line), the value is just everything (.*)
        if forward and suffix == "":
            valueRegex = ".*"
            simplifiedValueRegex = True                    
        # if we have a suffix and the value regex is too complicated (too long)
        # try to simplify the value regex if it can be via the suffix
        elif suffix != "" and len(valueRegex) > max_value_regex:
            endchar = suffix[0]
            # interesting suffix
            if endchar in VALUEPOST_CHARACTERS:
                # and  not in the extraction or is in extraction, but 
                if not endchar in extraction:
                    simplifiedValueRegex = True
                    valueRegex = "[^%s]*" % safeRegexLiteral(endchar)                    
                    oppositeRegex = ""
                    
            # if we haven't yet simplified the value regex and we are making a forward regex
            if not simplifiedValueRegex and forward:
                # see if start of suffix isn't in value regex, do non-greed match
                for suffixTasteLen in range(1,len(suffix)):
                    tasteOfSuffix = suffix[0:suffixTasteLen]
                    suffixTasteRegex = splitText(tasteOfSuffix)
                    # if new 'simple' suffix match more than 50% of
                    # the valueregex's complexity, forget it.  it's
                    # less effecient in terms of backtracking to not
                    # be worth the elegance of a shorter regex, unless
                    # it's really short.
                    if len(suffixTasteRegex) * 2 > len(valueRegex):
                        break
                    if suffixTasteRegex not in valueRegex:
                        oppositeRegex = suffixTasteRegex
                        valueRegex = ".*?"
                        break
                    
        return valueRegex, oppositeRegex

    def simpleOppositeRegex(self, oppositeChar, extraction):
        oppositeRegex = ""
        # if oppositeChar starts with well known punct/space char and that char is not in extraction use it as the oppositeText pattern
        if oppositeChar in VALUEPOST_CHARACTERS and not oppositeChar in extraction:
            oppositeRegex = oppositeChar
            if needsEsc(oppositeChar):
                oppositeRegex = "\\" + oppositeChar
            if oppositeRegex == ' ':
                oppositeRegex = "[ ]"  # too ugly and not worth it.  SPL-11254.  oppositeRegex = "[ ]" # conf file strips trailing space
        return oppositeRegex

    def getPrefixRegex(self, prefix, literal=False):
        if literal:
            return self.getLiteralPrefixRegex(prefix)
        else:
            return generateSimpleRegex(prefix, True)            

    def getLiteralPrefixRegex(self, text):
        # look for meaty term before match
        l = len(text)
        foundmeat = False
        end = -1
        for pos in range(l-1,-1,-1):
            punct = text[pos] in VALUEPOST_CHARACTERS
            if not punct:
                # if prefix meat is a number, we have failed
                if text[pos].isdigit():
                    break
                if not foundmeat:
                    end = pos
                foundmeat = True
            if foundmeat and punct:
                break
        if foundmeat:
            meatword = text[pos+1:end+1]
            # if known bad word
            if meatword.lower() in BAD_MEAT:
                return None
            else:
                meat = safeRegexLiteral(text[pos:])
                text = "" #text[:pos]
        else:
            return None
        return generateSimpleRegex(text, True) + meat
        
    def cleanupPattern(self, pattern):
        pattern = fixIdentifiers(pattern)
        self._wholePattern = pattern
        self._regex = re.compile(pattern)


    def getWholePattern(self):
        return self._wholePattern

    def getRE(self):
        return self._regex
    
    def setExamples(self, examples):
        self._examplesCount = {}
        for ke in examples:
            self._examplesCount[ke] = 0
            
    def getExamples(self):
        return self._examplesCount.keys()

    def getExamplesCount(self):
        return self._examplesCount

    def getLearnedCount(self):
        return self._learnedExtractionsCount
    
    def addExtraction(self, extraction):
        if extraction in self._examplesCount:
            self._examplesCount[extraction] += 1
        elif extraction in self._learnedExtractionsCount:
            self._learnedExtractionsCount[extraction] += 1
        else:
            self._learnedExtractionsCount[extraction] = 1


    def findExtractions(self, event):
        return self.getRE().findall(event)


