import re, sys, os
import splunk.searchhelp.describer as describer
import splunk.searchhelp.parser as parser
import splunk.searchhelp.utils as utils
import splunk.mining.dcutils as dcutils
from xml.sax import saxutils

HOME_PATH = 'wiki'
DIR_PATH = os.path.join(HOME_PATH, 'commands')
OUTPUT_PATH = os.path.join(DIR_PATH, '%s.wiki')

DATATYPES_PAGE = 'Documenation:SearchReference:Datatypes'

HTMLFORM = """
	<h1>%s</h1>
	<p>%s</p>
	<h3>Aliases:</h3>
	<p>%s</p>
	<h3>Related:</h3>
	<p>%s</p>
	<h3>Details:</h3>
	<p>%s</p>
	<h3>Syntax:</h3>
	<p>%s</p>
	<h3>Examples:</h3>
	<dl>
%s	</dl>
"""

OLDHTMLFORM = """
  <table border=1>
    <tr><td colspan=2<b><font size="+2">%s</font></b>%s%s</td></tr>		
    <tr><td><b>Details:</b></td><td>%s</td></tr>		
    <tr><td><b>Syntax:</b></td><td>%s</td></tr>		
    <tr><td><b>Examples:</b></td><td>%s<br/>
    </td></tr>		
  </table>
  <br/><br/>
"""

WIKIFORMARGUMENTNODATATYPE = """
; %(arg)s
: '''Datatype:''' %(datatype)s
: '''Description:''' %(description)s
"""

WIKIFORMARGUMENT = """
; %(arg)s
: '''Datatype:''' OPENL7L7""" + DATATYPES_PAGE + """#%(datatype)s | %(datatype)sCLOSEL7L7
: '''Description:''' %(description)s
"""

WIKIFORMNOARGUMENT = """
: '''Datatype:''' OPENL7L7""" + DATATYPES_PAGE + """#%(datatype)s | %(datatype)sCLOSEL7L7
: '''Description:''' %(description)s
"""


WIKIFORMDATATYPE = """
: '''Datatype:''' <%(datatype)s>
"""



WIKISUBSYNTAX = """
; OPENL7L7""" + DATATYPES_PAGE + """#%(arg)s | %(arg)sCLOSEL7L7
: '''Description:''' %(description)s
: '''Syntax:''' %(syntax)s
"""

WIKIFORMATEXAMPLE = """
'''Example %(count)s:''' %(comment)s
<search>%(example)s</search>
"""

WIKIFORMATALSOSEE = "[[Documentation:SearchReference:%(prettyname)s|%(name)s]] "

WIKI_CATEGORIES = ['4.0', '4.0beta', '4.0.1', '4.0.2']
WIKIFORMATCATEGORIES = "[[Category:%s]]"

WIKIFORM = """
= %(name)s =
== Synopsis ==
%(shortdesc)s

=== Syntax ===
%(syntax)s

=== Arguments ===
%(args)s

== Description ==
%(details)s

== Examples ==
%(examples)s

%(alsosee)s

%(categories)s
"""

WIKISUBARGUMENTS = """
===== Secondary Arguments =====
"""

WIKIDATATYPEDETAILS = """
===== Datatype Details =====
"""


########################## CHEATSHEET START

CHEAT_WIKI_FORM = """
= Search cheatsheet =

Here are some examples illustrating some useful things you can do with the search language.  

Learn more about the commands used in these examples by referring to the [[Documentation:user:SearchCommands|search command reference]].

%s
"""

CHEAT_WIKI_CATEGORY = """
%s
%s

{| cellpadding="5" cellspacing="0" border="1" width="100%%"
%s|}

"""

CHEAT_WIKI_ROW = """|-
|width="70%%"| %s
|<code>%s</code> 
"""


def addExample(categories, category, comment, example):
    catCommands = categories.get(category,None)
    if catCommands == None:
        catCommands = categories[category] = []
    example = example.replace("|", "&#124;")
    catCommands.append((comment, example))



EXAMPLE_SUFFIXES = ['',0,1,2,3,4,5]

def wikiCheatsheetCommands(stanzas):

    commands = utils.getAllCommands(stanzas)

    categories = {}
    for command in commands:
        try:
            key = command
            if not key.endswith("-command"):
                key += "-command"
            if key not in stanzas.stanzas:
                # these are aliases
                continue
            stanza = stanzas[key]
            category = stanza.get('category','Misc')
            added = False
            # add all cheat examples
            for i in EXAMPLE_SUFFIXES:
                exattr = "examplecheat%s" % i
                coattr = "commentcheat%s" % i
                if exattr in stanza and coattr in stanza:
                        example = stanza[exattr]
                        comment = stanza[coattr]
                        addExample(categories, category, comment, example)
                        added = True
                        #print "ADDED FOR %s -- %s -- %s " % (command, comment, example)
                        
            # if no valid cheat example, use the first example/command you can find
            if not added:
                for i in EXAMPLE_SUFFIXES:
                    exattr = "example%s" % i
                    coattr = "comment%s" % i
                    if exattr in stanza and coattr in stanza:
                        example = stanza[exattr]
                        comment = stanza[coattr]
                        addExample(categories, category, comment, example)
                        added = True                        
                        break
            if not added:
                print "Skipping cheat for %s command because there is no example with a comment!" % command
        except Exception, e:
            print "Error with command: %s -- %s " % (command, e)
            import traceback
            print traceback.format_exc()

    categoryNames = categories.keys()
    categoryNames.sort()
    catstext = ""
    lastMajorName = None
    for catName in categoryNames:
        items = categories[catName]
        catDescription = "" # no category descrptions yet
        rowtext = ""
        for item in items:
            rowtext += CHEAT_WIKI_ROW % (item[0], item[1])
        if rowtext != "":
            # handle SPL-24975. sophy wants subheadings
            catHeader = "== %s ==" % catName
            if "::" in catName:
                majorName, minorName = catName.split("::", 1)
                if majorName != lastMajorName:
                    catHeader = "== %s ==\n\n=== %s ===" % (majorName, minorName)
                else:
                    catHeader = "=== %s ===" % (minorName)
                lastMajorName = majorName
                
            catstext += CHEAT_WIKI_CATEGORY % (catHeader, catDescription, rowtext)

    catstext = CHEAT_WIKI_FORM % catstext
    filename = os.path.join(HOME_PATH, 'cheatsheet.wiki')
    if not os.path.isdir(HOME_PATH): os.makedirs(HOME_PATH)
    f = file(filename, "w")
    f.write(catstext)
    f.close()

########################## CHEATSHEET END


########################## DATATYPES START

WIKIDATATYPE_FORM = """
= Search datatypes =

Here are datatypes used to define the syntax of the search language.

Learn more about the commands used in these examples by referring to the [[Documentation:user:SearchCommands|search command reference]].

%s
"""

WIKIDATATYPE_STANZA = """

=== %(name)s ===
: '''Syntax:''' %(syntax)s
: '''Description:''' %(description)s"""
WIKIDATATYPE_EXAMPLE = """
: '''Example:''' %s"""

        
def linkDataTypes(syntax, stanzas, page=""):

    REGEX = re.compile('<([a-zA-Z0-9_/-]+)(?:(:[a-zA-Z0-9_/-]+))?>')
    while True:
        match = REGEX.search(syntax)
        if match == None: break
        datatype = match.group(1)
        variable = match.group(2)
        
        if datatype not in stanzas:
            if variable == None: # no name to datatype
                syntax = REGEX.sub('&lt;\\1&gt;', syntax, 1)
            else:
                syntax = REGEX.sub('&lt;\\1\\2&gt;', syntax, 1)
        else:
            if variable == None: # no name to datatype
                syntax = REGEX.sub('&lt;[[%s#\\1|\\1]]&gt;' % page, syntax, 1)
            else:
                syntax = REGEX.sub('&lt;[[%s#\\1|\\1]]\\2&gt;' % page, syntax, 1)
    return syntax

            
def wikiDataTypes(stanzas):

    out = ""
    for datatype in stanzas:
        if not datatype.endswith("-command"):
            stanza = stanzas[datatype]
            syntax = stanza.get('simplesyntax', stanza.get('syntax', None))
            if syntax == None: continue
            syntax = linkDataTypes(syntax, stanzas)
            description = stanza.get('description', 'None')
            description = re.sub("\s+", ' ', description)
            examples = ""      
            myoutput = WIKIDATATYPE_STANZA % { 'name':htmlToWiki(datatype), 'syntax':syntax, 'description':htmlToWiki(description) }
            for k,v in stanza.items():
                if k.lower().startswith('example'):
                    myoutput += WIKIDATATYPE_EXAMPLE % v
            out += myoutput

    form_out = WIKIDATATYPE_FORM % out
    filename = os.path.join(HOME_PATH, 'datatypes.wiki')
    if not os.path.isdir(HOME_PATH): os.makedirs(HOME_PATH)
    f = file(filename, "w")
    f.write(form_out)
    f.close()

########################## CHEATSHEET END






########################## LIST OF COMMANDS START
LIST_WIKI_FORM = """
= List of search commands =
'''This is a work-in-progress.'''

The table below lists all search commands with a short description and links to their individual reference pages. For a quick guide with examples for use of these search commands, refer to the [[Search Cheatsheet]].

{| cellpadding="5" cellspacing="0" border="1"
!Command
!Alias(es)
!Description
!See also
%s|
|}


%s

"""

LIST_WIKI_ROW = """|-
|<code>[[Documentation:SearchReference:%(command)s|%(command)s]]</code>
|%(aliases)s
|%(shortdesc)s
|<code>%(related)s</code>
"""

# normalize spaces and comma separated lists into ", " separated
def prettyListValue(stanza, attr):
    if attr not in stanza: return ""
    values = re.split("[, ]+", stanza[attr])
    return ", ".join(values)

def wikiListCommands(stanzas):

    commands = utils.getAllCommands(stanzas)
    commands.sort()
    outtext = ""
    for command in commands:
        try:
            key = command
            if not key.endswith("-command"):
                key += "-command"
            if key not in stanzas.stanzas: # these are aliases
                continue
            stanza = stanzas[key]
            aliases = prettyListValue(stanza, 'alias')
            related = prettyListValue(stanza, 'related')
            if 'shortdesc' not in stanza:
                shortdesc = stanza['description']
                if len(shortdesc) > 100:
                    shortdesc = shortdesc[:100] + "..."
            else:
                shortdesc = stanza['shortdesc']

            outtext += LIST_WIKI_ROW % ({'command':command, 'shortdesc':shortdesc, 'aliases':aliases, 'related':related}) 
        except Exception, e:
            print "Error with command: %s -- %s " % (command, e)
            import traceback
            print traceback.format_exc()

    categoriestext = ""
    for category in WIKI_CATEGORIES: 
        categoriestext += WIKIFORMATCATEGORIES % category

    outtext = LIST_WIKI_FORM % (outtext, categoriestext)
        
    filename = os.path.join(HOME_PATH, 'commands.wiki')
    if not os.path.isdir(HOME_PATH): os.makedirs(HOME_PATH)
    f = file(filename, "w")
    f.write(outtext)
    f.close()



########################## LIST OF COMMANDS END    




## TODO. replace "\s<no-ws>\s" with ""
## replace quotes with literals. e.g. "max=" -> <em>max</eb>
    


def htmlToWiki(text):
    text = text.replace("<code>",  "<tt>")
    text = text.replace("</code>", "</tt>")
    text = text.replace("<em>",    "''")
    text = text.replace("</em>",   "''")
    text = text.replace("&", "-=AND=-")
    text = describer.escapeAllBut(text, ['tt','b', 'search', 'br'] )
    text = text.replace("-=AND=-", "&")    
    # write [[val]] in <nowiki> to they aren't taken as a link
    text = re.sub("(\[\[.*?\]\])", "<nowiki>\\1</nowiki>", text)
    # OPENL7L7blahCLOSEL7L7 -> [[blah]]
    text = re.sub("OPENL7L7(.*?)CLOSEL7L7", "[[\\1]]", text)
    return text


def getDesc(stanzas, keyword, details):
    if len(keyword) == 0:
        return ""
    if describer.isLeafDataType(keyword):
        return ""
    stanza = keyword + "-command"
    if stanza in stanzas:
        return stanzas[stanza]['description']
    else:
        s = getSentence(details, keyword.upper())
        if s != "":
            return s
        s = getSentence(details, "'" + keyword + "'")
        if s != "":
            return s
        else:
            return getSentence(details, keyword)            
        

def getSentence(text, keyword):
        pos = text.find(" " + keyword + " ")
        #print "text pos keyword", text, pos, keyword
        
        if pos < 0:
            return ""
        start = text.rfind(". ", 0, pos)
        if start < 0:
            start = 0
        else:
            start += 2
        end = text.find(". ", pos)
        if end < 0:
            end = len(text)
        return text[start:end]

UNDEFINED_DATATYPES = set()

def getArguments(stanzas, command, desc):
    global UNDEFINED_DATATYPES

    stanza = stanzas[command + "-command"]
    syntax = describer.cleanSyntax(stanza.get('simplesyntax', stanza.get('syntax', 'No syntax')))
    exp = parser.getExp(syntax)
    syntax = exp.toSimpleRegex(True)        
    out = ""
    details = desc['details']

    datatypes = set()
    REGEX = re.compile('<([a-zA-Z0-9_/-]+)(?::([a-zA-Z0-9_/-]+))?>')                
    matches = REGEX.findall(syntax)
    for match in matches:
        variable = ''
        datatype = match[0]
        if datatype in datatypes:
            continue
        datatypes.add(datatype)
        if len(match[1]) > 0:
            variable = match[1]
        description = 'None'
        dtstanza = stanzas.get(datatype,None)
        if dtstanza != None:
            description = dtstanza.get('shortdesc', dtstanza.get('description',description))
        description = describer.removeWhitespaces(description)
        if description == 'None':
            continue
        form = WIKIFORMARGUMENT        
        if variable == '': form = WIKIFORMNOARGUMENT
        out +=  form % { 'arg': variable, 'datatype': datatype, 'description' : description }

    syntax = linkDataTypes(syntax, stanzas, DATATYPES_PAGE)
    syntax = syntax.replace("[[", "OPENL7L7").replace("]]", "CLOSEL7L7")

    
    return out,syntax


def removeDups(redundantList):
    dedupedList = []
    for item in redundantList:
        if item not in dedupedList:
            dedupedList.append(item)
    return dedupedList


def cleanText(text):
    return saxutils.escape(describer.removeWhitespaces(text))


SEP = re.compile("[, ]+")
def wikiDescribeCommands(stanzas, thiscommand=None, printPrivate=False, usesimplesyntax = True, toFile=False):

    commands = utils.getAllCommands(stanzas)

    wikinewcommandsfilename = os.path.join(DIR_PATH, 'newcommands.txt')
    if not os.path.isdir(DIR_PATH): os.makedirs(DIR_PATH)
    if os.path.exists(wikinewcommandsfilename): os.remove(wikinewcommandsfilename)

    for command in commands:
        if thiscommand != None and command != thiscommand:
            continue
        try:
            print command
            desc = describer.describeCommand(stanzas, command, usesimplesyntax)
            if desc == None:
                #print "NONE FOR %s" % command
                continue
            exampletext = ""
            i = 1
            for example,comment in desc.get('examples',[]):
                exampletext += WIKIFORMATEXAMPLE % { 'count': i, 'comment': cleanText(comment), 'example': cleanText(example)}
                i += 1
            relatedtext = ""
            if len(desc['related']) > 0:
                relatedtext = "== See also ==\n"
                for related in SEP.split(desc['related']):
                    relatedtext += WIKIFORMATALSOSEE % { 'prettyname': related.capitalize(), 'name' : related }

            categoriestext = ""
            # TAKEN OUT FOR NOW PER DOCS REQUEST: SEP.split(desc['category'])
            categories = removeDups([command] + WIKI_CATEGORIES) 
            for category in categories:  # should add to searchbnf
                categoriestext += WIKIFORMATCATEGORIES % category

            relatedtext    = relatedtext.replace("[[", "-=SQR=-")
            categoriestext = categoriestext.replace("[[", "-=SQR=-")

            desc['args'], desc['syntax'] = getArguments(stanzas, command, desc)
            desc['examples'] = exampletext
            desc['alsosee'] = relatedtext
            desc['categories'] = categoriestext
            
            out = WIKIFORM % desc
            if toFile:
                out = htmlToWiki(out)
                out = out.replace("-=SQR=-", "[[")
            
                filename = OUTPUT_PATH % command
                if not os.path.exists(filename):
                    f = file(wikinewcommandsfilename, "a")
                    f.write(filename + "\n")
                    f.close()
                    print "!!!!!! new search command file: %s..." % filename
                dcutils.writeText(filename, out)

        except Exception, e:
            print "Error with command: %s -- %s " % (command, e)
            import traceback
            print traceback.format_exc()
            

def describeCommands(stanzas, printPrivate=False, usesimplesyntax = True):
    commands = utils.getAllCommands(stanzas)
    out = ""
    for command in commands:
        try:
            desc = describer.describeCommand(stanzas, command, usesimplesyntax)
            exampletext = ""
            for example,comment in desc['examples']:
                exampletext += '\t<dt>%s</dt>\n\t<dd><code>%s</code></dd>\n' % (cleanText(comment), cleanText(example))
            out += HTMLFORM % ( desc['name'], desc['shortdesc'], desc['aliases'], desc['related'], desc['details'], desc['syntax'], exampletext)
        except Exception, e:
            pass #print "Error with command: %s -- %s " % (command, e)
    return out

    
if __name__ == '__main__':
    stanzas = utils.getStanzas("searchbnf", utils.TEST_SESSION(), utils.TEST_NAMESPACE())
    #print getLiterals(stanzas)
    command = None
    if len(sys.argv) < 2:
        print "Usage: python %s [wiki | command | all]" % sys.argv[0]
    else:
        command = sys.argv[1].strip().lower()
        if command == "wiki":
            print "\nWriting to '%s' subdirectory..." % DIR_PATH
            wikiDescribeCommands(stanzas, None, False, True, True)
            print "\nWriting out cheatsheet..."
            wikiCheatsheetCommands(stanzas)
            print "\nWriting out list of commands..."            
            wikiListCommands(stanzas)
            print "\nWriting out datatypes..."
            wikiDataTypes(stanzas)

            
        elif command == "all":
            print describeCommands(stanzas)
        else:
            print describer.describeCommand(stanzas, command, True)


