import unittest
import json
#SPL-18208
import splunk.auth as auth
from splunk.search.Transformer import *

###########################################################
### Unit tests

g_exit_code = 0

def exitingAssert(cond):
    global g_exit_code
    if not cond:
        print "FAILED ASSERT"
        import traceback
        print traceback.print_stack()
        g_exit_code = 1
        exit(1)

def exitingAssertEquals(cond1, cond2):
    exitingAssert(cond1 == cond2)

        
class TestIntentions(unittest.TestCase):

    _sessionKey = auth.getSessionKey('admin', 'changeme')
    _hostPath   = splunk.mergeHostPath()

    # searches 
    q = {
        'single'   : "search foo bar baz",
        'two'      : "search quux | diff 1 2",
        'audit'    : "search index=_audit login | audit",
        'fields'   : "search access | fields host,source, sourcetype",
        'twiki'    : "search main | top twikiuser",
        'sort1'    : "search error localhost | sort + host",
        'sort2'    : "search error localhost | sort 100 + host,source",
        'quotes': 'search twikiuser="ivan" | diff 1 2',
        'internal' : 'search index="_internal"',
        'paren'    : 'search sourcetype="syslog"'
    }

    def equalDicts(self, d1, d2):
        if len(d1.keys()) != len(d2.keys()): return False
        for k,v in d1.items():
            if d1[k] != d2[k]: return False
        for k,v in d2.items():
            if d1[k] != d2[k]: return False
        return True

    def testQuantTerms(self):
        """ Test adding in a term with a gt or lt sign """
        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        applyIntention(ps, 'addtermgt', {"three":1})
        exitingAssertEquals(str(ps), 'search foo bar baz three>1')

        applyIntention(ps, 'addtermgt', {"three":2})
        exitingAssertEquals(str(ps), 'search foo bar baz three>1 three>2')

        applyIntention(ps, 'addtermlt', {"one":2})
        exitingAssertEquals(str(ps), 'search foo bar baz three>1 three>2 one<2')

        base, decomposed = decomposeSearch(ps)
        exitingAssert( {'name': 'addtermgt', 'arg': {'three': 1}} in decomposed)
        exitingAssert( {'name': 'addtermgt', 'arg': {'three': 2}} in decomposed)
        exitingAssert( {'name': 'addtermlt', 'arg': {'one': 2}} in decomposed)
    
    def testFieldsSerialization(self):
        """ Test serialization of field params """ 
        ps = parseSearch(self.q['twiki'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        exitingAssertEquals( getClauseWithCommand(ps, "search").serialize(), 'search main')
        exitingAssertEquals( getClauseWithCommand(ps, "top").serialize(), 'top twikiuser limit="10"')

    def testFields(self):
        """ Test field transforms""" 
        # positive tests
        tests = [
            ("search * | fields + foo,bar,baz ",
                [{'name': 'setfields', 'arg': {'exclusive': True, 'fields': ['foo','bar','baz'] } } ] ),
            ("search * | fields foo,bar,baz ",
                [{'name': 'setfields', 'arg': {'exclusive': False, 'fields': ['foo','bar','baz'] } } ] ),
            ("search * | fields - foo,bar,baz ",
                [{'name': 'excludefields', 'arg': {'fields': ['foo','bar','baz'] } } ] ),
        ]

        for search, standard in tests:
            ps = parseSearch(search, hostPath=self._hostPath, sessionKey=self._sessionKey)
            (base, decomposed) = decomposeSearch(ps)
            for intent in decomposed:
                if intent['name'] in ['setfields', 'excludefields']:
                    for std in standard:
                        if intent['name'] == std['name']:
                            exitingAssert( self.equalDicts(intent['arg'], std['arg']) )

        # negative tests
            nTests = [
                "search * | fields -host",
                "search * | fields host-",
                "search * | fields +host",
                "search * | fields host+",
                "search * | fields +host,cat,fail",
                "search * | fields ,,,",
            ]
            for search in nTests:
                ps = parseSearch(search, hostPath=self._hostPath, sessionKey=self._sessionKey)
                #print "\n\n\nSEARCH:", search
                (base, decomposed) = decomposeSearch(ps)
                for intent in decomposed: 
                    if intent['name'] in ['setfields', 'excludefields']:
                        print "\n\tThis should fail %s -- %s" % (search, intent)
                    exitingAssert( intent['name'] not in ['setfields', 'excludefields'] )

    def testDecomposeSearch(self):
        """ Test untransforms """

        # test audit decomposition
        auditDecomposedStandard = [
            {'name': 'audit'},   # make the search audited
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'login'}, # add the login term
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'index': '_audit'} },
        ]
        auditBaseSearch = "search *"

        ps = parseSearch(self.q['audit'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        (base, decomposed) = decomposeSearch(ps)
        for i in auditDecomposedStandard:
            exitingAssert(i in decomposed)
        exitingAssertEquals( len(decomposed), len(auditDecomposedStandard) )
        exitingAssertEquals( str(base), auditBaseSearch )

        # test sort decomposition
        sort1DecomposedStandard = [
            {'name': 'sort', 'arg': { "ascending": True, "fields":['host']} },
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'localhost'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'error'},
        ]
        sort1BaseSearch = "search *"

        ps = parseSearch(self.q['sort1'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        (base, decomposed) = decomposeSearch(ps)

        for i in decomposed:
            exitingAssert(i in sort1DecomposedStandard)
        exitingAssertEquals( len(decomposed), len(sort1DecomposedStandard) )
        exitingAssertEquals( str(base), sort1BaseSearch )

        # test Not term decomposition
        notTermDecomposedStandard = [
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'x'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'NOT y'},
        ]
        notTermBaseSearch = "search *"

        ps = parseSearch("search x not y", hostPath=self._hostPath, sessionKey=self._sessionKey)
        (base, decomposed) = decomposeSearch(ps)

        for i in decomposed:
            exitingAssert(i in notTermDecomposedStandard)
        exitingAssertEquals( len(decomposed), len(notTermDecomposedStandard) )
        exitingAssertEquals( str(base), notTermBaseSearch )

        # test decomposition with 'sourcetype'
        ps = parseSearch('search sourcetype="twiki"', hostPath=self._hostPath, sessionKey=self._sessionKey)
        (base, decomposed) = decomposeSearch(ps)
        for intent in decomposed:
            if intent['arg'].has_key('sourcetype'):
                exitingAssertEquals( intent['arg']['sourcetype'], 'twiki')

        # make sure time terms are not expanded by the parser endpoint
        ps = parseSearch("search index=_internal startmonthsago=2 | top limit=5 group ", hostPath=self._hostPath, sessionKey=self._sessionKey)
        (base, decomposed) = decomposeSearch(ps)

        #noTimeExpansion = [
        #    {'name': 'plot',
        #     'arg': {
        #         'fields': ['group'],
        #         'normalizeOutliers': False,
        #         'suppressNull': False,
        #         'limit': '5',
        #         'mode': 'top',
        #         'suppressOther': False}},
        #    {'name': 'starttimeterm', 'arg': {'unit': 'months', 'quantity': '2'}},
        #    {'name': 'addterm', 'arg': {'index': '_internal'}},
        #]
        #for i in decomposed:
        #    exitingAssert(i in noTimeExpansion)
        #exitingAssertEquals( len(decomposed), len(noTimeExpansion) )
        #exitingAssertEquals( str(base), 'search *')


    def testJsonString(self):
        """ Test quoting of JSONified intention output """

        jsonStandard = '[{"flags": ["ESCAPED", "indexed"], "name": "addterm", "arg": {"twikiuser": "ivan"}}]'
        ps = parseSearch(self.q['quotes'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        (base, decomposed) = decomposeSearch(ps)
        exitingAssertEquals(json.dumps(decomposed), jsonStandard)

    def testSort(self):
        """ Test sorting """
        sortOptions = {
            "ascending":    False,
            "fields":       ['host', 'source'],
            "maxresults":   500
        }

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        transformed = applyIntention(ps, 'sort', sortOptions)
        exitingAssertEquals( transformed.clauses[-1].args['search'], "500 - host,source")

        del sortOptions['maxresults']

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        transformed = applyIntention(ps, 'sort', sortOptions)
        exitingAssertEquals( transformed.clauses[-1].args['search'], "- host,source")

    def testSimpleTerms(self):
        """ Test the adding and removing of single bare terms from a simple search """

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        # does the transformed search contain the new term?
        transformed = applyIntention(ps, 'addterm', 'quux')
        exitingAssert( transformed.serialize().find("quux") > -1 )
        exitingAssert( transformed.clauses[0].args['search'].find("quux") > -1 )

        transformed = applyIntention(ps, 'addterm', {"user":"ivan"})
        exitingAssert( transformed.serialize().find('user="ivan"') > -1 )
        exitingAssert( transformed.clauses[0].args['search'].find('user="ivan"') != -1 )

        transformed = applyIntention(ps, 'addterm', {"index":"_internal"})
        exitingAssert( transformed.serialize().find('index="_internal"') > -1 )

        transformed = applyIntention(ps, 'addterm', {"userid":"6"})
        exitingAssert( transformed.serialize().find('userid="6"') > -1 )

        transformed = applyIntention(ps, 'addterm', "( twikiaction=save OR twikiaction=edit )")
        exitingAssert( transformed.serialize().find("( twikiaction=save OR twikiaction=edit )") > -1 )

        # now let's remove terms
        transformed = applyIntention(ps, 'removeterm', 'foo')
        exitingAssert( transformed.serialize().find("foo") == -1 )
        transformed = applyIntention(ps, 'removeterm', "userid=6")
        exitingAssert( transformed.serialize().find('userid=6') == -1 )

        # toggle a term
        transformed = applyIntention(ps, 'toggleterm', 'togterm')
        exitingAssert( transformed.serialize().find("togterm") > -1 )
        transformed = applyIntention(ps, 'toggleterm', 'togterm')
        exitingAssert( transformed.serialize().find("togterm") == -1 )

        # term toggling is insensitive to quotes in kv-pairs
        transformed = applyIntention(ps, 'toggleterm', 'k="v"')
        transformed = applyIntention(ps, 'toggleterm', 'k=v')
        exitingAssert( transformed.serialize().find('"k=v"'))
        exitingAssert( transformed.serialize().find('k=\"v\"'))

        # test toggling terms that used to expand
        ps = parseSearch(self.q['paren'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        transformed = applyIntention(ps, 'toggleterm', 'sourcetype="syslog"')
        exitingAssertEquals( str(transformed), 'search *')

        # make sure the '*' is removed 
        ps = parseSearch('search *', hostPath=self._hostPath, sessionKey=self._sessionKey)
        transformed = applyIntention(ps, 'addterm', 'quux')
        exitingAssertEquals( transformed.clauses[0].args['search'], 'quux' )

        # if the user had a '*' in there, we should leave it alone
        ps = parseSearch('search * foo bar', hostPath=self._hostPath, sessionKey=self._sessionKey)
        transformed = applyIntention(ps, 'addterm', 'quux')
        exitingAssertEquals( transformed.clauses[0].args['search'], '* foo bar quux' )


    def testUnicodeTerms(self):
        """ Test the adding and removing of unicode terms from a simple search """

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        # try adding a unicode term
        transformed = applyIntention(ps, 'addterm', u'unicode')
        exitingAssert( transformed.serialize().find("unicode") > -1 )

        # ... try to remove the unicode one too
        transformed = applyIntention(ps, 'removeterm', u'unicode')
        exitingAssert( transformed.serialize().find("unicode") == -1 )

        # toggle a unicode term
        transformed = applyIntention(ps, 'toggleterm', u'togterm')
        exitingAssert( transformed.serialize().find("togterm") > -1 )
        transformed = applyIntention(ps, 'toggleterm', u'togterm')
        exitingAssert( transformed.serialize().find("togterm") == -1 )


    def testSimpleTermsInMultipleClauses(self):
        """ Test the adding and removing of single bare terms from a search with multiple clauses """

        ps = parseSearch(self.q['two'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        # does the transformed search contain the new term?
        transformed = applyIntention(ps, 'addterm', 'foo')
        exitingAssert( transformed.serialize().find("foo") > -1 )
        exitingAssert( transformed.clauses[2].args['search'].find("foo") > -1 )

        # now let's remove a term
        transformed = applyIntention(ps, 'removeterm', 'quux')
        exitingAssert( transformed.serialize().find("quux") == -1 )
        exitingAssert( transformed.clauses[0].args['search'].find("quux") == -1 )

    def testTermNegation(self):
        """ Test the negation of terms from a search """

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        # does the transformed search contain the new term?
        transformed = applyIntention(ps, 'negateterm', 'foo')
        exitingAssert( transformed.clauses[0].serialize().find("NOT foo") > -1 )

        transformed = applyIntention(ps, 'negateterm', {'origin':'example.org'} )
        exitingAssert( transformed.clauses[0].serialize().find('NOT origin="example.org"') > -1 )


    def testQuotedTerms(self):
        """ Test the adding and removing of single quoted terms from a search """

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        # does the transformed search contain the new term?
        transformed = applyIntention(ps, 'addterm', '"quux cruft"')
        exitingAssert( transformed.serialize().find('"quux cruft"') > -1 )
        exitingAssert( transformed.clauses[0].args['search'].find('"quux cruft"') > -1 )

        # now let's remove a term
        transformed = applyIntention(ps, 'removeterm', '"cruft quux"')
        exitingAssert( transformed.serialize().find('"cruft quux"') == -1 )
        exitingAssert( transformed.clauses[0].args['search'].find('"cruft quux"') == -1 )

    def testUnknownIntention(self):
        """ Test handling of unknown intentions """

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        # does the transformed search contain the new term?
        transformed = applyIntention(ps, 'noSuchIntention', '"quux cruft"')
        exitingAssertEquals( transformed, ps)

    def testAudit(self):
        """ Test making a search audited """

        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        # does the transformed search contain the new term?
        # TODO Make a util for equivalence testing of searches
        # FIXME i am supposed to fail for now
        transformed = applyIntention(ps, 'audit')
        exitingAssertEquals( transformed.clauses[-1].command, 'audit')
        exitingAssertEquals( transformed.serialize(), 'search foo bar baz | audit')


    def testPlot(self):
        """ Test plotting a chart. """

        # make a count(_raw) chart
        ps = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)

        args = {
            #"statop" :    "count",
            "fields" :    [['count','_raw']],
            "mode"   :    "timechart",
        }
        transformed = applyIntention(ps, 'plot', args)
        exitingAssertEquals( transformed.clauses[-1].command, 'timechart')
        exitingAssertEquals( transformed.clauses[-1].args['search'], 'count(_raw)')

        # make a distinct_count chart
        ps2 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args2 = {
            #"statop" :    "distinct_count",
            "fields" :    [['dc','_raw']],
            "normalizeOutliers" :    "true",
            "mode"   :    "timechart",
        }

        transformed2 = applyIntention(ps2, 'plot', args2)
        exitingAssertEquals( transformed2.clauses[-1].command, 'outlier')
        exitingAssertEquals( transformed2.clauses[-2].command, 'timechart')
        exitingAssertEquals( transformed2.clauses[-2].args['search'], 'dc(_raw)')

        # make a top chart
        ps3 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args3 = {
            "fields" :    ["_raw"],
            "limit" :    "10",
            "mode"   :    "top",
        }

        transformed3 = applyIntention(ps3, 'plot', args3)
        exitingAssertEquals( transformed3.clauses[-1].command, 'top')
        exitingAssertEquals( transformed3.clauses[-1].args['search'], '_raw')
        exitingAssertEquals( transformed3.clauses[-1].args['limit'], '10')

        # make a chart split
        ps4 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args4 = {
            #"statop" :    "count",
            "fields" :    [['count','date_wday']],
            "mode"   :    "chart",
        }

        transformed4 = applyIntention(ps4, 'plot', args4)
        exitingAssertEquals( transformed4.clauses[-1].command, 'chart')
        exitingAssertEquals( transformed4.clauses[-1].args['search'], 'count(date_wday)')

        # make a default chart
        ps5 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args5 = {
            "mode"   :    "chart",
        }

        transformed5 = applyIntention(ps4, 'plot', args5)
        exitingAssertEquals( transformed4.clauses[-1].command, 'timechart')
        exitingAssertEquals( transformed4.clauses[-1].args['search'], 'count')

        # should get an exception about splitting more than one field
        ps6 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args6 = {
            "mode"   :    "timechart",
            "fields" :    [['count','date_wday'], ['count','date_hour']],
            "splitby":    "someField",
        }

        self.assertRaises(splunk.SearchException, applyIntention,ps4, 'plot', args6 )

        # test all recognized timechart pre-op options
        timechartPreOptions = [
            ('bins', 20),
            ('span', '"23h"')
        ]
        for option in timechartPreOptions:
            ps7 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
            args7 = {
                "mode"   :    "timechart",
                "fields" :    [['count','date_wday']]
            }
            args7[option[0]] = option[1]
            transformed7 = applyIntention(ps7, 'plot', args7)
            exitingAssertEquals( str(transformed7), 'search foo bar baz | timechart %s=%s count(date_wday)' %  (option[0], option[1]))

        # test multiple recognized timechart pre-op options
        ps7 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args7 = {
            "mode"   :    "timechart",
            "fields" :    [['count','date_wday']]
        }
        preOpString = []
        for option in timechartPreOptions:
            args7[option[0]] = option[1]
            preOpString.append('%s=%s' % (option[0], option[1]))
        transformed7 = applyIntention(ps7, 'plot', args7)
        exitingAssertEquals( str(transformed7), 'search foo bar baz | timechart %s count(date_wday)' % ' '.join(preOpString))

        # test combo timechart pre-op and post-op options
        ps7 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args7 = {
            "mode"   :    "timechart",
            "fields" :    [['count','date_wday']],
            "span" :    '2d',
            "suppressNull": True
        }
        transformed7 = applyIntention(ps7, 'plot', args7)
        exitingAssertEquals( str(transformed7), 'search foo bar baz | timechart span="2d" count(date_wday) usenull="f"')

        # test combo timechart pre-op and post-op options with special case event count
        ps7 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args7 = {
            "mode"   :    "timechart",
            "fields" :    [['count','__events']],
            "span" :    '2d',
            "suppressNull": True
        }
        transformed7 = applyIntention(ps7, 'plot', args7)
        exitingAssertEquals( str(transformed7), 'search foo bar baz | timechart span="2d" count usenull="f"')


        # make a chart split
        ps8 = parseSearch(self.q['single'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        args8 = {
            "fields" :    [['count','__events']],
            "mode"   :    "timechart",
        }
        transformed8 = applyIntention(ps8, 'plot', args8)
        exitingAssertEquals( str(transformed8), 'search foo bar baz | timechart count')


    def testPlotDecompose(self):
        """ Test decomposing a charting search. """
        testChartingSearch = 'search source="syslog" | rare limit=20 date_second | outlier'

        chartingDecompositions = {
            'search source="syslog" | rare limit=20 date_second | outlier' : \
            [   {'name': 'plot', 'arg': {'mode': 'rare', 'fields': {'clauses': ['date_second']}, 'normalizeOutliers': True, 'limit': '20',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': True}},
                {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | chart avg(height) by age' : \
            [   {'name': 'plot', 'arg': {'splitby': 'age', 'fields': [['avg','height']], 'mode': 'chart',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False}},
                {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart count(martinis)' : \
            [   {'name': 'plot', 'arg': {'fields': [['count','martinis']], 'mode': 'timechart',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False}},
                {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart count(martinis) usenull=f' : \
                [   {'name': 'plot', 'arg': {'fields': [['count','martinis']], 'mode': 'timechart',
                    'suppressOther': False, 'suppressNull': True, 'normalizeOutliers': False}},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart count(martinis) usenull=f | outlier' : \
                [   {'name': 'plot', 'arg': {'fields': [['count','martinis']], 'mode': 'timechart',
                    'suppressOther': False, 'suppressNull': True, 'normalizeOutliers': True}},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search * | timechart dc(_raw) by sourcetype' : \
                [   {'name': 'plot', 'arg': {'fields': [['dc','_raw']], 'mode': 'timechart', 'splitby':'sourcetype',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False}}
                ],
            'search source="syslog" | timechart count(_raw) by twikipage' : \
                [   {'name': 'plot', 'arg': {'mode': 'timechart', 'fields': [['count','_raw']], 'splitby':'twikipage',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False }},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart count by twikipage' : \
                [   {'name': 'plot', 'arg': {'mode': 'timechart', 'fields': [['count','__events']], 'splitby':'twikipage',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False }},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart bins=12 count by twikipage' : \
                [   {'name': 'plot', 'arg': {'mode': 'timechart', 'fields': [['count','__events']], 'splitby':'twikipage',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False, 'bins': 12 }},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart span=1m count by twikipage' : \
                [   {'name': 'plot', 'arg': {'mode': 'timechart', 'fields': [['count','__events']], 'splitby':'twikipage',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False, 'span': '1m' }},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart span=1m bins=60 count by twikipage' : \
                [   {'name': 'plot', 'arg': {'mode': 'timechart', 'fields': [['count','__events']], 'splitby':'twikipage',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False, 'span': '1m', 'bins': 60 }},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search source="syslog" | timechart dc(hello) dc(world)' : \
                [   {'name': 'plot', 'arg': {'mode': 'timechart', 'fields': [['dc', 'hello'], ['dc', 'world']],
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False }},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
            'search index=_internal group=per_index_thruput | timechart per_second(kb)' : \
                [   {'name': 'plot', 'arg': {'suppressOther': False, 'fields': [['per_second', 'kb']],
                    'normalizeOutliers': False, 'mode': 'timechart', 'suppressNull': False}},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg':  {'index': '_internal'}},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'group': 'per_index_thruput'}},
                ],
            'search source="syslog" | chart count by assigneeID,sprint' : \
                [   {'name': 'plot', 'arg': {'mode': 'chart', 'fields': [['count','__events']], 'splitby':'assigneeID,sprint',
                    'suppressOther': False, 'suppressNull': False, 'normalizeOutliers': False }},
                    {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': {'source': 'syslog'}},
                ],
        }

        for search, intents in chartingDecompositions.items():
            chart = parseSearch(search, hostPath=self._hostPath, sessionKey=self._sessionKey)
            (base, decomposed) = decomposeSearch(chart)
            b = str(base)
            #print "BASE JOIN:", b
            #print "GOOD?", b == "search *"
            exitingAssertEquals( str(base), "search *")

            for pos, i in enumerate(intents):
                if not i in decomposed:
                    print "\n\n%s FAILED TO pDECOMPOSE INTENT: \n\ti: %s \n\tsearch: %s \n\tdecomposed: %s\n" % (pos, i, search, decomposed)
                    print "\nDECOMPOSE:", decomposed
                    print "INTENTS:", intents
                    print
                    
                    exitingAssert(i in decomposed)

            # test recomposition
            ps = parseSearch(base, hostPath=self._hostPath, sessionKey=self._sessionKey)
            try:
                for i in intents:
                    ps = applyIntention(ps, i['name'], i['arg'])
                    #print "\tApplied %s - %s and got %s" % (i['name'], i['arg'], str(ps) )
                #print "-----------"
            except Exception, e:
                print "\tTried %s - %s and got %s" % (i['name'], i['arg'], str(ps) )
                print e
                

    def testIndexInternal(self):
        """ Test handling search without a non-KV term (aka, implied *) """

        ps = parseSearch(self.q['internal'], hostPath=self._hostPath, sessionKey=self._sessionKey)
        transformed = applyIntention(ps, 'addterm', "anything" )
        exitingAssertEquals( str(transformed), 'search index="_internal" anything' )

    def testSearchesWithParens(self):
        """ Test handling of searches that have KV pairs in parentheses. """

        ps = parseSearch('search error OR failed OR severe OR ( sourcetype=access_* ( 404 OR 500 OR 503 ) )' , hostPath=self._hostPath, sessionKey=self._sessionKey)
        
        base, decomposed = decomposeSearch(ps)

        standard = [
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'error'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'OR'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'failed'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'OR'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'severe'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': 'OR'},
            {'name': 'addterm', 'flags':['ESCAPED', 'indexed'], 'arg': '( sourcetype=access_* ( 404 OR 500 OR 503 ) )'},
        ]

        #print "***", decomposed
        for i in standard:
            #print "**", i
            exitingAssert(i in decomposed)
        

    def testLackofSearchClause(self):
        """ Test handling of searches that don't have a search clause """

        ps = parseSearch("| metadata type=sources", hostPath=self._hostPath, sessionKey=self._sessionKey)
        exitingAssert( str(ps).startswith('|') )
        base, decomposed = decomposeSearch(ps)
        exitingAssertEquals( decomposed, [])

        # SPL-18847
        ps = parseSearch("| windbag", hostPath=self._hostPath, sessionKey=self._sessionKey)
        exitingAssert( str(ps).startswith('|') )
        ps = parseSearch("|windbag", hostPath=self._hostPath, sessionKey=self._sessionKey)
        exitingAssert( str(ps).startswith('|') )
        ps = parseSearch("             | windbag", hostPath=self._hostPath, sessionKey=self._sessionKey)
        exitingAssert( str(ps).startswith('|') )


    def testTokenReplacement(self):
        """ Test replacement of $tokens$. """

        ps = parseSearch("search $foo$ source=$bar$", hostPath=self._hostPath, sessionKey=self._sessionKey)
        transformed = applyIntention(ps, 'replace', {'target':'foo', 'replacement': 'session' } )
        transformed = applyIntention(transformed, 'replace', {'target':'bar', 'replacement': '/var/log/secure' } )
        exitingAssertEquals( str(transformed), 'search session source=/var/log/secure')

        ps = parseSearch("search $foo$ source=$bar$ | fields $baz$", hostPath=self._hostPath, sessionKey=self._sessionKey)
        transformed = applyIntention(ps, 'replace', {'target':'foo', 'replacement': 'session' } )
        transformed = applyIntention(transformed, 'replace', {'target':'bar', 'replacement': '/var/log/secure' } )
        transformed = applyIntention(transformed, 'replace', {'target':'baz', 'replacement': 'eventtype' } )
        exitingAssertEquals( str(transformed), 'search session source=/var/log/secure | fields eventtype')

    def testCrazyStars(self):
        """ Test to see we aren't getting superfluous stars. Where are these crazy stars coming from? """
        ps = parseSearch("search *CrazyStars" , hostPath=self._hostPath, sessionKey=self._sessionKey)
        (base, decomposed) = decomposeSearch(ps)
        exitingAssertEquals( str(base), "search *")

    def testFieldsRecompose(self):
        """ Test the fields intentions can be recomposed with what we have decomposed out """
        originalSearch = "search sourcetype=*apache* | fields host"
        searchPart = 'search sourcetype=*apache*'
        fieldsParts = ['fields host', 'fields host,source', 'fields + foo', 'fields - bar']

        for fieldsPart in fieldsParts:
            search = "%s | %s" % (searchPart, fieldsPart)
            ps = parseSearch(search, hostPath=self._hostPath, sessionKey=self._sessionKey)
            (base, decomposed) = decomposeSearch(ps)

            ps = parseSearch(base, hostPath=self._hostPath, sessionKey=self._sessionKey)
            for intent in decomposed:
                ps = applyIntention(ps, intent[INAME], intent[IARG])
            exitingAssertEquals( str(getClauseWithCommand(ps, 'fields')), fieldsPart)

    #def testRelTimeTerms(self):
    #        """ Test extraction of relative timeterms  """
    """
        # positive tests
        tests = [
            ("search nick index=_internal earliesttime=-1d@d ",
                [{'name': 'earliesttime', 'arg': {'count': '-1', 'snapUnit': 'd', 'unit': 'd'}} ] ),

            ("search ivan latesttime=-1d@h ",
                [{'name': 'latesttime', 'arg': {'count': '-1', 'snapUnit': 'h', 'unit': 'd'}} ] ),

            ("search ivan latesttime=-1d",
                [{'name': 'latesttime', 'arg': {'count': '-1', 'unit': 'd'}} ] ),

            ("search ivan earliest=@y",
                [{'name': 'earliesttime', 'arg': {'snapUnit': 'y'}} ] ),

            ("search ivan earliest=5mon@y",
                [{'name': 'earliesttime', 'arg': {'count': '5', 'unit': 'mon', 'snapUnit': 'y'}} ] ),

            ("search nick earliesttime=-7d@d latesttime=-1d@d",
                [{'name': 'earliesttime', 'arg': {'count': '-7', 'snapUnit': 'd', 'unit': 'd'}},
                 {'name': 'latesttime',   'arg': {'count': '-1', 'snapUnit': 'd', 'unit': 'd'}} ] ),
        ]

        for search, standard in tests:
            ps = parseSearch(search, hostPath=self._hostPath, sessionKey=self._sessionKey)
            (base, decomposed) = decomposeSearch(ps)
            for intent in decomposed:
                if intent['name'] in ['earliesttime', 'latesttime']:
                    for std in standard:
                        if intent['name'] == std['name']:
                            exitingAssert( self.equalDicts(intent['arg'], std['arg']) )

        # negative tests
            nTests = [
                "search * earliesttime=-1x@d",
                "search * earliesttime=-ad@d",
                "search * earliesttime=-1d@",
                "search * earliesttime=-1d@x",
                "search * earliesttime=112d@",
                "search * earliesttime=112",
                "search * latesttime=-1x@d",
                "search * latesttime=-ad@d",
                "search * latesttime=-1d@",
                "search * latesttime=-1d@x",
                "search * latesttime=112d@",
            ]
            for search in nTests:
                ps = parseSearch(search, hostPath=self._hostPath, sessionKey=self._sessionKey)
                (base, decomposed) = decomposeSearch(ps)
                for intent in decomposed: 
                    if intent['name'] in ['earliesttime', 'latesttime']:
                        print "\n\tThis should fail %s -- %s" % (search, intent)
                    exitingAssert( intent['name'] not in ['earliesttime', 'latesttime'] )

"""
    def NOTWORKINGYETtestStatsEval(self):
        """ ***INCOMPLETE*** Text Stats and Eval intentions (SPL-18324 and SPL-18515) """
        searchString = 'search eventtype=banner-js earliest=-1d@d latest=@d | rex field=uri "/js/(?<t>[^/]*)/(?<v>[^/]*)/login/(?<l>[^/]*)" | bin span=1d _time | stats count dc(clientip) as clientips by _time t v l | eval v=if(v=="", "NULL", v) | eval group="usage"'
        ps = parseSearch(searchString, hostPath=self._hostPath, sessionKey=self._sessionKey)

        exitingAssertEquals( str(ps), searchString )
        (base, decomposed) = decomposeSearch(ps)

    def testAddingSearchAfterStreaming(self):
        """ Do fielded terms correctly append at the end of streaming commands """
        ps = parseSearch('search error | kv access-extractions', hostPath=self._hostPath, sessionKey=self._sessionKey)
        ps = applyIntention(ps, 'addterm', {"userid":"6"})
        ps = applyIntention(ps, 'addterm', 'quux', ['indexed'])
        exitingAssert( str(ps.clauses[0]) == 'search error quux' )

        ps = parseSearch('search error | eval userid=something', hostPath=self._hostPath, sessionKey=self._sessionKey)
        ps = applyIntention(ps, 'addterm', {"userid":"6"})
        print str(ps)
        exitingAssert( str(ps.clauses[0]) == 'search error' )
        exitingAssert( str(ps.clauses[-1]) == 'search userid="6"' )

    def testScratch(self):
        """ ------SCRATCH------- """
        ps = parseSearch('search index="_internal" | chart sum(events) by hello,world', hostPath=self._hostPath, sessionKey=self._sessionKey)
        xf = applyIntention(ps, 'addtermgt', {"userid":"6"} )
        #print str(xf), "\n"
        #print str(xf.clauses[-1])
        #print xf.clauses[-1].args


    def testDrilldown(self):
    
    ##     searches = [
    ##         "stats avg(cpu)    by   host",
    ##         "stats count(user) by   host",
    ##         "chart count(user) over host",
    ##         "chart dc(date_minute) over group by series",
    ##         "timechart dc(date_second)",
    ##         "timechart dc(user) by group",
    ##         "top user,host,date_hour",
    ##         'search foo'
    ##        ]

        searches = [
            # sorkin wacky "stats | chart"
            ('search elvis | eval user = "foo" | stats min(_time) as _time first(user) as user max(total_run_time) as total_run_time first(search) as search by search_id | search search=search* search!=*_internal* search!=*_audit* | chart median(total_run_time) perc95(total_run_time) sum(total_run_time) count by user', { 'vals': [('user', 'admin'),('user', 'admin')]}, 'search elvis | eval user = "foo" | stats min(_time) as _time first(user) as user max(total_run_time) as total_run_time first(search) as search by search_id | search search=search* search!=*_internal* search!=*_audit* user="admin"'),
            
            # events
              # cell
              ('search foo', { 'vals': [('_time','1234567890'),('sourcetype','syslog')]}, 'search foo sourcetype="syslog"'),
              # column
              ('search foo', { 'vals': [('_time','1234567890'),('sourcetype','sourcetype')]}, 'search foo | sort +sourcetype'),
              # first column = row
              ('search foo', { 'vals': [('_time','1234567890'),('_time','1234567890')]}, 'search foo'), # nick handling time # _time="1234567890"'),
    
            # chart case is same as stats
              # cell
              ('search foo | chart count(user) over host', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, 'search foo host="localhost" | top user'),
              ('search foo | chart avg(user)   over host', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, 'search foo host="localhost"'),
              # click on func should be row click
              ('search foo | chart count(user) over host', { 'vals': [('host', 'localhost'),('count(user)', '1')]},  'search foo host="localhost" | top user'),
              ('search foo | chart avg(user)   over host', { 'vals': [('host', 'localhost'),('avg(user)', '1')]},    'search foo host="localhost"'),
            
              # cell in first column = row
              ('search foo | chart count(date_minute) over group by series', { 'vals': [('group','mpool'),('group','mpool')]}, 'search foo group="mpool" | top date_minute'),
              ('search foo | chart avg(date_minute) over group by series', { 'vals': [('group','mpool'),('group','mpool')]}, 'search foo group="mpool"'),
            
              # invalid cell NULL
              ('search foo | chart count(date_minute) over group by series', { 'vals': [('group','mpool'),('NULL','0')]}, None),
              ('search foo | chart avg(date_minute) over group by series', { 'vals': [('group','mpool'),('NULL','0')]}, None),
              # invalid cell OTHER
              ('search foo | chart count(date_minute) over group by series', { 'vals': [('group','mpool'),('OTHER','0')]}, None),
              ('search foo | chart avg(date_minute) over group by series', { 'vals': [('group','mpool'),('OTHER','0')]}, None),
            
              # middle cell
              ('search foo | chart count(date_minute) over group by series', { 'vals': [('group','mpool'),('netstat','60')]}, 'search foo series="netstat" group="mpool" | top date_minute'),
              ('search foo | chart avg(date_minute) over group by series', { 'vals': [('group','mpool'),('netstat','60')]}, 'search foo series="netstat" group="mpool"'),
            
              # column
              ('search foo | chart count(date_minute) over group by series', { 'vals': [('group','mpool'),('netstat','netstat')]}, 'search foo | chart count(date_minute) over group by series | sort +netstat'),
              ('search foo | chart avg(date_minute) over group by series', { 'vals': [('group','mpool'),('netstat','netstat')]}, 'search foo | chart avg(date_minute) over group by series | sort +netstat'),
            
            
            #timechart
    
              ('search foo | timechart count avg(date_minute)', { 'vals': [('_time','1234567890-1234567892'),('_time','1234567890-1234567892')]}, 'search foo'),
              # cell in first column = row
              ('search foo | timechart count(date_second)', { 'vals': [('_time','1234567890-1234567892'),('_time','1234567890-1234567892')]}, 'search foo'),
              ('search foo | timechart avg(date_second)', { 'vals': [('_time','1234567890-1234567892'),('_time','1234567890-1234567892')]}, 'search foo'),
              # cell 
              ('search foo | timechart count(date_second)', { 'vals': [('_time','1234567890-1234567892'),('count(date_second)','57')]}, 'search foo | top date_second'),
              ('search foo | timechart avg(date_second)', { 'vals': [('_time','1234567890-1234567892'),('_time','1234567890-1234567892')]}, 'search foo'),
              # can add test for sorting here
            
            # timechart with by clause
              # cell in first column = row
              ('search foo | timechart count(user) by group', { 'vals': [('_time','1234567890-1234567892'),('_time','1234567890-1234567892')]}, 'search foo | top user'),
              ('search foo | timechart avg(date_second)', { 'vals': [('_time','1234567890-1234567892'),('_time','1234567890-1234567892')]}, 'search foo'),
              # cell 
              ('search foo | timechart count(user) by group', { 'vals': [('_time','1234567890-1234567892'),('per_index_thruput','0')]}, 'search foo group="per_index_thruput" | top user'),
              ('search foo | timechart avg(user) by group', { 'vals': [('_time','1234567890-1234567892'),('per_index_thruput','0')]}, 'search foo group="per_index_thruput"'),
            
            # top
              # cell in first column = row
              ('search foo | top host,group,date_second', { 'vals': [('host','localhost'),('host','localhost')]}, 'search foo host="localhost"'),
              # cell 
              ('search foo | top host,group,date_second', { 'vals': [('host','localhost'),('group','pipeline')]}, 'search foo group="pipeline"'),
              # invalid cells
              ('search foo | top host,group,date_second', { 'vals': [('host','localhost'),('count','24')]}, 'search foo host="localhost"'),
              ('search foo | top host,group,date_second', { 'vals': [('host','localhost'),('percent','24')]}, 'search foo host="localhost"'),
    
            # stats
              # cell
              ('search foo | stats count(user) by host,group', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, 'search foo host="localhost"'),
              ('search foo | stats avg(user)   by host,group', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, 'search foo host="localhost"'),
              # cell/first row
              ('search foo | stats count(user) by host,group', { 'vals': [('host', 'localhost'),('count(user)', '1')]}, 'search foo host="localhost" | top user'),
              ('search foo | stats avg(user)   by host,group', { 'vals': [('host', 'localhost'),('avg(user)', '1')]}, 'search foo host="localhost"'),
              # column (sort)
              ('search foo | stats avg(user)   by host,group', { 'vals': [('host', 'localhost'),('host', 'host')]}, 'search foo | stats avg(user)   by host,group | sort +host'),
              ('search foo | stats avg(user)   by host,group | sort +host', { 'vals': [('host', 'localhost'),('host', 'host')]}, 'search foo | stats avg(user)   by host,group | sort -host'),
              ('search foo | stats avg(user)   by host,group | sort +elvis', { 'vals': [('host', 'localhost'),('host', 'host')]}, 'search foo | stats avg(user)   by host,group | sort +elvis | sort +host'),
              ('search foo | stats avg(user)   by host,group | sort +host | sort +elvis', { 'vals': [('host', 'localhost'),('host', 'host')]}, 'search foo | stats avg(user)   by host,group | sort +elvis | sort -host'),
        ]
    
        xsearches = [
    
            # I should be able to click on the numbers under
            # avg(date_minute) and it should fall back to the row click
            # drilldown, but it doesn't. It errors out.
            ('search * | chart avg(date_minute) by host', { 'vals': [('host', 'localhost'),('avg(date_minute)', '27.55')]}, 'search host="localhost"'),
    
            # I should definitely be able to click on the numbers under
            # dc(twikipage), but this errors as well.
            ('search * | stats dc(date_minute) by host', { 'vals': [('host', 'localhost'),('dc(date_minute)', '5')]}, 'search host="localhost" | top date_minute'),
    
            # perfectly legal syntax but it throws an exception if you
            # click on the 'count' values. Presumably the test cases all
            # have count(foo) but not the canonical count arg, which means
            # 'count of events'
            ('search * | timechart count avg(date_minute)', { 'vals': [('_time', '1234567890-1234567892'),('count', '0')]}, 'search *'), # should set time
    
            # # clicks on the count(_raw) cells end up with searches that
            # # have a " | top _raw" at the end.  Even if we do go with top
            # # on the end, it's probably worth a special case to make sure
            # # we never do top _raw...
            # # ('search * | timechart count(_raw) avg(date_minute)' , { 'vals': [('_time', '1234567890-1234567892'),('count(_raw)', '0')]}, None),
    
            # with args like this xfield=host
            # xval=AaronYoung column=twikiaction val=edit it should fold
            # in both host=AaronYoung and twikiaction=edit but it
            # only puts in twikiaction="edit"
            ('search * | stats avg(date_minute) by host,sourcetype', { 'vals': [('host', 'localhost'),('sourcetype', 'syslog')]}, 'search host="localhost" sourcetype="syslog"'),
            ('search * | stats avg(date_minute) by host,sourcetype', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, 'search host="localhost"'),
    
            # a) clicks on the avg(date_minute) values should fall back to
            # the row click case, but it errors out.
            ('search * | stats avg(date_minute) dc(date_hour) by host', { 'vals': [('host', 'localhost'),('avg(date_minute)', '2')]}, 'search host="localhost"'),
            #
            # b) clicks on the dc(date_hour) values should be well
            # defined, but it gives back a malformed search search
            # sourcetype="twiki" | search dc(twikipage)="7" which then
            # throws a parse error when I POST.
            ('search * | stats avg(date_minute) dc(date_hour) by host', { 'vals': [('host', 'localhost'),('dc(date_hour)', '2')]}, 'search host="localhost" | top date_hour'),
            
            # this is a silly search. With pure functor(field) args, stats
            # of course only returns a single row. However the clicks upon
            # these rows end up with a malformed search search
            # sourcetype="twiki" | search sum(date_hour)="219974"
            ('search * | stats avg(date_minute) sum(date_hour)', { 'vals': [('avg(date_minute)', '32.4242'),('avg(date_minute)', '32.4242')]}, 'search *'),
            ('search * | stats avg(date_minute) sum(date_hour)', { 'vals': [('avg(date_minute)', '32.4242'),('sum(date_hour)', '2323')]}, 'search *'),
            
            # the drilldown clicks yield incorrect searches
            # sourcetype=twiki | search twikiaction="save" host="AshleyJones" | top None
            ('search * | chart count over date_minute by host', { 'vals': [('date_minute', '3'),('date_minute', '3')]}, 'search date_minute="3"'),
            ('search * | chart count over date_minute by host', { 'vals': [('date_minute', '3'),('OTHER', '3')]}, None),
            ('search * | chart count over date_minute by host', { 'vals': [('date_minute', '3'),('localhost', '1554')]}, 'search host="localhost" date_minute="3"'),
              ]
        renameSearches = [
            ('search * | stats avg(date_minute) by host,sourcetype | kmeans', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, None),
            ('search * | stats avg(date_minute) by host,sourcetype | dedup host', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, 'search host="localhost"'),
            ('search * | stats avg(date_minute) by host,sourcetype | head 10', { 'vals': [('host', 'localhost'),('host', 'localhost')]}, 'search host="localhost"'),
            ('search * | stats avg(date_minute) by host,sourcetype | rename host as elvis', { 'vals': [('elvis', 'localhost'),('elvis', 'localhost')]}, 'search host="localhost"'),
            ('search foo | stats count(user) as elvis by host,group', { 'vals': [('host', 'localhost'),('elvis', '1')]}, 'search foo host="localhost" | top user'),
            ('search foo | chart count(user) as elvis over host', { 'vals': [('host', 'localhost'),('elvis', '1')]},  'search foo host="localhost" | top user')
        ]

        legendSearches = [
            ('search foo | timechart avg(user) by group', { 'vals': [(None, None),('per_index_thruput',None)]}, 'search foo group="per_index_thruput"'),
            ('search foo | chart avg(size) over source by host', { 'vals': [(None, None),('localhost', None)]}, 'search foo host="localhost"'),
            ('search foo | timechart avg(user)',                                           { 'vals': [(None, None),('avg(user)',None)]},       None),
            ('search foo | chart avg(size)    over host',                                  { 'vals': [(None, None),('avg(size)', None)]},      None),
            ('search sourcetype="audittrail" | head 1000| chart avg(linecount) by action', { 'vals': [(None, None),('avg(linecount)', None)]}, None),
        ]            

    
        searches.extend(xsearches)
        searches.extend(renameSearches)
        searches.extend(legendSearches)
        print len(searches), "test cases"

        
        for search,args,expected in searches:
            p = Parser.parseSearch(search)
            try:
                result = applyIntention(p, 'drilldown', args, None)
            except Exception, e:
                if expected != None:
                    print "\nS: %s\nP:%s\nA:%s" % (search, p.rawJsonable(), args)
                    print "Unexpected:", e
                    import traceback
                    print traceback.format_exc()
                    exitingAssert(False)                    
                    break 
                else:
                    pass #print "SUCCESS.  I WANTED A THROW EXCEPTION", e
                continue
            if str(result) == str(expected):
                pass #print "Success:\n\tResult:   %s\n\tExpected: %s\n" % (result, expected)            
            else:
                print "\nS: %s\nP:%s\nA:%s" % (search, p.rawJsonable(), args)                
                print "!"*20, "Fail:\n\tResult:   %s\n\tExpected: %s\n" % (result, expected)
                exitingAssert(False)

        

        
if __name__ == "__main__":
    # Execute test suite.
    transformSuite = unittest.TestLoader().loadTestsFromTestCase(TestIntentions)
    unittest.TextTestRunner(verbosity=3).run(transformSuite)
    # stupid global variable for error code becausee unit test infrastructure seems to each up system.exit and return 0
    exit(g_exit_code)
