import re
import splunk.bundle
import splunk.auth
import logging as logger

# cache confs, as splunk.bundle is really, really slow
confs = {}
def getStanzas(confFile, sessionKey, namespace):
    global confs
    key = confFile + sessionKey
    if key not in confs:
        try:
            confs[key] = splunk.bundle.getConf(confFile, sessionKey, namespace)
        except:
            msg = "Unable to get '%s' configuration for '%s' namespace." % (confFile, namespace)
            logger.error(msg)
            raise Exception(msg)
    return confs[key]


def isPublic(stanza):
    if "usage" not in stanza:
        logger.warn("The %s command is missing a 'usage' value." % stanza.name)
        return False
    usage = stanza["usage"].strip().lower()
    return "public" in usage and not "deprecated" in usage

def isGenerating(stanza):
    return stanza.get("generating", "false").lower().startswith("t")

def getAttr(stanzas, command, attr, defaultval):
    if not command.endswith("-command"):
        command += "-command"
    if command not in stanzas.stanzas:
        return defaultval
    stanza = stanzas[command]
    return stanza.get(attr, defaultval)

PUBNAMES = []
def getAllCommands(stanzas, onlyGenerating = False):
    global PUBNAMES
    if len(PUBNAMES) > 0:
        return PUBNAMES
    publicnames = []
    stanzanames = stanzas.stanzas.keys()
    SEP = re.compile("[, ]+")
    for key in stanzanames:
        val = stanzas[key]
        if key.endswith("-command") and isPublic(val) and (not onlyGenerating or isGenerating(val)):
            name = key[0:-8]
            publicnames.append(name)
            # add aliases, if any
            aliasstr = stanzas[key].get('alias', None)
            if aliasstr != None:
                aliases = SEP.split(aliasstr)
                publicnames.extend(aliases)
    publicnames.sort()
    PUBNAMES = publicnames
    return PUBNAMES

ALIASMAP = {}
# return a dict mapping alias to preferred name
def getAliasMap(stanzas):
    global ALIASMAP
    if len(ALIASMAP) > 0:
        return ALIASMAP
    aliasMap = {}
    SEP = re.compile("[, ]+")
    stanzanames = stanzas.stanzas.keys()
    for key in stanzanames:
        val = stanzas[key]
        if key.endswith("-command") and isPublic(val):
            name = key[0:-8]
            # make an alias from command to itself
            aliasMap[name] = name
            # if any aliases, map them to the preferred name
            aliasstr = stanzas[key].get('alias', None)
            if aliasstr != None:
                aliases = SEP.split(aliasstr)
                for alias in aliases:
                    aliasMap[alias] = name
    ALIASMAP = aliasMap
    return ALIASMAP

def getJustCommands(search, aliasMap):
    commands = getCommands(search, aliasMap)
    return [c.strip() for search in commands for c,a in search ]

# removes quotedparts
def removeQuotedParts(search):
    search = search.replace('\\"', "ESCAPED_QUOTE")
    # remove all quote pairs or last quote
    quotes = re.findall('".*?(?:"|$)', search)
    for quote in quotes:
        search = search.replace(quote,'')
    search = search.replace("ESCAPED_QUOTE", '\\"')
    return search.strip()


def getCommands(search, aliasMap):
    seq = []
    args = []
    search = search.strip()
    if not search.startswith("search") and not search.startswith('|'):
        search = "|search " + search
    if not search.startswith('|'):
        search = "| " + search
    search = removeQuotedParts(search)    
    subsearches = re.findall('\[(.*)\]', search) # greedy to get outtermost ]
    for subsearch in subsearches:
        search = search.replace(subsearch,'')
        subcommandseqs = getCommands("| " + subsearch, aliasMap)
        for sub in subcommandseqs:
            if len(sub) > 0:
                seq.append(sub)
    commandsAndArgs = re.findall("\|\s*([a-zA-Z0-9_]+)([^|\[\]]*)", search)
    seq.append(commandsAndArgs)
    
    if aliasMap != None:
        for search in seq:
            for c,a in search:
                if c in aliasMap:
                    val = (c,a)
                    pos = search.index(val)
                    search.remove(val)
                    search.insert(pos, (aliasMap[c],a))
    return seq
    
def escapePipeBrackets(search, unescape = False):
    replacements = (("|", "-=BAR=-"), ("[", "-=OPEN=-"), ("]", "-=CLOSE=-"))
    for replacement in replacements:
        if unescape:
            search = search.replace(replacement[1], replacement[0])            
        else:
            search = search.replace(replacement[0], replacement[1])
    return search

def escapeQuotedParts(search, unescape = False):
    search = search.replace('\\"', "-=ESCAPED_QUOTE=-")
    # escape all quote pairs or last quote
    quotes = re.findall('".*?(?:"|$)', search)
    for quote in quotes:
        search = search.replace(quote,escapePipeBrackets(quote, unescape))
    search = search.replace("-=ESCAPED_QUOTE=-", '\\"')
    return search.strip()


    
# basically the same as getCommands()[-1] but tolerant over unmatched ['s
# as the user is typing in
#    e.g. "|search 404 | sort [ metadata fields " --> metadata
#    e.g. "|search 404 | sort [ metadata fields] " --> sort
def getLastCommand(search, aliasMap):

    search = escapeQuotedParts(search)
    # remove all complete [subsearches] (we couldn't be inside one if it's complete)
    subsearches = re.findall('(\[.*?\])', search)
    for subsearch in subsearches:
        search = search.replace(subsearch,'')    
    lbar = search.rfind('|')
    # we might be in a partial subsearch -- e.g. "|search [crawl"
    lsqr = search.rfind('[')
    last = None
    if lbar >= 0:
        last = lbar
    if lsqr >= 0 and lsqr > lbar:
        last = lsqr
    if last == None:
        return None
    lastcommand = re.findall("\s*([a-zA-Z0-9_]+)(.*)", search[last:])
    if len(lastcommand) > 0:
        command = lastcommand[0][0]
        args = lastcommand[0][1].strip()
        args = escapeQuotedParts(args, True)
        
        if aliasMap != None and command in aliasMap:
            command = aliasMap[command]
        return command, args
    return None

def allButLast(search):
    comm = getLastCommand(search, None)
    if comm == None:
        return ""
    command,arg = comm
    try:
        match = re.search("[|[]\s*%s\s*%s" % (command, arg), search)
    except Exception, e:
        logger.error("Unable to find last command due to regex problem for command: '%s' arg: '%s': %s" % (command, arg, e))
        match = None
        
    if match == None:
        return search
    return search[:match.start()]

def readText(filename):
    text = ""
    try:
        f = open(filename, 'r')
        text = f.read()
        f.close()
    except Exception, e:
        logger.error('Cannot read file %s because: %s' % (filename, e))
    return text


def prettyList(l, conj="and"):
    output = ""
    llen = len(l)
    if llen == 0:
        return ""
    if llen == 1:
        return l[0]
    if llen == 2:
        return "%s %s %s" % (l[0], conj, l[1])
    for i, v in enumerate(l):
        if output != "":
            output += ", "
        if i == llen-1:
            output += conj +  " " 
        output += str(v)
    return output
    
def TEST_SESSION():
    return splunk.auth.getSessionKey('admin','changeme')

def TEST_NAMESPACE():
    return "search"
