import os
import re, splunk.admin as admin
import wmi, json
import splunk.entity as en
import splunk.rest
import win32api, win32con, pywintypes
import logging as logger
import re

# a wild card string representing all instances that will show in the UI
ALL_INSTANCES = "-- All instances --"

"""
  WQL is in the form:
  SELECT fld1[,...] FROM src [WHERE predicate]
"""
class WQLParser:

	"""
	Suppose you want to represent:
	  "SELECT fld1, fld2 FROM aClass WHERE x = 1"

	You could initialize it as:
	p = WQLParser(fields = ["fld1", "fld2"], clas = "aClass", where = "x = 1")
	... or, if you want to omit the where clause:
	p = WQLParser(fields = ["fld1", "fld2"], clas = "aClass")
	print p.

	Also, you could give it a WQL statement and parse it:
	p = WQLParser()
	try:
	  p.parse("SELECT fld1, fld2 FROM aClass WHERE x = 1")
	except ValueError, e:
	  print "Error parsing: %s" % e
        
	Note that where is optional
	"""
	def __init__(self, fields = None, clas = None, where = None):
		self._fields = fields
		if fields is not None and clas is None:
			raise ValueError, "Invalid use: if fields is specified, class must be, too"

		self._class = clas
		self._where = where

	def getFields(self):
		if self._fields is None:
			raise ValueError, "WQL has not been parsed"
		return self._fields

	def getClass(self):
		if self._class is None:
			raise ValueError, "WQL has not been parsed"
		return self._class

	def getWhere(self):
		return self._where

	"""
	Given a fieldName and valuesList = ["one", "two"], will compose a
	where clause that looks like 'fieldName="one" OR fieldName="two"'.
	"""
	def setOrWqlWhereClauseValues(self, fieldName, valuesList):
		clauses = []
		for val in valuesList:
			clauses.append('%s="%s"' % (fieldName, self._escape(val)))
		if len(clauses) > 0:
			self._where = " OR ".join(clauses)
		else:
			self._where = None

	"""
	given a WHERE clause that looks like
	  'fieldName="one" OR fieldName="two"'
	will return a list ["one", "two"].  Will return empty list if there
	are no fields to parse.  Will raise ValueError if unable to parse.
	"""
	def tokenizeOrWqlWhereClauseValues(self, fieldName):
		if self._where is None or self._where == "":
			return []

		# used to match 'fldName="value"'
		retval = []
		compr = re.compile('^\s*%s\s*=\s*"([^"]+)"\s*$' % fieldName, re.IGNORECASE)
		for kv in re.split("(?i)\s+OR\s+", self._where, re.IGNORECASE):
			valres = compr.search(kv)
			if valres and len(valres.groups()) > 0:
				retval += [self._unescape(valres.group(1))]
			else:
				raise ValueError, "The where clause '%s' is not parsable" % self._where

		return retval


	def parse(self, wql):
		self.__init__()

		try:
			# ex: "SELECT PercentProcessorTime FROM Win32_PerfFormattedData_PerfOS_Processor WHERE x = 1"
			# res1 = ['', ' PercentProcessorTime FROM Win32_PerfFormattedData_PerfOS_Processor WHERE x = 1']
			res1 = re.split(r'(?i)select\s', wql)
			if len(res1) != 2:
				raise ValueError

			# res2 = [' PercentProcessorTime ', ' Win32_PerfFormattedData_PerfOS_Processor WHERE x = 1']
			res2 = re.split(r'(?i)\sfrom\s', res1[1])
			if len(res2) != 2:
				raise ValueError

			# res3 = [' Win32_PerfFormattedData_PerfOS_Processor ', ' x = 1']
			# without "WHERE": [' Win32_PerfFormattedData_PerfOS_Processor']
			res3 = re.split(r'(?i)\swhere\s', res2[1])

			# extract comma-separated fields into a list
			self._fields = [i.strip() for i in re.split(",", res2[0])]
			self._class = res3[0].strip()

			# see if the WQL beyond the class name appears to be valid
			m = re.search(r'\s', self._class)
			if m is not None:
				# there are spaces and the query appears unpraseable beyond the class
				raise ValueError

			# the "WHERE" clause is optional
			if len(res3) == 2:
				self._where = res3[1].strip()
			else:
				self._where = None

		except:
			raise ValueError, "Invalid WQL specified (%s)" % wql

	def __str__(self):
		if self._fields is None or self._class is None:
			raise ValueError, "WQL has not been set"

		wql = " ".join(["SELECT", ", ".join(self._fields), "FROM", self._class])
		if self._where is not None and len(self._where) > 0:
			wql += " WHERE %s" % self._where
		return wql

	def _escape(self, s):
		return s.replace('\\', '\\\\')

	def _unescape(self, s):
		return s.replace('\\\\', '\\')

def test_WQLParser():
	wqls = [
		"SELECT fld1 FROM class",
		" select fld2,fld2 from class ",
		"SELECT fld2, fld2 FROM class where x = 1",
		"SELECT * FROM class where x = 1",
		"SELECT * FROM class blah blah blah",
	]

	for wql in wqls:
		print "* '%s'" % wql
		p = WQLParser()
		try:
			p.parse(wql)
			print "  '%s'" % p
			print "    flds %s" % p.getFields()
			print "    clss '%s'" % p.getClass()
			print "    wher '%s'" % p.getWhere()
		except ValueError, e:
			print "  ERROR: %s" % e

	print
	p = WQLParser(fields = ["fld1", "fld2"], clas = "class", where = "x = 2")
	print p
	p = WQLParser(fields = ["fld1"], clas = "class")
	print p

"""
Test code for WQLParser.setOrWqlWhereClauseValues() and WQLParser.tokenizeOrWqlWhereClauseValues()
"""
def test_tokenizeOrWqlWhereClauseValues():
	tdata = [
		('name = "igor"', True),
		('Nname = "igor"', False),
		('blah = "igor"', False),
		('name = "igor" OR Name="ozra"', True),
		('name = "igor" or Name="ozra"', True),
		('name = "igor" ORName="ozra"', False),
		('name = "igor"OR Name="ozra"', False),
		('  name = "igor" OR Name="ozra"  ', True),
		('OR name = "igor" OR Name="ozra"  ', False),
		('  name = "igor" OR Name="ozra"  OR ', False),
		(' OR  name = "igor"', False),
		('name="igor" OR', False),
		('name="igor" AND', False),
		('name = "igor" AND Name="ozra"', False),
		('name = "igor" OR Name="ozra" and ', False),
		('qwe', False),
		#('', True),
		#(None, True),
	]
	verifyValues = ["igor", "ozra"]

	key = "Name"
	fields = "*"
	clas = "Test"

	print "Testing WQLParser.tokenizeOrWqlWhereClauseValues()...\n"
	for tdatum in tdata:
		success = True

		# method 1 using the constructor
		#p = WQLParser(fields = fields, clas = clas, where = tdatum[0])

		# method 2 using .parse()
		p = WQLParser()
		if tdatum[0] is not None:
			wql = "SELECT %s FROM %s WHERE %s" % (fields, clas, tdatum[0])
		else:
			wql = "SELECT %s FROM %s WHERE" % (fields, clas)
		p.parse(wql)

		print "'%s'" % tdatum[0]
		try:
			orClauses = p.tokenizeOrWqlWhereClauseValues(key)
			print "  %s" % orClauses
			for i in orClauses:
				if i not in verifyValues:
					raise Exception, "Unexpected value '%s' found" % i
		except Exception, e:
			print "  ERROR: %s" % str(e)
			success = False

		if success != tdatum[1]:
			raise Exception, "Expected %s, got %s" % (tdatum[1], success)
	print "GOOD.\n"

	tdata2 = [
		[],
		["one"],
		["one", "two"],
		["esc\\", "two"],
	]

	print "Testing WQLParser.setOrWqlWhereClauseValues()...\n"
	for tdatum in tdata2:
		p = WQLParser(fields = fields, clas = clas)
		p.setOrWqlWhereClauseValues(key, tdatum)
		print tdatum
		print "  %s" % str(p)
		clauses = p.tokenizeOrWqlWhereClauseValues(key)
		if clauses != tdatum:
			raise Exception, "Clause value error: expected '%s' got '%s'" % (tdatum, c)

	print "GOOD.\n"

"""
Reads the wmi config file and fills in a ConfInfo object
"""
def getWmiConf(handler, confInfo):
	def transformName(name):
		return name[4:]
	confDict = handler.readConfCtx("wmi")

	if None != confDict:
		for stanza, settings in confDict.items():
			# only process WMI: stanzas that contain a wql key
			if stanza[:4] == "WMI:" and len(stanza) > 4 and "wql" in settings:
				name = transformName(stanza)  # strip away leading "WMI:"
				confInfo[name].append("disabled", settings.get("disabled", "0"))

				# if missing assume localhost
				confInfo[name].append("server", settings.get("server", "localhost"))

				# must be present
				confInfo[name].append("interval", settings["interval"])
				confInfo[name].append("wql", settings["wql"])

				instances = ""

				try:
					p = WQLParser()
					p.parse(settings["wql"])
					confInfo[name].append("class", p.getClass())

					#confInfo[name].append("fields", p.getFields())
					confInfo[name].append("fields", ", ".join(p.getFields()))

					try:
						# get the values of all 'Name=".."' kv pairs under the where clause
						instances = ", ".join(p.tokenizeOrWqlWhereClauseValues("Name"))
					except ValueError, e:
						# cannot do it
						raise e

				except ValueError, e:
					# unable to parse WQL; simply set to empty string
					confInfo[name].append("class", "")

					#confInfo[name].append("fields", [])
					confInfo[name].append("fields", "")

				confInfo[name].append("instances", instances)
		confInfo.copyMetadata(confDict, transformName)
                
"""
Given a class name and optional host, will retrieve all "instances" of that
class and return a list of the values of the "Name" property for each instance.
On error or if unable to retireve, will simply return an empty list.
"""
def getWmiInstances(className, host = None):
	retval = []
	c = None
	try:
		# to retrieve "instances", return property "Name" for all available instances of the class
		if host is not None:
			c = wmi.WMI(host, find_classes = False)
		else:
			c = wmi.WMI(find_classes = False)

		# try with un-normalized name first
		retval = [str(i.Name) for i in c.__getattr__(unnormWmiClassName(className))(["Name"]) if i.Name is not None]
	except AttributeError, e:
		# that failed: try with the class name verbatim
		try:
			retval = [str(i.Name) for i in c.__getattr__(className)(["Name"]) if i.Name is not None]
		except:
			# unable to get any results for the property "Name": assume empty
			pass
	except:
		# unable to get any results for the property "Name": assume empty
		pass
	return retval


"""
Only list classes which begin with this prefix
"""
WMI_FILTER_IN_PREFIX = "Win32_PerfFormattedData_"

"""
Normalizes a WMI class name into a "user-friendly" format.
"""
def normWmiClassName(name):
	l = len(WMI_FILTER_IN_PREFIX)
	if name[:l] == WMI_FILTER_IN_PREFIX:
		return name[l:]
	else:
		return name

"""
Given a name which is assumed to have been normalized using normWmiClassName(),
retrieves the un-normalized form of it.
"""
def unnormWmiClassName(name):
	return "%s%s" % (WMI_FILTER_IN_PREFIX, name)

"""
Retrieves all classes on which we filter on the local or remote machine.  Then
names will be normalized using the normWmiClassName() function
"""
def getWmiClasses(host = None):
	try:
		c = None
		if host is None:
			c = wmi.WMI()
		else:
			try:
				c = wmi.WMI(host)
			except wmi.x_wmi, e:
				# unable to obtain classes from remote host: try locally
				c = wmi.WMI()
	except Exception, e:
		raise admin.InternalException, "Unable to get wmi classes from host '%s': %s.  " \
						"Make sure WMI is configured correctly." % (host, str(e))	# TK mgn 06/19/09

	# return only a list of classes that begin with WMI_FILTER_IN_PREFIX;
	# strip the leading WMI_FILTER_IN_PREFIX when returning

	l = len(WMI_FILTER_IN_PREFIX)

	#return [normWmiClassName(i) for i in c.classes if i[:l] == WMI_FILTER_IN_PREFIX]
	return [i for i in c.classes if i[:l] == WMI_FILTER_IN_PREFIX]

"""
Given a class name, attempts to retrieve all available fields
"""
def getWmiFields(clsName, host = None):
	pp = None
	try:
		c = None
		if host is None:
			c = wmi.WMI()
		else:
			try:
				c = wmi.WMI(host)
			except wmi.x_wmi, e:
				# unable to obtain fields from remote host: try locally
				c = wmi.WMI()

		try:
			# first try with un-normalized class name
			pp = c.__getattr__(unnormWmiClassName(clsName))
		except wmi.x_wmi, e:
			raise e
		except:
			# this failed; try with the verbatim form of the name
			pp = c.__getattr__(clsName)

	except Exception, e:
		raise admin.InternalException, "Unable to get wmi fields of class '%s': %s.  " \
						"Make sure WMI is configured correctly." % (clsName, str(e))	# TK mgn 06/19/09

	return pp.properties

"""
Encodes a currentName into JSON, depending on whether currentName is in
selectedItemNames.  It applies normalizationFunc to every encoded item.

Example output: "{"Tcpip_ICMPv6": 0}"
"""
def createJsonEncodedItem(currentName, selectedItemNames, normalizationFunc = lambda x: x):
	d = None
	l = [normalizationFunc(i).lower() for i in selectedItemNames]
	if normalizationFunc(currentName).lower() in l:
		d = {currentName: 1}
	else:
		d = {currentName: 0}
	return json.JSONEncoder().encode(d)

"""
@param procInputs - list of input processors endpoint paths to be reloaded
@param scriptInputs - list of scripted input endpoint paths to be reloaded

Example endpoint path: admin/win-eventlogs/_reload
"""
def reloadConf(self, procInputs=[], scriptInputs=[]):
	for procInput in procInputs:
		en.getEntities( procInput,
		  	sessionKey = self.getSessionKey() )

	#TODO: Handle restart of individual cripted
	for scriptInput in scriptInputs:
		en.getEntities( scriptInput,
		  	sessionKey = self.getSessionKey() )

"""
Given a "disabled" config value as a string, returns True or False
"""
def isDisabled(s):
	s = str(s)
	s = s.lower().strip()
	if s == "true" or s == "1" or s == "yes" or s == "":
		return True
	return False

"""
Returns a string of representing a value disabled config value.  Looks at
the existing value s, and tries to use the same convention.
"""
def setDisabled(s, disabled = 0):
	convTable = (("0", "1"), ("false", "true"), ("no", "yes"))
	ls = s.lower().strip()
	oposite = (disabled + 1) % 2

	for conv in convTable:
		if conv[disabled] == ls:
			# already properly set: just return it in the prefered case
			return ls
		elif conv[oposite] == ls:
			# set to oposite value: return the mathing oposite
			return conv[disabled]

	return convTable[0][disabled]

"""
Deletes a registry a key and all of the subkeys under it in local machine
registry hive, under Software.  Currently used by the cli to delete splunk
registry keys when the user does "clean all"
"""
def DeleteSplunkRegistryKeys(splunkKey):
    hKeyRoot = win32con.HKEY_LOCAL_MACHINE

    RegDeleteKeyRecurse(hKeyRoot, os.path.join("Software", splunkKey))

"""
Deletes a registry key and all the subkeys under it
"""
def RegDeleteKeyRecurse(hKeyRoot, hSubKey):
    hKey = 0
    enumKeys = ""
    rootKeyStr = "HKEY_LOCAL_MACHINE"

    try:
        win32api.RegDeleteKey(hKeyRoot, hSubKey)
        logger.info("\tCleaning registry key %s\%s" % (rootKeyStr, hSubKey))
        return 0
    except pywintypes.error, e:
        pass

    try:
        hKey = win32api.RegOpenKeyEx(hKeyRoot, hSubKey, 0, win32con.KEY_READ)
    except pywintypes.error, e:
        logger.debug("Could not open registry key=%s\%s: %s" % (rootKeyStr, hSubKey, e[2]))
        return 1

    try:
        enumKeys = win32api.RegEnumKeyEx(hKey);
    except pywintypes.error, e:
        logger.debug("Could not enum key=%s\%s: %s" % (rootKeyStr, hSubKey, str(e[2])))
        return 1

    for enumKey in enumKeys:
        RegDeleteKeyRecurse(hKeyRoot, os.path.join(hSubKey, enumKey[0]))

    try:
        logger.info("\tCleaning registry key %s\%s" % (rootKeyStr, hSubKey))
        win32api.RegDeleteKey(hKeyRoot, hSubKey)
    except pywintypes.error, e:
        logger.error("Failed to delete key='%s\%s': %s" % (rootKeyStr, hSubKey,str(e[2])))

    win32api.RegCloseKey(hKey)

"""
Restarts the WMI scripted input (splunk-wmi) if it's already running, noop if
it is not.
"""
def reloadWmi(sessionKey):
	try:
		postArgs = { '_action': 'restart', 'script': 'splunk-wmi'}
		response, content = splunk.rest.simpleRequest(
			'admin/script',
			sessionKey = sessionKey,
			method = 'POST',
			postargs = postArgs)

		if response.status != 200:
			raise admin.InternalException, \
				"Error restarting WMI scripted input (splunk-wmi): " \
				"Expected status code 200, got %d" % response.status  # TK mgn 06/19/09
	except Exception, e:
		raise admin.InternalException, \
				"Unable to restart WMI scripted input (splunk-wmi): %s" % str(e)  # TK mgn 06/19/09

