/**
* JS routines used by manager/EAI
*/

Splunk.namespace("Splunk.EAI");

Splunk.EAI = $.klass({
    elements_loaded: false,
    eaicfg: {},
    endpoint_base: '',
    eai_attributes: {},
    admin_namespace: '',
    initial_form_fields: {},
    entity_name: false,
    saving: 0,
    target_has_acl: false,
    success_message: false,
    redirect: true,
    use_messenger_on_error: true,
    confirmSubmission: false,
    confirmQuestion: _("Are you sure you want to do this?"),

    initialize: function() {
        this.logger = Splunk.Logger.getLogger("Splunk.EAI");
    },

    setup: function(eaicfg, endpoint_base, eai_attributes) {
        this.eaicfg = eaicfg;
        this.endpoint_base = endpoint_base;
        this.eai_attributes = eai_attributes;
    },


    /**
    * Keep track of the initial form values so we can detect which have been 
    * deleted later
    */
    elementsLoaded: function(namespace, entity_name) {
        this.elements_loaded = true;
        this.admin_namespace = namespace;
        this.entity_name = entity_name;
        var f = this.getFormValues(this.endpoint_base);
        for (var elname in f) {
            this.initial_form_fields[elname] = true;
        }
    },


    /**
    * Find all onChange events for an element and execute them in order
    *
    * Onchange handles are methods of this class prefixed with "action_"
    */
    doElementOnChange: function(el, element_name) {
        var elcfg =  this.eaicfg[element_name];
        if (!elcfg['onChange']) 
            return;

        return this._execChangeAction(el, element_name, elcfg, elcfg['onChange']);
    },

    _execChangeAction: function(el, element_name, elcfg, actions) {
        var actions = actions instanceof Array ? actions : [ actions ];
        for(var i=0; i<actions.length; i++) {
            //var action = 'this._action_'+actions[i]['_action'];
            if (this.elements_loaded || actions[i]['_triggerOnLoad'] != false) {
                var action = this['action_'+actions[i]['_action']];
                if (!action) {
                    this.logger.error("Invalid manager onChange action configured: "+actions[i]['_action']);
                    return;
                }
                action.call(this,el, element_name, elcfg, actions[i]);
            }
        }
    },


    /**
    * Display all but selected elements when an element action is triggered
    */
    action_showonly: function(el, element_name, elcfg, occfg) {
        // hide all except the element to show
        var type = elcfg['type'];
        if (!occfg['_groupset'])
            return;
        var groupset = occfg['_groupset'] instanceof Array ? occfg['_groupset'] :  [ occfg['_groupset'] ];

        if (type=='select')
            var selected = $(el).children('[selected]').val();
        else if (type=='checkbox')
            var selected = $(el).attr('checked') ? 1 : 0;
        else 
            var selected = $(el).val();

        var show = occfg[selected];
        if (!show)
            show = 'ALL';

        for(var i=0; i<groupset.length; i++) {
            var elname = groupset[i];
            if ($('#item-'+elname+'[type="hidden"]').length) {
                continue; // never show hidden fields
            }
            if (show == 'ALL') {
                $('#item-'+elname).show();
            } else if ((show instanceof Array && show.indexOf(elname)==-1) || (show != elname)) {
                $('#item-'+elname).hide();
            }
        }

        if (show instanceof Array) {
            for (i=0; i<show.length; i++) {
                if ($('#item-'+show[i]+'[type="hidden"]').length)
                    continue; // never show hidden fields
                $('#item-'+show[i]).show();
            }
        } else {
            if (!$('#item-'+show+'[type="hidden"]').length)
                $('#item-'+show).show();
        }
    },

    /**
    * Modify the list of attributes marked as 'required'
    */
    action_updateRequired: function(el, element_name, elcfg, occfg) {
        var modifylist = occfg['value_map'];
        if (!modifylist) {
            this.eai_required = undefined;
            return;
        }
        var req = {};
        for (var i=0, fc=this.eai_attributes.requiredFields.length; i<fc; i++) {
            req[this.eai_attributes.requiredFields[i]] = 1;
        }
        var current = $(el).val();
        if (!current in modifylist) {
            this.logger.error('updateRequired onChange handler does not define a value in the value_map for '+current);
            return;
        }
        modifylist = modifylist[current].split(/ +/);
        for (var i=0; i<modifylist.length; i++) {
            if (!modifylist[i])
                continue;
            var op = modifylist[i].substr(0, 1);
            var field = modifylist[i].substr(1);
            switch(op) {
                case '+':
                    req[field] = 1;
                    break;
                case '-':
                    delete req[field];
                    break;
                default:
                    this.logger.error('Field "'+op+field+'" present in updateRequired onChange handler that is not prefixed with + or -');
            }
        }
        this.eai_required  = [];
        for (var field in req) 
            this.eai_required.push(field);
    },

    action_triggerChangeHandler: function(el, element_name, elcfg, occfg) {
        target = $('#item-'+occfg['target']);
        if (occfg['makeVisible']) 
            target.show();
        $('*[onchange]', target).trigger('onchange');
        $('input[type="radio"][onclick][checked]', target).trigger('onclick');
        $('input[type="checkbox"][onclick]', target).trigger('onclick');
    },

    /**
    * Refresh an element by re-requesting it from the server using the current form values
    */
    _triggerElementRefresh_active_count: 0,
    action_triggerElementRefresh: function(el, element_name, elcfg, occfg) {
        var self = this;
        var target = occfg['target'];
        var makeVisible = occfg['makeVisible'] ? true : false;
        var element_overlay = occfg['updateElement'] ? occfg['updateElement'] : {};

        if (!this._triggerElementRefresh_active_count++) {
            $('.splOverlay-white').show();
            $('#loadingmessage').show();
        }
        var xhr = $.ajax({
            type: 'POST',
            url: Splunk.util.make_url('manager', self.admin_namespace, self.endpoint_base, '_element', target),
            data: { 
                form_defaults: JSON.stringify(self.getFormValues(self.endpoint_base)),
                eai_attributes: JSON.stringify(self.eai_attributes),
                element_overlay: JSON.stringify(element_overlay),
                entity_name: JSON.stringify(self.entity_name)
            },
            dataType: 'html',
            success: function(htmlresult, textStatus) {
                if (!--self._triggerElementRefresh_active_count) {
                    $('.splOverlay-white').hide();
                    $('#loadingmessage').hide();
                }
                var newel = $('#item-'+target).replaceWith(htmlresult);
                if ($('#item-'+target).hasClass('.accumulator')) 
                    self.initAccumulators('item-'+target);
                if (occfg['_triggerOnRefresh']) {
                    newel = $('#item-'+target);
                    $('*[onchange]', newel).trigger('onchange');
                    $('input[type="radio"][onclick][checked]', newel).trigger('onclick');
                    $('input[type="checkbox"][onclick]', newel).trigger('onclick');
                }
                if (occfg['_chained_action']) {
                    self._execChangeAction(el, element_name, elcfg, occfg['_chained_action']);
                }
                if (makeVisible) 
                    $('#item-'+target).show();
            },
            error: function(xhr, textStatus, errorThrown) {
                if (!--self._triggerElementRefresh_active_count) {
                    $('.splOverlay-white').hide();
                    $('#loadingmessage').hide();
                }
                Splunk.Messenger.System.getInstance().send('error', 'elementRefresh', sprintf(_('Splunk failed to fetch the element from the server: ')+textStatus)); 
            }
        });
    },


    /**
    * Set the value of an element to a new value 
    * Automatically trigger's the target elements onChange handlers once set
    */
    action_setElementValue: function(el, element_name, elcfg, occfg) {
        var target = occfg['target'];
        var value = occfg['value'];
        var makeVisible = occfg['makeVisible'] ? true : false;
        var value_map = occfg['value_map'];
        if (value_map) {
            switch(el.type) {
                case 'checkbox':
                    var current = $(el).filter(':checked').val();
                    break;
                case 'select':
                    var current = $(el).filter(':selected').val();
                    break;
                default:
                    var current = $(el).val();
            }
            if (!(current in value_map)) {
                // don't do anything if not explictly asked to
                return;
            }
            value = value_map[current];
        }
        var targetcfg =  this.eaicfg[target];
        switch(targetcfg.type) {
            case 'radio':
                var target_el = $('#item-'+target).find('input[type="radio"][value="'+value+'"]');
                target_el.val(1);
                break;
            case 'checkbox':
                var target_el = $('#item-'+target).find('input[type="checkbox"][value="'+value+'"]');
                target_el.val(1);
                break;
            case 'select':
            case 'multiselect':
                var target_el = $('#item-'+target).find('select[name="'+target+'"]');
                target_el.val(value);
                break;
            case 'textfield':
                var target_el = $('#item-'+target).find('input[name="'+target+'"]');
                target_el.val(value);
                break;
            case 'textarea':
                var target_el = $('#item-'+target).find('textarea[name="'+target+'"]');
                target_el.val(value);
                break;
            case 'hidden':
                var target_el = $('#item-'+target);
                target_el.val(value);
                break;
        }
        if (makeVisible)  {
            $('#item-'+target).show();
        }
        this.doElementOnChange(target_el, target);
    },


    getFormValues: function(form_name) {
        var f = document.forms[form_name];
        if (!f) {
            Splunk.Messenger.System.getInstance().send('error', 'getFormValues', sprintf(_('Splunk failed to fetch form values for form "%s"'), form_name));
            return;
        }
        var result = {};
        for(var i=0; i<f.elements.length; i++) {
            var el = f.elements[i];
            var elname = el.name;
            if (elname)
                result[elname] = $(el).val()
        }
        return result;
    },

    /**
    * Initialize passed accumulators
    * 
    * @accList {Array} : array containing the ids of the accumulators to initialize
    */
    initAccumulators: function(accList) {
        if (!accList)
            return;
        var accumulators = accList instanceof Array ? accList :  [ accList ];
        
        for(var i=0; i<accumulators.length; i++) {      
            var elname = accumulators[i].substring(5);
            (function(elname) {
                $('#' + accumulators[i]).bind("click", function(evt){
                    var t = evt.target;
                    
                    if ( $(t).is('li') )
                        $li = $(t);    
                    else if ( $(t).is('span') )
                        $li = $(t).parent('li'); 
                    else if ( $(t).is('a.addAllLink') ) {
                        addAll(this);
                        doElementOnChange(t, elname);
                        return false;
                    } else if ( $(t).is('a.removeAllLink') ) {
                        removeAll(this);
                        doElementOnChange(t, elname);
                        return false;
                    } else
                        return;
                    
                    /* if the clicked item is in the available options ul and has the class selected OR if it's in the selectedOptions ul */        
                    if ( ($li.parents('.availableOptions').length && $li.hasClass('selected')) || $li.parents('.selectedOptions').length  ) {  
                        removeFromSelected($li, this);
                    } else {  // it must be in available options and not have the selected class
                        addToSelected($li, this);       
                    } 
                    
                    doElementOnChange(t, elname);
                    return false;
                });
            })(elname);
        };
        
        var removeFromSelected = function($li, context) {
            var value = $li.attr('name');
            
            // remove from selected options list and hidden select element
            $('.selectedOptions li[name=' + value + ']', context).remove();
            $('.hiddenSelectedOptions option[value=' + value + ']', context).remove();
            
            $('.availableOptions li[name=' + value + ']', context).removeClass('selected');           
        };
        
        var addToSelected = function(arr, context) {
            var liArray = arr instanceof Array ? arr :  [ arr ];
            
            for(var i=0; i<liArray.length; i++) {
                $li = $(liArray[i]);
                value = $li.attr('name');
                fieldValue = $li.text();
                label = $li.text();
                            
                // add new li to selected options ul
                $('.selectedOptions', context).append("<li name='" + value + "'><span class='splIcon splIcon-arrow-w'></span>" + fieldValue + "</li>");
            
                // add to hidden selected options select
                $('.hiddenSelectedOptions', context).append("<option selected='selected' value='" + value + "'>" + label + "</option>"); 
                
                // set selected class on selected option
                $li.addClass('selected');
            }
            
            // sort alphabetically
            var selItems = $('.selectedOptions li', context).get();
            selItems.sort(function(a,b){ 
              var keyA = $(a).text();
              var keyB = $(b).text();
            
              if (keyA < keyB) return -1;
              if (keyA > keyB) return 1;
              return 0;
            });
            var selItemsUl = $('.selectedOptions', context);
            $.each(selItems, function(i, li){
              selItemsUl.append(li);
            });  
        };
        
        var addAll = function(context) {
            var liArray = $('.availableOptions', context).find('li').not('.selected').get();
            addToSelected(liArray, context);
        };
        
        var removeAll = function(context) {
            $('.selectedOptions', context).find('li').remove();
            $('.hiddenSelectedOptions', context).find('option').remove();
            
            $('.availableOptions li.selected', context).removeClass("selected");   
        };
    },

    /**
    * Dispatcher function to gather all accumulators into an array and pass it to the setup function
    */
    dispatchInitAccumulators: function() {
        var accumulators = [];
        $('.accumulator').not('.accDisabled').each(function(){
            accumulators.push($(this).attr('id'));    
        });
        if ( accumulators.length > 0 )
            this.initAccumulators(accumulators);   
    },

    /**
    * Set up admin tabs
    *
    * @tabID {string} : id for the tabwrapper defining the tabs themselves.  
    * @selectedTab {string} (optional) : defines which tab to intialize as 'selected', value is the id of tab to turn on
    */
    adminTabify: function(tabID, selectedTab) { 
        var $tabID = (tabID.substr(0,1) == '#') ? $(tabID) : $('#' + tabID);
        if (selectedTab)
            selectedTab = (selectedTab.substr(0,1) == '#') ? selectedTab : '#' + selectedTab;
         
        //make sure all tabs are off except for either the selected one passed, or default to the first tab
        $tabID.find('li').not('.linkTab').each(function(i){
            $a = $(this).children('a');
            if ((!selectedTab && i==0) || (selectedTab && selectedTab == $a.attr('href'))) {
                $(this).addClass('selected');
                $($a.attr('href')).show();
            } else {
                $($a.attr('href')).hide();    
            }
        });

        $tabID.bind('click', function(evt){
            var t = evt.target;
            
            var $elem = false;
                    
            if ( $(t).is('a') && !$(t).parent('li').hasClass('linkTab') ) {
                $elem = $(t);    
            } else if ( $(t).is('li') && !$(t).hasClass('linkTab') ) {
                $elem = $(t).children('a');
            }
            
            if ( $elem ) {
                $($('li.selected').children('a').attr('href')).hide();
                $('li.selected').removeClass('selected');
                $elem.parent('li').addClass('selected');
                $( $elem.attr('href') ).show();
                
                return false;
            }        
        }); 
        
        // attach a class to the admin content wrapper for use with cascading, so we know we're in a tabbed env
        $('.adminContent').addClass('hasTabs');      
    },

    setError: function(errmsg) {
        if (this.use_messenger_on_error) {
            var messenger = Splunk.Messenger.System.getInstance();
            // Any existing error messages on screen are now obsolete.
            messenger.clear();
            messenger.send('error', 'splunk.eai', errmsg);
        } else {
            $('.EAIError').text(errmsg).show();
        }
    },

    /**
     * Main handler for EAI property form submission
     *
     */
    saveForm: function(formObject, successCallback, errorCallback) {
        if (this.saving++)
            return false;
        
        if (this.confirmSubmission){
            allowSubmit = confirm(this.confirmQuestion);
            if (!allowSubmit){
                this.saving--;    
                return false;
            }
        }
            
            
        var self = this;

        var messenger = Splunk.Messenger.System.getInstance();
        this.clearFieldErrors();

        if (!this.validateRequiredFields(formObject)) {
            this.setError(_('Please enter all required fields'));
            if (errorCallback)
                errorCallback();
            this.saving--;
            return false;
        }   

        try {
            // deal with unselected checkboxes; only process checkboxes that
            // have been marked as proxiable
            var checkboxSuccess = $('input[type="checkbox"]', formObject).each(function() {
                // don't do this for disabled inputs!
                if ($(this).hasClass('proxiable') && $(this).attr('disabled') != true) {
                    self._generateProxyCheckbox(this, formObject);
                }
            });
        } catch (e) {
            alert(_('Error occurred during submission: unable to get checkbox state'));
            return false;
        }
        
        try {
            if (!this.fieldMappingSubmitHandler())
                return false;
            this.handleDeletedFields(); // must be last.
        } catch (e) {
            alert(_('Error occurred processing form submission: ')+e);
            return false;
        }


        // Backup button nanmes
        $('.jmFormActions button.splButton-primary span').each(function() {
            $(this).data('org', $(this).text());
        });
        $('.jmFormActions button.splButton-primary span').text(_('Saving...'));

        $('input[name="__redirect"]', formObject).val(this.redirect ? '1' : '0')

        $(formObject).ajaxSubmit({
            dataType: 'json',
            success: function(response) {
                var callback = false;
                switch(response.status) {
                    case 'OK':
                        if (response.redirect) {
                            window.location = response.redirect;
                            return;
                        }

                        messenger.send(
                            'info',
                            'splunk.eai',
                            self.success_message ? self.success_message : response.msg
                        );
                        callback = successCallback;
                        break;
                    case 'FIELD_ERRORS':
                        self.setError(_('Please correct the errors below'));
                        for(var fieldname in response.fields) {
                            self.setFieldError(fieldname, response.fields[fieldname]);
                        }
                        callback = errorCallback;
                        break;
                    case 'ERROR':
                        self.setError(response.msg);
                        callback = errorCallback;
                        break;
                    default:
                        self.logger.error('Unknown status response received from EAI submission endpoint: '+response.status);
                        break;
                }
                // Restore button names
                $('.jmFormActions button.splButton-primary span').each(function() {
                    $(this).text($(this).data('org'));
                });
                self._revertProxiedCheckboxes(formObject);
                window.scroll(0, 0); // make sure user can see any messages
                if (callback)
                    callback();
                self.saving = 0;
            },
            error: function(xhr, status, error) {
                self.setError(sprintf(_("Your entry was not saved.  The following error was reported: %(error)s."), {'error': error}));
                $('.jmFormActions button.splButton-primary span').each(function() {
                    $(this).text($(this).data('org'));
                });
                self._revertProxiedCheckboxes(formObject);
                if (errorCallback)
                    errorCallback();
                self.saving = 0;
            }
        });

        return false; // inhibit submit action
    },

    setFieldError: function(element_name, error_msg) {
        $('#item-'+element_name+' .widgeterror').text(error_msg).show();
    },

    clearFieldErrors: function() {
        $('.widgeterror').text('').hide();
    },

    _isFieldSet: function(fieldvalues, fieldname) {
        if (fieldvalues.hasOwnProperty(fieldname)
            && ( fieldvalues[fieldname]===undefined || fieldvalues[fieldname]===null || !Splunk.util.trim(fieldvalues[fieldname]).length || !fieldvalues[fieldname])) {
                return false;
        }
        return true;
    },

    validateRequiredFields: function(frm) {
        var required = this.eai_required==undefined ? this.eai_attributes.requiredFields : this.eai_required;
        if (!required)
            return true; // nothing to validate
        var ok = true;
        var values = this.getFormValues($(frm).attr('name'));
        for(var i=0, fc=required.length; i<fc; i++) {
            var fieldname = required[i];
            if (!this._isFieldSet(values, fieldname)) {
                this.setFieldError(fieldname, _("Required field"));
                ok = false;
            }
        }

        // Check required-if-visible fields
        for(var fieldname in this.eaicfg) {
            // Check if the element if marked requiredIfVisible
            if (!this.eaicfg[fieldname].hasOwnProperty('requiredIfVisible'))
                continue;
            // Check if the element is hidden; if so we don't care about it
            if (!values.hasOwnProperty(fieldname) || $('#item-'+fieldname).is(':hidden'))
                continue;
            if (!this._isFieldSet(values, fieldname)) {
                // Bad user; go to your room.
                this.setFieldError(fieldname, _("Required field"));
                ok = false;
            }
        }
        return ok;
    },

    /**
     * Deleted fields need an empty hidden input field so the endpoint knows to remove them
     */
    handleDeletedFields: function() {
        var container = $('#deleted');
        if (container.length) {
            container.empty();
        } else {
            container = $('<div id="deleted" style="display:none"></div>').appendTo('form[name='+this.endpoint_base+']');
        }
        var current = this.getFormValues(this.endpoint_base);
        for (var fieldname in this.initial_form_fields) {
            if (current[fieldname] == undefined) {
                $('<input type="hidden" name="'+fieldname+'" value="" />').appendTo(container);
            }
        }
    },

    /**
     * Update fieldmapping widgets so that the elements are named correctly
     */
    fieldMappingSubmitHandler: function() {
        $('.fieldmapping').filter('.left').each(function() {
            var fieldname = $(this).val();
            var fieldval = $(this).next();
            if (fieldname) {
                var prefix = $(this).parents('.repeatable').data('repeatable').prefix;
                $(fieldval).attr('name', prefix+fieldname);
            } else {
                $(fieldval).attr('name', 'spl-ctrl_null');
            }
        });
        return true;
    },

    initRepeatables: function(frm) {
        $('.repeatable a.addnew', frm).click(this.repeatableAddNew.bind(this));
        $('.repeatable a.delete', frm).click(this.repeatableDeleteField.bind(this));
    },

    /**
     * Add an entry to a repeatable list of elements
     */
    repeatableAddNew: function(ev) {
        var target = ev.target;
        var container = $(target).parent().parent();
        var maxcount = $(target).parents('.repeatable').data('repeatable').maxcount;
        if (maxcount>-1) {
            if ($(container).find('.widget').length-1 >= maxcount) { // the template is an extra widget, hence -1
                alert(sprintf(ungettext("Cannot add more than %s entry", "Cannot add more than %s entries", maxcount), maxcount));
                return false;
            }
        }
        var template = container.find('.repeatabletemplate > div');
        var newentry = template.clone();
        newentry.attr('id', template.attr('id')+Math.round(Math.random()*1000000000));
        $(target).parent().before(newentry);
        var del = this.repeatableDeleteField.bind(this);
        $('.delete', newentry).click(del);
        this.setupKeyOverride(newentry);
        return false;
    },

    /**
     * Remove an entry from a repeatable list of elements
     */
    repeatableDeleteField: function(ev) {
        var target = ev.target;
        var container = $(target).parents('.repeatable');
        var mincount = container.data('repeatable').mincount;
        if (mincount>0) {
            if ($(container).find('.widget').length-1 <= mincount) { // the template is an extra widget, hence -1
                alert(sprintf(ungettext("Must have at least %s entry", "Must have at least %s entries", mincount), mincount));
                return false;
            }
        }
        var widget = $(target).parents('.widget');
        widget.remove();
        return false;
    },

    /**
     * Generates a proxy hidden input for checkboxes that are submitted while in
     * their unselected state
     */
    _generateProxyCheckbox: function(checkboxObject, formObject, falseValue) {
        
        if ($(checkboxObject).attr('checked')) return;

        if (falseValue == null) falseValue = '0';

        var expectedName = checkboxObject.name;
        var displacedName = '_spl_proxied_' + expectedName;

        checkboxObject.name = displacedName;
        $(sprintf('<input type="hidden" name="%s" value="%s" />', expectedName, falseValue)).appendTo(formObject);
    },

    /**
     * Undo proxied inputs
     */
    _revertProxiedCheckboxes: function(formObject) {
        $('input[type="checkbox"]', formObject).each(function() {
            var name = $(this).attr('name');
            if (name.indexOf('_spl_proxied_')==0) {
                var original_name = name.substr(13);
                $('input[name="'+original_name+'"]').remove();
                $(this).attr('name', original_name);
            }
        });
    },

    /**
     * Override return key
     */
    setupKeyOverride: function(frm) {
        if (!frm)
            frm = $('#eaiform');
        var self = this;

        // most input fields should submit the form
        $('input', frm).not('.withbutton').keypress(function(ev) { 
            if (ev.which==13) {
                frm.submit();
                return false;
            }
            return true;
        });

        // except for text fields that have a button attached; those should trigger
        // the appropriate onchange handler
        $('input.withbutton', frm).keypress(function(ev) { 
            if (ev.which==13) {
                self.doElementOnChange(ev.target, $(ev.target).attr('name'));
                //$(ev.target).change(); // pretend the user clicked the button
                return false;
            }
            return true;
        });
 
    },

    /**
     * Execute run-time inits
     *
     */
    handleAdminReady: function(namespace, entity_name) {

        this.dispatchInitAccumulators();
        this.initRepeatables();
        
        //
        // leverage the base input event handlers to initialize the widget states
        // so that rendered page state matches the current settings
        //
        $('*[onchange]').trigger('onchange');
        $('input[type="radio"][onclick][checked]').trigger('onclick');
        $('input[type="checkbox"][onclick]').trigger('onclick');

        // handle the return key correctly
        this.setupKeyOverride();
        
        // mark page as loaded
        this.elementsLoaded(namespace, entity_name);

    }
});
Splunk.EAI.instance = null;
Splunk.EAI.getInstance = function() {
    if (arguments.length) {
        // Could conceivably be called more than once to re-initialize EAI with new params
        Splunk.EAI.instance = new Splunk.EAI();
        Splunk.EAI.instance.setup.apply(Splunk.EAI.instance, arguments);
    } else if(!Splunk.EAI.instance) {
        alert("Attempted to fetch instance of Splunk.EAI prior to initialization");
    }
    return Splunk.EAI.instance;
};

    

/**
 * Restart splunkd/splunkweb and udpate the UI with progress
 */
function restart_server() {
    var start_time;
    var restart_timeout = false;
    var restart_tries = 0;
    var restart_url_base;
    var restart_done = false;
    var restart_notified = false;
    var img_list = [];
    var RESTART_MAX_TRIES = 180; // Wait max of 3 minutes for restart
    var PING_INTERVAL = 2000;

    var restart_fail = function(msg) {
        $('.splOverlay').hide();
        $('#restartstatus').hide();
        $('#statusmsg').hide();
        Splunk.Messenger.System.getInstance().send('error', 'restart_server', msg);
    };

    var restart_succeeded = function() {
        if (!restart_notified) {
            alert(_('Restart successful - click OK to log back into Splunk'));
            window.location.href = restart_url_base + Splunk.util.make_url('/');
            restart_notified = true;
        }
        return;
    };

    var restart_in_progress = function(display_ssl_warn) {
        $('.splOverlay').show();
        $('#statusmsg').hide();
        $('#restartstatus').show();
        if (display_ssl_warn)
            $('.sslwarn').show();
    };

    var restart_firefox3_notice = function() {
        $('.splOverlay').show();
        $('#statusmsg').hide();
        $('#restartff3warn').show();
        $('#restartff3warn a').attr('href', restart_url_base + Splunk.util.make_url('/'));
    };

    /**
    * Track when the current instance of the appserver was started; if it changes then
    * the restart has finished.
    */
    var restart_check_status = function () {
        if (restart_done) 
            return;
        if(restart_tries++ >= RESTART_MAX_TRIES) {
            return restart_fail(_('Timed out waiting for restart'));
        }

        $.ajax({
            type: 'GET',
            dataType: 'jsonp',
            cache: false,
            url: restart_url_base + Splunk.util.make_url('/config/UI_UNIX_START_TIME?fmt=jsonp&callback=?'),
            success: function(data) {
                if (data.start_time > start_time) {
                    restart_done = true;
                    clearTimeout(restart_timeout);
                    return restart_succeeded();
                }
            }
        });
        // if the script fails to load (ie. because splunkweb is still down)
        // jquery won't give us an error; instead nothing will happen hence we
        // need to refire the poll
        restart_timeout = setTimeout(restart_check_status, PING_INTERVAL);
    };

    var restart_img_loaded = function() {
        restart_done = true;
        restart_succeeded();
    };

    /**
    * Img checking works better than timestamp checking if the protocol has been changed
    */
    var restart_check_status_img = function() {
        if (restart_done)
            return;
        var im = new Image();
        im.onload = restart_img_loaded;
        im.src = restart_url_base + Splunk.util.make_url('/static/img/skins/default/a.gif?_=') + Math.random();
        img_list.push(im);
        restart_timeout = setTimeout(restart_check_status_img, PING_INTERVAL);
    };

    if (!confirm(_("Are you sure you want to restart Splunk?")))
        return false;

    restart_in_progress();
    $.post(Splunk.util.make_url('/api/manager/control'), {operation: 'restart_server'}, function(data) {
        if (data.status=='OK') {
            start_time = data.start_time;
            restart_url_base = (data.ssl ? 'https://' : 'http://')
                + window.location.hostname
                + (data.port==80 || (data.ssl && data.port==443) ? '' : ':'+data.port);

            // Turn off stuff that continues to try to communicate with the appserver
            Splunk.Logger.mode.Default = Splunk.Logger.mode.None;
            Splunk.Messenger.System.getInstance().abortRequests = true;

            restart_tries = 0;
            var current_port = window.location.port || (window.location.protocol=='http:' ? 80 : 443);
            var current_ssl = window.location.protocol == 'https:';
            if ( ($.browser.mozilla && $.browser.version > "1.9") && ((!current_ssl && data.ssl) || (data.ssl && current_port!=data.port)) ) {
                // FF3 can repeatedly display a hard error about invalid certificates
                return restart_firefox3_notice();
            }
            if (data.ssl!=current_ssl || current_port!=data.port) {
                restart_check_status_img();
            } else {
                restart_check_status();
            }
            return restart_in_progress(data.ssl && (!current_ssl || current_port!=data.port));
        } else if (data.status == 'PERMS') {
            return restart_fail(_('Permission Denied - You are not authorized to restart the server'));
        } else if (data.status == 'AUTH') {
            return restart_fail(_('Restart failed'));
        } else if (data.status == 'FAIL') {
            return restart_fail(_('Restart failed: '+data.reason));
        } else {
            return restart_fail(_('Restart failed'));
        }
    }, 'json');
    return false;
};


/**
 * Request a resync of the authentication system
 */
function resync_auth() {
    var msg = Splunk.Messenger.System.getInstance();
    msg.send('info', 'resync_auth', _('Syncing..'));
    $.ajax({
        type: 'POST',
        data: {operation: 'resync_auth'},
        dataType: 'text',
        cache: false,
        url: Splunk.util.make_url('/api/manager/control'),
        success: function(data) {
            if (data.substr(0,2)=='OK') {
                msg.send('info', 'resync_auth', _('Authentication synched successfully.'));
            } else if(data.substr(0,4)=='AUTH') {
                msg.send('error', 'resync_auth', _('Synch failed - authentication rejected - Log in again to retry.'));
            } else if(data.substr(0,5)=='PERMS') {
                msg.send('error', 'resync_auth', _('Permission denied. You are not authorized to reload the authentication system'));
            } else {
                msg.send('error', 'resync_auth', _('Synch failed: ')+data);
            }
        },
        error: function() {
            msg.send('error', 'resync_auth', _('Failed to contact server'));
        }
    });
};


///////////////////////////////////////////////////////////////////////////////
// Global event handlers
///////////////////////////////////////////////////////////////////////////////

