// we listen to the jobResurrected event (triggered in Splunk.SearchContext.resurrect())
// and because it passes the containing group title of the module resurrecting the job, 
// we are able to map the job.getCreateTime() values into 'last refreshed:' header for each panel.
Splunk.DashboardManager = $.klass({
    dateDict : {},

    initialize: function() {
        $(document).bind('jobResurrected', this.displayRefreshedTimeForGroup.bind(this));
        $(document).bind('jobDispatched',  this.displayRefreshedTimeForGroup.bind(this));
        
        // setup the headers to auto-truncate long titles
        this.titleHeaders = $('.layoutCell .splHeader h2');
        this.handlePanelResize();
        $(window).bind('resize', this.handlePanelResize.bind(this));
        $(document).bind('allModulesLoaded', this.handlePanelResize.bind(this));
        
        //do equal heights
        this.equalizeHeights();
    },
    
    /**
     * for the given group (aka panel title), we return a date object with the timestamp of the oldest job it contains.
     */
    getEarliestCreateTimeForGroup: function(group) {
        var earliest = null;
        for (var i=0,l=this.dateDict[group].length; i<l; i++) {
            if (!earliest) {
                earliest = this.dateDict[group][i]; 
                continue;
            }
            if (this.dateDict[group][i] < earliest) {
                earliest = this.dateDict[group][i];
            }
        }
        return earliest;
    },
    
    /**
     * On every jobDispatched and jobResurrected event, we'll file this away and update the panel headers.
     */
    displayRefreshedTimeForGroup: function(event, job, group) {
        if (group && job.getCreateTime()) {
            // 1 We keep a list of all of the times received per group.
            if (!this.dateDict.hasOwnProperty(group)) {
                this.dateDict[group] = [];
            }
            this.dateDict[group].push(job.getCreateTime());
            
            // 2 we get the earliest one for this group.
            var earliest = this.getEarliestCreateTimeForGroup(group);
            
            // 3 detecting the special case for 'today at foo'
            var today = new Date();
            today.setHours(0);
            today.setMinutes(0);
            today.setSeconds(0);

            // 4 get the correct localized string. 
            var dateText = null;
            if (today<earliest) {
                dateText = sprintf(_("today at %(earliest_time)s."), {earliest_time: format_time(earliest, "medium")}) ;
            } else {
                dateText = format_datetime(earliest, "medium");
            }
            var lastRefreshedSpan = $("<span>")
                .addClass("splLastRefreshed")
                .html(sprintf(_("refreshed: <b>%(dateText)s</b>"), {dateText: dateText}));

            // 5 go find the correct panel title
            $(".splHeader-dashboard").each(function(i){
                
                if ($(this).find("h2").attr('title') == group) {
                    $(this).find("span.splLastRefreshed").remove();
                    $(this).append(lastRefreshedSpan);
                }
            })
        }
    },
    
    /**
     * Auto truncates the panel headers based on available width
     */
    handlePanelResize: function() {
        this.titleHeaders.each(function() {
            // without this check it has the neat effect of nuking the contents of all $(".splHeader h2")
            // including those in modules like ResultsHeader.
            if ($(this).attr('title')) {
                // this is just a trial and error calculation; could be smarter
                var charWidth = parseInt(Math.pow($(this).parent().width() / 12 - 15, 1.15), 10);
                $(this).text(Splunk.util.smartTrim($(this).attr('title'), charWidth));
            }
        });
        // set equal heights 
        this.equalizeHeights();
    },
    
    /**
     * This method catches the case where the user is looking at a new simple
     * dashboard that has no panels, and displays an accelerator link
     */
    showDashboardPrompts: function() {
        if (Splunk.util.getCurrentViewConfig()['view']['objectMode'] != 'SimpleDashboard') {
            return false;
        }
        var panelCount = $('.dashboardCell').length;
        if (panelCount == 0) {
            var link = $(
                _('<p class="dashboardPromptMessage">This dashboard is empty. <a href="#">Edit the dashboard</a> to add a panel.</p>')
            ).bind('click', function() {
                Splunk.Globals.Viewmaster.openDashEditForm(Splunk.util.getCurrentView());
                return false;
            }).appendTo($('.layoutRow.firstRow'));
        }
    },
    /**
     * This method equalizes heights of dashboard cells within the same panel
     */
    equalizeHeights: function() {
        $(".equalHeightRow").each(function(){
    		$(this).find('.layoutCellInner').css({'min-height': 0}); 
            if ($.browser.msie && $.browser.version == 6.0) { 
                $(this).children().css({'height': 0}); 
            }
            var max = 0;
    		$(this).find('.layoutCellInner').each(function(i){
    			if ($(this).height() > max) { max = $(this).height(); }
    		});
            if ($.browser.msie && $.browser.version == 6.0) { $(this).find('.layoutCellInner').css({'height': max}); }
    		$(this).find('.layoutCellInner').css({'min-height': max}); 
    	});;
    }
    
});

