/* Menu building class */
/*
 *    Usage:
 *     var foo = new Splunk.MenuBuilder({options array}); 
 *
 *    Options:
 *         - containerDiv : reference to the container div (this.container usually)
 *         - menuDict : JSON dictionary of menu contents/structure
 *         - activator : Jquery selector for menu activator, usually an <a> tag ex: $('#_menuActivator')
 *         - menuOpenClass : class to apply to the activator when open
 *         - menuClasses : String, list of classes to apply to the top-level outer wrapper of the menu.  Can be multiple classes, space seperated. ex: 'class1 class2 ..'
 *         - autoInit : automatically build the menu. Defaults to true;
 *         - showOnInit : build menu and immediately show it, defaults to false
 *         - fadeEffects : use a fadeIn effect for when showing the menu, defaults to true
 *
 *    Menu Dictionary:
 *      The menu is built from the menuDict object passed in.  The variety of options is confusing, this is a first pass at doc'ing.
 *      Menu Dict Options:
 *          label: {string} text to display in the menu item (ex: Save Search)
 *          uri:   {string} becomes the href of the menu item
 *          style: {string} list of classes (separated by spaces) to apply to this menu item
 *          items: {array} used to define a submenu
 *              ex: {
 *                      label: 'sub menu 1',
 *                      items: [
 *                                { 
 *                                  label: 'sub menu item 1'
 *                                  uri: 'http://foo.com'
 *                                },
 *                                { 
 *                                  label: 'sub menu item 2'
 *                                  uri: 'http://bar.com'
 *                                }
 *                      ]
 *                  }
 *          callback: {function} defines a callback function to fire when menu item is clicked                     
 *          attrs: {array} array of attributes and their values to apply to the menu item
 *              ex: {
 *                      label: 'foo',
 *                      attrs: {
 *                                  'attr1' : 'attr1Value',
 *                                  'attr2' : 'attr2Value'
 *                              }
 *                   } 
 *          showInFilter: If a filter is defined for this menu, show only menu items that have that filter set in their 'showInFilter' property.
 *          divider: {string} creates a divider div, value is the class to apply to the divider (splDivider is pre-defined in default skin)
 *          menuType: {string} optionally specify special menu type:
 *                     - 'htmlBlock' : specifies that for this menu (or submenu), the contents will be a simple html block.  used in conjuction with 'element' and optionally 'action'
 *                          ex: {
 *                                  menuType: 'htmlBlock',
 *                                  element: '.specialElement',
 *                                  action: 'clone'  
 *                               }
 *                     - 'grouping' : specifies that this is a menu grouping header, which can be used to group menu items under a common header. 'label' becomes header
 *                          ex: {
 *                                  menuType: 'grouping',
 *                                  label: 'Group 1'
 *                               }
 *                           
 *           element:  {JQuery selector} used in conjuction with menuType: 'htmlBlock', this JQuery selector indicates the element to put into the html block menu or submenu
 *           action: {string} defines a special action to use with menuTyp: 'htmlBlock'.  currently, the only option is 'clone', which causes the element specified with the 'element' property to be cloned into the html block.  default is to remove the element from its current position and move it within the dom
 *           popup : if property is present, the uri of the menu item will open in a new window
 *           remove: if property is present, menu will be removed on click
 *
 *          Examples:
 *              Basic Menu :
 *                   var newMenu = [
 *                                      {
 *                                          label : 'Item 1',
 *                                          uri: 'http://www.splunk.com',
 *                                          callback : function(){ alert('foo'); }
 *                                      },
 *                                      {
 *                                           label : 'Item 2',
 *                                           uri: 'http://splunkbase.com'
 *                                           style: 'specialFoofyClass'
 *                                      }
 *                                  ]
 *                            
 *             Menu with submenu :
 *                  var newMenu = [
 *                                    {
 *                                        label : 'Item 1',
 *                                        uri : 'http://www.splunk.com'
 *                                    },
 *                                    {
 *                                        label : 'Item 2',
 *                                        items : [
 *                                              {
 *                                                  label : 'Submenu Item 1',
 *                                                  uri : 'http://www.splunk.com'
 *                                              },
 *                                              {
 *                                                  label : 'Submenu Item 2',
 *                                                  uri : 'http://www.splunk.com'
 *                                              }                                                    
 *                                        ]
 *                                    }
 *                                  ]
 *              Menu with divider and groupings :
 *                  var newMenu = [
 *                                  {
 *                                      menuType: 'grouping',
 *                                      label: 'Group 1'
 *                                  },
 *                                  {
 *                                      label: 'Item 1',
 *                                      uri: 'http://www.splunk.com'
 *                                  },
 *                                  {
 *                                      label: 'Item 2',
 *                                      uri: 'http://www.splunk.com'
 *                                  },
 *                                  {
 *                                      divider: 'splDivider'
 *                                  },
 *                                      menuType: 'grouping',
 *                                      label: 'Group 2'
 *                                  },
 *                                  {
 *                                      label: 'Item 1',
 *                                      uri: 'http://www.splunk.com'
 *                                  },
 *                                  {
 *                                      label: 'Item 2',
 *                                      uri: 'http://www.splunk.com'
 *                                  } 
 *                                ]                      
 *    
 */


Splunk.MenuBuilderGlobalCounter = 1;
Splunk.MenuBuilderCurrentGuid = 0;

Splunk.MenuBuilder = $.klass({
    _containerDiv    : null, //reference to the container
    _menuOpen         : null, //flag: is menu open
    _menuActivator    : null, //reference to menu activator
    _menuTimer        : null, //timer to be used for timeouts
    _timeoutDelay    : null, //default timeout delay
    _menu            : null, //reference for the menu  
    _guid            : null, //unique guid for each menubuilder instance
    
    // define the max number of characters to display per menu item   
    MAX_ITEM_LENGTH: 50,
    
    // number of milliseconds to animate fades
    FADE_DURATION: 100,
    
    initialize: function(options) {       
        this.logger = Splunk.Logger.getLogger("menu_builder.js");

        //set up menuOpening event broadcast and binding
        if (!Splunk.MenuBuilder.bodyOnclickLazilyInitialized) {
            $('body').click(function() {
                $('body').trigger("menuOpening");
            });
               
            Splunk.MenuBuilder.bodyOnclickLazilyInitialized = true;
        }

        $('body').bind('menuOpening', this.onMenuOpening.bind(this));

        //defaults
        this._menuOpen = false;
        this._timeoutDelay = 5000;
        this._menu = false;
        this._menuOpen = false;
        this._guid =  Splunk.MenuBuilderGlobalCounter++; //Math.random() * 100000;
       
        this._options = {
            containerDiv  : false,
            menuDict      : false,
            activator     : false,
            menuOpenClass : false,
            menuClasses   : '',
            autoInit      : true,
            showOnInit    : false,
            fadeEffects   : false,
            filter        : false
        };
       
        // Set the options using the defaults
        if (options) $.extend(this._options, options);
       
        // Ensure there is a place to put the menu
        if ( this._options['containerDiv'] ) {
            this._containerDiv = $(this._options['containerDiv']); // note $() cast since it is assumed elsewhere
        } else {
            this.logger.error('Splunk.MenuBuilder: No container specified');
        }
       
        // Build the menu on initialize if asked to do so.
        if (this._options.autoInit) this.menuInit();
    },

    menuInit: function() {
        // build the dictionary into a menu
        this._menu = this.buildMenu(this._options['menuDict'], 1);

        // add menu class and any classes they specified
        var extraClasses;
        if ( this._options['menuClasses'] ) {
            extraClasses = this._options['menuClasses'];   
        } else {
            extraClasses = "splMenu-primary";   
        }
        this._menu.addClass('splMenu ' + extraClasses);

        // attach menu to container
        this._containerDiv.append(this._menu);

        // used to reduce the .bind(this) calls
        var moduleInstance = this;

        // attach click to menu activator
        if ( this._options['activator'] ) {
            this._menuActivator = $(this._options['activator']); // again, cast with $() to be sure. this also allows jquery selectors to be valid
            this._menuActivator.click( function(evt){
                evt.stopPropagation();
                moduleInstance._onActivatorClick(evt);
                return false;
            });
        } else {
            this.logger.error('Splunk.MenuBuilder: No activator specified');
        }

        //add hovers for menu
        this._menu.hover(function() {
                clearTimeout(moduleInstance._menuTimer);
            },
            function() {
                clearTimeout(moduleInstance._menuTimer);
                moduleInstance._menuTimer = setTimeout( function() {
                        moduleInstance.hideMenu();
                    }, moduleInstance._timeoutDelay);
            }
        );

        //add top-level click event handler
        $(this._menu).bind('click', function(evt){
            this.onMenuClick(evt);
        }.bind(this));

        // hard stopping clicks from escaping the html block in menus.  for some reason testing if we're in the menu block in a top-level handler fails to catch all clicks
        $('.htmlBlock', this._menu).click(function(evt){
             evt.stopPropagation();
        });

        //add submenu arrows span and hover actions
        this._menu.children('ul').children('div').children('li:has(div.outerMenuWrapper)')    //all list items in the first-level that have submenus
            .hover( function(){
                clearTimeout(moduleInstance._menuTimer);
                moduleInstance.menuOver(this);
            }, function(){
                clearTimeout(moduleInstance._menuTimer);
                moduleInstance._menuTimer = setTimeout(function(){ moduleInstance.menuOut(this) }.bind(this),400);
            });
        
        //show the menu now.
        if(this._options.showOnInit){
            setTimeout(this.showMenu.bind(this), 0);
        }
    },
    /* top-level click handler */
    onMenuClick: function(evt){
        var t = evt.target;

        //stop clicks on submenu parent items
        if ( $(t).is('a') && $(t).parent('li.hasSubMenu').length ) {
            evt.stopPropagation();
            return false;
        } else {
            $('a', t).click();
        }
    },
    /* function for hovering over an li containing a submenu */
    menuOver: function(orig) {
        this.hideSubMenus(orig); 
        
        $(orig).addClass('sfhover');
       
        var submenu = $(orig).children('div.outerMenuWrapper');

        var hangRight = ($(orig).offset().left + $(orig).width() + submenu.width() > $(window).width())
        var submenuLeft = (hangRight) ? -submenu.width() : $(orig).parent('div').parent('ul').width();
        submenu.css({left:submenuLeft});

        submenu.show();
    },
   
    /* function for mouseout of submenu, hides submenu */
    menuOut: function(orig) { 
        $(orig).removeClass('sfhover');
        $(orig).children('div.outerMenuWrapper').hide();
    },
   
    /* build menu structure from the JSON dict passed */
    buildMenu: function(menuDict, menuLevel) {
        var menu = $('<div class="outerMenuWrapper splShadow splMenuGuid-' + this._guid + '"><ul></ul></div>');
        $.each(menuDict, function(i,menuitem) {

            // If a filter is defined for this menu, show only menu items that have that filter set in their 'showInFilter' property.
            if (this._options.filter !== false) {
                if ( menuitem.hasOwnProperty("showInFilter") && (menuitem.showInFilter != undefined) ) {
                    var l = menuitem.showInFilter.length;
                    var hasMatch = false;
                    for ( var i=0; i<l; i++ ) {
                       if ( menuitem.showInFilter[i] == this._options.filter ) hasMatch = true;
                    }
                    if ( !hasMatch ) return true; // If nothing matches, onto the next...
                } else {
                    return true; // If the menu has a filter set and the menuitem does not, toss out the menuitem.
                }
            }

            var itemClasses = '';
       
            // apply any styles to this item if specified in the menu dictionary
            if (menuitem.hasOwnProperty('style')){
                itemClasses += menuitem.style;
            }

            if ( menuitem.hasOwnProperty('divider') ) {  // NOTE: Need to figure this out, I don't really like this approach.  we need a way to specify a divider element,
                                                         //       but not tie implementation to the JSON dict
                newNode = $(['<div class="',menuitem.divider,' ',itemClasses,'"></div>'].join(''));  // currently using a div with the class passed
            } else if ( menuitem.hasOwnProperty('menuType') && menuitem.menuType == 'htmlBlock') { // htmlBlocks are to drop in the contents of a div, or possibly html returned remotely via ajax
                if ( menuitem.hasOwnProperty('element') ) {
                    // html block to place in menu comes from an existing element with class elementClass
                    // grab the element and shove it into the menu structure
                    newNode = $('<li class="' + itemClasses + ' htmlBlock"></li>');

                    if ( menuitem.hasOwnProperty('action') && menuitem.action == 'clone' ) { //clone element and add it in
                        $(menuitem.element).clone(true).appendTo(newNode);
                    } else { //actually insert the dom node itself
                        $(menuitem.element).appendTo(newNode);
                    }
                }
            } else {
                var label = '';

                //NOTE: again, probably not the best approach to this, we need a way to specify grouping headers.
                if ( menuitem.hasOwnProperty('menuType') && menuitem.menuType == 'grouping') { //treat label as a header, further uls as the items to be grouped
                    label = $('<p class="splunkMenuGroupingHeader">').text(menuitem.label);
                } else {

                    var href = 'javascript:void(0);';
                    var attrs = '';
                   
                    if (menuitem.hasOwnProperty('uri')){
                        href = menuitem.uri;
                    }
                    label = $('<a>')
                        .attr("href", href)
                        .addClass("menuItemLink")
                        .text(Splunk.util.smartTrim(menuitem.label, this.MAX_ITEM_LENGTH));

                    if (menuitem.hasOwnProperty('attrs')) {
                        for (var key in menuitem.attrs) {
                            label.attr(key, menuitem.attrs[key]);                               
                        }
                    }
                    
                    if (menuitem.hasOwnProperty('popup')) {
                        label.click(function(){
                            Splunk.window.open(this.href, menuitem.popup);
                            return false;
                        });
                    }

                    if (menuitem.hasOwnProperty('remove')) {
                        label.click(function(){
                            this.removeMenu();
                        }.bind(this));
                    }
                    
                }
                
                var newNode = $('<li class="' + itemClasses + '">').append(label);

                

                if ( menuitem.hasOwnProperty('items') ) {
                    newNode
                          .append('<span class="splIcon splIcon-triangle-4-e"></span>')
                      .addClass('hasSubMenu')
                      .append( this.buildMenu(menuitem.items) );
                }
           
                // Allow callbacks using jQuery's typical callback binding ('this' refers to calling element in callback)
                if (menuitem.hasOwnProperty('callback')) {
                    newNode.click(function(event){
                        this.hideMenu(menuitem.callback, event);
                        event.stopPropagation();
                        return false;
                    }.bind(this));
                }
            }
               menu.children('ul').append(newNode);
        }.bind(this));

        // inner wrapper element for extra styling needs.  unfortunate, but necessary for additional styling options
        menu.children('ul').children('*').wrapAll('<div class="innerMenuWrapper"></div>');

        return menu;
    },
   
    /**
     * Removes the menu and deactivates the activator control.
     *
     */
    removeMenu: function() {
        // Remove click actions from the menu
        this._menuActivator.unbind('click');
       
        // Remove the event handlers from the menu itself
        this._menu.unbind('click').unbind('hover');
       
        // kill the menu
        this._menu.remove();
    },
   
    /**
     * Removes the current menu and builds a new one
     * using a new menu dictionary.
     *
     * TODO better updating of menu items without rebuilding the entire menu structure
     */
    updateMenu: function(menuDict) {
        this.removeMenu();
        this._options.menuDict = menuDict;
        this.menuInit();
    },
   
    /* function to position and show menu */
    showMenu: function() {
        //position it
        this.setPosition();

        //using Brandon Aaron's bgiframe plugin to fix select element bleed-through in IE6
        this._menu.bgiframe();
      
        this._menuOpen = true;
        
        Splunk.MenuBuilderCurrentGuid = this._guid;

        if ( this._options["menuOpenClass"] ) {
            this._menuActivator.addClass(this._options["menuOpenClass"]);
        } else {
            this._menuActivator.addClass('menuOpen');
        }

        if ( this._options['fadeEffects'] ) {
            this._menu.fadeIn(this.FADE_DURATION);
        } else {
            this._menu.show();
        }
    },
   
    setPosition: function() {
        var t, o, posX, posY, 
            wst, //window scroll top
            wsl, //window scroll left
            ww = $(window).width(),
            wh = $(window).height(),
            height = this._menu.outerHeight(),
            width = this._menu.width();

        //basic positioning
        t = $(this._menuActivator);
        o = t.position();  
        posX = o.left;
        posY = o.top + t.outerHeight();

        var positionedParent = t.offsetParent();

        if ( $(positionedParent).css('overflow') == 'hidden') {  // check we're not gettng cut off by a positioned parent with overflow hidden
            //x-pos
            pw = $(positionedParent).width();
            if ( $.fn.scrollLeft )
            {           
                psl = $(positionedParent).scrollLeft();
                if ( pw + psl < posX + width )  //offscreen to the right?
                {
                    posX = o.left - width + t.outerWidth(); //fine, open to the left instead
                }
            }
            //y-pos
            ph = $(positionedParent).height();
            if ( $.fn.scrollTop )
            {           
                pst = $(positionedParent).scrollTop();
                if ( ph + pst < posY + height )  //offscreen at the bottom?
                {
                    if ( posY > height ) {
                        posY = o.top - height + t.outerHeight(); //fine, open on top instead
                    } else { // this sucks. the positioned parent is shorter than the menu, and is overflow:hidden (clipping).  do gymnastics  
                        if ( $.browser.msie && $.browser.version == "6.0" ) {
                           $(positionedParent).css('padding-bottom',height);   
                        } else {
                            $(positionedParent).css('min-height', posY + height);
                        }
                        if ( $.browser.mozilla && $.browser.version.substr(0,3) < "1.9" ){ // this is a fix for a bug in ff2 (187435)
                            this._menu.css('overflow','auto');   
                        }
                    }
                }
            }        
        } else {
            //x-pos  NOTE: need better positioning options
            if ( $.fn.scrollLeft )
            {           
                wsl = $(window).scrollLeft();
                if ( ww + wsl < posX + width )  //offscreen to the right?
                {
                    posX = o.left - width + t.outerWidth(); //fine, open to the left instead
                }
            }
        }
        
        
        

        //set the position
        this._menu.css({top:posY, left:posX});
    },
   
    /* function to hide all submenus
            NOTE: not sure we want to go down this path, where everything is so heavily DOM-dependent.
     */
    hideSubMenus: function(orig) {
        this._menu.find('.hasSubMenu').each(  //loop over all menu items that have submenus
            function(){
                if ( this != orig ) { //if this hover is triggered on the currently open menu, don't hide it's submenus
                   $(this).children('div.outerMenuWrapper').hide();
                }
            }
        );
    },
   
    /* function to hide the menu */
    /* @param {function} callback - if they assigned a callback to a menu item, fire it after the menu closes */
    /* @param {Object} event - a JQuery normalized event */
    hideMenu: function(callback, event) {
        this._menuOpen = false;
        if ( this._options["menuOpenClass"] ) {
            this._menuActivator.removeClass(this._options["menuOpenClass"]);
        } else {
            this._menuActivator.removeClass('menuOpen');
        }
        this.hideSubMenus();
        
        // if this menu is being closed because of another menu being opened,
        // do a fast hide
        if (!this._menuOpen && !callback)
            this._menu.fadeOut(this.FADE_DURATION);
        else {
            this._menu.hide();   
            return callback(event);
        }
    },
   
    /* function to handle broadcasted 'menuOpening' event */
    onMenuOpening: function(e, clickedGUID) {
        if ( this._guid == clickedGUID ) {
            //return false;
        } else {
            this.hideMenu();
        }
    },
   
    /* function to handle click on the menu activator */
    _onActivatorClick: function(evt) {
        if ( !this._menuOpen ) {
            $('body').trigger('menuOpening', [this._guid]);
            this.showMenu();
        } else {
            this.hideMenu();
        }
    },
      
    /* getter for a handle on the menu */
    getMenu: function(){
        return this._menu;
    },
   
    /* getter for the menu activator */
    getActivator: function () {
        return this._menuActivator;
    }
})


