/* Popup building class
 *
 *   @param {Object} contents A jQuery element reference to the content to inject
 *   @param {Object} (Optional) options An optional object literal having non-required settings.
 *
 *    Usage:
 *       var popupHandle = new Splunk.Popup({ options array })
 *
 *    Options:
 *        title: (string) Title to display in the header bar of the popup (ex: 'My Popup')
 *        contents: (JQuery object or html) The contents of the popup.  (ex: $('.popupContainer')).  NOTE: if html is used, cloneFlag should be set to false
 *        pclass: (string) Class to apply to the popup, used mainly for styling (ex: 'popupClass')
 *        cloneFlag: (bool) Flag to tell popup whether to clone the passed contents into the popup (true) or move them.  This should generally only be left as true when
 *                          passing a reference to an existing element that you wish to leave untouched, for example a dom scaffolding you intend to reuse.
 *        buttons: (array/object lit) Array to define buttons and their handlers.
 *                 properties:
 *                    label: (string) button text to display (ex: 'Ok')
 *                    type: (string) type of button.  expected value is either 'primary' or 'secondary'
 *                    callback: (function) callback to fire on button click.  callback should return true to close popup, false to leave it open.
 *        inlineMode: setting this to true invokes 'minimally invasive' mode. see more explanation below
 *
 *    Methods:
 *        getPopup()
 *            desc: getter for a reference to the popup container
 *
 *        destroyPopup()
 *            desc: destroys popup
 *
 *
 *    Example usage:
 *       var popup = new Splunk.Popup($('.popupScaffolding'), {
 *           title : "My Popup",
 *           pclass : "myPopupClass",
 *           buttons : [
 *               {
 *                   label: 'Cancel',
 *                   type : 'secondary',
 *                   callback: function(){
 *                       return this.popupCancel();
 *                   }.bind(this)
 *               },
 *               {
 *                   label: 'Apply',
 *                   type : 'primary',
 *                   callback: function(){
 *                       return this.popupAccept();
 *                   }.bind(this)
 *               }
 *           ]
 *       });
 *
 *      Inline Mode :
 *              In inlineMode, Popup will provide no chrome whatsoever, the only services it provides are the modal transparent overlay,
 *              popup positioning and showing, and closing the popup.  If the dom element provided (popup) has any item with class="splIcon-close",
 *              Popup will attach a close handler to it, as well as the default binding to esc key presses.  destroyPopup simply hides the dom with display:none;
*/

Splunk.Popup = $.klass({
    
    // reference to DOM instance
    _popup: null,
    
    // indicates if popup is currently active
    isActive: false,
    
    initialize: function(contents, options) {
        this.logger = Splunk.Logger.getLogger("popup_manager.js");

        if ( !contents ) {
            this.logger.error('Splunk.Popup: No popup contents specified');
        }

        this.popupContents = contents;

        //defaults
        this._options = {
            title       : '',
            pclass      : false,
            cloneFlag   : true,
            buttons     : false,
            inlineMode  : false,
            onDestroy   : function() {}
        };

        // Set the options using the defaults
        if (options) $.extend(this._options, options);

        //check they passed a title and contents
        if ( !this._options['title'] ) {
            this.logger.warn('splunk.Popup: No popup title specified, leaving blank');
        }

        //create the overlay
        this.createOverlay();

        //create the popup
        if ( this._options['inlineMode'] ) {
            this._popup = this.popupContents;
        } else {
            this._popup = $('<div class="popupContainer"></div>')
                .append('<div class="splHeader splHeader-popup"><a href="javascript:void(0)" class="splIcon splIcon-close"></a><h2>' + this._options['title'] + '</h2></div>')
                .append('<div class="popupContent"></div>')
                .append('<div class="popupFooter"></div>')
                .prependTo('body');

            //add class for styles if they specified one
            if ( this._options['pclass'] ) {
                this._popup.addClass( this._options['pclass'] );
            } else {
                this._popup.addClass('ieBugFixer'); //this fixes a crazy ie6 bug, noted and a better solution is being investigated
            }

            //set the contents
            this.setPopupContents(this.popupContents);

            //set up buttons
            if ( this._options['buttons'] ) {
                this.setPopupButtons(this._options['buttons']);
            }
        }

        //position it
        this.positionPopup();

        //ok, show the damn thing
        $(this._popup).css({visibility:'visible',display:'block'});

        //make the popup draggable
        if ( $(this._popup).children('.splHeader-popup').length ) {
            $(this._popup).draggable({
                handle: ".splHeader-popup"
            });
        }

        // setup event handlers
        this._setupEventHandlers();

        // fire generic onPopupLoaded function
        // this.onPopupLoaded();

        // flag as active
        this.isActive = true;
        
        Splunk.Popup._globalPopupCount += 1;
        
        return this._popup;
    },
    /* deprecated, leaving to warn legacy code */
    createPopup: function(popupParent, popupTitle, popupContents, handleOK, handleCancel, handleAccept) {
       this.logger.error('splunk.Popup: Function signature changed, please see /exposed/js/popup.js for details');
    },
    _setupEventHandlers: function(){
         // handle close button click
        $(this._popup).find('.splHeader .splIcon-close').click(function(){
            this.destroyPopup();
            return false;
        }.bind(this));

        // Create an escape key binding to close the popup
        $(document).bind('keydown.Popup', function(event) {
            if (event.keyCode == 27) {
                this.destroyPopup();
            }
        }.bind(this));
    },
    setPopupContents: function(popupContents){
            //this is terribly destructive, throws away what was in the popup if there is anything then pushes the new content in.  dangerous...
            $(this._popup).find('.popupContent').empty();

            var newContents = $(popupContents);

            if ( this._options['cloneFlag'] ) {
                //$('.popupContent', this._popup).append(newContents.clone(true));
                newContents = $(popupContents).clone(true).appendTo($(this._popup).find('.popupContent')).show();
            } else {
                //$('.popupContent', this._popup).append(newContents);
                newContents = $(popupContents).appendTo($(this._popup).find('.popupContent')).show();
            }

             //for IE, set the width explicitly
            if ( $.browser.msie ) {
                this.setPopupDimensions(newContents);
            }
    },
    setPopupButtons: function(buttons) {
        var context = this;

        $.each(buttons, function(i,button) {
            if ( button.hasOwnProperty('label') ) {
                buttonText = button.label;
            } else {
                buttonText = '';
            }
            if ( button.hasOwnProperty('type') ) {
                buttonType = button.type;
            } else {
                buttonType = 'primary';
            }
            if ( button.hasOwnProperty('callback') ) {
                buttonHandler = button.callback;
            } else {
                context.logger.warn('Splunk.Popup: Button has no callback defined');
                buttonHandler = function(){ return true };
            }

            context.addButton(buttonText, buttonType, buttonHandler);
        });

    },
    addButton: function(buttonText, buttonType, buttonHandler) {
        buttonType = buttonType.toLowerCase();

        var newbutton = $('<button></button>')
            .addClass('splButton-' + buttonType)
            .append('<span>' + buttonText + '</span>')
            .appendTo($(this._popup).children('div.popupFooter'))
            .click(function(){
                if ( buttonHandler() ) {
                    this.destroyPopup();
                }
            }.bind(this));
    },
    createOverlay: function(popupParent) {
        if (Splunk.Popup._globalPopupCount > 0) return;
        $('<div class="splOverlay"></div>')
            .prependTo('body').show();

        // For FF2 mac only, there are at least two bugs with flash, opacity, and position:fixed items.  this fixes it by hiding the flash timeline.
       if ( $.browser.mozilla && $.browser.version.substr(0,3) < "1.9" ){
               $('.FlashWrapperContainer').css('z-index','-1');
        }

        if ( $.browser.msie && $.browser.version == '6.0' ) {
            var h, w;
            h = $('body').height();
            w = $('body').width();
            $('.splOverlay')
                .css({height: h, width: w})
                .bgIframe();
        }
    },
    destroyOverlay: function() {
        if (Splunk.Popup._globalPopupCount > 0) return;
        $('.splOverlay').remove();
    },
    setPopupDimensions: function(popupContents) {
        //var popupContentsHeight = $(popupContents).outerHeight();
        var popupContentWidth = $(popupContents).outerWidth();

        //var popupChromeHeight = parseInt($(this._popup).find('.splHeader-popup').outerHeight(),10) + parseInt($(this._popup).find('.popupFooter').outerHeight(),10);

        //var totalPopupHeight = parseInt(popupContentsHeight, 10) + parseInt(popupChromeHeight,10);

        //$(this._popup).css({width:popupContentWidth});
        $(this._popup).css({width:popupContentWidth});
    },
    positionPopup: function() {
        var posX, posY,
            height = this._popup.outerHeight(),
            width = this._popup.outerWidth(),
            wst, //window scroll top
            wsl, //window scroll left
            ww = $(window).width(),
            wh = $(window).height();

        //basic positioning top left if all else fails
        posX = 0;
        posY = 0;

        //check for any problems (window shorter than popup, etc.)

        //y-pos
        if ( $.fn.scrollTop )
        {
            wst = $(window).scrollTop();
            if ( wh > height ) //verify window is taller than the popup
            {
                posY = (wh - height)/2;
            }
        }

        //x-pos
        if ( $.fn.scrollLeft )
        {
            wsl = $(window).scrollLeft();
            if ( ww > width )  //verify window is wider than the popup
            {
                posX = (ww - width)/2;
            }
        }

        //set the position
        this._popup.css({top:posY, left:posX});
    },
    onPopupLoaded: function() {
        if ( $('#runMe').length ) {
           eval($('#runMe').text());
        }
    },
    //getter for popup container
    getPopup: function() {
        return this._popup;
    },
    destroyPopup: function(){
        Splunk.Popup._globalPopupCount -= 1;

        if ( this._options['inlineMode'] ) {
            $(this._popup).hide();
        } else {
            $(this._popup).remove();
        }
        this.destroyOverlay();

		if ( $.browser.mozilla && $.browser.version.substr(0,3) < "1.9" ){
               $('.FlashWrapperContainer').css('z-index','auto');
        }
        
        this.isActive = false;

        //remove handlers
        $(document).unbind('keydown.Popup');

        // Fire the pseudo callback
        if (this._options['onDestroy']) this._options['onDestroy']();
    }
})

Splunk.Popup.createExportResultsForm = function(formContainer, searchId) {
    // another weird variable that we can pass in as a closure.
    // see comment in createSavedSearchForm.  This can be eliminated by
    // making the popup class pass the button callbacks an argument
    // thats a reference back to the container div of the popup contents.
    var exportPopupHandle = null;
    var exportPopup = new Splunk.Popup(formContainer, {
        title: _('Export Results'),
        buttons: [
            {
                label: _('Cancel'),
                type: 'secondary',
                callback: function(){
                    return true;
                }
            },
            {
                label: _('Export'),
                type: 'primary',
                callback: function(){
                    // Export dialog functions
                    return $(exportPopupHandle).find(".exForm").submit();
                }
            }
        ]
    });
    exportPopupHandle = exportPopup.getPopup();
    var exportForm = $(exportPopupHandle).find(".exForm")[0];
    exportForm.action = Splunk.util.make_url('api/search/jobs/' + searchId + '/results');
}

Splunk.Popup.createEventtypeForm = function(formContainer, formTitle, eventtypeSearch, tags) {
   // Taken straight from the TitleBar.js file.
   // Get the form from EAI
   

   // We need to keep the html for the "Loading..." popup separate from the actual
   // form popup so that we never copy/clone the original form we get from the endpoint
   // to keep any data bound to tags by javascript in the form intact.
   var loadingContainer = $('<div></div').appendTo(formContainer);
   var loadedFormContainer = $('<div style="display:block"></div>').appendTo(formContainer);

   // Insure that the intermediate popup always says loading
   // This is a bit of a necessary hack... Right now EAI forms come with
   // some js that gets interpreted when the form is requested asynchronously.
   // By not resetting this html content, the form appears to be preloaded,
   // but may not be correctly rendered because the js that originally came
   // with it is not re-evaluated.
   $(loadingContainer).html('<div class="popupLoading">Loading...</div>');
   
   if ($.browser.msie) {
       $(loadingContainer).css({"overflow":"", "height":""});
   }
   
   // Create the popup just for now.
   var eventtypePopup = new Splunk.Popup(loadingContainer, {
       title: formTitle,
       pclass: 'TitleBarEventtypePopup',
       cloneFlag: false,
       onDestroy: function() {
           Splunk.Popup.createEventtypeForm.opened = false;
       }
   });
   
   // Hide the footer in the intermediary popup
   $('.popupFooter', eventtypePopup.getPopup()).hide();
   
   var messenger = Splunk.Messenger.System.getInstance(); 

   // Cancel handler for the form
   var cancel = function(){ return true; };
   
   // Save button handler for the form
   var handleSaveSubmit = function() {
       var form = $('form.entityEditForm', eventtypePopup.getPopup());
       var eai = Splunk.EAI.getInstance();
       eai.redirect = false;

       // pull out the name the user gave it, cause we pass that as the argument
       // to the client's callback, which is called if the POST succeeds.
       var name = $('input[name="name"]', form).val();
       eai.success_message = sprintf(_("Your eventtype '%(eventtypeName)s' was saved."), {eventtypeName: name});

       // Alert the user that their save is being processed
       $('div.popupFooter button.splButton-primary span', this._popup).text(_('Saving...'));
       $('div.popupFooter button.splButton-primary', this._popup).unbind('click').removeClass('primary').addClass('secondary');

       // submit the form and close the popup
       eai.saveForm(form, 
          // Success
           function() {
               eventtypePopup.destroyPopup();
           },
           // Error; leave the popup on screen so the user can correct the problem
           function() {
               $('div.popupFooter button.splButton-primary span', this._popup).text(_('Save'));
               $('div.popupFooter button.splButton-primary', this._popup).bind('click', handleSaveSubmit).removeClass('secondary').addClass('primary');
           }
       );
       return false;
   };
   
   // Sets the savedSearchPopup up correctly
   var setupPopup = function() {
       // Kill the "Loading..." popup
       if (eventtypePopup) eventtypePopup.destroyPopup();

       eventtypePopup = new Splunk.Popup(loadedFormContainer, {
           title: formTitle,
           pclass: 'TitleBarEventtypePopup',
           cloneFlag: false,
           onDestroy: function() {
               Splunk.Popup.createEventtypeForm.opened = false;
           },
           buttons: [
               {
                   label: _('Cancel'),
                   type: 'secondary',
                   callback: cancel
               },
               {
                   label: _('Save'),
                   type: 'primary',
                   callback: handleSaveSubmit
               }
           ]
       });

       // Set the focus to the first element in the form
       $('form.entityEditForm input[name="name"]', eventtypePopup.getPopup()).focus();

       // We want error messages to stay inside the popup instead of going 
       // to to the masked messenger at the top of the page
       Splunk.EAI.getInstance().use_messenger_on_error = false;
       
       if (eventtypeSearch) {
           $('form.entityEditForm textarea[name="search"]', eventtypePopup.getPopup()).val(eventtypeSearch);
       }
       
       if (tags) {
           $('form.entityEditForm input[name="tags"]', eventtypePopup.getPopup()).val(tags);           
       }

       // TODO - this is probably good functionality across all popups,
       // catch form.onsubmit(which errant return-keypresses can trigger)
       // and bind it to the callback for the 'primary' action button
       // if there's only one, and return false on it, if there's more than one.
       // Override the onsubmit handler for the form as it will be triggered explictly by our handler
       $("form", loadedFormContainer).unbind('submit');
       $("form", loadedFormContainer).submit(handleSaveSubmit);
   };

   // Modal marked as opened
   Splunk.Popup.createEventtypeForm.opened = true;

   var form_url = Splunk.util.make_url('manager', Splunk.util.getCurrentApp(), '/saved/eventtypes/_new?action=edit&noContainer=2&viewFilter=modal&eleOnly=1');

   // Get the form from EAI
   $.ajax({
       url: form_url,
       dataType: 'html',
       error: function() {
           if (eventtypePopup) eventtypePopup.destroyPopup();
           messenger.send('error',
               'splunk.eventtype',
               _("Splunk encountered an error when it attempted to retrieve the eventtype form. Try again or contact an admin.")
           );
       },

       complete:function(resp, status) {
           // Javascript contained in the form is executed at this point
           $(loadedFormContainer).html(resp.responseText);

           /***************
            * Resizing the form element in THIS FORM'S CASE ONLY.
            * if we need this elsewhere, this logic should be pulled up into popup class itself.
            ***************/
           var safeHeight = $(window).height() - 200;
           if ($.browser.msie) {
               // overflow:scroll fixes a bug that could only be reproduced on IE8 on zhang's machine.
               $(loadedFormContainer).css("overflow", "scroll")
               if ((parseInt($.browser.version) <7) ) {
                   $(loadedFormContainer).css("height", safeHeight + "px");
               } else {
                   $(loadedFormContainer).css("max-height", safeHeight + "px");
               }
           } else {
               $(loadedFormContainer).css("overflow", "auto")
               $(loadedFormContainer).css("max-height", safeHeight + "px");
           }

           // Initialize the actual popup
           setupPopup();
       }
   });
};


Splunk.Popup.createSavedSearchForm = function(formContainer, title, searchContext) {

    // Prevent multiple popups
    if (Splunk.Popup.createSavedSearchForm.opened) return;
    
    var logger = Splunk.Logger.getLogger("Splunk.Popup.createSavedSearchForm");
    
    // We need to keep the html for the "Loading..." popup separate from the actual
    // form popup so that we never copy/clone the original form we get from the endpoint
    // to keep any data bound to tags by javascript in the form intact.
    var loadingContainer = $('<div></div').appendTo(formContainer);
    var loadedFormContainer = $('<div style="display:block"></div>').appendTo(formContainer);

    // Insure that the intermediate popup always says loading
    // This is a bit of a necessary hack... Right now EAI forms come with
    // some js that gets interpreted when the form is requested asynchronously.
    // By not resetting this html content, the form appears to be preloaded,
    // but may not be correctly rendered because the js that originally came
    // with it is not re-evaluated.
    //$(formContainer).html('<div class="popupLoading">Loading...</div>');
    $(loadingContainer).html('<div class="popupLoading">Loading...</div>');
    if ($.browser.msie) {
        $(loadingContainer).css({"overflow":"", "height":""});
    }

    // for now I have to define this var out here, cause it gets bound in closures
    // in our callbacks, and it does an evil snake-eating-own-tail thing.
    // TODO - it would be a lot nicer and I think quite sensible to rework the popup.js callbacks
    // so that they are given an argument that is a reference to the popup's form object.
    // after which the need for this messy closure would go away.

    // Create the popup just for now.
    var savedSearchPopup = new Splunk.Popup(loadingContainer, {
        title: title,
        pclass: 'TitleBarSavedSearchPopup',
        cloneFlag: false,
        onDestroy: function() {
            Splunk.Popup.createSavedSearchForm.opened = false;
        }
    });

    // Hide the footer in the intermediary popup
    $('.popupFooter', savedSearchPopup.getPopup()).hide();

    var messenger = Splunk.Messenger.System.getInstance(); 

    // Cancel handler for the form
    var cancel = function(){ return true; };

    // Save button handler for the form
    var handleSaveSubmit = function() {
        var form = $('form.entityEditForm', savedSearchPopup.getPopup());
        var eai = Splunk.EAI.getInstance();
        eai.redirect = false;

        // Check that the form's submit handler gives the go ahead to submit the form
        //if (!handleEditFormSubmit(form[0]))
            //return false;

        // first write out the shared viewstate and get the viewstate id;
        // inject the ID into the EAI form
        var vs_id = Splunk.Globals.ModuleLoader.commitViewParams(null, true);
        var vs_input = $('input[name="vsid"]', form);
        if (!vs_input) {
            context.logger.error('handleSaveSubmit - Cannot find vsid element; viewstate will not be saved');
        } else {
            vs_input.val(vs_id);
        }
        // pull out the name the user gave it, cause we pass that as the argument
        // to the client's callback, which is called if the POST succeeds.
        var name = $('input[name="name"]', form).val();
        eai.success_message = sprintf(_("Your search '%(savedSearchName)s' was saved."), {savedSearchName: name});

        // Alert the user that their save is being processed
        $('div.popupFooter button.splButton-primary span', this._popup).text(_('Saving...'));
        $('div.popupFooter button.splButton-primary', this._popup).unbind('click').removeClass('primary').addClass('secondary');

        // submit the form and close the popup
        eai.saveForm(form, 
            function() {
                // Success
                Splunk.Globals.ModuleLoader.refreshViewData();
                if(window.opener && !window.opener.closed){
                    try{
                        window.opener.Splunk.Globals.ModuleLoader.refreshViewData();
                    }catch(e){
                        logger.warn("Could not refresh view data for parent opener", e);
                    }
                }
                savedSearchPopup.destroyPopup();
            },
            function() {
                // Error; leave the popup on screen so the user can correct the problem
                // scroll to the top so they can see the error
                loadedFormContainer.scrollTop(0);
                
                $('div.popupFooter button.splButton-primary span', this._popup).text(_('Save'));
                $('div.popupFooter button.splButton-primary', this._popup).bind('click', handleSaveSubmit).removeClass('secondary').addClass('primary');
            }
        );
        return false;
    };

    // Sets the savedSearchPopup up correctly
    var setupPopup = function() {
        // Kill the "Loading..." popup
        if (savedSearchPopup) savedSearchPopup.destroyPopup();

        savedSearchPopup = new Splunk.Popup(loadedFormContainer, {
            title: title,
            pclass: 'TitleBarSavedSearchPopup',
            cloneFlag: false,
            onDestroy: function() {
                Splunk.Popup.createSavedSearchForm.opened = false;
            },
            buttons: [
                {
                    label: _('Cancel'),
                    type: 'secondary',
                    callback: cancel
                },
                {
                    label: _('Save'),
                    type: 'primary',
                    callback: handleSaveSubmit
                }
            ]
        });

        // Set the focus to the first element in the form
        $('form.entityEditForm input[name="name"]', savedSearchPopup.getPopup()).focus();

        // We want error messages to stay inside the popup instead of going 
        // to to the masked messenger at the top of the page
        Splunk.EAI.getInstance().use_messenger_on_error = false;

        if (searchContext) {
            var search = searchContext.getSearch();
            var timeRange = searchContext.getTimeRange();
            var earliestTime = timeRange.getEarliestTimeTerms();
            var latestTime = timeRange.getLatestTimeTerms();

            // pre-populate the search string if we were given one.
            if (search) {
                $('form.entityEditForm textarea[name="search"]').val(Splunk.util.stripLeadingSearchCommand(search));
            }

            if (earliestTime) {
                $('form.entityEditForm input[name="dispatch.earliest_time"]').val(earliestTime);
            }

            if (latestTime) {
                $('form.entityEditForm input[name="dispatch.latest_time"]').val(latestTime);
            }
        }

        // Save the dispatching view
        $('form.entityEditForm input[name="displayview"]', savedSearchPopup.getPopup()).val(Splunk.util.getCurrentDisplayView());

        $('form.entityEditForm input[name="request.ui_dispatch_view"]', savedSearchPopup.getPopup()).val(Splunk.util.getCurrentDisplayView());

        // TODO - this is probably good functionality across all popups,
        // catch form.onsubmit(which errant return-keypresses can trigger)
        // and bind it to the callback for the 'primary' action button
        // if there's only one, and return false on it, if there's more than one.
        // Override the onsubmit handler for the form as it will be triggered explictly by our handler
        $("form", loadedFormContainer).unbind('submit');
        $("form", loadedFormContainer).submit(handleSaveSubmit);
    };

    // Modal marked as opened
    Splunk.Popup.createSavedSearchForm.opened = true;

    // Get the form from EAI
    $.ajax({
        url: Splunk.util.make_url('manager', Splunk.util.getCurrentApp(), "/saved/searches/_new?action=edit&noContainer=1&viewFilter=modal&eleOnly=1"),
        dataType: 'html',
        error: function() {
            if (savedSearchPopup) savedSearchPopup.destroyPopup();
            messenger.send('error',
                'splunk.savedsearches',
                _("Splunk encountered an error when it attempted to retrieve the save search form. Try again or contact an admin.")
            );
        },

        complete:function(resp, status) {
            // Javascript contained in the form is executed at this point
            $(loadedFormContainer).html(resp.responseText);

            /***************
             * Resizing the form element in THIS FORM'S CASE ONLY.
             * if we need this elsewhere, this logic should be pulled up into popup class itself.
             ***************/
            var safeHeight = $(window).height() - 200;
            if ($.browser.msie) {
                // overflow:scroll fixes a bug that could only be reproduced on IE8 on zhang's machine.
                $(loadedFormContainer).css("overflow", "scroll")
                if ((parseInt($.browser.version) <7) ) {
                    $(loadedFormContainer).css("height", safeHeight + "px");
                } else {
                    $(loadedFormContainer).css("max-height", safeHeight + "px");
                }
            } else {
                $(loadedFormContainer).css("overflow", "auto")
                $(loadedFormContainer).css("max-height", safeHeight + "px");
            }

            // Initialize the actual popup
            setupPopup();
        }
    });
};

Splunk.Popup._globalPopupCount = 0;



Splunk.Popup.createShareLinkForm = function(formContainer, title, context) {
    var messenger = Splunk.Messenger.System.getInstance(); 

    // the callback to run if we successfully save the job.
    var onSaveSuccess = function() {

        var app = Splunk.util.getCurrentApp();
        var view = Splunk.util.getCurrentDisplayView();
        var args = {"sid" : context.job.getSearchId()};

        args["vs"] = Splunk.Globals.ModuleLoader.commitViewParams(null, true);

        url = [];
        url.push(document.location.protocol);
        url.push("//");
        url.push(document.location.host);
        url.push(Splunk.util.make_url("app", app, view));

        url.push("?" + Splunk.util.propToQueryString(args))
        $("input[name='linkText']", formContainer)
            .val(url.join(""))
            .focus(function() {this.select()})
            .focus();
        
        // this is called both by the cancel button on the popup layer
        // as well as by the little blue 'undo save'  link in the text.
        var undoEverything = function(event) {
            var undoSaveSuccess = function() {
                context.logger.debug("results are now unsaved again.");
            };
            var undoSaveFailure = function() {
                messenger.send('error', 'splunk.search', _("Splunk encountered an error while trying to undo the saving of your results. Try again or contact an admin."));
            }
            context.job.unsave(undoSaveSuccess, undoSaveFailure);
            
            var undoACLSuccess = function() {
                context.logger.debug("results are now again marked as NOT world readable.");
            }
            var undoACLFailure = function() {
                messenger.send('error', 'splunk.search', _("Splunk encountered an error while trying to make your results not world-readable. Try again or contact an admin."));
            }
            context.job.undoWorldReadable(undoACLSuccess, undoACLFailure);
        };
        // workaround - if i attach the undoEverything handler AFTER creating the popup, 
        // some cloning nonsense manages to remove it.
        // therefore i create a null reference, bind it, and only then fill it with a reference
        // to the Splunk.Popup instance.

        var linkPopup = null;
        $("a.undoEverything", formContainer).click(function() {
            undoEverything();
            linkPopup.destroyPopup();
        });
    
        var linkPopup = new Splunk.Popup(formContainer, {
            title: title,
            buttons: [
                /*{
                    label: _('Undo'),
                    type: 'secondary',
                    callback: function(){
                        undoEverything();
                        return true;
                    }
                },*/
                {
                    label: _('Close'),
                    type: 'primary',
                    callback: function(){
                        return true;
                    }
                }
            ]
        });
    };
    var onSaveFailure = function(e) {
        messenger.send('error', 'splunk.search', _("Splunk encountered an error when it tried to save your results. It cannot generate a link. Try again or contact an admin.") + e);
    }
    context.job.save(onSaveSuccess, onSaveFailure);

    var onACLSuccess = function() {
        context.logger.debug("results are now marked as world readable.");
    }
    var onACLFailure = function() {
        messenger.send('error', 'splunk.search', _("Splunk encountered an error when it tried to make your results world-readable. Try again or contact an admin."));
    }
    context.job.makeWorldReadable(onACLSuccess, onACLFailure);
}
// This is unfortunate but until we have a popup manager
// or we internalize popup state in the popup class we'll
// have to manage open/closed states manually.
Splunk.Popup.createSavedSearchForm.opened = false;
Splunk.Popup.createEventtypeForm.opened = false;
