//
// profiler.js - All-platform function profiling
//

Function.prototype.profile = function() {

    var __method = this, args = $A(arguments), name = args.shift();
    if (__method.isProfiled) return __method;

    var fn = function() {
        var start = new Date().getTime();
        var rv = __method.apply(this, args.concat($A(arguments)));
        Profiler.recordExecution(__method, name, new Date().getTime() - start);
        return rv;
    }
    if (__method.prototype) fn.prototype = __method.prototype;
    fn.isProfiled = true;
    return fn;
}

var Profiler = {

    statistics: {},
    logger: Splunk.Logger.getLogger("profiler.js", Splunk.Logger.mode.Firebug),
    start: function()
    {
        this.profilerRunning = false;
        alert('Starting profiler: please wait...');

        setTimeout(this._profileAll.bind(this), 1000);
    },

    _profileAll: function()
    {
        // register all functions in our namespace
        var namespace = globalVars.getOwn();

        // add some native objects that we'd like to profile
        //var natives = ['String', 'Array', 'Number', 'Date', 'Math', 'RegExp'];
        //natives.each(function(n) { namespace[n] = eval(n); });

        for (var name in namespace) // first, profile all prototyped methods
        {
            var global = namespace[name];
            if ((typeof global == 'function') && global.prototype)
            {
                //this.logger.debug('global is ' + name);
                this._profileObject(name + '.prototype', global.prototype);
            }
        }

    for (var name in namespace) // then, profile all non-prototyped methods of objects in the namespace, if they haven't already been profiled.
    {
        var global = namespace[name];
        switch(typeof global)
        {
            case 'function':
            // profile the function
                switch(name)
                {
                    // ignore some that cause 'too much recursion' errors
                    case '$A':
                    case 'String':
                    case 'Array':
                    case 'Object':
                    case 'Number':
                    case 'Math':
                    case 'Date':
                    case 'RegExp':
                    break;
                    default:
                      if (eval(name + '.profile'))
                      {
                        eval(name + ' = ' + name + '.profile("' + name + '")');
                      }
                      break;
                }
                break;
               
            case 'object':
            // profile any functions the object holds
            switch(name)
            {
                case 'Profiler': break;
                default: this._profileObject(name, global); break;
            }
            break;
        }
    }

        this.profilerRunning = true;
        alert('Profiler is running... call Profiler.report() to see results.');
    },

    _profileObject: function(name, object)
    {
        if (object.isProfiled) return;
        object.isProfiled = true;

        for (var property in object)
        {
            switch(property)
            {
                case 'profile': break;
                default:
                    try
                    {
                        switch(typeof object[property])
                        {
                            case 'function':
                                if (object[property].profile) object[property] = object[property].profile(name + '.' + property);
                            break;
                        }
                    }   
                catch(e)
                {
                this.logger.debug('Profiler exception (probably normal): ' + e);
                }
                break;
            }
        }
    },

    recordExecution: function(fn, name, ms)
    {
        if (!this.profilerRunning) return;

        //if (console) console.log('recorded execution of ' + name + ' in ' + ms + ' ms');
        var rec;
        if (rec = this.statistics[fn])
        {
          rec[1]++;
          rec[2] += ms;
          if (ms < rec[3]) rec[3] = ms;
          if (ms > rec[4]) rec[4] = ms;
        }
        else
        {
          // Use indexed arrays rather than named members for speed (may help to mimimize Heisenberg problem)
          this.statistics[fn] = [name, 1, ms, ms, ms];  // 0: name, 1: calls, 2: total time, 3: min time, 4: max time
        }
    },

    report: function(comparedReport)  // comparedReport is optional
    {
        this.profilerRunning = false;
       

        var html = '<table style="font: 11px arial;"><tr><th style="text-align: left;">Function &nbsp; </th><th style="text-align: left;">Calls &nbsp; </th><th style="text-align: left;">Time &nbsp; </th><th style="text-align: left;">Avg &nbsp; </th><th style="text-align: left;">Min &nbsp; </th><th style="text-align: left;">Max &nbsp; </th></tr>';
        this.reportStats = [];

        for (var fn in this.statistics)
        {
          this.reportStats.push(this.statistics[fn]);
        }

        this.reportStats.sort(function(a, b) { return b[2] - a[2]; });

        var columnOrder = [0,1,2,5,3,4];
        var lim = this.reportStats.length;
        for (var i=0; i<lim; i++)
        {
          var item = this.reportStats[i];
          item[5] = parseInt(item[2] / item[1]);
          html += '<tr>'
          columnOrder.each(function(j){

            html += '<td>' + item[j];
            if ((j > 0) && comparedReport && comparedReport[item[0]])
            {
              var compared = comparedReport[item[0]][j];
           
              // bold the compared value if it's significantly different
              if (this._significantlyDifferent(item[j], compared))
              {
                html += ' <b>(' + compared + ')</b>';
              }
              else
              {
                html += ' (' + compared + ')';
              }
            }
            html += ' &nbsp; </td>';
          }.bind(this));
          html += '</tr>';
        }
        html += '</table>';

        if (!window.LOG_POPUP) setPopupLogger(true, true);
        this.logger.debug(html);
        this.logger.debug('<br><br>Profile JSON:<br>' + JSON.stringify(this.reportStats) + '<br><br>');
    },

    _significantlyDifferent: function(a, b)
    {
        return (Math.abs(a - b) / ((a + b) / 2)) > .2;  // if the difference is greater than this percentage of the average value
    },

    compare: function()
    {
        var json = prompt('Paste JSON from another report');
        var comparedReport = eval(json);
        try
        {
            // organize the compared report as a hash, keys are function names.
            var h = {};
            comparedReport.each(function(r){
                h[r[0]] = r;
            });
         
          this.report(h);
        }
        catch(e)
        {
          alert('Exception (is the pasted JSON valid?): ' + e);
        }
    }
};








/*
globalVars v 1.0
copyright 2006 Thomas Frank

This program is free software under the terms of the
GNU General Public License version 2 as published by the Free
Software Foundation. It is distributed without any warranty.

Note (jthane) This *INCREDIBLE* hack is necessary to enumerate the global namespace in IE.
See http://www.thomasfrank.se/global_namespace.html for details.
If the machines ever do take over, they are going to be pissed.
*/

globalVars={
    init:function(){
        this.initially=[];
        for(var i in window){this.initially[i]=true};
    },
    get:function(x){
        var win=window;
        if(navigator.userAgent.toLowerCase().indexOf("msie")>=0 && window["ActiveXObject"]){
            win=this.ieFix()
        };
        var obj={};
        var ffx=',addEventListener,location,document,navigator,event,frames,';
        for(var i in win){
            if(i=="fullScreen"){continue};
            var a=true;
            if(this.initially[i] && typeof window[i]=="function"){this.initially[i]=false};
            a=a && !(this.initially[i]);
            a=a && !(window[i]===undefined);
            a=a && !(window[i]===null);
            a=a && (i.indexOf("[[")<0);
            a=a && (window[i]+"").indexOf("[native code]")<0;
            var b=a;
            a=a && (window[i]+"").indexOf("[object HTML")!=0;
            if(a!=b){b=false} else {b=true};
            a=a && (window[i]+"").indexOf("[object Window")!=0;
            a=a && (window[i]+"").indexOf("[object]")!=0;
            a=a && ((window[i]+"")!="[function]");
            a=a && ((window[i]+"")!="(Internal Function)");
            a=a && (!(typeof window[i]=="function" && (window[i]+"").indexOf("<Logger")==0));
            a=a && (i!="NaN");
            a=a && (i!="0" && i!="1");       
            a=a && (i!="Infinity");
            a=a && (i!="Math");
            a=a && ffx.indexOf(','+i+',')<0;
            if(x && b){a=true};
            if (a){obj[i]=window[i]};
        };
        obj["globalVars"]=globalVars;
        if(window["onload"]){obj.onload=onload};
        return obj
    },
    getAll:function(){return this.get(true)},
    getOwn:function(){return this.get()},
    getNative:function(){
        var x=this.getAll();
        var y=this.getOwn();
        var z={};
        for(var i in x){if(y[i]===undefined){z[i]=x[i]}};
        return z
    },
    ieFix:function(){
        var allt="";
        var x=document.scripts;
        for (var i=0;i<x.length;i++){
            if (x[i].innerHTML){allt+=x[i].innerHTML}
            else {if((x[i].src+"").indexOf("undefined")<0){allt+=this.getsrc(x[i].src)}}
        };
        allt=allt.replace(/\W/g," ").split(" ");
        var obj={};
        for(var i=0;i<allt.length;i++){if(window[allt[i]]!==undefined){obj[allt[i]]=true}};
        for(var i in window){obj[i]=true};
        return obj
    },
    getsrc:function(x) {
        var httpRequest = new ActiveXObject('Msxml2.XMLHTTP');
        httpRequest.open('GET', x, false);
        httpRequest.send(null);
        return httpRequest.responseText;
    }
}
globalVars.init()
