////////////////////////////////////////////////////////////////////////////////
// TimeRange objects
////////////////////////////////////////////////////////////////////////////////

/**
 * General purpose object to hold both absolute and relative time ranges.
 */
Splunk.TimeRange = $.klass({
    _absoluteArgs : {},
    _relativeArgs : {},

    _fallbackAbsoluteTimeFormat: "%s.%Q",

    _unitMap : {},
    YEAR        : 0,
    MONTH       : 1,
    DAY         : 2,
    HOUR        : 3,
    MINUTE      : 4,
    SECOND      : 5,
    MILLISECOND : 6,

    DATE_METHODS  : [
        // Dont wrap any of these year, month, day values in _(). They are part of the logic here and never displayed.
        // instead its this._unitMap below that gets localized.
        {name: "year",   getter : "getFullYear",     setter: "setFullYear", minValue: "1974"},
        {name: "month",  getter : "getMonth",        setter: "setMonth",    minValue: "0"},
        {name: "day",    getter : "getDate",         setter: "setDate",     minValue: "1"},
        {name: "hour",   getter : "getHours",        setter: "setHours",    minValue: "0"},
        {name: "minute", getter : "getMinutes",      setter: "setMinutes",  minValue: "0"},
        {name: "second", getter : "getSeconds",      setter: "setSeconds",  minValue: "0"},
        {name: "millisecond", getter : "getMilliseconds", setter: "setMilliseconds",  minValue: "0"}
    ],


    initialize: function(earliestArg, latestArg, earliestOffset, latestOffset) {
        // keeping copy() up to date is a pain. This is either cheating, or a
        // smart way of with that problem. See usage within copy()
        this._constructorArgs = [earliestArg, latestArg, earliestOffset, latestOffset];

        // looks crazy, but im just trying to cover the full range of abbreviations
        // that are supported by the backend.
        var u = this._unitMap;
        u.s = u.sec = u.secs = u.second = u.seconds = _("second");
        u.m = u.min = u.mins = u.minute = u.minutes = _("minute");
        u.h = u.hr  = u.hrs  = u.hour   = u.hours   = _("hour");
        u.d = u.day = u.days   = _("day");
        // it looks like weeks were removed entirely at some point.  See src/util/TimeParser.cpp
        //u.w = u.week           = _("week");
        u.mon = u.month = u.months = _("month");
        u.y = u.yr = u.yrs   = u.year   = u.years   = _("year");

        this._absoluteArgs = {
            "earliest" : false,
            "latest" : false
        }
        this._relativeArgs = {
            "earliest" : false,
            "latest" : false
        }
        if (earliestArg=="now") earliestArg="0s";
        if (latestArg=="now")   latestArg="0s";
        
        this.logger = Splunk.Logger.getLogger("time_range.js");
        // we assume when a timerange is created directly that its the full range.
        this._isSubRangeOfJob = false;
        
        
        this._absoluteArgs["earliest"] = this.parseAbsoluteArgs(earliestArg);
        this._absoluteArgs["latest"]   = this.parseAbsoluteArgs(latestArg);
        
        // if earliestOffset wasnt passed we assume latestOffset was not passed either.
        if (this.isAbsolute() && !Splunk.util.isInt(earliestOffset)) {
            var offsetNow = Splunk.util.getServerTimezoneOffset();
            // although the offset of now() might be different from the offset at the start and at the end of the timerange
            // and although the offset at the start of the timerange might differ from the end,  
            // the offset now() is the best we have.
            // Bottom line -- 
            // a) when your TZ has DST, and this time range spans 1 DST boundary (or 3 or 5), 
            // and 
            // b) the offsets were not filled in by the constructor, 
            // then one of the two sides will be off by an hour.
            if (Splunk.util.isInt(offsetNow)) {
                earliestOffset = - Splunk.util.getServerTimezoneOffset() / 60;
                latestOffset   = earliestOffset;
            } else {
                this.logger.error("no offset provided to Splunk.TimeRange constructor, and Splunk.util.getServerTimezoneOffset unavailable");
            }
        } else {
            this.logger.debug("offsets provided by constructor " + earliestOffset + ", " + + latestOffset);
        }
        this.earliestServerOffsetThen = earliestOffset;
        this.latestServerOffsetThen   = latestOffset;
        

        // if its not an absolute term, try it as a relative term.
        if (!this._absoluteArgs["earliest"]) {
            this._relativeArgs["earliest"] = this.parseRelativeArgs(earliestArg);
        }
        if (!this._absoluteArgs["latest"]) {
            this._relativeArgs["latest"] = this.parseRelativeArgs(latestArg);
        }
        // if that failed, then we are left with everything, false, which is fine it just means
        // an 'All Time' search.
    },
    /**
     * returns true if this timeRange contains the given date.  For relative
     * timeranges this will not be entirely accurate.
     * For instance it assumes currently that there are 30 days in every month.
     */
    containsTime: function(dateObj) {
        var earliestTime = this.getAbsoluteEarliestTime();
        var latestTime   = this.getAbsoluteLatestTime();
        if (!earliestTime && !latestTime) return true;
        return (earliestTime <= dateObj  && latestTime > dateObj);
    },
    /**
     * returns true if the given range is contained by or equal to this instance.
     * In the case where this is a relative range, it will always assume that it contains
     * absolute ranges.  Because of how the timeline is implemented, this never seems to
     * come up as a shortcoming.
     * TODO - fix using this.containsTime
     */
    containsRange: function(range) {
        // if this is a range over all time, then it contains all time ranges. (including other "all time" ranges )
        // All time contains everybody, including other All Time
        if (this.isAllTime()) return true;
        // Nobody can contain All Time, except others which hit the case above.
        if (range.isAllTime()) return false;
        // if these ranges are identical, then true.
        if (this.equalToRange(range)) return true;

        // Currently relative ranges will ALWAYS claim to contain absolute ranges.
        if (this.relativeTerm  && !range.relativeTerm) return true;

        // trivial but expensive case of them being identical.  happens a lot.
        if (
            this.getEarliestTimeTerms() == range.getEarliestTimeTerms() &&
            this.getLatestTimeTerms() == range.getLatestTimeTerms()
            ) {
            return true;
        }

        if (this.getAbsoluteEarliestTime() > range.getAbsoluteEarliestTime())  return false;
        else if (this.getAbsoluteLatestTime() < range.getAbsoluteLatestTime()) return false;

        return true;
    },
    /**
     * returns true if the given range is exactly equal to this instance.
     */
    equalToRange: function(range) {
        if (this===range) return true;

        if (this.getRelativeEarliestTime() != range.getRelativeEarliestTime()) return false;
        if (this.getRelativeLatestTime() != range.getRelativeLatestTime()) return false;

        if (typeof(this.getAbsoluteEarliestTime()) != typeof(range.getAbsoluteEarliestTime())) return false;
        if (typeof(this.getAbsoluteLatestTime()) != typeof(range.getAbsoluteLatestTime())) return false;

        var rangeEarliestTime = range.getAbsoluteEarliestTime();

        if (this.getAbsoluteEarliestTime() && rangeEarliestTime &&  this.getAbsoluteEarliestTime().getTime() != rangeEarliestTime.getTime()) return false;

        var rangeLatestTime   = range.getAbsoluteLatestTime();
        if (this.getAbsoluteLatestTime() && rangeLatestTime && this.getAbsoluteLatestTime().getTime() != rangeLatestTime.getTime()) return false;
        return true;
    },

    /**
     * @returns -1 if the timeRange isnt governed by a relative searchTerm at all.
     * Otherwise an integer representing the number of milliseconds between earliestTime and latestTime.
     */
    getDuration: function() {
        if (this.relativeTerm) return -1;
        else if (this.getAbsoluteLatestTime() && this.getAbsoluteEarliestTime()) {
            return this.getAbsoluteLatestTime() - this.getAbsoluteEarliestTime();
        }
        else return -1;
    },
    getTimezoneOffsetDelta: function(serverOffsetThen, d) {
        if (!Splunk.util.isInt(serverOffsetThen)) {
            return 0;
        }
        // what JS thinks the timezone offset is at the time given by d. This WILL INCLUDE DST
        var clientOffsetThen = d.getTimezoneOffset() * 60;
        // what splunkd told is the actual timezone offset.
        serverOffsetThen     = serverOffsetThen * -60;

        return 1000 * (serverOffsetThen - clientOffsetThen);
    },
    isAllTime: function() {
        return (this.getAbsoluteLatestTime() || this.getAbsoluteEarliestTime() || this._relativeArgs["earliest"] || this._relativeArgs["latest"]) ? false : true;
    },
    isSubRangeOfJob: function() {
        return this._isSubRangeOfJob;
    },
    setAsSubRangeOfJob: function(isSubRange) {
        this._isSubRangeOfJob = isSubRange;
    },
    getEarliestTimeTerms: function() {
        var timeTerms = [];
        if (this.getAbsoluteEarliestTime()) {
            var utc = this.getAbsoluteEarliestTime().getTime() / 1000;
            timeTerms.push(utc);
        } else if (this._relativeArgs["earliest"]) {
            timeTerms.push(this.getRelativeEarliestTime());
        }
        return timeTerms.join(" ");
    },
    getLatestTimeTerms: function() {
        var timeTerms = [];
        if (this.getAbsoluteLatestTime()) {
            var utc = this.getAbsoluteLatestTime().getTime() / 1000;
            timeTerms.push(utc);
        } else if (this._relativeArgs["latest"]) {
            timeTerms.push(this.getRelativeLatestTime());
        }
        return timeTerms.join(" ");
    },

    /*
     * We currently do no intelligent snapping on generic zoom in cases although
     * we certainly could in the future and it would be nice for our users.
     */
    zoomIn: function() {
        var duration = this.getDuration();
        if (!duration) this.logger.error("Assertion failed - TimeRange.zoomOut not implemented for relative time terms range=" + this.toConciseString());

        this._absoluteArgs["earliest"].setTime(this.getAbsoluteEarliestTime().getTime() + Math.round(duration/2));
        return new Splunk.TimeRange(this._absoluteArgs["earliest"], this._absoluteArgs["latest"], this.earliestServerOffsetThen, this.latestServerOffsetThen);
    },
    /*
     * Basically zooms out to double the duration, leaving the latest side of
     * the range unchanged.  However then if the range is really ragged and there's
     * nice human-understandable boundaries nearby like 'midnight', or '6:00:00',
     * we snap both the earliest and latest sides of the range out to that level
     * of granularity.   Best way to understand more is to go to the unit tests
     * and read the testcases.
     */
    zoomOut: function() {
        var earliest = this._absoluteArgs["earliest"];
        var earliestDelta = this.getTimezoneOffsetDelta(this.earliestServerOffsetThen, earliest);
        earliest.setTime(earliest.valueOf() - earliestDelta);
        
        var latest = this._absoluteArgs["latest"];
        var latestDelta   = this.getTimezoneOffsetDelta(this.latestServerOffsetThen, latest);
        latest.setTime(latest.valueOf() - latestDelta);

        var getValueFor = function(dateObj, scaleConst) {
            return dateObj[this.DATE_METHODS[scaleConst]["getter"]]();
        }.bind(this);
        
        var setValueFor = function(dateObj, scaleConst, value) {
            return dateObj[this.DATE_METHODS[scaleConst]["setter"]](value);
        }.bind(this);

        var getRoundingLevel = function(duration) {
            var flattenedLevel;
            // Note: we never flatten the years.
            //       That would be setting the year to 1970 and is not a useful thing to do in any case.
            // more than 6 months and we flatten the months
            if (duration > 1000 * 6 * 30 * 24 * 60 * 60) return this.MONTH;
            // more than 15 days and we flatten the days
            if (duration > 1000 * 15 * 24 * 60 * 60) return this.DAY;
            // more than 12 hours and we flatten the hours
            if (duration > 1000 * 12 * 60* 60) return this.HOUR;
            // more than 30 minutes and we flatten the minutes
            if (duration > 1000 * 30 * 60) return this.MINUTE;
            // more than 30 seconds and we flatten the seconds
            if (duration > 1000 * 30 ) return this.SECOND;
            // flatten milliseconds
            return this.MILLISECOND
        }.bind(this);


        var duration = this.getDuration();
        if (!duration) this.logger.error("Assertion failed - TimeRange.zoomOut not implemented for relative time terms range=" + this.toConciseString());

        // Fortune favors the bold. Just set earliest so the interval is quadrupled.
        earliest.setTime(earliest.getTime() - (3*duration));

        // clone the original values in case DST garbles our rounding into an invalid range.
        var unroundedEarliest = new Date(earliest);
        var unroundedLatest   = new Date(latest);
        
        // the rest of the method is trying to clean up the mess made by our nice simple barbarism.

        var flattenedLevel = getRoundingLevel(duration);

        this.logger.warn('zero out all the ones at the flattenedLevel and higher (' + flattenedLevel + ')');
        // zero out all the ones AT THE flattenedLevel AND HIGHER.
        for (var i=flattenedLevel; i<this.DATE_METHODS.length; i++) {
            var methodDict = this.DATE_METHODS[i];
            setValueFor(earliest, i, methodDict["minValue"]);

            // if the flattened level is already at a minValue, we do nothing.
            if (getValueFor(latest, i) != methodDict["minValue"]) {
                // might seem odd, but the simplest way is to round *down*,
                // and then increment one unit up, by one.  The Date class will
                // do the hard work for us.
                setValueFor(latest, i, methodDict["minValue"]);
                if (i==flattenedLevel) {
                    var valueForNextHighestUnit = getValueFor(latest, i-1);

                    setValueFor(latest, i-1, valueForNextHighestUnit+1);
                }
            }
        }
        if (earliest.valueOf() + earliestDelta >= latest.valueOf() + latestDelta) {
            this.logger.warn("DST corner case where assumptions made by the rounding code were broken by the offset change at 2AM");
            earliest = unroundedEarliest;
            latest   = unroundedLatest;
        } 
        earliest.setTime(earliest.valueOf() + earliestDelta);
        latest.setTime(latest.valueOf()     + latestDelta);
        return new Splunk.TimeRange(earliest, latest, this.earliestServerOffsetThen, this.latestServerOffsetThen);
    },
    
    /**
     * PRIVATE method to not duplicate code between getRelativeEarliestTime and
     * getRelativeLatestTime()
     */
    genericGetForRelativeArg:function(which) {
        if (!this._relativeArgs.hasOwnProperty(which)) return false;
        var dict = this._relativeArgs[which];
        var str = [];
        if (dict.hasOwnProperty("count")) {
            if (dict["count"] >= 0) str.push("+");
            str.push(dict["count"])
        }
        if (dict.hasOwnProperty("units")) {
            str.push(dict["units"]);
        }
        if (dict.hasOwnProperty("snapUnits")) {
            str.push("@" + dict["snapUnits"]);
        }

        return str.join("");
    },
    /**
     * PRIVATE gets the earliestTime when it is a relative time term.
     */
    getRelativeEarliestTime: function() {
        return this.genericGetForRelativeArg("earliest");
    },
    /**
     * PRIVATE gets the latestTime when it is a relative time term.
     */
    getRelativeLatestTime: function() {
        return this.genericGetForRelativeArg("latest");
    },
    /**
     * PRIVATE returns the earliestTime as a JS Date object.
     */
    getAbsoluteEarliestTime: function() {
        return this._absoluteArgs["earliest"];
    },
    /**
     * PRIVATE returns the latestTime as a JS Date object.
     */
    getAbsoluteLatestTime: function() {
        return this._absoluteArgs["latest"];
    },
    isAbsolute: function() {
        return (this._absoluteArgs["earliest"] || this._absoluteArgs["latest"])? true:false;
    },
    isRelative: function() {
        return (this._relativeArgs["earliest"] || this._relativeArgs["latest"])? true:false;
    },
    /**
     * PRIVATE tries to parse the argument as a relative time string, into its
     * internal representation used by TimeRange.
     */
    parseRelativeArgs: function(arg) {
        if (!arg) return false;
        if (arg.indexOf(" ")!=-1) this.logger.error("Assertion failed - Currently we can only deal with a single relative term at a time. ");
        var relativeDict = {};
        // split off the @ term, if there is one.
        var splitArgs = arg.split("@");
        arg = splitArgs[0] || false;
        if (splitArgs.length > 1) {
            relativeDict["snapUnits"] = splitArgs[1];
        }
        if (arg) {
            relativeDict["count"] = parseInt(arg, 10);
            relativeDict["units"] = Splunk.util.trim(arg.replace(relativeDict["count"], ""), "+");
        }
        return relativeDict;
    },
    /**
     * PRIVATE tries to parse the argument as either a Date object literal, or a
     * int or float representing seconds since the epoch (aka UTC)
     * NOTE: UTC values are always in the server's timezone.  It is illegal
     * to grab a js Date Object,  and use it's UTC.  What you need can always be found
     * in the TimeRange class.
     */
    parseAbsoluteArgs: function(arg, serverOffsetThen) {
        if (!arg) return false;
        if (arg instanceof Date) {
            return arg;
        } else if (arg && ("" + arg).match(Splunk.TimeRange.UTCRegex)) {
            var d = new Date();
            d.setTime(arg * 1000);
            return d;
        } else {
            //  try and parse them as strings containing absolute dates
            var parsedDate = Splunk.util.parseDate(arg, this.getTimeFormat());
            return parsedDate;
        }
        return false;
    },
    /**
     * Serializes the timeRange. Generally for debug statements. 
     * For human-readable purposes, use toConciseString()
     */
    toString: function() {
        var str = [];
        if (this.getAbsoluteEarliestTime()) {
            str.push(this.getAbsoluteEarliestTime());
        }
        if (this.getAbsoluteLatestTime()) {
            str.push(this.getAbsoluteLatestTime());
        }
        if (this.getRelativeEarliestTime()) {
            str.push(this.getRelativeEarliestTime());
        }
        if (this.getRelativeLatestTime()) {
            str.push(this.getRelativeLatestTime());
        }
        return str.join(" ");
    },
   
    /**
     * returns a compact localized description of the range.
     */
    toConciseString: function() {        
        if (this.isAbsolute()) {
            // open ended ranges are supported so be careful not to use new Date() by default (which would result in the browser's current time).
            var earliestTime = null;
            var latestTime   = null;
            if (this.getAbsoluteEarliestTime()) {
                earliestTime = new Date();
                earliestTime.setTime(this._absoluteArgs["earliest"].valueOf());
                // correct timezone offsets.
                if (Splunk.TimeRange.CORRECT_OFFSET_ON_DISPLAY) {
                    var earliestDelta = this.getTimezoneOffsetDelta(this.earliestServerOffsetThen, earliestTime);
                    earliestTime.setTime(earliestTime.valueOf() - earliestDelta);
                }
            }
            if (this.getAbsoluteLatestTime()) {
                latestTime = new Date();
                latestTime.setTime(this._absoluteArgs["latest"].valueOf());
                // correct timezone offsets.
                if (Splunk.TimeRange.CORRECT_OFFSET_ON_DISPLAY) {
                    var latestDelta   = this.getTimezoneOffsetDelta(this.latestServerOffsetThen, latestTime);
                    latestTime.setTime(latestTime.valueOf()     - latestDelta);
                }
            }        
            // defers to i18n format_daterange function.
            return format_datetime_range(Splunk.util.getConfigValue('LOCALE', "NONE"), earliestTime, latestTime);
        } else if (this.isRelative()) {
            return this.formatRelativeRange();
        } else {
            return _("over all time");
        }
    },
    formatRelativeRange: function() {
        var relEarliest = this._relativeArgs["earliest"];
        var relLatest   = this._relativeArgs["latest"];
        if (relEarliest.hasOwnProperty("snapUnits")) {
            if (relLatest.hasOwnProperty("snapUnits") != relLatest.hasOwnProperty("snapUnits")) {
                throw("Assertion failed - we dont support cases where one side has snapUnits and the other does not.");
            }
            // Special cases for 'today' 'yesterday', 'this week', 'previous week'
            if (this._unitMap[relEarliest["snapUnits"]] == "day"
                && this._unitMap[relLatest["snapUnits"]] == "day") {
                if (!relEarliest.hasOwnProperty("count")
                    && relLatest.hasOwnProperty("count")
                    && relLatest["count"] == 1) {
                    return _("today");
                }
                else if (!relLatest.hasOwnProperty("count")
                    && relEarliest.hasOwnProperty("count")
                    && relEarliest["count"] == -1) {
                    return _("yesterday");
                }
            }
            // check for the form @something  +1something@something,  which is 'This Something'
            if (!relEarliest.hasOwnProperty("count")
                    && relLatest.hasOwnProperty("count")
                    && relLatest["count"] == 1) {
                return sprintf(_("during this %(singleUnitOfTime)s"), {singleUnitOfTime: this._unitMap[relEarliest["snapUnits"]]})
            }
            // if this is in the form -1something@something  @something,  ie 'Last Something'
            if (!relLatest.hasOwnProperty("count")
                    && relEarliest.hasOwnProperty("count")
                    && relEarliest["count"] == -1) {
                return sprintf(_("during last %(singleUnitOfTime)s"), {singleUnitOfTime: this._unitMap[relEarliest["snapUnits"]]})
            }
        }
        // catches cases like "in the last month",  "in the last hour", regardless of snap term
        if (relEarliest.hasOwnProperty("count") && relEarliest["count"] == -1) {

            // TRANS: in this particular case the range is exactly one month (or one hour or one minute).
            return sprintf(_("in the last %(unitOfTime)s"), {unitOfTime: this._unitMap[relEarliest["units"]]});
        }
        // catches cases like "in the last 4 months",  "in the last 7 hours", also regardless of snap term
        else if (relEarliest.hasOwnProperty("units") && relEarliest.hasOwnProperty("count")) {
            // TRANS: TODO - still need to ungettext this plural form and remove the hardcoded S below... However since the value is dynamic in the code it requires further thought.
            return sprintf(_("in the last %(count)s %(unitOfTime)ss"), {count: -relEarliest["count"], unitOfTime: this._unitMap[relEarliest["units"]]})
        }
        // if we didnt match any other cases, returning the terms themselves is better than nothing...
        return sprintf(_("over custom relative time range (%(earliestTerm)s %(latestTerm)s)"), {earliestTerm: this.getRelativeEarliestTime(), latestTerm: this.getRelativeLatestTime()});
    },

    /**
     * Copies a new TimeRange instance and returns.  Useful for when you
     * want to pass a TimeRange property, but you dont entirely trust the
     * receiver to leave it unchanged.
     */
    copy: function() {
        var range = new Splunk.TimeRange(this._constructorArgs[0], this._constructorArgs[1], this._constructorArgs[2], this._constructorArgs[3]);
        range.setAsSubRangeOfJob(this.isSubRangeOfJob());
        return range;
    },
    strftime: function(date, timeFormat) {
        return date.strftime(timeFormat || this.getTimeFormat())
    },
    /*
    isLastDayOfMonth : function(oDate)  {
        oLocalDateCopy = new Date(oDate)
        var month = oLocalDateCopy.getMonth();
        var day   = oLocalDateCopy.getDate();
        oLocalDateCopy.setDate(oLocalDateCopy.getDate() + 1);
        this.logger.debug("toConciseString - isLastDayOfMonth", oLocalDateCopy.getMonth(), "==", month);
        return (oLocalDateCopy.getMonth() != month );
    },
    */
    getTimeFormat: function() {
        return Splunk.util.getConfigValue('DISPATCH_TIME_FORMAT', this._fallbackAbsoluteTimeFormat);
    }
});
Splunk.TimeRange.UTCRegex = new RegExp("^[0-9]*(\.[0-9]+)?$");
Splunk.TimeRange.CORRECT_OFFSET_ON_DISPLAY = true;
Splunk.TimeRange.relativeArgsToString = function(count, units, snapUnits) {
    // doesnt matter but we pick earliest.
    var which = 'earliest';
    var range = new Splunk.TimeRange();
    range._relativeArgs[which] = {count: count, units:units, snapUnits:snapUnits};
    return range.genericGetForRelativeArg(which);
}
/*
Splunk.TimeRange.getCustomTimeLabel = function(earliest, latest) {
    for (var i=0, l=Splunk.TimeRange.customTimes.length; i<l; i++) {
        var dict = Splunk.TimeRange.customTimes[i];
        if (earliest == dict.earliest && latest == dict.latest) {
            return dict.label;
        }
    }
    return false;
}
Splunk.TimeRange.customTimes = [];
*/