
/**
 * Splunk.Module is the abstract base class for Splunk UI modular components.
 * Sensible default values and behaviors are provided whever possible, and
 * instructions for overriding correctly are provided inline.
 *
 */
Splunk.Module = $.klass({
	/**
	 * Defines the URI segment that identifies the default viewstate
	 * persistence set name used for normal module 'stickiness'.
	 */
	CURRENT_VIEWSTATE_KEY: '_current',

    /**
     * Constructor.  If you override it, 99.9% of the time you should in your
     * first line call $super(container).
     */
    initialize: function(container) {

        this.logger = Splunk.Logger.getLogger("AbstractModule.js");

        if (this.parent || this.container || this._children) {
            this.logger.error(this.moduleType, ".initialize - Assertion Failed. Modules may not be reinitialized.");
        }

        this.container = $(container);
        this.resultsContainer = $('<div/>').appendTo(this.container);

        this.parent = null;
        this._children = [];
        this.baseContext = null;
        this._baseSettingsMap = {};

        this._lastSuccessfulDispatchedContext = null;
        
        // Holding place for the last param set collected by getResults
        this._previousResultParams = null;


        // TODO - this could be very useful, but I think should be avoided -ncm
        //this.container.module = this;

        //this._enablePropertyWatch = false;
        //if (this._enablePropertyWatch) $("<input class='debugInspector'/>").prependTo(this.container);
        // get the current DOM ID for this module
        this.moduleId = this.container.attr('id');

        // initialize this._params for modules

        // I have no idea what to call this, but a module can be shown or hide
        // for a couple different reasons, and we cant just do show() and hide()
        // directly, or else we'd only be able to have one 'mode' of invisibility
        // per view.
        this._invisibilityModes = {};


        // getResultsXHRObject holds a reference to the XHR object generated when
        // getResults is called.  If a request to reset the module is called,
        // abort will be called on the xhr object this is assigned to.
        this.getResultsXHRObject = null;
        // init params
        this.importParams();

        // observe any view data change events
        $(document).bind('viewDataChange', this.onViewDataChange.bind(this));
    },


    /*
     * Any module that returns false from requiresAsynchronousLoading() will be
     * considered to be 'INITIALIZED'
     * ie  'Splunk.util.moduleLoadStates.INITIALIZED'
     * as soon as it returns from it's constructor.
     * Conversely any module that returns true here must then be responsible for
     * doing it's own asynchronous loading, and making sure when its over that
     * the corresponding moduleLoadStatusChange event gets fired.
     */
    requiresAsynchronousLoading: function() {return false;},

    onLoadStatusChange: function(statusInt) {},


    /////////////////////////////////////////
    // Methods which involve pushing non-search settings to child modules
    // below the current module.
    /////////////////////////////////////////

    /*
     * This method will take the settings created by this module
     *  (courtesy of this.getSettings())
     * and combine them with the provided cumulativeSettingsMap
     * settings, and push them to all immediate children, and then
     * those children will repeat the process down to the leaves
     * of the tree.
     */
    pushSettingsToChildren: function(cumulativeSettingsMap, pushSilently) {
        //this.logger.debug(this.moduleType + " calling pushSettingsToChildren(", cumulativeSettingsMap, ", ", pushSilently, ")");
        cumulativeSettingsMap = $.extend({}, cumulativeSettingsMap);
        //TODO - check against this.getParam
        var localSettings = this.getSettings();
        
        for (var key in localSettings) {
            if (this._stickyParamList.hasOwnProperty(key) && this._stickyParamList[key] != localSettings[key]) {
                this.logger.error("Assertion failed ", key, " is declared to be sticky, but the key in this.getSettings() does not match the internal value");
            }
            if (this._persistableParamList.hasOwnProperty(key) && this._persistableParamList[key] != localSettings[key]) {
                this.logger.error("Assertion failed ", key, " is declared to be persistable, but the key in this.getSettings() does not match the internal value");
            }
            if (localSettings.hasOwnProperty(key)) {
                cumulativeSettingsMap[key] = localSettings[key];
            }
        }
        

        // TODO - we have so many for() loops.
        //        IMO we should write a withEachChildModule(someFunction) method.
        for (var i=0,len=this._children.length; i<len; i++) {

            // extend the local copy of the baseSettingsMap.
            $.extend(this._children[i]._baseSettingsMap, cumulativeSettingsMap);
            
            if (!pushSilently) {
                // call the handler to notify the module their settings have changed.
                // TODO - only call this if the map is different from their last one?
                this._children[i].onSettingsChange(cumulativeSettingsMap);
            }

            // continue the cascade down into the descendants.
            this._children[i].pushSettingsToChildren(cumulativeSettingsMap, pushSilently);
        }
    },
    /*
     * This will be called whenever a module above you in the hierarchy, calls
     * pushSettingsToChildren.  (There is a silent mode that was recently added,
     * whereby these handlers will NOT be called, but the underlying settings
     * will be updated.
     */
    onSettingsChange: function(settingsMap) {},
    /*
     * get key value pairs for any states that need to be pushed to descendant modules
     * in the view's functional hierarchy.
     * Commonly implemented, by modules that accept user input, but the input has no
     * effect on the underlying SearchContext, and then need to convey that user input
     * to modules below.
     */
    getSettings: function() {return {}},
    /*
     * returns a dictionary of all settings that have ever been received from above.
     */
    getBaseSettings: function() {
        return $.extend({}, this._baseSettingsMap);
    },
    /*
     * returns a boolean indicating whether a value for this setting has ever been received.
     */
    hasSetting: function(name) {
        return this._baseSettingsMap.hasOwnProperty(name);
    },
    /*
     * returns the value for a single setting name.
     */
    getSetting: function(name) {
        return this._baseSettingsMap[name];
    },
    useParamValuesAsSettings: function(paramNames) {
        for (var i=0; i<paramNames.length; i++) {
            var key = paramNames[i];
            if (this._params.hasOwnProperty(key)) {
                this._baseSettingsMap[key] = this._params[key];
            }
        }
    },




    /////////////////////////////////////////
    // Methods which involve how SearchContext.dispatchJob gets called
    /////////////////////////////////////////

    /*
     * Public function, but not to be commonly overridden.  It's intended as
     * a safer way to check for isSubclassOf(Splunk.Module.DispatchingModule)
     * returning false basically says a module doesnt need its context to be
     * running and exposing data endpoints when the user runs the
     * search.
     */
    requiresDispatch: function(context) {return false;},

    /*
     * Essentially an event handler, this will be called when the underlying
     * SearchContext instance available in this.getBaseContext(), has changed.
     */
    onContextChange: function() {},
    /*
     * method that actually kicks off the dispatch process.
     * the hard work is delegated to SearchContext.dispatchJob though.
     */
    _fireDispatch: function(explicitContext) {
        // assumption -- that history repeats itself.
        // Namely that at time A, the module that decided it had to call
        // _fireDispatch,
        // is the same module that at Time B decided it has to call _fireDispatch
        // Thus its the same module and will have a reference to the previous one here, which it can cancel.
        var lastContext = this._lastSuccessfulDispatchedContext;
        //  if we end up detecting the need to autocancel the previous job, 
        // we pass the reference in as a closure, because cancelling explicitly before 
        // the next POST produces noticeable flicker in the UI, as modules detect the cancellation
        // of their current job and reset themselves.
        var previousJobToCancel = null;

        // Check that we're not already dispatching from this very point in the hierarchy.
        // the user might be holding down the return key in SearchBar
        // If we are already dispatching, then we just bail. All will be well. 
        // The former dispatch will either succeed or fail. In either case all will be well. 
        // Note - it was tempting to put this on Job or SearchContext, but they get 
        //        cloned/re-initialized a little bit and this seemed less sketchy + trivial.
        // Note - this also makes the assumption about history repeating itself.
        
        if (this.dispatchAlreadyInProgress) {
            return false;
        } else if (lastContext
            && lastContext.isJobDispatched()
            && !lastContext.job.canBeAutoCancelled()) {
            Splunk.Globals['Jobber'].removeJob(lastContext.job.getSearchId());
        } else if (lastContext
            && lastContext.isJobDispatched()
            && lastContext.job.canBeAutoCancelled()) {
            previousJobToCancel = lastContext.job;
        } 

        var failHandler = function(failedContext) {
            this.dispatchAlreadyInProgress = false;
            this.logger.error(this.moduleType, " Context failed to dispatch job for search=", failedContext.getSearch());
            if (previousJobToCancel) previousJobToCancel.cancel();
            // in the name of providing immediate feedback when a search POST goes out, 
            // some modules (notably JobStatus) will have changed state onBeforeJobDispatched
            // this reset is required to reset them back to their null state.
            var descendants = this.getDescendants();
            for (var i=0, len=descendants.length; i<len; i++) {
                descendants[i].reset();
            }
            
        }.bind(this);

        var successHandler = function(contextDispatched) {
            this.logger.debug("success - context dispatched for search=", contextDispatched.getSearch());
            this.pushContextToChildren(contextDispatched);
            this._lastSuccessfulDispatchedContext = contextDispatched;
            this.dispatchAlreadyInProgress = false;
            if (previousJobToCancel) previousJobToCancel.cancel();
        }.bind(this)
        var contextToDispatch = explicitContext || this.getModifiedContext();
        
        this.dispatchAlreadyInProgress = true;
        contextToDispatch.dispatchJob(successHandler, failHandler, this.getGroupName());
    },
    /*
     * method used during SearchContext calculations, that tells us
     * whether there are any *immediate* children which require the
     * context objects to be dispatched before getting them.
     */
    someChildrenRequireDispatch: function(context) {
        for (var i=0; i<this._children.length; i++) {
            if (this._children[i].requiresDispatch(context)) {
                return true;
            }
        }
        return false;
    },
    
    /**
     * A generic mechanism for any module in the tree to make a change to the context right before it's 
     * dispatched.   Only usage as of this writing is to set the required_field_list and status_buckets
     * so that the job gets dispatched with sufficient statistics to give each module what it requires.
     */
    onBeforeJobDispatched: function(context) {},
   
    /**
     * This method should really never be overridden.
     * This method is the recommended way of getting the current
     * SearchContext that is the sum total of search information
     * provided by the modules ancestors further up the page.
     */
    getBaseContext: function(preserveJobReference) {
        if (preserveJobReference) this.logger.error("the preserveJobReference arg has been removed from Module.getBaseContext");
        if (!this.baseContext) {
            // for views that havent run any searches yet, this will cause an upward
            // cascade of getModifiedContext whereby they all get set on demand.
            if (this.parent) {
                this.baseContext = this.parent.getModifiedContext();
            } else {
                // this case will now be hit only by modules at the top level of the
                // view hierarchy.
                return new Splunk.SearchContext("*");
            }
        }
        if (this.baseContext) {
            return this.baseContext.clone();
        }
        this.logger.error("Assertion Failed in ", this.moduleType, " module. getBaseContext fell into an invalid case.");
    },

    /**
     * This method should really not be overridden by subclassed modules.
     * Instead either override:
     *
     * getModifiedContext when your module is receiving a new Context that
     * has the potential to create a new search job or
     *
     * applyContext when your module is receiving a Context that
     * has been created from an old UI state or from a previously run and
     * finished job, or from a user interaction further down the page.
     *
     */
    pushContextToChildren: function(explicitContext) {

        var childContext = explicitContext || this.getModifiedContext();

        // If the context received is null then stop the chain
        if (childContext === null) {
            this.logger.debug(this.moduleType, 'getModifiedContext returned "null", stopping the pushContextToChildren call.');
            return;
        }

        
        if (!childContext.isJobDispatched()) {
            if (this.someChildrenRequireDispatch(childContext)) {
                var descendants = this.getDescendants();
                for (var i=0, len=descendants.length; i<len; i++) {
                    descendants[i].onBeforeJobDispatched(childContext);
                }
                this._fireDispatch(childContext);
                // BAIL OUT. and currently _fireDispatch will result in another
                // call to pushContextToChildren.  TODO - rethink/improve this.
                return;

            } else {
                //this.logger.info(this.moduleType, ".onContextChange - no children require dispatch");
            }
        }
        // small change that simplifies a lot of complexity.
        // pushSettingsToChildren now takes a second optional arg, that will silently
        // update everyone, but not trigger any event handlers.
        // this relieves modules of the burden of worrying about the settings data
        // when the times comes to push a context.
        // the 2 specific places where this alleviates the most problems, is
        //  a) UI resurrection and b) the switcher logic.
        // the assumption that this change rests on,  is as follows:
        // "anything that onSettingsChange would have blown away,  will be blown away by the more comprehensive onContextChange."
        this.pushSettingsToChildren(false, true);
        for (var i=0; i<this._children.length; i++) {
            this._children[i].baseContext = childContext;
            this._children[i].onContextChange();
            this._children[i].pushContextToChildren();
        }
    },

    /**
     * Resets the context to null and removes the job from the jobber manifest.
     * Not usually overriden.  Called by the more general 'reset' method.
     */
    resetContext: function() {
        if (this.baseContext && this.baseContext.isJobDispatched()) {
            Splunk.Globals['Jobber'].removeJob(this.baseContext.job.getSearchId());
        }
        this.baseContext = null;
        this._lastSuccessfulDispatchedContext = null;
    },


    /////////////////////////////////////////
    // Methods which involve Selection State and things that modify the search
    // for the modules children.
    /////////////////////////////////////////
    /**
     * override me to pass a modified - or refined - context to children
     * rather than the context this module was given.
     */
    getModifiedContext: function() {
        if (!this.baseContext) {
            this.logger.warn(this.moduleType, "is calling getModified",
            "context and has no baseContext object. This will probably return a blank Context");
        }
        return this.getBaseContext();
    },

    /**
     * NEEDS DOCUMENTATION
     */
    applyContext: function(context) {
        //$("input.debugText", this.container).attr("value",'sid=' + context.job.getSearchId());
        if (context.isResurrected()) {
            this.logger.warn(this.moduleType, " does not implement applyContext");
        }
    },


    /////////////////////////////////////////
    // Methods which involve modifying the state of the module's parents.
    /////////////////////////////////////////


    /**
     * a child module calls this method on its parent to request that the
     * parent modify it's state
     * the module may choose to either
     * a) accept the request,
     * b) pass the request along to it's parent, or ignore the request.
     */
    passContextToParent: function(context) {
        // default behavior is to pass the request to parents until we find one
        // that wants to handle it.
        if (this.parent) {
            // resurrection proceeds up the entire tree no matter what.
            if (context.isResurrected()) {
                this.parent.applyContext(context)
                this.parent.passContextToParent(context);
            }
            // but the upstream-click contexts get stopped at the first return true.
            else if (!this.parent.applyContext(context)) {
                return this.parent.passContextToParent(context);
            } 
        } else if (context.isResurrected()) {
            // The topmost module gets the remnant of resurrection, as its base context.");
            var newBaseContext = context.clone();
            newBaseContext.setResurrected(false);
            this.baseContext = newBaseContext;
        }
        // if the propagation continues all the way to the top 
        // and we're not resurrecting, then
        // we are left without anything to do. 
        this.logger.warn("an upstream interaction made it to the top of the tree (to ", this.moduleType, ") because none of its downstream modules were able to take any action.")
        return false;
    },



    /*
     * Adopt the given module instance as a child.
     */
    addChild: function(child) {
        if (child.parent) {
            this.logger.error(this.moduleType, ".addChild - Assertion failed. this", child.moduleType, " instance already has a parentModule. There Can Be Only One.");
        }
        this._children.push(child);
        child.parent = this;
    },

    /*
     * impl. would be fancier with splice() but somehow this seems simpler/better.
     *
     * could be:
     * var idx = this._children.indexOf(childModule);
     * if (idx >= 0) this._children.splice(this._children.indexOf(childModule),1);
     */
    removeChild : function(childModule) {
        var newChildrenList = [];
        for (var i=0; i<this._children.length;i++) {
            if (childModule == this._children[i]) {
                this._children[i].parent = null;
            } else {
                newChildrenList.push(this._children[i]);
            }
        }
        this._children = newChildrenList;
    },

    /**
     * Methods to handle passing things to the tree scoped by the current module.
     * Plural methods returns a flat array with a generally predictable, though not
     * guaranteed, order.  Singular methods only return an object reference or null.
     *
     * Example use:
     *
     * * Cancel all jobs in the current module tree:
     *        $.each(this.getModulesInTree(), function(module){
     *             if (module.baseContext && module.baseContext.isJobDispatched()) module.baseContext.job.cancel();
     *        });
     *
     * * Hide all modules below the current module:
     *         $.each(this.getDescendants(), function(module) {
     *             module.container.css('display', 'none');
     *         })
     */

    /**
     * Find the root module in a module's tree.
     * @returns Splunk.Module object
     * @returns null
     */
    getRootAncestor: function() {
        var pointer = this, resp = null;
        while(pointer) {
            resp = pointer;
            pointer = pointer.parent;
        }
        return resp;
    },

    /**
     * Return an ordered list of all of the ancestors.
     * @returns Array
     */
    getAncestors: function() {
        var pointer = this.parent, resp = [];
        while(pointer) {
            resp.unshift(pointer);
            pointer = pointer.parent;
        }
        return resp;
    },
    /* walks up the tree looking for a 'group'.  Although group is reflected as a regular _param 
     * value, this functionality is not intended for wider usage.  Used only on dashboards currently.
     * Specifically HiddenSavedSearch broadcasts an event that has its sid and also the groupName
     * it is contained within.  This event is listened to by code in dashboard.html  to 
     * display the 'last refreshed' text for the panel. 
     */
    getGroupName: function() {
        if (this._params.hasOwnProperty("group")) {
            return this._params["group"] 
        } 
        return (this.parent) ? this.parent.getGroupName() : false;
    },

    /**
     * Return an array of all the calling module's decendants.
     * @returns Array
     */
    getDescendants: function() {
        var resp=this._children.slice();
        for(var i=0; i<resp.length; i++) {
            resp = resp.concat(resp[i]._children);
        }
        return resp;
    },

    /**
     * Return an array of all the modules in the calling module's tree.
     * @returns Array
     */
    getModulesInTree: function() {
        var root = this.getRootAncestor();
        return [root].concat(root.getDescendants());
    },

    show: function(invisibilityMode) {
        invisibilityMode = invisibilityMode || "global";
        //this.logger.debug(this.moduleType , '.show( ' , invisibilityMode, ')');

        if (this._invisibilityModes.hasOwnProperty(invisibilityMode)) {
            delete this._invisibilityModes[invisibilityMode];
        }
        this._changeVisibility();
    },
    hide: function(invisibilityMode) {
        invisibilityMode = invisibilityMode || "global";
        //this.logger.debug(this.moduleType , '.hide( ' , invisibilityMode, ')');
        this._invisibilityModes[invisibilityMode] = 1;
        this._changeVisibility();
    },
    /**
     * Will show the module and all it's descendant modules.
     * clients should pass a mode arg if they want their visibility
     * changes to be respected regardless of other clients that might
     * call these methods.
     * iow - if anyone out there thinks it's invisible, it's invisible.
     *       if nobody does, it's visible.
     */
    showDescendantTree: function(invisibilityMode) {
        // this isSwitcherLeaf property is an idiosyncracy introduced for the
        // SerializedSwitcher modules, which modify the hierarchy for their own
        // evil purposes.
        if (!this.isSwitcherLeaf) {
            for (var i=0; i<this._children.length;i++) {
                this._children[i].showDescendantTree(invisibilityMode);
            }
        }
        this.show(invisibilityMode);
    },
    /**
     * Will hide the module and all it's descendant modules.
     */
    hideDescendantTree: function(invisibilityMode) {
        // see comment about isSwitcherLeaf on the showDescendantTree method.
        if (!this.isSwitcherLeaf) {
            for (var i=0; i<this._children.length;i++) {
                this._children[i].hideDescendantTree(invisibilityMode);
            }
        }
        this.hide(invisibilityMode);
    },
    /**
     * Utility method used by both hide and show.
     * TODO - needs a less dumb implementation.
     */
    _changeVisibility: function() {
        var visible = true;
        for (var mode in this._invisibilityModes) {
            if (this._invisibilityModes.hasOwnProperty(mode)) {
                visible = false;
            }
        }
        if (visible) this.container.show();
        else this.container.hide();
    },
    /**
     * Event triggered when view controller data has updates. Implement this method to observe event.
     *
     * @param {Object} event A jQuery event.
     * @param {Object} data An object literal representation of all templateArgs returned from the view.py render method.
     */
    onViewDataChange: function(event, data) {},

    /**
     * Call this method to broadcast an onViewDataChange event.
     */
    pushViewDataChange: function(){
        Splunk.Globals.ModuleLoader.refreshViewData();
    },

    ///////////////////////////////////////////////////////////////////////////
    // Module parameter persistence methods
    ///////////////////////////////////////////////////////////////////////////


    /**
     * Import the startup parameters for this module, as passed in by via the
     * main HTML template.
     *
     */
    importParams: function() {

        this._params = {};
        this._stickyParamList = [];
        this._persistableParamList = [];

        if (!Splunk.Module.loadParams) {
            this.logger.info('importParams - view does not have any parameters defined');
            return;
        }

        if (Splunk.Module.loadParams.hasOwnProperty(this.moduleId)) {
            this._params = $.extend(true, {}, Splunk.Module.loadParams[this.moduleId]);
        }

        if (!Splunk.Module.paramConfig) {
            this.logger.info('importParams - view does not have parameter config defined');
            return;
        }

        if (Splunk.Module.paramConfig.sticky.hasOwnProperty(this.moduleId)) {
            this._stickyParamList = Splunk.Module.paramConfig.sticky[this.moduleId];
        }
        if (Splunk.Module.paramConfig.persistable.hasOwnProperty(this.moduleId)) {
            this._persistableParamList = Splunk.Module.paramConfig.persistable[this.moduleId];
        }
    },


    /**
     * Returns the current module's value for parameter 'key'.  If the value is
     * not set, then 'null' is returned.
     *
     * @param {string} key The name of the parameter to retrieve.
     * @param {string} fallbackValue The value to return if the module does
     *      not have an explicit value set; will supercede the default param
     *      value set in the .conf file.
     *
     */
    getParam: function(key, fallbackValue) {

        if (!key) {
            throw Error(_('getParam: Splunk cannot get param; null key name passed'));
        }

        if (this._params.hasOwnProperty(key)) {
            return this._params[key];
        } else if (fallbackValue != null) {
            return fallbackValue;
        } else {
            return null;
        }

    },


    /**
     * Persists a module parameter.  This is used to support default sticky
     * state on views without a viewstate specified.
     *
     * @param {string} key The name of the parameter to persist.
     * @param {string} value The value of the parameter to persist.
     * @param {bool} isSessionOnly Indicates if the parameter change should
     *		only persist for the current session. Default is false.
     *
     */
    setParam: function(key, value, isSessionOnly) {

        this._params[key] = value;

        if (this.getBaseContext().getViewStateId()) {
            this.logger.debug('setParam - skipping param write; inside viewstate');
            return;
        }

        if ($.inArray(key, this._stickyParamList) == -1) {
            this.logger.info('setParam - skipping param write: "' + key + '" is not sticky');
            return;
        }

        if (!isSessionOnly) {

            if (!(typeof(value) == 'string') && !(typeof(value) == 'boolean') && isNaN(value)) {
                this.logger.debug(
                    'setParam - Cannot persist non-primitive value: key='
                    + key + ' type=' + typeof(value));
                return;
            }

            $.ajax({
                type: 'POST',
                url: this._buildParamUri(),
                data: this._buildParamName(key) + '=' + encodeURIComponent(value),
                dataType: 'json',
                success: this._setParamCallback.bind(this),
                error: this._setParamErrorback.bind(this)
            });
        }

    },

    /**
     * Returns the current parameter state of this module.  Only parameters
     * that are configured to be persistable are included.
     *
     */
    snapshotParamset: function() {

        var output = {};
        var value = null;
        for (var key in this._params) {
            if (this._params.hasOwnProperty(key)
                && $.inArray(key, this._persistableParamList) > -1)
            {
                value = this._params[key];
                if (!(typeof(value) == 'string') && !(typeof(value) == 'boolean') && isNaN(value)) {
                    this.logger.warn(
                        'snapshotParamset - Cannot persist non-primitive value: key='
                        + key + ' type=' + typeof(value));
                    continue;
                }
                output[key] = this._params[key];
            }
        }
        return output;

    },


    /**
     * Private method.  Handles the callback from setParam()
     *
     */
    _setParamCallback: function(data, textStatus) {
        this.logger.debug('_setParamCallback - response: ' + data);
    },


    /**
     * Private method.  Handles error conditions from the setParam() callback.
     *
     */
    _setParamErrorback: function(xhr, status, error) {
         this.logger.error('_setParamCallback - error: ' + xhr.responseText);
    },


    /**
     * Private method.  Returns the relative URI of the endpoint that accepts
     * parameter persistence.
     *
     */
    _buildParamUri: function(viewstate_id) {

        viewstate_id = viewstate_id || '_current';

        return Splunk.util.make_url(
            'app',
            Splunk.util.getCurrentApp(),
            Splunk.util.getCurrentView(),
            viewstate_id
        );

    },


    /**
     * Private method.  Returns the parameter key name that is accepted by the
     * parameter persistence URI, as returned by _buildParamUri()
     *
     */
    _buildParamName: function(key) {
        if (key == null) {
            throw Error('_buildParamName - Cannot build name; null key');
        }
        return encodeURIComponent(this.moduleId + '.' + key);
    },



    ///////////////////////////////////////////////////////////////////////////
    // Default content renders
    ///////////////////////////////////////////////////////////////////////////

    /**
     * By default, DispatchingModule instances will always attempt to ask the python
     * server for HTML content at the following location:
     *
     *         /module/<host_app>/<module_name>/render?client_app=<client_app>
     *
     * where <host_app> is the application that contains the code for the module,
     * while <client_app> is the user context from which this module has been requested
     *
     */
    getResultURL: function(params) {
        var uri = Splunk.util.make_url('module', Splunk.util.getConfigValue('SYSTEM_NAMESPACE'), this.moduleType, 'render');
        params = params || {};
        if (!params.hasOwnProperty('client_app')) {
            params['client_app'] = Splunk.util.getCurrentApp();
        }
        uri += '?' + Splunk.util.propToQueryString(params);
        return uri;
    },

    getResultParams: function () {
        return {};
    },

    getResultsErrorHandler: function(xhr, textStatus, exception) {
        if (typeof exception !== 'undefined') {
            throw exception;
        }
        else {
            this.logger.warn(this.moduleType, '.getResults returned exception=',
                exception,
                'and the following text status message: ',textStatus);
        }
    },

    getResultsCompleteHandler: function(xhr, textStatus) {},

    /**
     * Fetch the HTML contents for this module from server
     *
     */
    getResults: function() {
        var params = this.getResultParams();
        // have to be careful and make a copy of the params.
        // Otherwise once it's a property and not just a free variable,
        // when it gets passed into getResultURL, it would get passed by reference and not by
        // value.  This causes a little bit of havoc so to avoid that we just deep copy the params.
        this._previousResultParams = $.extend(true, {}, params);

        if (Splunk._testHarnessMode) {
            this.logger.debug(this.moduleType, ".getResults -- we are in _testHarnessMode. If you're not running unit tests right now, something is horribly wrong.")
            return false;
        }

        var resultUrl = this.getResultURL(params);
        if (!resultUrl) {
            this.logger.warn("getResultsURL() appears to be unimplemented or returning null for this instance.");
        }

        var callingModule = this.moduleType;

        // we augment all requests to pass along the module name, by default
        this.getResultsXHRObject = $.ajax({
            type: "GET",
            cache: ($.browser.msie ? false : true),
            url: resultUrl,
            beforeSend: function(xhr) {
                xhr.setRequestHeader('X-Splunk-Module', callingModule);
            },
            success: function(htmlFragment) {
                if (params["sid"]) {
                    var job = this.getBaseContext().job;
                    if (job && (params["sid"] != job.getSearchId())) {
                        this.logger.warn(this.moduleType, "while a request to a particular sid was in-flight, it appears the sid changed underneath us. We are dropping the response and NOT calling renderResults. old sid is", params["sid"], " and the new sid is ", job.getSearchId());
                        return;
                    }
                }
                this.renderResults(htmlFragment); 
            }.bind(this),
            complete: this.getResultsCompleteHandler.bind(this),
            error: this.getResultsErrorHandler.bind(this)
        });
    },

    haveResultParamsChanged: function() {
        var currentResultParams = this.getResultParams();
        return (!Splunk.util.objectSimilarity(this._previousResultParams, currentResultParams));
    },

    /**
     * Calls abort on the xhr object handling a getResult request.
     * This is often used when reseting a module's heirarchy to ensure
     * outgoing async requests don't end up populating a module's view erroneously.
     */
    abortGetResults: function() {
        // If the xhr request is not DONE, then we abort it.
        if (this.getResultsXHRObject && this.getResultsXHRObject.readyState < 4) {
            this.logger.info('Aborting getResults request for', this.moduleType);
            this.getResultsXHRObject.abort();
        }
    },

    /**
     *
     * Inserts content into module.  Default behavior is to ask appserver for content.
     *
     * override me to render the result of a getResults
     * Feel free to either replace or supplement the DOM elements that were
     * created by initialize().
     *
     * @param {String} htmlFragment The html to inject.
     * @param {Boolean} turbo Use shallow clone for faster html injection. CAUTION: All event listeners will be lost in clone.
     */
    renderResults: function(htmlFragment, turbo) {
        if (!htmlFragment) {
            this.resultsContainer.html('No content available.');
            return;
        }

        if (turbo) {
            this.resultsContainer = $(Splunk.util.turboInnerHTML(this.resultsContainer[0], htmlFragment));
        } else {
            this.resultsContainer.html(htmlFragment);
        }

        // fire any finalizing methods
        if (this.onResultsRendered) {
            this.onResultsRendered();
        }
    },

    /**
     * Calls the potentially overridden resetContext and resetUI for a total
     * refresh of the module and UI.
     */
    reset: function() {
        this.abortGetResults();
        this.resetContext();
        this.resetUI();
    },
    /**
     * Should be overridden by each module to reset its interface back to a
     * default state.
     */
    resetUI: function() {
        this.logger.warn(this.moduleType, 'has not implemented a resetUI method.');
    },


    /**
     * adjusts current module DIV width to its parent width
     *
     */
    resizeWidthToParent: function() {
        this.container.width(this.container.parent().width());
    },


    /**
     * Display an error message in the content of the module.
     * @param {String} message The message to display.
     */
    displayInlineErrorMessage: function(message) {
        this.logger.error(message);
        $('div.error', this.container).remove();
        $("<div>")
            .addClass("error")
            .text(message)
            .appendTo(this.container);
        this.container.show();
    },

    /////////////////////////////////////////
    //  Testing/Debugging Methods
    /////////////////////////////////////////
    /*
     * print a module hierarchy (diagnostic)
     */
    printHierarchy: function(indentation) {
        if (!indentation) indentation = "";
        this.logger.debug(indentation, this.moduleType);

        for (var i=0; i<this._children.length;i++) {
            this._children[i].printHierarchy(indentation + "    ");
        }
    }
});


