Splunk.Module.JobManager = $.klass(Splunk.Module, {
       
    PAUSE_CLASS: "pause",
    UNPAUSE_CLASS: "unpause",
    FINALIZE_CLASS: "finalize",
    CANCEL_CLASS: "cancel",
    SAVE_CLASS: "save",

    PAUSE_ACTION: "pause",
    UNPAUSE_ACTION: "unpause",
    FINALIZE_ACTION: "finalize",
    CANCEL_ACTION: "cancel",
    SAVE_ACTION: "save",

    /**
     * Endpoint to POST to and create batch changes in jobs
     */
    BATCH_EDIT_ENDPOINT: Splunk.util.make_url('api/search/jobs/control'),
    
    
    /**
     * Filters on the results.
     */
    _filters: {},

    initialize: function($super, container) {
        $super(container);

        // And he said "You shall call this window splunk_job_manager." and it was good.
        // Allows the job link in AccountBar to target the right window if someone opens JobManager directly.
        if (window.name == '') window.name = 'splunk_job_manager';

        this.logger = Splunk.Logger.getLogger("job_manager.js");
        this.messenger = Splunk.Messenger.System.getInstance();
        
        /**
         * Internal reference to the job management's X-Splunk-List-Length
         * header that reports the total number of jobs.
         * Maintaining an internal length allows us to reduce the # or calls
         * we need to make to the pagination endpoint.
         */
        this.length = 0;
       
        /**
         * Stores a reference to the paginator's callback, allowing the job manager
         * to update the pagination when its internal settings change.
         */ 
        this.totalCountCallback = null;

        /**
         * Stores an internal map of which rows have been checked.
         */
        this._checkedMap = {};

        /**
         * Setup the behavior on table;
         */
        this._setupEventHandlers();
        this._buildTableHeaderSorting();

        /**
         * Get the params from the query string.
         * This is pretty sketchy but helps with pre-filtering for some scenarios.
         */
        var search = Splunk.util.queryStringToProp(window.location.search.substring(1));
        if (search['filters']) {
            var filter = search['filters'].split(':');
            if (filter.length != 2) return;
            this._filters[filter[0]] = filter[1];
        }
        else if (search['savedSearch']) {
            this._filters['label'] = search['savedSearch'];
        }

        $(document).one('allModulesLoaded', function() {this.getResults()}.bind(this));    
    },

    /**
     * Get an ancestor Count module if any exist.  This is to allow Job Manager
     * to toggle visibility of the Count module correctly.  See SPL-21956.
     */
    toggleCountModules: function(action) {
        jQuery.each(
            jQuery.map(this.getAncestors(), function(module) {
                if (module.moduleType == 'Splunk.Module.Count') return module;
            }),
            function() {
                this[action]();
            }
        );
    },

    /**
     * Handles all the internal pagination updates and finally updates the table.
     */    
    onSettingsChange: function() {
        var owner = this.getSetting(this.getParam('ownerSetting'));
        var namespace = this.getSetting(this.getParam('namespaceSetting'));

        if (owner) {
            this._filters['user'] = owner;
        }
        if (namespace) {
            this._filters['app'] = namespace;
        }

        this.getResults();
    },
    
    /**
     * Generates the parameters used to sort and filter the job listing.
     * Settings in the config file overwrite those in the defaults.
     * @returns Object
     */
    getResultParams: function() {
        var params = {
            'count': this.getParam('count'),
            'offset': this.getParam('offset'),
            'sort_by': this.getParam('sortKey'),
            'sort_dir': this.getParam('sortDir'),
            'search': ''
        };

        if ('user' in this._filters || 
            'app' in this._filters ||
            'label' in this._filters) {
            params['filters'] = this.getFilters();
        }

        var baseSettings = this.getBaseSettings();
        for (var settingName in params) {
            if (settingName in baseSettings) {
                params[settingName] = baseSettings[settingName];
            }
        }
        return params;
    },
    
    /**
     * Returns an array of filter options.
     * Example:
     *      filter for jobs created by user admin => user:admin
     * @returns Array
     */
    getFilters: function() {
        var filters = [];
        for (var key in this._filters) {
            filters.push(key + ':' + this._filters[key]);
        }
        return filters;
    },
    
    /**
     * Implemented to handle updating the paginator.
     */ 
    getResultsCompleteHandler: function(xhr) {
        var length = xhr.getResponseHeader('X-Splunk-List-Length');
        var callback = this.getSetting('totalCountCallback');

        if (this.totalCountCallback == null && typeof callback == 'function') {
            this.totalCountCallback = callback;
        }
        if (length && this.length != length && this.totalCountCallback) {
            this.length = length;
            this.totalCountCallback(length);
        }
    },

    /**
     * Custom getResults method to handle update vs refresh requests.
     */
    getResults: function(callback, sids) {
        if (Splunk._testHarnessMode) return false;
        
        var resultUrl = this.getResultURL();
        var params = this.getResultParams();

        if (sids)
            params['sid'] = sids;
            
        var callingModule = this.moduleType;
        
        var options = {
            type: "GET",
            url: resultUrl,
            data: params,
            beforeSend: function(xhr) {
                xhr.setRequestHeader('X-Splunk-Module', callingModule);
            },
            complete: this.getResultsCompleteHandler.bind(this),
            error: this.getResultsErrorHandler.bind(this)
        };
        
        if (callback)
            options['success'] = callback.bind(this);
        else
            options['success'] = this.replaceResults.bind(this);
        
		$.ajax(options);		
   },
    
    _updateHTMLFragment: function(html) {
        html.filter('tr').each(this._addJobEventListener.bind(this));
        html.find('input[type=checkbox][name=sid]').each(function(i,elem) {
            if (elem.value in this._checkedMap) elem.checked = this._checkedMap[elem.value];
        }.bind(this));
    },
    
    updateResults: function(html) {
        html = $(html);
        this._updateHTMLFragment(html);
        html.filter('tr.parentRow').each(function(idx, elem) {
            elem = $(elem);
            var matchedRows = this.getJobRows(elem.attr('sid'));
            matchedRows.filter('tr.parentRow').unbind().replaceWith(elem);
        }.bind(this));
        html.filter('tr.childRow').each(function(idx, elem) {
            elem = $(elem);
            var matchedRows = this.getJobRows(elem.attr('sid'));
            matchedRows.filter('tr.childRow').unbind().replaceWith(elem);
        }.bind(this));
    },
    
    replaceResults: function(html) {
        $('h2.jobManagerStatus').hide();

        // Empty string or white space can't be parsed by jQuery
        if (html == '' || /^\s+$/.test(html)) { 
            return this.displayNoResultsFound();
        }

        try {
            html = $(html);
            if (html.length > 0) {
                this._updateHTMLFragment(html);
                $('table.jobManagerGrid tbody').html(html);
                $('table.jobManagerGrid').show();
                this.toggleCountModules('show');
                return;
            } else {
                return this.displayNoResultsFound();
            }
        // Assume a bad selector == zero length response
        } catch (e) {
            return this.displayNoResultsFound();
        }
    },

    /**
     * Hides the main grid of jobs,
     * shoots off a message to let the user know there is nothing to display.
     */
    displayNoResultsFound: function() {
        $('table.jobManagerGrid').hide();
        $('h2.jobManagerStatus.nojobs').show();
        this.toggleCountModules('hide');
    },

    /**
     * Return the array of sids based on checked checkboxes.
     * @returns Array
     */
    getCheckedSids: function() {
        var sids = [];
        $('table.jobManagerGrid input[name=sid]:checked').each(function() {
            sids.push(this.value);
        });
        return sids;
    },

    /**
     * Retrieve the rows currently in the table based on the job's sid.
     * @returns jQuery list object
     */
    getJobRows: function(sid) {
        return $("table.jobManagerGrid tr", this.container).filter(function(index) {
            return $(this).attr('sid') == sid;
        });
    },

    /**
     * Called either when an individual job is acted upon (canceled, paused, etc.)
     * or a group of jobs are acted upon.
     */
    _handleJobStatusChanged: function(event, sids, action) {
        sids = jQuery.makeArray(sids);
        switch(action) {
            case 'cancel':
                var offset = this.getParam('offset');
                var count = this.getParam('count');
                if (this.isLastPage() && ((this.length - sids.length) <= offset)) {
                    this.setParam('offset', (offset - count));
                }
                this.getResults();
                for(var i=0, j=sids.length; i<j; i++) {
                    if (this._checkedMap[sids[i]]) {
                        delete this._checkedMap[sids[i]];
                    }
                }
            break;
            case 'pause':
            case 'unpause':
            case 'finalize':
            case 'done':
            case 'save':
                this.getResults(this.updateResults, sids);
            break;
        }
    },

    isLastPage: function() {
        var length = this.length;
        var offset = this.getParam('offset');
        var count = this.getParam('count');
        return ((offset + count) >= length);
    },
    
    /**
     * Actually preforms the call to the batch job control endpoint.
     */
    _execControl: function(action, sid) {
        var module = this;
        var callback = function (resp) {
            var sids = [];
            var action = null;
            $.each(resp['data'], function(i, item) {
                if (item['response']) {
                    sids.push(item['sid']);
                    if (action == null) action = item['action'];
                }
            });
            module._handleJobStatusChanged(null, sids, action);
        };
        $.post(this.BATCH_EDIT_ENDPOINT, {action: action, sid: sid}, callback, 'json');
    },
    
    /**
     * Applies sorting behavior on the JobManager table.
     * **Using jQuery's native event binding (this == intended event target)
     * overcomes some ie6 compatibility issues, hence the reassociation of
     * this to module.
     */
    _buildTableHeaderSorting: function() {
        var module = this;
        $('table.jobManagerGrid th.sortable', this.container).click(function(event) {
            var elem = $(this);
            var abbr = elem.attr('abbr');
            if (!abbr) return;
            $('table.jobManagerGrid th.sortable[abbr='+module.getParam('sortKey')+']').find('span').removeClass('splSortAsc splSortDesc');
            module.setParam('sortKey', elem.attr('abbr'));
            if (module.getParam('sortDir') == 'desc') {
                module.setParam('sortDir', 'asc');
                elem.find('span').addClass('splSortAsc');
            }
            else {
                module.setParam('sortDir', 'desc');
                elem.find('span').addClass('splSortDesc');
            }
            module.getResults();
        });
    },
    
    /**
     * Attaches the various event handlers that handle the dynamic operations of the ui
     */
    _setupEventHandlers: function() {
        // Ensure the job rows are updated.
        $(document).bind('jobStatusChanged', this._handleJobStatusChanged.bind(this));
        $(document).bind('jobDone', function(event, job) {
            this._handleJobStatusChanged(event, job.getSID(), 'done');
        }.bind(this));

        
        // Batch edit button actions
        $('div.jmBatchActions button', this.container).click(function(event) {
            var elem = null;
            if ($.browser.msie) {
                elem = event.target;
            } else {
                elem = event.currentTarget;
            }
            elem = $(elem);

            if (elem.hasClass(this.PAUSE_CLASS)) {
                this._execControl(this.PAUSE_ACTION, this.getCheckedSids());
            }
            else if (elem.hasClass(this.UNPAUSE_CLASS)) {
                this._execControl(this.UNPAUSE_ACTION, this.getCheckedSids());
            }
            else if (elem.hasClass(this.FINALIZE_CLASS)) {
                this._execControl(this.FINALIZE_ACTION, this.getCheckedSids());
            }
            else if (elem.hasClass(this.CANCEL_CLASS)) {
                var sids = this.getCheckedSids();
                // if (sids.length > 0 && confirm(sprintf(_('Are you sure you want to delete %i job(s)?'), sids.length))) {
                if (sids.length > 0 && confirm(sprintf( ungettext('Are you sure you want to delete %i job?', 'Are you sure you want to delete %i jobs?', sids.length), sids.length))) {
                    this._execControl(this.CANCEL_ACTION, sids);
                }
            }
            else if (elem.hasClass(this.SAVE_CLASS)) {
                this._execControl(this.SAVE_ACTION, this.getCheckedSids());
            }
        }.bind(this));
        
        // Select one
        $('table.jobManagerGrid', this.container).click(function(event) {
            var elem = event.target;
            if (elem.tagName.toUpperCase() == 'INPUT' && elem.type.toUpperCase() == 'CHECKBOX') {
                this._mapInput(elem);
            }
        }.bind(this));
        
        // Select all
        var module = this;
        $('div.jmSelect a.selectAll', this.container).click(function(event) {
            $('table.jobManagerGrid tr.parentRow input').each(function() {
                this.checked = true;
                module._mapInput(this);
            });
        });
        $('div.jmSelect a.selectNone', this.container).click(function(event) {
            $('table.jobManagerGrid tr.parentRow input').each(function() {
                this.checked = false;
                module._mapInput(this);
            });
        });
    },

    /**
     * Function provided to a jQuery .each method.
     * Attaches the appropriate event triggers to each job row.
     */
    _addJobEventListener: function(idx, row) {
        var module = this;
        $(row).click(function(event) {
            var sid = $(row).attr('sid');
            if (!sid) return;

            var target = event.target;
            if (target.tagName.toUpperCase() == "A") {
                target = $(target);
                if (target.hasClass('viewJob')) {
                    // Adding in a quick call to touch so that the job doesn't expire inbetween the user
                    // clicking the view link and the view opening.                    
                    var tmpJob = new Splunk.Job('*');
                    tmpJob.setSearchId(sid);
                    tmpJob.touch();
                    
                    // Open the job in a new window
                    Splunk.window.open(target.attr('href'), sid, {
                        'width': 1024,
                        'height': 768
                    });
                    
                    event.preventDefault();
                    event.stopPropagation();
                    return false;
                }
                else {

                    var errorHandler = function(xhr, type, error, action) {
                        if (xhr.status == 404) {
                            module.messenger.send('error', 'splunk.job_manager', sprintf(_('Splunk cannot %(action)s this job. It may have expired. Try refreshing this page.'), {'action': action}));
                        }
                        else {
                            module.messenger.send('error', 'splunk.job_manager', sprintf(_('Received unexpected error: %(error)s'), {'error': error}));
                        }
                    }

                    var tmpJob = new Splunk.Job('*');
                    tmpJob.setSearchId(sid);

                    if (target.hasClass('cancelJob') && !target.hasClass('cancelJobDisabled')) {
                        tmpJob.cancel(null, function(xhr, type, error, action) {
                            if (xhr.status == 404) {
                                module.messenger.send('error', 'splunk.job_manager', _('Splunk cannot delete this job. It may have expired. Try refreshing this page.'));
                            }
                            else {
                                module.messenger.send('error', 'splunk.job_manager', sprintf(_('Received unexpected error: %(error)s'), {'error': error}));
                            }
                        });
                    }
                    else if(target.hasClass('finalizeJob') && !target.hasClass('finalizeJobDisabled')) {
                        tmpJob.finalize(null, errorHandler);
                    }
                    else if (target.hasClass('pauseJob') && !target.hasClass('pauseJobDisabled')) {
                        tmpJob.pause(null, errorHandler);
                    }
                    else if (target.hasClass('unpauseJob') && !target.hasClass('unpauseJobDisabled')) {
                        tmpJob._isPaused = true;
                        tmpJob.unpause(null, errorHandler);
                    }
                    else if (target.hasClass('saveJob')) {
                        tmpJob.save(null, errorHandler);
                    }
                }
            }
            
            // Allow clicking on a row to check / uncheck the checkbox
            else if (target.tagName.toUpperCase() != 'INPUT'){
                var input = module.getJobRows(sid).find('input');
                if (input.attr('checked')) {
                    input.attr('checked', false);
                }
                else {
                    input.attr('checked', true);
                }
                module._mapInput(input[0]);
            }
        }).hover(function(){
            module._rowOver(this);
        }, function(){
            module._rowOut(this);
        });
         
        $(row).next('tr.childRow').hover(function(){
            module._rowOver(this);
        }, function() {
            module._rowOut(this);
        });
    },
    
    /**
     * Update the styles of the focused row.
     */
    _rowOver: function(orig) {
        if ( $(orig).hasClass('parentRow') ) {
            $(orig).addClass('jmRowHighlight')
                .next('tr.childRow').addClass('jmRowHighlight');
        } else if ( $(orig).hasClass('childRow') ) {
            $(orig).addClass('jmRowHighlight')
                .prev('tr.parentRow').addClass('jmRowHighlight');
        }
    },

    /**
     * Update the styles of the focused row.
     */
    _rowOut: function(orig) {
        if ( $(orig).hasClass('parentRow') ) {
            $(orig).removeClass('jmRowHighlight')
                .next('tr.childRow').removeClass('jmRowHighlight');
        } else if ( $(orig).hasClass('childRow') ) {
            $(orig).removeClass('jmRowHighlight')
                .prev('tr.parentRow').removeClass('jmRowHighlight');
        }
    },
    
    _mapInput: function(elem) {
        if (elem.checked) {
            this._checkedMap[elem.value] = elem.checked;
        }
        else {
            delete this._checkedMap[elem.value];
        }
    }
});
