/**
 * Message twice it's a long way to the bay!
 */
Splunk.Module.Message = $.klass(Splunk.Module, {
    EXCLUDE_LEVELS: [],
    /**
     * Define regex used to convert a mediawiki-style link token to a proper <a href> tag.  The basic pattern is:
     * [<external_link> <label>]
     * [[<internal_link>|<label>]]
     *
     * The target window can be specified by prefixing a (!):
     * [!<external_link> <label>]
     * [[!<internal_link>|<label>]]
     *
     * Example:
     * [http://www.google.com Google Search]
     * [[/manager/settings|Edit server settings]]
     *
     * TODO: If this ever needs more wiki syntax, we'll have to move to a proper tokenizer; this is just quick and dirty.
     */
    EXTERNAL_LINK_REX: /\[(\!?)(\w+\:\/\/[^\s]+)\s+([^\]]+)\]/g,
    INTERNAL_LINK_REX: /\[\[(\!?)([^\|]+)\|([^\]]+)\]\]/g,
    TARGET_BLANK_TOKEN_REX: /#~#!#~#/g,
    TARGET_SAME_TOKEN_REX: /\s*#~##~#/g,
    /**
     * Constructor
     */
    initialize: function($super, container){
        $super(container);
        this.logger = Splunk.Logger.getLogger("message.js");
        this.messages = [];
        this.maxSize = parseInt(this._params['maxSize'], 10);
        this.clearOnJobDispatch = Splunk.util.normalizeBoolean(this._params['clearOnJobDispatch']);
        this.defaults = this._params['default'] || [];
        this.filter = this._params['filter'];
        this.list = $(".MessageList", this.container);
        this.messenger = Splunk.Messenger.System.getInstance();
        this.messenger.receive(this.filter, this.unshift.bind(this), undefined, true);//Listen to all including control messages
        this.sendDefaults();//Send default configured messages to the UI.
        if(this.clearOnJobDispatch){
            $(document).bind('jobDispatched', this.onJobDispatched.bind(this));
        }
    },
    /**
     * Clear all message list elements from the DOM.
     *
     * @param {Array} arguments[0] Optional array of levels to exclude from clear process.
     */
    clear: function(){
        var levels = arguments[0] || [];
        var messages = [];
        this.logger.warn("Clearing messages with the exclusion of levels", levels);
        for(var i=0; i<this.messages.length; i++){
            var message = this.messages[i];
            if($.inArray(message.level, levels)!=-1){
                messages.push(message);
            }
        }
        this.messages = messages;
        this.list.html(this.getHTMLTransform());
    },
    /**
     * Simple handler for control messages.
     * 
     * @param {Object} message Standard messenger message object ({level:String, content:String, className:String, date:String}).
     */
    handleControlMessage: function(message) {
        switch(message.content) {
            case 'CLEAR':
                this.clear();
                break;
        }
    },
    /**
     * Triggered when a module receives a new job dispatched.
     *
     * @param {Object} event A jquery event.
     */
    onJobDispatched: function(event, context){
        this.clear(this.EXCLUDE_LEVELS);
    },
    /**
     * Generate the HTML content for display.
     * 
     * @type String
     * @return An HTML formatted string of all messages.
     */
    getHTMLTransform: function(){
        var html = [];
        for(var i=0; i<this.messages.length; i++){
            var message = this.messages[i];
            html.push('<li class="'+message.level+'">');
            html.push(this.getWikiTranform(message.content));
            html.push('</li>');
        }
        return html.join('');
    },
    /**
     * Transform a string against wiki formatting rules.
     *
     * @param {String} str The string to format.
     * 
     * @type String
     * @return A HTML formatted string. 
     */
    getWikiTranform: function(str){
        var text = Splunk.util.escapeHtml(str);
        // we tokenize the conditional '!' prefix and replace it after
        text = text.replace(this.EXTERNAL_LINK_REX, '<a href="$2" #~#$1#~#>$3</a>');
        text = text.replace(this.INTERNAL_LINK_REX, '<a href="$2" #~#$1#~#>$3</a>');
        text = text.replace(this.TARGET_BLANK_TOKEN_REX, 'target="_blank"');
        text = text.replace(this.TARGET_SAME_TOKEN_REX, '');
        return text;
     },
    /**
     * Send default configured messages to the UI.
     */
    sendDefaults: function(){
        for(var i=0; i<this.defaults.length; i++){
            var level = this.defaults[i].level || null;
            var content = this.defaults[i].content || null;
            var className = this.defaults[i].className || null;
            if(level && content && className){
                this.messenger.send(level, className, content);
            }else{
                this.logger.error("Invalid default settings for entry with level:", level, "content:", content, "className:", className);
            }
        }
    },
    /**
     * Adds one or more messages to the beginning of an ordered list and
     * returns the new length (behaves like the Array unshift method).
     *
     * @param {Arguments} arguments.length>0 Optional message objects to be added to the display ({level:String, content:String, className:String, date:String}).
     *
     * @type Number
     * @return The new length of messages.
     */
    unshift: function(){
        var previousMessages = [].concat(this.messages);
        for(var i=arguments.length-1; i>-1; i--){
            var message = arguments[i];
            if(message.control){
                this.handleControlMessage(message);
                continue;
            }
            for(var j=0; j<this.messages.length; j++){
                var content = this.messages[j].content;
                var level = this.messages[j].level;
                if(message.content==content && message.level==level){
                    this.messages.splice(j, 1);
                    j--;
                }
            }
            this.messages = [{level: message.level, content: message.content}].concat(this.messages);
        }
        this.messages.splice(this.maxSize, this.messages.length-this.maxSize);
        if(!Splunk.util.objectSimilarity(previousMessages, this.messages)){
            this.list.html(this.getHTMLTransform());
        }
        return $("li", this.list).length;
    }
});