Splunk.Module.Paginator = $.klass(Splunk.Module.DispatchingModule, {
    EVENTS_ENTITY_NAME: "events",
    RESULTS_ENTITY_NAME: "results",
    SETTINGS_MAP_ENTITY_NAME: "settings",
    /**
     * Paginate twice it's a long way to the bay!
     */
    initialize: function($super, container){
        $super(container);
        this.logger = Splunk.Logger.getLogger("paginator.js");

        this.useParamValuesAsSettings(["count", "maxPages"]);
        this.entityName = this._params['entityName'];
        this.length = 0;
        this.offset = 0;
        this.resultsContainer = $(".ResultsContainer", this.container);
        this.bindUIEvents();//Bootstrap top level event dispatcher.
    },
    onBeforeJobDispatched: function(context) {
        // we need at least 1 if the pager is set to look at events.
        if (this.entityName == this.EVENTS_ENTITY_NAME) {
            context.setMinimumStatusBuckets(1);
        }
    },
    /**
     * Bind a generic UI event listener.
     */
    bindUIEvents: function(){
        this.container.bind("click keydown", this.onUIEvent.bind(this));
    },
    /**
     * Return the baseContext item count based on a module entityName value (EVENTS_ENTITY_NAME, RESULTS_ENTITY_NAME or SETTINGS_MAP_ENTITY_NAME).
     */
    getEntityCount: function(){
        var count;
        var context = this.getBaseContext();
        switch(this.entityName){
            case this.EVENTS_ENTITY_NAME:
                //SearchContext now has it's own getEventAvailableCount
                //that will return the correct answer even when the user has 
                //selected a subset of the timerange  
                count = context.getEventAvailableCount();
                break;
            case this.RESULTS_ENTITY_NAME:
                count = context.job.getResultCount();
                break;
            case this.SETTINGS_MAP_ENTITY_NAME:
                count = this.length;
                break;
            default:
                this.logger.error("Invalid module entityName value of", this.entityName);
                count = 0;
        }
        return count;
    },
    /**
     * Override default.
     */
    getSettings: function(){
        var settings = {
            count:    this.getSetting("count"),
            maxPages: this.getSetting("maxPages"),
            offset:   this.offset
        };
        if (this.entityName == this.SETTINGS_MAP_ENTITY_NAME) {
            settings['totalCountCallback'] = function(length) {
                this.length = (length != null) ? length : this.length;
                this.getResults();
            }.bind(this);
        }
        // pass a reference to onOffsetChange down, for any second paginators we may find below.
        settings["upstreamPaginator"] = this;
        return settings;
    },
    /**
     * Override default.
     */
    getResultParams: function(){
        return {
            "length":    this.getEntityCount(),
            "count":     this.getSetting("count"),
            "max_pages": this.getSetting("maxPages"),
            "offset":    this.offset
        };
    },
    /**
     * Does room exist for more pages to be displayed (Used for request throttling).
     */
    hasCapacity: function(){
        return ($("li.page", this.container).length < this.getSetting("maxPages"));
    },
    /**
     * Handle job complete and retrieve new results if required.
     */
    onJobDone: function(event){
        if(this.getEntityCount()==0){
            this.resultsContainer.html("");
        }
    },
    /**
     * Handle job progress notification and retrieve new results if required.
     *
     * @param {Object} event A jQuery event.
     */
    onJobProgress: function(event){
        if(this.hasCapacity() && this.getEntityCount()>0){
            this.getResults();
        }
    },
    /**
     * Handle a UI event and retrieve results response.
     *
     * @oaram {Object} element The DOM element that triggered the event.
     */
    onOffsetChange: function(element){
        element = $(element);//Cast to a jquery element.
        var resource = element.attr("href");
        var query = resource.split("?")[1];
        try{
            this.offset = parseInt(Splunk.util.queryStringToProp(query).offset, 10);
        }catch(err){
            this.logger.error("Could not parse offset from uri.", err);
            return false;
        }
        this.pushSettingsToChildren();
        this.getResults();
        return false;//Cancel the click/keyboard event from making a request.
    },
    /**
     * Handles a new search.
     */
    onContextChange: function(){
        this.offset = 0;
        if(this.getBaseContext().isJobDispatched()){
            if(this.getEntityCount()>0){
                this.getResults();
            } else{
                this.resultsContainer.html('');
            }
        }
    },
    /**
     * Handle setting changes.
     *
     * @param {Object} settings An object literal containing various types for values.
     */
    onSettingsChange: function(settings){
        // if there is an upstream paginator, we catch it's offset and if it's different, we update ourselves. 
        var hasUpstreamPaginator = (this.getSetting("upstreamPaginator") != null);
        // this means the upstream paginator has sent us an offset value.
        if (hasUpstreamPaginator && this.getSetting("offset") != null) {
            // the offset they sent is different from our internal offset.
            if (this.offset != this.getSetting("offset")) {
                this.offset = this.getSetting("offset");
            }
        }
        if (this.haveResultParamsChanged()){
            //Reset offset when count change is greater than offset.
            if (this.offset != 0 && settings.hasOwnProperty('count') && parseInt(settings.count, 10)>this.offset){
                this.offset = 0;
            }
            this.getResults();
        }
    },
    /**
     * Top level UI event listener and dispatcher.
     *
     * @param {Object} event A jQuery event.
     */
    onUIEvent: function(event){
       var eventType = event.type;
       var eventTarget = event.target;//What was the source element of the event.
       if((eventType==="click" || eventType==="keydown") && $(eventTarget).is("a")){
           if ($(eventTarget).hasClass('disabled')) return false;
           var upstreamPaginatorReference = this.getSetting("upstreamPaginator")
           if (upstreamPaginatorReference) {
                upstreamPaginatorReference.onOffsetChange(eventTarget);
                // unless the upper paginator is already visible onscreen, scroll up to it.
                $(window).scrollTop(Math.min($(window).scrollTop(), upstreamPaginatorReference.container.offset().top));
           }
           return this.onOffsetChange(eventTarget);
       }
    },
    /**
     * Override render results so the message is set to an empty string if no content is available.
     */
    renderResults: function($super, htmlFragment){
        if(!htmlFragment){
            this.resultsContainer.html("");
        }else{
            $super(htmlFragment);
        }
    },
    /**
     * Reset the UI to its original state.
     */
    resetUI: function(){
        this.offset = 0;
        this.length = 0;
        this.pushSettingsToChildren();
        this.resultsContainer.html("");
    },

    requiresDispatch: function() {
        var entityName = this.getParam('entityName');
        if (entityName == this.SETTINGS_MAP_ENTITY_NAME) return false;
        return true;
    }
});
