//put Module in the namespace if it isnt already there.
Splunk.namespace("Module.DispatchingModule");

// A simple ascii representation of the Splunk timeline
Splunk.Module.AsciiTimeline = $.klass(Splunk.Module.DispatchingModule, {
    _selectionTimeRange : false,
   
    /*
     * overriding initialize to set up references and event handlers.
     */
    initialize: function($super, container) {
        $super(container);
        this.logger = Splunk.Logger.getLogger("ascii_timeline.js");
        this._chartContainer   = $(".timelineChart", this.container);
        this.selectionHeaderContainer = $(".selectionHeader", this.container);       
    },
    onBeforeJobDispatched: function(context) {
        // needs at least one status bucket in it's dispatched searches.
        context.setMinimumStatusBuckets(300);
    },
    getResultURL: function() {
        var context = this.getBaseContext();
        return context.getUrl('timeline');
    },
    onJobProgress: function(event) {
        this.logger.debug(this.moduleType + " onJobProgress");
        this.getResults();
    },
    onContextChange: function() {
        this._selectionTimeRange = false;
    },
    /* 
    We override getModifiedContext, because this class implements a 'selection' state. In this case
    it implements a "Selection Time" behaviour.  Children will display data from the selected time range.
    */
    getModifiedContext: function() {
        if (this._selectionTimeRange) {
            this.logger.debug(this.moduleType, ".getModifiedContext - has a selectionRange so it will return a modified context.");

            // TODO - when it's children are eventsViewers, fieldViewers, fieldPickers, etc... 
            //        Then the same searchId can be reused.
            //        However when there's a child like FlashChart or ResultsTable
            //        that gets results, we need to redispatch.
            //        Currently the plan is to make requiresDispatch take a
            //        context argument, and become literally context-dependent,
            //        but this hasnt been done yet.
            var context = this.getBaseContext();
            context.setTimeRange(this._selectionTimeRange);
            return context;
        }
        else {
            return this.getBaseContext();
        }
    },
    renderResults: function(resultXML) {
        this._chartContainer.html('');

        var table = $('<table width="95%" />');
        var tr = $('<tr />').appendTo(table);

        var maxCount = 0;
        var maxChars = 10;
       
        // one loop to find the maxCount.
        $(resultXML).find('timeline').find("bucket").each(function() {
            maxCount = Math.max(maxCount, $(this).attr("c"));
        });
       
        $('<td class="leftAxis">' + maxCount + '--</td>').appendTo(tr);
       
        // another loop to draw the bars.
        // Since we need both a) to reference the object in context, with 'this', and b) to reference this Module instance,
        //     we need to create a small closure here.
        var moduleInstance = this;
        $(resultXML).find('timeline').find("bucket").each(function() {
           
            var count = $(this).attr("c");
            var charCount = Math.ceil(maxChars * count / maxCount);
           
            var td = $('<td class="bar" />');
            for (var i=0; i<charCount; i++) {
                $('<div />').html('X').appendTo(td);
            }
           
            td.attr("count",count);
            td.attr("starttime",$(this).text());
            td.attr("duration",$(this).attr("d"));
            td.mousedown(function(mozEvent) {
                moduleInstance._handleMouseDown(mozEvent,this);
            });
            td.mouseup(function(mozEvent) {
                moduleInstance._handleMouseUp(mozEvent,this);
            });
            td.appendTo(tr);
           
        });
       
        $('<td class="rightAxis">--' + maxCount + '</td>').appendTo(tr);
       
        // highlight histogram by selectionTimeRange
        if (this._selectionTimeRange) {
            this.logger.warn("selectionTimeRange =", this._selectionTimeRange.toConciseString());
            this._highlightTimeRange(tr);
        }
        table.appendTo(this._chartContainer);
       
       
    },
   
    _handleMouseDown: function(evt,tdNode) {
        // prevent the ascii bars from getting selected by dragging.
        evt.preventDefault();
        // if this is a shift click, then we do nothing on mousedown.
        if (evt.shiftKey) return false;
        this._mouseDownTimeRange = this._getTimeRangeForBucket(tdNode);
        return false;
    },
    _handleMouseUp: function(evt,tdNode) {
        evt.preventDefault();
        if (!this._mouseDownTimeRange) return false;
        var mouseUpTimeRange = this._getTimeRangeForBucket(tdNode);
        this.logger.debug(this.moduleType, "mouseUp timeRange=", mouseUpTimeRange.toConciseString());
        this.logger.debug(this.moduleType, "mouseDown timeRange=", this._mouseDownTimeRange.toConciseString());

        var selectionTimeRange;
        if (this._mouseDownTimeRange.getAbsoluteEarliestTime() >= mouseUpTimeRange.getAbsoluteLatestTime()) {
            selectionTimeRange = new Splunk.TimeRange(mouseUpTimeRange.getAbsoluteEarliestTime(),this._mouseDownTimeRange.getAbsoluteLatestTime());
        }
        else {
            selectionTimeRange = this._mouseDownTimeRange.copy();
        }
        // set the internal property that will change the job inherited by children.
        this._selectionTimeRange = selectionTimeRange;

        // highlight histogram by selectionTimeRange
        this._highlightTimeRange();

        // push the modified search to our children.
        this.pushContextToChildren();
        return false;
    },
   
    _getTimeRangeForBucket: function(start, end) {
        if (!start && !end) {
            this.logger.debug(this.moduleType, "._getTimeRangeForBucket given undefined buckets. exiting and returning 'all time'");
            return new Splunk.TimeRange();
        }
        else if (!start) {
            this.logger.error(this.moduleType, "._getTimeRangeForBucket - start undefined. Using end");
            start = end;
        }
        // quite a normal case.  For convenience end argument is often omitted.
        else {
            if (!end) end = start;
        }

        // this block retains this bakedFoo optimization from oxiclean.
        if (!start.bakedStartTime || !end.bakedEndTime) {
            var startTimeStr = start.getAttribute("starttime");
            var timeFormatStr = Splunk.util.getConfigValue('DISPATCH_TIME_FORMAT');
            start.bakedStartTime = Splunk.util.parseDate(startTimeStr,timeFormatStr);
            var endStartTimeStr = end.getAttribute("starttime");
            if (endStartTimeStr) {
                endDuration = parseInt(end.getAttribute("duration"));
                endStartTime = Splunk.util.parseDate(endStartTimeStr,timeFormatStr);
                try {
                    end.bakedEndTime = new Date(endStartTime.getTime()   + ((endDuration)*1000));
                } catch(e) {
                    this.logger.error(this.moduleType, ".getTimeRangeForBucket - parsing exception around date='", endStartTimeStr, "'\n", e);
                }
            }
            else {
                this.logger.warn("TimelineManager.getTimeRangeForBucket - no timerange defined for node, which implies 'All time'.");
                start.bakedStartTime = null;
                start.bakedEndTime = null;
            }
        }
        var range = new Splunk.TimeRange(start.bakedStartTime, end.bakedEndTime);
        return range;
    },
    _highlightTimeRange: function(chartContainer) {
        if (!chartContainer) chartContainer = this._chartContainer;
        this.logger.debug(this.moduleType, ".highlightHistogramByTimeRange -", this._selectionTimeRange);
        var totalCount = 0;
        var bars =  $('td.bar', chartContainer);
        for (var i=0; i<bars.length;i++) {
            var bucketTimeRange = this._getTimeRangeForBucket(bars[i]);   
            var bar = $(bars[i]);
            if (this._selectionTimeRange.containsRange(bucketTimeRange))   {
                bar.addClass("highlight");
                totalCount += parseInt(bar.attr("count"));
            }
            else bar.removeClass("highlight");
        }
        // update the selection date display.
        this._updateSelectionHeader(totalCount);
    },
    _updateSelectionHeader: function(totalCount) {
        var headerStr = ["selected:"];
        headerStr.push(Splunk.util.getCommaFormattedNumber(totalCount));
        headerStr.push("events");
        headerStr.push(this._selectionTimeRange.toConciseString());
       
        this.selectionHeaderContainer.text(headerStr.join(" "));
    },
   
    resetUI: function() {
        this._chartContainer.empty();
    }
});
