import cherrypy, logging, math, os, splunk.search, splunk.bundle, splunk.util
import controllers.module as module
from operator import itemgetter
from splunk.appserver.mrsparkle.lib import i18n, util
import datetime

logger = logging.getLogger('splunk.modules.events_viewer')

MAX_EVENTS_CONSTRAINT = 100 # Browser crash control
MAX_LINES_CONSTRAINT_DEFAULT = 500 # Browser crash control (currently set to backend limit)
MIN_LINES_CONSTRAINT = 5
TRUNCATION_MODE_DEFAULT = "abstract"
RENDERER_DEFAULT = 'default'
TEMPLATE_DEFAULT = 'EventsViewer_default_renderer.html'
RENDERER_SELECT_FIELD = 'eventtype'
SEGMENTATION_DEFAULT = "none"
SHOW_SOURCE_REQUIRED_FIELDS = ['_cd', 'source', 'host', 'index']
HTML_TEMPLATE = 'results/EventsViewer_generate_results.html'

class EventsViewer(module.ModuleHandler):
    
    def generateResults(self, host_app, client_app, sid, count=10, display_row_numbers=1, enable_event_actions=1, enable_field_actions=1, earliest_time=None, latest_time=None, field_list=None, offset=0, has_layout=1, max_lines=None, max_lines_constraint=None, min_lines=None, replace_newlines=0, segmentation=None, truncation_mode=None, entity_name='events', **args):
        '''
        HTML formatted set of events for a given sid.
        
        Arg:
        host_app: The name of the application in which this module is hosted.
        client_app: The name of the application from where the request originated.
        display_row_numbers: Controls the display of positional numbers or not, 0 or 1.
        sid: The job sid.
        count: The maximum amount of events to show.
        earliest_time: The earliest time constraint.
        field_list: Comma separated list of fields.
        has_layout: Control events_viewer_generate_results.html layout wrapper.
        latest_time: The latest time constraint.
        max_lines: The maximum event lines to show, None, 0 or >max_lines_constraint will display max_lines_constraint
        min_lines: The minimum event lines to show, None or <MIN_LINES_CONSTRAINT will display MIN_LINES_CONSTRAINT
        max_lines_constraint: The physical constraint for visible max lines, used for browser crash control defaults to MAX_LINES_CONSTRAINT_DEFAULT if None.
        truncation_mode: Job related truncation mode for splunkd, defaults to TRUNCATION_MODE_DEFAULT.
        replace_newlines: Control replacement of newline characters.
        offset: The offset starting point of the event series.
        '''

        
        # get handle to job object, return error if not accessible.
        response_data = {}
        try:
            job = splunk.search.getJob(sid)
        except splunk.ResourceNotFound, e:
            response_data['error'] = _('The job appears to have expired. Splunk could not retrieve events for this search.<!-- Exception: %s -->') % e
        except Exception, e:
            response_data['error'] = _('An error occurred. Splunk could not retrieve events for this search.<!-- Exception: %s -->') % e
        if response_data.get('error', None):
            return self.controller.render_template(HTML_TEMPLATE, response_data)

        # get display position setting
        display_row_numbers = splunk.util.normalizeBoolean(display_row_numbers)

        # get enable event actions
        enable_event_actions = splunk.util.normalizeBoolean(enable_event_actions)
        
        # get enable field actions
        enable_field_actions = splunk.util.normalizeBoolean(enable_field_actions)
        
        # get fields for event
        if field_list is None:
            field_list = []
        else:
            field_list = splunk.util.stringToFieldList(field_list)

        # cast has_layout to Boolean
        has_layout = splunk.util.normalizeBoolean(has_layout)

        # get segmentation for events
        if segmentation is None:
            segmentation = SEGMENTATION_DEFAULT

        # get max_lines_constraint or default constraint
        if max_lines_constraint is None:
            max_lines_constraint = MAX_LINES_CONSTRAINT_DEFAULT
        else:
            max_lines_constraint = int(max_lines_constraint)

        # get max_lines setting or default constraint
        if max_lines is None:
            max_lines = max_lines_constraint
        else:
            max_lines = min(int(max_lines), max_lines_constraint)
            #0 value delimits everything up to safe browser constraint
            if max_lines is 0:
                max_lines = max_lines_constraint

        # get min_lines setting or default constraint
        if min_lines is None:
            min_lines = MIN_LINES_CONSTRAINT
        else:
            min_lines = max(int(min_lines), MIN_LINES_CONSTRAINT)

        # get replace new lines
        replace_newlines = splunk.util.normalizeBoolean(replace_newlines)

        if truncation_mode is None:
            truncation_mode = TRUNCATION_MODE_DEFAULT
            
        # set search sdk fetch options
        # note: python sdk timeFormat setFetchOption breaks events[0:].time SPL-18484
        job.setFetchOption(
            time_format = cherrypy.config.get('DISPATCH_TIME_FORMAT'),
            earliestTime=earliest_time, 
            latestTime=latest_time, 
            maxLines=max_lines, 
            segmentationMode=segmentation, 
            truncationMode=truncation_mode,
            # part 1 of SPL-22806 add click support to timestamp in events viewer
            # output_time_format was formerly i18n.ISO8609_MICROTIME
            # however to get epochTime to the client, and workaround various problems in SDK/dt2epoch, 
            # the simplest way was to get the epochTime value from the API directly.
            output_time_format=i18n.ISO8609_MICROTIME
            
        )

        # validate offset
        offset_start = max(int(offset), 0)

        # pick up the correct result range, depending on the resource requested
        if count: 
            offset_end = min(offset_start + int(count), job.eventAvailableCount)
        elif entity_name == 'results': 
            offset_end = job.resultCount
        else: 
            offset_end = job.eventAvailableCount

        if offset_end - offset_start > MAX_EVENTS_CONSTRAINT:
            logger.warn('number of results requested exceeded limit; truncating to %s results' % MAX_EVENTS_CONSTRAINT)
            offset_end = offset_start + MAX_EVENTS_CONSTRAINT
            
        # support pulling result data into event viewer
        if entity_name == 'results':
            # results typically have a different field list from event feed;
            # reset the requested field list to include eventtype field,
            # otherwise the custom renderer will never trigger
            field_list = job.results.fieldOrder
            events = job.results[offset_start:offset_end]
        else:
            events = job.events[offset_start:offset_end]

        # retrieve available event renderers
        rendererMap, priorityList = self._getRenderers(client_app)
        
        # augment the event object to have additional attributes
        for i, event in enumerate(events):
            
            # part 1 of SPL-22806 add click support to timestamp in events viewer
            event.epoch = splunk.util.dt2epoch(splunk.util.parseISO(str(event.time)))

            # note that since getTZOffsetMinutes is a little unusual, it's in splunk.appserver.mrsparkle.lib.util
            # aka the 'other util'
            event.timezoneOffset = util.getTZOffsetMinutes(float(event.epoch))
            event.time = datetime.datetime.fromtimestamp(float(event.epoch))


            event.offset = i + offset_start
            event = self._setRenderer(event, RENDERER_SELECT_FIELD, rendererMap, priorityList)
            event.event_menu = self._getEventMenu(job, event, client_app, latest_time=latest_time)
            
            
            

        # template args
        response_data = {
            "host_app": host_app,
            "client_app": client_app,
            "display_row_numbers": display_row_numbers,
            "earliest_time": earliest_time,
            "enable_event_actions": enable_event_actions,
            "enable_field_actions": enable_field_actions,
            "events": events,
            "field_list": field_list,
            "has_layout":has_layout,
            "job":job,
            "latest_time": latest_time,
            "min_lines": min_lines,
            "max_lines": max_lines,
            "max_lines_constraint": max_lines_constraint,
            "offset_end": offset_end,
            "offset_start": offset_start,
            "replace_newlines": replace_newlines,
            "segmentation": segmentation
        }
        logger.info("rendering for %s" % client_app)
        return self.controller.render_template(HTML_TEMPLATE, response_data)

    def _setRenderer(self, event, selectField, rendererMap, rendererPriorityList):
        # set defaults
        event.renderer = RENDERER_DEFAULT
        event.renderer_template = TEMPLATE_DEFAULT

        # get the events field values and turn it into a python list
        fieldValues = event.get(selectField, None)
        if fieldValues is None: return event
        fieldValuesList = [x.value for x in fieldValues._fieldValue]

        # walk down the list and find the renderer with the highest priority
        # note that the renderer's name should match a field value
        renderer = None
        fieldValue = None
        for renderer in rendererPriorityList:
            if renderer in fieldValuesList:
                fieldValue = renderer
                break

        # if not custom renderer was found, return the event with default rendering
        if fieldValue == None: return event

        # assign the custom renderer to this event
        event.renderer_template = rendererMap[ fieldValue ]['path']
        event.renderer = fieldValue

        return event

    def _getRenderers(self, fromApp = None):
        '''
        build the map between the a field and their custom renderers
        renderMap = {<eventtype>: <template_path>, ... }
        
        cache this because the config read is expensive
        '''
        TEMPLATE_APPSCOPE_SEPARATOR = ':'
            
        # build the map between the a field and their custom renderers
        # renderMap = {<eventtype>: <template_path>, ... }
        rendererMap = {}
        renderConf = splunk.bundle.getConf('event_renderers', namespace=fromApp)
        logger.info("only loading renderers from %s" % fromApp)

        for stanza in renderConf.findStanzas().keys():
            if renderConf[stanza].has_key(RENDERER_SELECT_FIELD):
                rendererMap[ renderConf[stanza][RENDERER_SELECT_FIELD] ] = {
                    'path'      : '/event_renderers/%s' % renderConf[stanza]['template'],
                    'priority'  : int(renderConf[stanza].get('priority', 100) )
                }
                appScope = renderConf[stanza].get('template_app', False)
                if appScope:
                    path = rendererMap[ renderConf[stanza][RENDERER_SELECT_FIELD] ]['path']
                    rendererMap[ renderConf[stanza][RENDERER_SELECT_FIELD] ]['path'] = "/%s%s%s" % (appScope, TEMPLATE_APPSCOPE_SEPARATOR, path )
            else:
                rendererMap['default'] = {
                    'path'      : renderConf[stanza]['template'],
                    'priority'  : 100
                }

        # make a sorted list of renderers by priority, highest to lowest
        priorityList = sorted([ (k,v['priority']) for k,v in rendererMap.items() ], key=itemgetter(1), reverse=True )
        priorityList = [ k for k,v in priorityList ]

        logger.info('_getRenderers - loaded custom event renderers: rendererMap=%s priorityList=%s' % (rendererMap, priorityList))
        return (rendererMap, priorityList)
    
    def _getEventMenu(self, job, event, client_app, latest_time=None):
        '''
        A list of menu data dictionaries items following menu builder convention for an event.
        TODO: Extract this to configuration.

        job: SDK job object.
        event: SDK event object.
        client_app: The current application.
        latest_time: Optional latest time value for the event.
        
        Menu Data dictionary format:
        Required
        'label': 'The event menu label',
        'url': 'The fully qualified url to open',
        
        Optional
        'popup': 'nameedpup' #Lanch in a named popup window
        'remove': True #Destroys the created menu onclick
        '''
        data = []

        # show source
        if set(SHOW_SOURCE_REQUIRED_FIELDS).issubset(set(event.fields.keys())):
            show_source_target = '/app/%s/show_source' % client_app
            show_source_qs = {
                'sid': job.sid,
                'offset': event.offset
            }
            if latest_time:
                show_source_qs['latest_time'] = latest_time
            show_source = {
                'label': _('Show Source'),
                'uri': self.controller.make_url(show_source_target, _qs=show_source_qs),
                'popup': 'showsource',
                'remove': True,
            }
            data.append(show_source)
        
        # interactive field extraction
        extract_fields_target = '/ifx'
        extract_fields_qs = {
            'sid': job.sid,
            'offset' :event.offset,
            'namespace': client_app
        }
        extract_fields = {
            'label': _('Extract Fields'),
            'uri': self.controller.make_url(extract_fields_target, _qs=extract_fields_qs),
            'popup': 'ifx',
            'remove': True,
        }
        data.append(extract_fields)

        '''
        # interactive eventtype builder
        build_eventtype_target = '/etb'
        build_eventtype_qs = {
            'sid': job.sid,
            'offset' :event.offset,
            'namespace': client_app
        }
        build_eventtype = {
            'label': _('Build Eventtype'),
            'uri': self.controller.make_url(build_eventtype_target, _qs=build_eventtype_qs),
            'popup': 'ifx',
            'remove': True,
        }
        data.append(build_eventtype)

        # this time link
        this_time_target = '/app/%s/flashtimeline' % client_app

        logger.error('if this feature is ever uncommented beware there is a subtle bug below that is easy to miss. The times you end up with are off.')
        logger.error('also the cast to float, and ALSO use of dt2epoch will introduce rounding errors in subseconds.')
        earliest = datetime.datetime.fromtimestamp(float(event.epoch))
        latest  = earliest + datetime.timedelta(seconds=1)
        earliestEpoch = splunk.util.dt2epoch(earliest)
        latestEpoch = splunk.util.dt2epoch(latest)

        this_time_qs = {
            'q': job.search,
            'latest': latestEpoch,
            'earliest' : earliestEpoch
        }
        this_time = {
            'label': _('This Time'),
            'uri': self.controller.make_url(this_time_target, _qs=this_time_qs),
            'remove': True,
        }
        data.append(this_time)
        '''
        
        return data
