#
# Splunk UI module python renderer
# This module is imported by the module loader (lib.module.ModuleMapper) into
# the splunk.appserver.mrsparkle.controllers.module.* namespace.
#


# required imports
import cherrypy
import controllers.module as module

# common imports
import splunk, splunk.search, splunk.util, splunk.entity
import lib.util as util
import lib.i18n as i18n

# logging setup
import logging
logger = logging.getLogger('splunk.appserver.controllers.module.SimpleResultsTable')

import math
import cgi

# define standard time field name        
TIME_FIELD = '_time'

# define separator for rendering multi-value fields
MULTI_VALUE_SEPARATOR = '<br />'

# define hard limit for displaying multi-value fields
MAX_MULTI_VALUE_COUNT = 50

class SimpleResultsTable(module.ModuleHandler):
    '''
    Provides module content for the SimpleResultsTable.  The arguments supported 
    are any params supported by the /splunk/search/jobs/<sid>/results endpoint, i.e.,
    
    count
    offset
    field_list
    search
    '''
    
    def generateResults(self, host_app, client_app, sid, count=1000, 
            earliest_time=None, latest_time=None, field_list=None, 
            offset=0, max_lines=None, entity_name='results',
            postprocess=None, display_row_numbers='True', show_preview='0'):
    
        
        # check inputs
        count = max(int(count), 0)
        offset = max(int(offset), 0)
        display_row_numbers = splunk.util.normalizeBoolean(display_row_numbers)
        if not sid:
            raise Exception('SimpleResultsTable.generateResults - sid not passed!')

        # get job
        try:
            job = splunk.search.getJob(sid, sessionKey=cherrypy.session['sessionKey'])
        except splunk.ResourceNotFound:
            return _("job sid=%s not found") % sid
        
        # pass in any field list
        if (field_list) :
            job.setFetchOption(fieldList=field_list, show_empty_fields=False)
        
        if postprocess:
            job.setFetchOption(search=postprocess)

        if show_preview == '1' and entity_name == 'results' and job.resultPreviewCount and not job.isDone:
            entity_name = 'results_preview'
            
        # set formatting
        job.setFetchOption(
            time_format=cherrypy.config.get('DISPATCH_TIME_FORMAT'),
            earliestTime=earliest_time, 
            latestTime=latest_time,
            output_time_format=i18n.ISO8609_MICROTIME
        )

        # build output
        shash = hash(sid)
        output = []
        output.append('<div class="simpleResultsTableWrapper"><table class="simpleResultsTable splTable">')

        # displayable fields; explicitly pull the _time field into the first column
        fieldNames = [x for x in getattr(job, entity_name).fieldOrder if (not x.startswith('_') or x == TIME_FIELD)]
        try:
            timePos = fieldNames.index(TIME_FIELD)
            fieldNames.pop(timePos)
            fieldNames.insert(0, TIME_FIELD)
        except ValueError:
            pass
        
        # cache local dataset
        dataset = getattr(job, entity_name)[offset: offset+count]
        
        # determine the relative values for heatmapping
        localMin = 0
        localMax = 0
        adjustedMaxDelta = 0
        ordered = []
        for row in dataset:
            if 'TOTAL' not in row.values():
                for fieldName in fieldNames:
                    if fieldName != 'TOTAL':
                        try:
                            ordered.append(float(row[fieldName][0].value))
                        except:
                            pass
        ordered.sort()
        if len(ordered):
            localMin = ordered[0]
            localMax = ordered[-1]

            # bracket min/max to 95th percentile
            adjustedMin, adjustedMax = util.getPercentiles(ordered, .05, .95)
            adjustedMaxDelta = max(abs(adjustedMin), abs(adjustedMax))

            logger.debug('SimpleResultsTable - localMin=%s localMax=%s adjustedMin=%s adjustedMax=%s' % (localMin, localMax, adjustedMin, adjustedMax))
            
            
        # generate headers
        output.append('<tr>')
        if display_row_numbers:
            output.append('<th class="pos"></th>')
        for field in fieldNames:
            output.append('<th>%s</th>' % cgi.escape(field))
        
        # generate data
        for i, result in enumerate(dataset):

            # check for the TOTAL row (not column)
            omitFromHeatmap = False
            for val in result.values():
                for item in val:
                    try:
                        if 'TOTAL' in item.getValue():
                            omitFromHeatmap = True
                            break
                    except:
                        pass
            
            rowClass = []
            if i % 5 == 4: rowClass.append('s')

            if len(rowClass):
                output.append('<tr class="%s">' % ' '.join(rowClass))
            else:
                output.append('<tr>')
                
            if display_row_numbers:
                output.append('<td class="pos">%s</td>' % i18n.format_number(i+1+offset))

            for field in fieldNames:
                
                heatValue = None
                isMaxValue = False
                isMinValue = False
                if not omitFromHeatmap and field != 'TOTAL':
                    v = result.get(field, None)
                    try:
                        v = float(v[0].value)
                        heatValue = min(max(math.ceil(v / adjustedMaxDelta * 1000) / 1000, -1), 1)
                        if v == localMax: isMaxValue = True
                        if v == localMin: isMinValue = True
                    except:
                        pass
                output.append('<td')
                if heatValue != None:
                    output.append(' heat="%s"' % heatValue)
                if isMaxValue:
                    output.append(' isMax="1"')
                if isMinValue:
                    output.append(' isMin="1"')
                
                fieldValues = result.get(field, None)

                # render field values; multi-value as a list
                # cap display count to prevent blowout
                if field=='_time' and result.time:
                    output.append('>%s</td>' % i18n.format_datetime_microseconds(result.time))
                elif fieldValues:
                    renderedValues = [cgi.escape(x.value) for x in fieldValues[:MAX_MULTI_VALUE_COUNT]]
                    if len(fieldValues) > MAX_MULTI_VALUE_COUNT:
                        clipCount = len(fieldValues) - MAX_MULTI_VALUE_COUNT
                        renderedValues.append(_('[and %d more values]') % clipCount)
                    output.append('>%s</td>' % MULTI_VALUE_SEPARATOR.join(renderedValues))
                    
                else:
                    output.append('></td>')
                
            output.append('</tr>')
            
        output.append('</table></div>')
        
        if entity_name == 'results' and job.resultCount == 0:
            if job.isDone:
                output = _('No results found')
            else:
                output = _('Waiting for search to complete...')
        else:
            output = ''.join(output)
                
        return output

