Splunk.Module.FlashChart = $.klass(Splunk.Module.FlashWrapper, {

    /**
     * Defines the prefix for settings that should be automatically passed
     * into the charting library.
     *
     * Ex: to set font size, set "charting.fontSize"
     *
     */
    SETTING_PREFIX: 'charting.',

    /**
     * define settingsMap keys that this module should listen to.  These
     * keys are older-style keys that do not have the standard chart
     * setting prefix defined in SETTING_PREFIX.
     *
     */
    LEGACY_SETTINGS: {
        "chart": 1,
        "chart.stackMode": 1,
        "secondaryAxis.scale": 1,
        "chart.nullValueMode": 1,
        "legend.placement": 1,
        "primaryAxisTitle.text": 1,
        "secondaryAxisTitle.text": 1,
        "chart.showMarkers": 1,
        "secondaryAxis.minimumNumber": 1,
        "secondaryAxis.maximumNumber": 1
    },

    USE_FLASH_POLLER: false,


    initialize: function($super, container){
        $super(container);
        this.logger = Splunk.Logger.getLogger("flash_chart.js");
        
        this.legend_onSetLabels = this.legend_onSetLabels.bind(this);
        this.legend_onLabelIndexMapChanged = this.legend_onLabelIndexMapChanged.bind(this);
    },


    initializeBridge: function() {
        this.bridge.addProperty("legend.numLabels", this.legend_numLabels.bind(this));
        this.bridge.addMethod("legend.setLabels", this.legend_setLabels.bind(this));
        this.bridge.addMethod("legend.getLabelIndex", this.legend_getLabelIndex.bind(this));
        this.bridge.addEvent("legend.setLabels");
        this.bridge.addEvent("legend.labelIndexMapChanged");
    },

    legend_numLabels: function() {
        return Splunk.Legend.numLabels();
    },
    legend_setLabels: function(labels) {
        Splunk.Legend.setLabels(this.bridge.id(), labels);
    },
    legend_getLabelIndex: function(label) {
        return Splunk.Legend.getLabelIndex(label);
    },
    legend_onSetLabels: function() {
        this.bridge.dispatchEvent("legend.setLabels");
    },
    legend_onLabelIndexMapChanged: function() {
        this.bridge.dispatchEvent("legend.labelIndexMapChanged");
    },


    onConnect: function($super, isReconnect) {
        
        Splunk.Legend.register(this.bridge.id());
        Splunk.Legend.addEventListener("setLabels", this.legend_onSetLabels);
        Splunk.Legend.addEventListener("labelIndexMapChanged", this.legend_onLabelIndexMapChanged);

        $super(isReconnect); // called last since it sets _isBridgeConnected == true
    },
    

    onClose: function($super) {
        Splunk.Legend.removeEventListener("labelIndexMapChanged", this.legend_onLabelIndexMapChanged);
        Splunk.Legend.removeEventListener("setLabels", this.legend_onSetLabels);
        Splunk.Legend.unregister(this.bridge.id());

        $super();
    },

    /**
     * Workaround to fix SPL-27325.  Only needed in 4.0.X code.
     * in a nutshell we have fixed the bug by no longer even telling the swf what the search id is
     * until we know there are some results or something worth requesting. 
     */
    jobHasData: function(job) {
        return (job.getResultCount() > 0 || (!job.areResultsTransformed() && job.getEventCount > 0));
    },

    onContextChange: function() {
        if (!this._isBridgeConnected) {
            this.logger.debug("bridge is not connected onContextChange. Exiting. onContextChange will fire onConnect.");
            return;
        }

        var context = this.getBaseContext();
        var plotIntention = context.getIntentionReference("plot");
        var isTopOrRare = (plotIntention && plotIntention["arg"]["mode"] in {"top":1, "rare":1})
        
        // if the job is already done there will be no progress events, and right here the jobId assignment
        // will trigger the final render. 
        // In order for PageStatus to be notified of these renders, we have to set up a monitor here.
        if (context.job.isDone()) {
            // Notifying PageStatus that a render is beginning. 
            this.renderMonitor = Splunk.Globals['PageStatus'].register(this.moduleType + ' - rendering final data - ' + this.container.attr('id'));
        }
        if (this.jobHasData(context.job)) {
            this.setBridgeProperty("jobID", context.job.getSearchId());
            this.callBridgeMethod("update");
        }
    },
    determineAxisType: function(intentionArg) {
        var fieldNames = [];
        
        //var fieldsList   = plotIntention["arg"]["fields"]
        // TODO - plot intentions in mode="chart" unfortunately extract the x-axis arg as though it was a 'splitby', which its not.
        var splitBy = intentionArg["splitby"]
        // TODO See above comment about mode="chart" and splitBy
        if (splitBy) fieldNames.push(splitBy);
        // TODO - when the above is fixed, the correct thing to do will be to get the [1] element from the fields themselves.
        //for (var i=0;i<fieldsList.length; i++) {
        //    fieldNames.push(fieldsList[i][1]);
        //}
        
        var args = {
            field_list:  fieldNames,
            top_count: 0,
            min_freq: 0
        };
        // we want a context that was run with statusBuckets>0.  
        // If there's a base_sid attribute in the URL, that is exactly the general sort of thing
        // that guy is there for.  If he's there we use him. 
        var context; 
        var qsDict = Splunk.util.queryStringToProp(document.location.search);
        if (qsDict.hasOwnProperty("base_sid")) {
            context = Splunk.SearchContext.resurrectFromSearchId(qsDict["base_sid"]);
        } else {
            // if no base_sid was found, we try and use our own sid, but it's quite likely 
            // it was run with status_buckets=0 so our summary request will fail.
            context = this.getBaseContext().clone();
        }
        $.get(context.getUrl('summary', args), function(resultXML) {
            var eventCount = parseInt($(resultXML).find("summary").attr("c")); 
            var moduleInstance = this;
            $.each($(resultXML).find("field"), function(index) {
                var fieldElt = $(this);
                // treat as numeric if HALF or more of the occurences are considered numeric
                var isNumeric = (fieldElt.attr("nc") > fieldElt.attr("c")/2);
                if (isNumeric) {
                    moduleInstance.bridge.callMethod("setValue", "primaryAxis", "numeric");
                    moduleInstance.bridge.callMethod("setValue", "primaryAxis.fitZero", "false");
                }
            });
            
        }.bind(this));        
    },
    onJobProgress: function($super) {
        $super();
        if (this.USE_FLASH_POLLER) return false;
        var context = this.getBaseContext();
        if (this.jobHasData(context.job)) {
            this.setBridgeProperty("jobID", context.job.getSearchId());
            this.callBridgeMethod("update");
        }
    },
    onJobDone: function($super) {
        $super();
        if (this.USE_FLASH_POLLER) return false;
        var context = this.getBaseContext();
        this.setBridgeProperty("jobID", context.job.getSearchId());
        this.callBridgeMethod("update");
    },

    /**
     * Returns the proper key to pass into the flash charting library.  Keys
     * that are prefixed with SETTING_PREFIX are stripped.  Returns null if
     * the key does not have the correct prefix or is not listed in
     * LEGACY_SETTINGS.
     *
     */
    getNormalizedSetting: function(settingName) {
        if (!settingName) return null;
        if (this.LEGACY_SETTINGS.hasOwnProperty(settingName)) {
            return settingName;
        }
        if (settingName.substring(0, this.SETTING_PREFIX.length) == this.SETTING_PREFIX) {
            return settingName.substring(this.SETTING_PREFIX.length);
        }
        return null;
    },


    onSettingsChange: function(settingsMap) {
        
        // set the flash chart properties
        for (var key in settingsMap) {
            if (settingsMap.hasOwnProperty(key) && this.getNormalizedSetting(key)) {
                this.logger.debug('FlashChart - set ' + key + '=' + settingsMap[key]);
                try {
                    this.callBridgeMethod("setValue", 
                        this.getNormalizedSetting(key), 
                        settingsMap[key]
                    );
                } catch (e) {
                    this.logger.error("catching a jabridge exception '", e, "'. Basically this is the SPL-17758 problem. catching it so it doesnt derail everything. ");
                }
            }
        }
        var context = this.getBaseContext();
        var plotIntention = context.getIntentionReference("plot");
        var isTopOrRare = (plotIntention && plotIntention["arg"]["mode"] in {"top":1, "rare":1})
        
        // #1 - special case for top/rare, where we tell the charting system to only render the top N rows.
        if (this._params["maxRowsForTop"] && isTopOrRare && (this.getSetting("chart") != "pie")) {
            this.setBridgeProperty("resultsCount", this._params["maxRowsForTop"]);
        } else {
            this.setBridgeProperty("resultsCount", this.getParam('maxResultCount'));
        }
        
        // #2 - another special case for top/rare, where we suppress the 'percent' field.
        try {
            if (isTopOrRare) {
                this.setBridgeProperty("fieldHideList", ["percent"]);
            } else {
                this.setBridgeProperty("fieldHideList", null);
            }
        } catch(e) {
            this.logger.error("error trying to set the fieldHideList");
        }
        if (plotIntention && plotIntention["arg"]["mode"]=="chart") {
            this.determineAxisType(plotIntention["arg"]);
        }
        // set the ancillary props
        if (settingsMap.hasOwnProperty('chartTitle') && settingsMap['chartTitle']) {
            $('.chartTitle', this.container).text(settingsMap['chartTitle']).show();
        } else {
            $('.chartTitle', this.container).hide();
        }
        
    }
});
