Splunk.namespace("Module");

/**
 * FlashTimeline displays a timeline of events from a search.  It allows
 * time range selection queries.
 *
 * AS bridge properties:
 *  viewMinimum: lower most bound in a time range
 *  viewMaximum: uppoer most bound in a time range
 *  selectionMinimum: lower most bound in a selected part of a time range
 *  selectionMaximum: upper most bound in a selected part of a time range
 *
 * AS bridge events:
 *  viewChanged: fired when the range changes from zoom ins or outs
 *  selectionChanged: fired when the selected time area changes
 *  chartClicked: fired when a mouse click is captured on the chart
 */

Splunk.Module.FlashTimeline = $.klass(Splunk.Module.FlashWrapper, {
    initialize: function($super, container){
        $super(container);
        this._controlLinksContainer = $(".controlLinks",this.container);
        $("a", this._controlLinksContainer).hide();
        this.logger = Splunk.Logger.getLogger("flash_timeline.js");
        

        $("a.zoomIn",this.container).click(this.zoomIn.bind(this));
        $("a.zoomOut",this.container).click(this.zoomOut.bind(this));
        $("a.linScale",this.container).click(this.linScale.bind(this));
        $("a.logScale",this.container).click(this.logScale.bind(this));
        $("a.selectAll",this.container).click(this.selectAll.bind(this));
        
        this.isEntireRangeSelected = true;
    },
    onBeforeJobDispatched: function(context) {
        context.setMinimumStatusBuckets(this._params["statusBuckets"]);
    },
    initializeBridge: function() {
        this.bridge.addMethod("formatTooltip", this.formatTooltip.bind(this), [ "earliestTime", "latestTime", "eventCount" ], "String");
    },
    
    /**
     * Augment the presentation control for flash timeline to support series colors.
     */
    setPresentation: function($super){
        var seriesColors = this.getCSSColor([this.container], "border-right-color");
        if(seriesColors){
            try{
                this.bridge.callMethod("setValue", "seriesColors", "["+seriesColors.replace("#", "0x")+"]");
                this.logger.info("Set foregroundColor", seriesColors);
            }catch(e){
                this.logger.error("Could not setValue for seriesColors");
            }
        }
        $super();
    },
    
    getModifiedContext: function() {
        var context = this.getBaseContext();
        if (this.isEntireRangeSelected) return context;

        var selectedBuckets = this.bridge.callMethod("getSelectedBuckets");
        // for reasons unknown,  callBridgeMethod doesnt work with getSelectedBuckets
        //var selectedBuckets = this.callBridgeMethod("getSelectedBuckets");

        var continueSummingAvailableEvents = true;
        if (!selectedBuckets) {
            this.logger.warn(this.moduleType, " Assertion failed - getModifiedContext was called but getSelectedBuckets is null.");
            return context;
        }
        var numberOfBuckets = selectedBuckets.length;

        var totalEvents = 0;
        var totalEventsAvailable = 0;
        for (var i=numberOfBuckets-1; i>-1; i--) {
            var b = selectedBuckets[i];
            totalEvents += b["eventCount"];
            if (continueSummingAvailableEvents) {
                totalEventsAvailable += b["eventAvailableCount"];
            }
            if (b["eventAvailableCount"] < b["eventCount"]) {
                // TODO mark the completness/incompleteness of the set, so other modules can know its not complete.
                //      currently everything is still using onJobDone to determine >= or exact vals, which is incorrect
                //      in all selected range cases
                continueSummingAvailableEvents = false;
            }
        }
        context.setSelectedEventAvailableCount(totalEventsAvailable);
        context.setSelectedEventCount(totalEvents);
        var range = this.getSelectionRange(selectedBuckets);
        context.setTimeRange(range);
        return context;
    },


    onContextChange: function() {
        // from this point, until the timelineDataAvailable event is fired, we must
        // disable or ignore all interaction.
        $("a",this._controlLinksContainer).hide();
        // when getting a new context, reset the selection range.
        this.clearSelectionRange();
        this.setBridgeProperty("jobID", this.getBaseContext().job.getSearchId());
    },

    onConnect: function($super, isReconnect) {
        this.bridge.addEventListener('selectionChanged', this.onSelectionChanged.bind(this));
        $super(isReconnect); // called last since it sets _isBridgeConnected == true
    },

    /**
     * After an unpause, manually tell the timeline to start monitoring
     * the sid again.
     */
    onJobStatusChange: function($super, event, status) {
        $super(event, status);
        var context = this.getBaseContext();
        var sid = context.job.getSearchId();
        if ((status == 'unpause' || status == 'finalize') && sid) {
            this.setBridgeProperty('jobID', sid);
        }
    },
    _getTimeRangeFromUTC: function(earliestPropertyName, latestPropertyName) {
        try {
            var earliestUTC = this.getBridgeProperty(earliestPropertyName);
            var latestUTC   = this.getBridgeProperty(latestPropertyName);

            if (Splunk.util.isInt(Math.floor(earliestUTC)) && Splunk.util.isInt(Math.floor(latestUTC))) {
                var buckets = this.bridge.getProperty("timelineData").buckets;
                var earliestOffset = null;
                var latestOffset   = null;
                
                if (buckets.length>0) {
                    earliestOffset = buckets[0].earliestTimezoneOffset;
                    latestOffset = buckets[buckets.length-1].latestTimezoneOffset;
                }
                
                return new Splunk.TimeRange(earliestUTC, latestUTC, earliestOffset, latestOffset)
            } else {
                this.logger.error("undefined values " + earliestUTC + ", " + latestUTC);
            }
        } catch (e) {
            this.logger.error(this.moduleType, " exception getting earliest and latest selected times - " + e);
        }
        return new Splunk.TimeRange();
    },
    getSelectionRange: function(selectedBuckets) {
        selectedBuckets = selectedBuckets || this.bridge.callMethod("getSelectedBuckets");
        var numberOfBuckets = selectedBuckets.length;
        if (numberOfBuckets==0) {
            this.logger.error(this.moduleType, " getSelectionRange returned an empty selectedBuckets. Returning 'All Time'");
            return new Splunk.TimeRange();
        }
        var earliestBucket = selectedBuckets[0];
        var latestBucket = selectedBuckets[numberOfBuckets-1];
        var range = new Splunk.TimeRange(earliestBucket["earliestTime"], latestBucket["latestTime"], earliestBucket["earliestTimezoneOffset"], latestBucket["latestTimezoneOffset"]);
        range.setAsSubRangeOfJob(!this.isEntireRangeSelected);

        return range;
    },
    clearSelectionRange: function() {
        this.isEntireRangeSelected = true;
        this.setBridgeProperty("selectionMinimum", NaN);
        this.setBridgeProperty("selectionMaximum", NaN);
    },

    getTimelineRangeFromUTC: function() {
        return this._getTimeRangeFromUTC("viewMinimum", "viewMaximum");
    },
    onSelectionChanged: function() {
        this.isEntireRangeSelected = false;
        $("a.selectAll",this.container).show();
        if (this.getBaseContext().isJobDispatched()) {
            this.pushContextToChildren();
        }
    },
    getScaleDescription: function() {
        var timelineScale = this.getBridgeProperty("timelineScale");
        if (!timelineScale) return "";
        var unit = timelineScale.unit;
        var value= timelineScale.value;
        if (parseFloat(value) < 1) {
            if (unit == "second") {
                value = value * 1000;
                unit = "millisecond"
                 
            }
            else this.logger.error("error - timelineScale has a fractional unit but not in seconds.");
        }
        if (unit == "millisecond") {
            return sprintf(ungettext(_('1 bar = %s millisecond'), _('1 bar = %s milliseconds'), value), value);
        } else if (value==1) {
            switch (unit) {
                case "year":
                    return _('1 bar = 1 year');
                case "month":
                    return _('1 bar = 1 month');
                case "day":
                    return _('1 bar = 1 day');
                case "hour":
                    return _('1 bar = 1 hour');
                case "minute":
                    return _('1 bar = 1 minute');
                case "second":
                    return _('1 bar = 1 second');
                default: 
                    this.logger.error("received uncaught unit of ", unit);
            }
        } else {
            this.logger.error("received a timelineScale that has >1 unit per bucket. (" + value + "). This should not happen");
        }
        return "";
    },
    onDataUpdated: function($super) {
        $super();
        if (this.getBaseContext().isJobDispatched()) {
            // we have an axis, so we can zoom in and out.
            $("a.zoomIn",this.container).show();
            $("a.zoomOut",this.container).show();
            $("a.linScale",this.container).show();
            $("a.logScale",this.container).show();
            
            $(".bucketSize", this.container).html(this.getScaleDescription()); 
        }
    },
    onJobProgress: function($super) {
        $super();
        this.callBridgeMethod("update");
    },
    _passTimeRangeToParent: function(range) {
        var context = new Splunk.SearchContext();
        context.setTimeRange(range);
        this.passContextToParent(context);
    },
    zoomIn: function() {
        var range = this.getSelectionRange();
        
        // if no selection is recorded, or if the entire range is selected,
        // then we defer to the generic zoom methods on the TimeRange class.
        if (!range || (range && range.equalToRange(this.getTimelineRangeFromUTC()))) {
            // TODO - there is further logic that 3.X had, that we'll need to put
            // back someday,  where we check whether the resulting recent-half
            // of the range is unpopulated
            //   If it is, we instead snap into the populated data..
            // unless the *entire* timeline is unpopulated, in which case we
            // scratch our heads but let the generic zoomIn method zoom into the
            // recent half even though it's empty.
            
            this._passTimeRangeToParent(range.zoomIn());
        // otherwise the user has selected a range and the range is a subset.
        // In this case we zoom into exactly that subset range.
        } else {
            this._passTimeRangeToParent(range);
        }

        // when we get the next jobId, the whole range will be selected again.
        $("a.selectAll",this.container).hide();
    },
    zoomOut: function() {
        var timelineRange = this.getTimelineRangeFromUTC();
        
        if (timelineRange) {
            this._passTimeRangeToParent(timelineRange.zoomOut());
        } else {
            this.logger.error(this.moduleType, " could not zoom out cause we had an undefined timelineRange");
        }
    },
    logScale: function(evt) {
        $("a.linScale", this.container).removeClass("linScaleSelected");
        var t = evt.target;
        if ( $(t).is('a') ) {
            $(t).addClass("logScaleSelected");
        } else if ( $(t).is('span') ) {
            $(t).parent('a').addClass("logScaleSelected");
        }
        this.callBridgeMethod("setValue","secondaryAxis.scale","log")
    },
    linScale: function(evt) {
        $("a.logScale", this.container).removeClass("logScaleSelected");
        var t = evt.target;
        if ( $(t).is('a') ) {
            $(t).addClass("linScaleSelected");
        } else if ( $(t).is('span') ) {
            $(t).parent('a').addClass("linScaleSelected");
        }
        this.callBridgeMethod("setValue","secondaryAxis.scale","linear")
    },
    selectAll: function() {
        $("a.selectAll",this.container).hide();
        this.clearSelectionRange();
        if (this.getBaseContext().isJobDispatched()) {
            this.pushContextToChildren();
        }
    },
    resetUI: function() {
        this.setBridgeProperty('jobID', '');
        $("a",this._controlLinksContainer).hide();
    },
    formatTooltip: function(earliestTime, latestTime, earliestOffset, latestOffset, eventCount) {
        var range = new Splunk.TimeRange(earliestTime, latestTime, earliestOffset, latestOffset);
        var tooltip = _(sprintf(_("%(eventCount)s events %(timeRangeString)s"), {eventCount: eventCount, timeRangeString: range.toConciseString()}));
        return tooltip;        
    }
});
